/******************************************************************
 *
 *  File:   wan_pm.c
 *
 *  Author: Nick Vaccaro <nvaccaro@lab126.com>
 *
 *  Date:   05/22/06
 *
 *  Copyright 2006, Lab126, Inc.  All rights reserved.
 *
 *  Description:
 *
 *      WAN driver support for the Fiona Power Management System
 *
 ******************************************************************/
#if 0
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/completion.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/syscalls.h>
#include <asm/uaccess.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/suspend.h>
#include <asm/system.h>
#endif

#include <linux/kernel.h>
#include <asm/arch/system.h>
#include <asm/arch/fiona.h>
#include <asm/arch/fiona_eventmap.h>
#include <asm/arch/fpow.h>	// FPOW-related constants
#include <net/anydata.h>    // For set_wan_power and wan_status_t

//extern fiona_hw g_fiona_hw;
#define WAKE_ON_WAN_GPIO    (g_fiona_hw.wake_on_wan_gpio)
#define WAN_SWITCH_GPIO     (g_fiona_hw.wan_switch_gpio)

// Static global holding current FPOW power state of the driver
FPOW_POWER_MODE	g_wan_power_mode = FPOW_MODE_INVALID; 

// Static global holding fpow device entry returned from fpow_register()
struct fpow_component *g_wan_fpow_component_ptr = NULL;

static void
wake_on_wan_ring(int enable)
{
    if (enable)  {
        // Set the "wake on WAN ring" source, rising edge only
		fiona_set_wake_source(WAKE_ON_WAN_GPIO);
		fiona_set_rising_edge_wake_source(WAKE_ON_WAN_GPIO);
        powdebug("wake_on_wan_ring(enable) - pwer after = %x\n",g_fiona_hw.pwer);
    }
    else  {
        // clear the "wake on WAN ring" source, rising edge only
		fiona_clear_wake_source(WAKE_ON_WAN_GPIO);
		fiona_clear_rising_edge_wake_source(WAKE_ON_WAN_GPIO);
        powdebug("wake_on_wan_ring(disable) - pwer after = %x\n",g_fiona_hw.pwer);
    }
}

void 
wake_on_wan_switch(int enable)
{
    if (enable)  {
        // Set the "wake on WAN switch transitions" source, 
        //  trigger on rising and falling edge
		fiona_set_wake_source(WAN_SWITCH_GPIO);
		fiona_set_rising_edge_wake_source(WAN_SWITCH_GPIO);
		fiona_set_falling_edge_wake_source(WAN_SWITCH_GPIO);
        powdebug("wake_on_wan_switch(enable) - pwer after = %x\n",g_fiona_hw.pwer);
    }
    else  {
        // clear the "wake on WAN ring" source, rising and falling edge
		fiona_clear_wake_source(WAN_SWITCH_GPIO);
		fiona_clear_rising_edge_wake_source(WAN_SWITCH_GPIO);
		fiona_clear_falling_edge_wake_source(WAN_SWITCH_GPIO);
        powdebug("wake_on_wan_switch(disable) - pwer after = %x\n",g_fiona_hw.pwer);
    }
}


/*********************************************************
 * Routine:     wan_fpow_setmode
 *
 * Purpose:     Handles Fiona Power Management callbacks
 *              for doing WAN power management.
 *
 * Paramters:   private - value passed at registration - ignored
 *              state   - ignored
 *              mode    - Power mode to switch to
 *                          Currently supported Modes:
 *                              FPOW_MODE_ON
 *                              FPOW_MODE_SLEEP
 *                              FPOW_MODE_OFF
 *
 * Returns:     
 *  FPOW_NO_ERR                     - Transitioned to mode with no errors.
 *  -FPOW_NOT_ENOUGH_POWER_ERR      - not enough power
 *  -FPOW_COMPONENT_TRANSITION_ERR  - Failed to transition to mode
 *  -EINVAL                         - Invalid mode
 *
 * Assumptions: none
 *********************************************************/

static FPOW_ERR 
wan_fpow_setmode(void *private, u32 state, u32 mode)
{
	int	retVal = FPOW_NO_ERR;

    powdebug("wan_fpow_setmode(%s)\n",fpow_mode_to_str(mode));

	// Handle the "wake on ring" case
    //
    // NOTE:  It doesn't matter what mode he's setting us to, as we need
    // to dispatch the TPH if it came in so it doesn't get lost, irrespective
    // of what power level the caller currently wants.  This lets the brains
    // of the state machine decide what to do about the TPH (if WAN is
    // forced to an OFF or SLEEP state, it will save it off until it can respond
    // to it later).
    //
    // This is OK to do everytime because the PEDR bit is cleared after doing
    // it the first time, assuring that if it happens again, it was because
    // a TPH actually came in again while PXA was sleeping and woke the device.
    // 
	if (PEDR & WAN_WAKE_PEDR_BIT) {
		// "WAN ring" woke up the device; clear the indicator
		PEDR = WAN_WAKE_PEDR_BIT;
		powdebug("wan_fpow_setmode(%s) sent WAN_RING_EVENT\n",fpow_mode_to_str(mode));
		printk("wan_pm: tph wake event occurred; sending WAN_RING_EVENTCODE notification\n");
		handle_fiona_event(WAN_RING_EVENTCODE);
	}

	// If we're already in the proper mode, no need to do anything else
	if (g_wan_power_mode == mode)
		goto exit;

	// Switch to the appropriate power mode
	switch (mode)  {
		case FPOW_MODE_ON:
            // Based on discussion / direction from Ken/Shahriar, the
            // Framework is now going to be responsible for WAN power.
            // Commented out power transition calls. -njv

			// Turn on the module and all wake sources
			// retVal = set_wan_power(WAN_ON);
			if (retVal == FPOW_NO_ERR) {
				g_wan_power_mode = mode;
                wake_on_wan_ring(1);
                wake_on_wan_switch(1);
			}
			break;

		case FPOW_MODE_SLEEP:
			// Set the "wake on WAN ring" source
            wake_on_wan_ring(1);

            // Wake on WAN switch transitions
            wake_on_wan_switch(1);

            // For now, don't know about an intermediate mode, so we're going
            // completely off for now.  May add as I learn more about chip. -njv
			// retVal = set_wan_power(WAN_OFF); 
			if (retVal == FPOW_NO_ERR)
				g_wan_power_mode = mode;
			break;

		case FPOW_MODE_OFF:
			// Clear the "wake on WAN ring" source
            wake_on_wan_ring(0);

			// retVal = set_wan_power(WAN_OFF); 
			if (retVal == FPOW_NO_ERR)
				g_wan_power_mode = mode;
			break;

		default:
			retVal = -EINVAL;
			break;
	}

exit:
	powdebug("WAN: driver now in %s mode\n",fpow_mode_to_str(g_wan_power_mode));
    return(retVal);
}

static int
wan_fpow_getmode(void *private) 
{
    powdebug("wan_fpow_getmode(0x%x) returning %s\n",(int)private,fpow_mode_to_str(g_wan_power_mode));
	return (g_wan_power_mode);
}

int 
wan_register_with_fpow(struct fpow_component **fpow_dev, void *private)
{
    struct fpow_registration_rec   reg;
    int rc = 0;

	powdebug("wan_register_with_fpow(0x%x, 0x%x)\n",(int)fpow_dev, (int)private);
    memclr(&reg,sizeof(reg));

    strcpy(reg.name,"wan");
    reg.device_class = FPOW_CLASS_WAN;
    reg.supported_modes = ( FPOW_MODE_ON_SUPPORTED | \
                            FPOW_MODE_SLEEP_SUPPORTED | \
                            FPOW_MODE_OFF_SUPPORTED);
    reg.private = private;
    reg.getmode = wan_fpow_getmode;
    reg.setmode = wan_fpow_setmode;
    *fpow_dev = fpow_register(&reg);
    if (*fpow_dev == NULL) {
        printk("WAN: Failed to register with fpow !!\n");
        rc = -1;
    }
    return(rc);
}

void
wan_unregister_with_fpow(struct fpow_component **fpow_dev)
{
	fpow_unregister(*fpow_dev);
    *fpow_dev = NULL;
}
