/*
 * linux/sound/soc.h -- ALSA SoC Layer
 *
 * Author:		Liam Girdwood
 * Created:		Aug 11th 2005
 * Copyright:	Wolfson Microelectronics. PLC.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
 
#ifndef __LINUX_SND_SOC_H
#define __LINUX_SND_SOC_H

#include <linux/device.h>
#include <linux/types.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/control.h>
#include <sound/ac97_codec.h>

#ifdef CONFIG_PM
// This SoC layer was backported from 2.6.17.  This kernel does not
// have this message type yet....
typedef unsigned int pm_message_t;
#endif

/*
 * Convenience kcontrol builders
 */

#define SOC_SINGLE_VALUE(reg,shift,mask,invert) ((reg) | ((shift) << 8) | ((shift) << 12) | ((mask) << 16) | ((invert) << 24))
#define SOC_SINGLE(xname, reg, shift, mask, invert) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, .name = xname, .info = snd_soc_info_volsw, \
  .get = snd_soc_get_volsw, .put = snd_soc_put_volsw, \
  .private_value =  SOC_SINGLE_VALUE(reg, shift, mask, invert) }
#define SOC_DOUBLE(xname, reg, shift_left, shift_right, mask, invert, power) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, .name = (xname), .info = snd_soc_info_volsw, \
  .get = snd_soc_get_volsw, .put = snd_soc_put_volsw, \
  .private_value = (reg) | ((shift_left) << 8) | ((shift_right) << 12) | ((mask) << 16) | ((invert) << 24) }
#define SOC_ENUM_DOUBLE(xreg, xshift_l, xshift_r, xmask, xtexts) \
{ .reg = xreg, .shift_l = xshift_l, .shift_r = xshift_r, \
  .mask = xmask, .texts = xtexts }
#define SOC_ENUM_SINGLE(xreg, xshift, xmask, xtexts) \
	SOC_ENUM_DOUBLE(xreg, xshift, xshift, xmask, xtexts)
#define SOC_ENUM(xname, xenum) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, .name = xname, .info = snd_soc_info_enum_double, \
  .get = snd_soc_get_enum_double, .put = snd_soc_put_enum_double, \
  .private_value = (unsigned long)&xenum }

 
/*
 * Audio interface types
 */
#define SND_SOC_AC97		0x1
#define SND_SOC_I2S			0x2
#define SND_SOC_SSP			0x4

/*
 * hw interface formats
 */
#define SND_SOC_HWFMT_I2S			(1 << 0)	/* I2S mode */
#define SND_SOC_HWFMT_RIGHT_J		(1 << 1)	/* Right justified mode */
#define SND_SOC_HWFMT_LEFT_J		(1 << 2)	/* Left Justified mode */
#define SND_SOC_HWFMT_DSP_A			(1 << 3)	/* L data msb after FRM */
#define SND_SOC_HWFMT_DSP_B			(1 << 4)	/* L data msb during FRM */

/*
 * hw signal inversions
 */
#define SND_SOC_HWFMT_NB_NF			(1 << 24)	/* normal bit clock + frame */
#define SND_SOC_HWFMT_NB_IF			(1 << 25)	/* normal bclk + inv frm */
#define SND_SOC_HWFMT_IB_NF			(1 << 26)	/* invert bclk + nor frm */
#define SND_SOC_HWFMT_IB_IF			(1 << 27)	/* invert bclc + frm */

/*
 * hw clock masters
 * This is wrt the codec, the inverse is true for the interface
 * i.e. if the codec is clk and frm master then the interface is 
 * clk and frame slave.
 */
#define SND_SOC_HWFMT_CBM_CFM		(1 << 28)	/* codec clk & frm master */
#define SND_SOC_HWFMT_CBS_CFM		(1 << 29)	/* codec clk slave & frm master */
#define SND_SOC_HWFMT_CBM_CFS		(1 << 30)	/* codec clk master & frame slave */
#define SND_SOC_HWFMT_CBS_CFS		(1 << 31)	/* codec clk & frm slave */


#define SND_SOC_FORMAT_MASK		0x000000ff
#define SND_SOC_INV_MASK		0x0f000000
#define SND_SOC_CLOCK_MASK		0xf0000000

#define SND_SOC_HWBITS(x)		(1 << (x - 1))	/* sample size in bits */

/*
 * Codec IO
 */
#define snd_soc_read(codec, reg) codec->read(codec, reg)
#define snd_soc_write(codec, reg, value) codec->write(codec, reg, value)

struct snd_soc_pcm_stream;
struct snd_soc_ops;
struct snd_soc_hw_mode;
struct snd_soc_hw_bus;
struct snd_soc_pcm_codec;
struct snd_soc_pcm_interface;
struct snd_soc_codec;
struct snd_soc_platform;
struct snd_soc_machine_config;
struct snd_soc_machine;
struct soc_enum;

/*
 * SoC registration objects:-
 * 
 *  o Codec (n)
 *  o Platform (1)
 *  o Interface (n)
 *  o Machine (1)
 * 
 * Only 1 Platform and Machine may be registered with the SoC layer. However there is
 * no limit on the number of codecs and interfaces that can be registered.
 */
int snd_soc_register_codec(struct snd_soc_codec *codec);
void snd_soc_unregister_codec(struct snd_soc_codec *codec);

int snd_soc_register_platform(struct snd_soc_platform *platform);
void snd_soc_unregister_platform(struct snd_soc_platform *platform);

int snd_soc_register_interface(struct snd_soc_pcm_interface *iface);
void snd_soc_unregister_interface(struct snd_soc_pcm_interface *iface);

int snd_soc_register_machine(struct snd_soc_machine *machine);
void snd_soc_unregister_machine(struct snd_soc_machine *machine);

/* pcm <-> interface connect */
int snd_soc_pcm_connect(struct snd_soc_machine *machine, int nconfig);
int snd_soc_pcm_register(struct snd_soc_machine *machine, char *name);

/* set runtime hw params */
int snd_soc_set_runtime_hw(snd_pcm_substream_t *substream, 
	const snd_pcm_hardware_t *hw);

/* codec register bit access */
int snd_soc_update_bits(struct snd_soc_codec *codec, unsigned short reg,
				unsigned short mask, unsigned short value);
				
int snd_soc_test_bits(struct snd_soc_codec *codec, unsigned short reg,
				unsigned short mask, unsigned short value);

/*
 *Controls
 */
snd_kcontrol_t *snd_soc_cnew(const snd_kcontrol_new_t *_template, 
	void *data, char *long_name);

int snd_soc_info_enum_double(snd_kcontrol_t * kcontrol, snd_ctl_elem_info_t * uinfo);
int snd_soc_get_enum_double(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol);
int snd_soc_put_enum_double(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol);
int snd_soc_info_volsw(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t * uinfo);
int snd_soc_get_volsw(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol);
int snd_soc_put_volsw(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol);
	
/* SoC supported stream types */
struct snd_soc_pcm_stream {
	char *sname;
	unsigned int rates;		/* SNDRV_PCM_RATE_* */
	u64 formats;			/* SNDRV_PCM_FMTBIT_* */
	unsigned int rate_min;		/* min rate */
	unsigned int rate_max;		/* max rate */
	unsigned int channels_min;	/* min channels */
	unsigned int channels_max;	/* max channels */
};

/* SoC audio ops */
struct snd_soc_ops {
	int (*startup)(snd_pcm_substream_t *);
	void (*shutdown)(snd_pcm_substream_t *);
	int (*hw_params)(snd_pcm_substream_t *, snd_pcm_hw_params_t *);
	int (*hw_free)(snd_pcm_substream_t *);
	int (*prepare)(snd_pcm_substream_t *);
	int (*trigger)(snd_pcm_substream_t *, int);
};

/* SoC hardware mode */
struct snd_soc_hw_mode {
	u32 hformat;	/* SND_SOC_HWFMT_* */
	u32 hbits;		/* SND_SOC_HWBITS() */
	u32 rate;		/* sample rate */
	u32 fs;			/* ratio of rate:bclk */ 
	long priv1;		/* private mode data 1 */
	long priv2;		/* private mode data 2 */
	long priv3;		/* private mode data 3 */
};

/* SoC hardware bus and all modes */
struct snd_soc_hw_bus {
	int num_hmodes;				/* number of hw modes */
	struct snd_soc_hw_mode *hmodes;	/* array of supported audio hw modes */	
};

/* SoC Codec pcm stream */
struct snd_soc_pcm_codec {
	char *name;

	/* codec pcm capabilities */
	struct snd_soc_pcm_stream playback;
	struct snd_soc_pcm_stream capture;
	struct snd_soc_hw_bus hw;
	int nplayback;
	int ncapture;
	
	/* runtime info */
	struct snd_soc_hw_mode hw_runtime;
	struct snd_soc_ops ops;
	struct snd_soc_pcm_interface *pcm_i;
	struct snd_soc_codec *codec;
	
	/* private data */
	void *priv;
	void (*private_free)(snd_pcm_runtime_t *runtime);
};

/* SoC platform audio port interface */
struct snd_soc_pcm_interface {
	
	/* interface description */
	char *name;
	unsigned int id;
	unsigned char type;
	
	int probed;
	int (*probe)(struct snd_soc_pcm_interface *interface);
	void (*remove)(struct snd_soc_pcm_interface *interface);
	int (*suspend)(struct snd_soc_pcm_interface *interface);
	int (*resume)(struct snd_soc_pcm_interface *interface);
	
	/* interface pcm capabilities */
	struct snd_soc_pcm_stream capture;
	struct snd_soc_pcm_stream playback;
	struct snd_soc_hw_bus hw;
	
	/* runtime info */
	struct snd_soc_hw_mode hw_runtime;
	struct snd_soc_pcm_codec *pcm_c;
	struct snd_soc_ops ops;
	snd_card_t *card;
	ac97_t *ac97;
	
	/* private data */
	void *platform_data;
	
	struct list_head list;
};

/* SoC Codec interface */
struct snd_soc_codec {
	char *name;
	char *longname;
	struct module *owner;
	
	/* callbacks */
	int (*probe)(struct snd_soc_codec *codec);
	void (*remove)(struct snd_soc_codec *codec);
	int (*suspend)(struct snd_soc_codec *codec);
	int (*resume)(struct snd_soc_codec *codec);
	int (*dpm_event)(struct snd_soc_codec *codec, int event);
	
	/* runtime */
	snd_card_t *card;
	struct platform_device pdev;
	void *private_data;
	void *machine_data;
	int active_streams;
	int pcm_devs;
	int probed;
	int dpm_state;
	
	/* codec IO */
	void *control_data; /* codec control (i2c/3wire) data */
	unsigned int (*read)(struct snd_soc_codec *, unsigned int);
	int (*write)(struct snd_soc_codec *, unsigned int, unsigned int);
	void *reg_cache;
	
	struct list_head list;
	struct list_head dpm_widgets;
	struct list_head dpm_paths;
	
	/* codec pcms */
	struct snd_soc_pcm_codec *pcms;
	int	npcms;
};
	
/* SoC platform interface */
struct snd_soc_platform {
	char *name;
	
	int	(*probe)(struct device * dev);
	int	(*remove)(struct device * dev);
	struct bus_type	* bus;
	int (*pcm_new)(snd_card_t *, struct snd_soc_pcm_codec *, snd_pcm_t *);
	void (*pcm_free)(snd_pcm_t *);
	snd_pcm_ops_t *pcm_ops;
};

/* SoC machine configuration */
struct snd_soc_machine_config {
	char *name;			/* Codec name */
	char *sname;		/* stream name */
	int interface;		/* interface type */
	int int_num;		/* interface number */
	u32 connect:1;		/* interface and codec have been connected */
	u32 card_reg:1;		/* codec card registered */
	u32 flags;			/* preference flags */
	
	/* codec/machine specific init - e.g. add machine controls */
	int (*init)(struct snd_soc_codec *codec); 
};

/* SoC machine interface */
struct snd_soc_machine {
	char *name;
	
	int	(*probe)(struct snd_soc_machine *machine);
	int	(*remove)(struct snd_soc_machine *machine);
	int (*suspend)(struct snd_soc_machine *machine);
	int (*resume)(struct snd_soc_machine *machine);
	
	struct snd_soc_ops *ops;
	struct snd_soc_machine_config *config;
	int nconfigs;
};

/* enamerated kcontrol */
struct soc_enum {
	unsigned short reg;
	unsigned char shift_l;
	unsigned char shift_r;
	unsigned int mask;
	const char **texts;
	void *dpm;
};

#endif
