/*
 * wm9713-voice.c  --  PXA WM9713 PCM audio driver 
 *
 * Copyright 2005 Wolfson Microelectronics PLC.
 * Author: Liam Girdwood
 *         liam.girdwood@wolfsonmicro.com or linux@wolfsonmicro.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Revision history
 *    1st Aug 2005   Initial version.
 * 
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/wm97xx.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/initval.h>

#define NAME		"wm9713-voice"
#define VERSION		"0.1"

/*
 * Debug
 */ 
#if 0
#define dbg(format, arg...) printk(KERN_DEBUG TS_NAME ": " format "\n" , ## arg)
#else	/* 
 */
#define dbg(format, arg...)
#endif	/* 
 */
#define err(format, arg...) printk(KERN_ERR TS_NAME ": " format "\n" , ## arg)
#define info(format, arg...) printk(KERN_INFO TS_NAME ": " format "\n" , ## arg)
#define warn(format, arg...) printk(KERN_WARNING TS_NAME ": " format "\n" , ## arg)

#define WM9713_PCM_8K		(0xb << 8)
#define WM9713_PCM_12K		(0x7 << 8)
#define WM9713_PCM_16K		(0x5 << 8)
#define WM9713_PCM_24K		(0x3 << 8)
#define WM9713_PCM_32K		(0x2 << 8)
#define WM9713_PCM_48K		(0x1 << 8)
#define WM9713_PCM_96K		(0x0 << 8)

#define WM9713_VOICE_HWFMT \
	(SND_SOC_HWFMT_I2S | SND_SOC_HWFMT_RIGHT_J | SND_SOC_HWFMT_LEFT_J | \
	SND_SOC_HWFMT_DSP_A | SND_SOC_HWFMT_DSP_B | SND_SOC_HWFMT_CBM_CFM | \
	SND_SOC_HWFMT_CBM_CFS | SND_SOC_HWFMT_CBS_CFS | SND_SOC_HWFMT_NB_NF | \
	SND_SOC_HWFMT_NB_IF | SND_SOC_HWFMT_IB_NF | SND_SOC_HWFMT_IB_IF)

/*
 * priv1 is pcm clock divider
 */
static struct snd_soc_hw_mode wm9713_voice[] = {
	{WM9713_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		48000,	256,	WM9713_PCM_48K, },
	{WM9713_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		32000,	384,	WM9713_PCM_32K,},
	{WM9713_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		24000,	256,	WM9713_PCM_24K,},
	{WM9713_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		16000,	256,	WM9713_PCM_16K,},
	{WM9713_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		8000,	256,	WM9713_PCM_8K,},
};
	
struct wm9713_pcm {
	int pcm_rate;
	struct wm97xx *wm;
	snd_pcm_t *pcm_voice;
};

static struct wm9713_pcm wm9713;

static int wm9713_voice_startup(snd_pcm_substream_t *substream)
{
	int reg;

	/* power up clock control and voice DAC */
	reg = wm97xx_reg_read(wm9713.wm, AC97_POWERDOWN); 
	wm97xx_reg_write(wm9713.wm, AC97_POWERDOWN, reg & ~0x2000);
	reg = wm97xx_reg_read(wm9713.wm, AC97_EXTENDED_MID); 
	wm97xx_reg_write(wm9713.wm, AC97_EXTENDED_MID, reg & ~0x1000);

	return 0;
}

static void wm9713_voice_shutdown(snd_pcm_substream_t *substream)
{
}

static int wm9713_voice_prepare(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	u16 reg = 0x8000;

	switch(pcm_c->hw_runtime.hformat & SND_SOC_CLOCK_MASK){
		case SND_SOC_HWFMT_CBM_CFM:
			reg |= 0x4000;
			break;
		case SND_SOC_HWFMT_CBM_CFS:
			reg |= 0x6000;
			break;
		case SND_SOC_HWFMT_CBS_CFS:
			reg |= 0x0200;
			break;
	}
	/* clock inversion */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_INV_MASK) {
		case SND_SOC_HWFMT_IB_IF:
			reg |= 0x00c0;
			break;
		case SND_SOC_HWFMT_IB_NF:
			reg |= 0x0080;
			break;
		case SND_SOC_HWFMT_NB_IF:
			reg |= 0x0040;
			break;
	}
	
	switch(pcm_c->hw_runtime.hformat & SND_SOC_FORMAT_MASK) {
		case SND_SOC_HWFMT_I2S:
			reg |= 0x0002;
			break;
		case SND_SOC_HWFMT_RIGHT_J:
			break;
		case SND_SOC_HWFMT_LEFT_J:
			reg |= 0x0001;
			break;
		case SND_SOC_HWFMT_DSP_A:
			reg |= 0x0003;
			break;
		case SND_SOC_HWFMT_DSP_B:
			reg |= 0x0043;
			break;
	}
	switch(pcm_c->hw_runtime.hbits) {
		case SND_SOC_HWBITS(16):
			break;
		case SND_SOC_HWBITS(20):
			reg |= 0x0004;
			break;
		case SND_SOC_HWBITS(24):
			reg |= 0x0008;
			break;
		case SND_SOC_HWBITS(32):
			reg |= 0x000c;
			break;
	}
	/* enable PCM interface in master mode */
	wm97xx_reg_write(wm9713.wm, AC97_CENTER_LFE_MASTER, reg);
	
	/* sample rate */
	reg = wm97xx_reg_read(wm9713.wm, AC97_HANDSET_RATE) &0xe0ff;
	wm97xx_reg_write(wm9713.wm, AC97_HANDSET_RATE, reg | pcm_c->hw_runtime.priv1);
	return 0;
}

static struct snd_soc_pcm_codec wm9713_voice_client = {
	.name = "WM9713 Voice",
	.playback = {
		.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 | SNDRV_PCM_RATE_48000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,
		.channels_min = 1,
		.channels_max = 1,},
	.capture = {
		.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 | 
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |
		SNDRV_PCM_RATE_96000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,
		.channels_min = 1,
		.channels_max = 1,},
	.ops = {
		.startup = wm9713_voice_startup,
		.shutdown = wm9713_voice_shutdown,
		.prepare = wm9713_voice_prepare,},
	.hw = {
		.num_hmodes = ARRAY_SIZE(wm9713_voice),
		.hmodes = &wm9713_voice[0],
		},
};

static void wm9713_device_release(struct device * dev)
{
}

static struct snd_soc_codec wm9713_soc_codec;

static int wm9713_soc_probe(struct snd_soc_codec *codec)
{
	int ret;

	//ret = snd_soc_pcm_new(codec, 0, &wm9713_voice_client, 
	//		&wm9713.pcm_voice, 1, 1, SND_SOC_SSP, 1);
	return ret;
}

static void wm9713_soc_remove(struct snd_soc_codec *codec)
{
}

static struct snd_soc_codec wm9713_soc_codec = {
	.name = "WM9713 Voice",
	.longname = "WM9713 Portable Voice Codec",
	.owner = THIS_MODULE,
	.pdev = {
		.name		= "alsa-wm9713",
		.id		= -1,
		.dev		= {.release = wm9713_device_release},
	},
	.probe = wm9713_soc_probe,
	.remove = wm9713_soc_remove,
};

static int wm97xx_wm9713_probe(struct wm97xx* wm)
{	
	if (wm->ac97 == NULL)
		return -ENODEV;
	
	wm9713.wm = wm;
	return snd_soc_register_codec(&wm9713_soc_codec);
}

static void wm97xx_wm9713_remove(struct wm97xx* wm)
{
	snd_soc_unregister_codec(&wm9713_soc_codec);
	wm9713.wm = NULL;
}

static struct wm97xx_platform_drv pxa_wm9713 = {
	.id = 0x4c13,
	.mask = 0xffff,
	.probe = 	wm97xx_wm9713_probe,
	.remove = 	wm97xx_wm9713_remove,
};

static int __init pxa_wm9713_init(void)
{
	wm97xx_register_platform_drv(&pxa_wm9713);
	return 0;
}

static void __exit pxa_wm9713_exit(void)
{
	wm97xx_unregister_platform_drv(&pxa_wm9713);
}

/* Module information */
MODULE_AUTHOR("Liam Girdwood <liam.girdwood@wolfsonmicro.com>");
MODULE_DESCRIPTION("pxa2xx WM9713 PCM driver");
MODULE_LICENSE("GPL");

module_init(pxa_wm9713_init);
module_exit(pxa_wm9713_exit);
