/*
 * pxa2xx-i2s.c  --  ALSA Soc Audio Layer
 *
 * Copyright 2005 Wolfson Microelectronics PLC.
 * Author: Liam Girdwood
 *         liam.girdwood@wolfsonmicro.com or linux@wolfsonmicro.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  Revision history
 *    12th Aug 2005   Initial version.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/delay.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>
#include <sound/soc.h>

#include <asm/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/audio.h>

#include "pxa2xx-pcm.h"

#define USE_OLD_PM_CODE     // Use original PM code -njv

#define PXA_I2S_HWFMT \
	(SND_SOC_HWFMT_I2S | SND_SOC_HWFMT_LEFT_J  | SND_SOC_HWFMT_CBS_CFS | \
	SND_SOC_HWFMT_CBM_CFS | SND_SOC_HWFMT_NB_NF)
	
#define PXA_I2S_HWFMT_96K \
	(SND_SOC_HWFMT_I2S | SND_SOC_HWFMT_LEFT_J  | \
	SND_SOC_HWFMT_CBM_CFS | SND_SOC_HWFMT_NB_NF)

static DECLARE_MUTEX(pxa_i2s_sem);
static int i2s_open = 0;

/*
 * priv1 is sadiv
 */
static struct snd_soc_hw_mode pxa2xx_i2s[] = {
	{PXA_I2S_HWFMT, 	SND_SOC_HWBITS(16),		8000,	64,		0x00000048},
	{PXA_I2S_HWFMT, 	SND_SOC_HWBITS(16),		11025,	64,		0x00000034},
	{PXA_I2S_HWFMT, 	SND_SOC_HWBITS(16),		16000,	64,		0x00000024},
	{PXA_I2S_HWFMT, 	SND_SOC_HWBITS(16),		22050,	64,		0x0000001a},
	{PXA_I2S_HWFMT, 	SND_SOC_HWBITS(16),		44100,	64,		0x0000000d},
	{PXA_I2S_HWFMT, 	SND_SOC_HWBITS(16),		48000,	64,		0x0000000c},
	/* 96k is not in yellow book - ymmv */
	{PXA_I2S_HWFMT_96K, 	SND_SOC_HWBITS(16),	96000,	64,		0x00000006},
};

static pxa2xx_pcm_dma_params_t pxa2xx_i2s_pcm_stereo_out = {
	.name			= "I2S PCM Stereo out",
	.dev_addr		= __PREG(SADR),
	.drcmr			= &DRCMRTXSADR,
	.dcmd			= DCMD_INCSRCADDR | DCMD_FLOWTRG |
				  DCMD_BURST32 | DCMD_WIDTH4,
};

static pxa2xx_pcm_dma_params_t pxa2xx_i2s_pcm_stereo_in = {
	.name			= "I2S PCM Stereo in",
	.dev_addr		= __PREG(SADR),
	.drcmr			= &DRCMRRXSADR,
	.dcmd			= DCMD_INCTRGADDR | DCMD_FLOWSRC |
				  DCMD_BURST32 | DCMD_WIDTH4,
};

static pxa2xx_pcm_dma_params_t pxa2xx_i2s_pcm_mono_out = {
	.name			= "I2S PCM Mono out",
	.dev_addr		= __PREG(SADR),
	.drcmr			= &DRCMRTXSADR,
	.dcmd			= DCMD_INCSRCADDR | DCMD_FLOWTRG |
				  DCMD_BURST16 | DCMD_WIDTH2,
};

static pxa2xx_pcm_dma_params_t pxa2xx_i2s_pcm_mono_in = {
	.name			= "I2S PCM Mono in",
	.dev_addr		= __PREG(SADR),
	.drcmr			= &DRCMRRXSADR,
	.dcmd			= DCMD_INCTRGADDR | DCMD_FLOWSRC |
				  DCMD_BURST16 | DCMD_WIDTH2,
};

static pxa2xx_gpio_t gpio_bus[] = {
	{ /* I2S SoC Slave */
		.rx = GPIO29_SDATA_IN_I2S_MD,
		.tx = GPIO30_SDATA_OUT_I2S_MD,
		.clk = GPIO28_BITCLK_IN_I2S_MD,
		.frm = GPIO31_SYNC_I2S_MD,
	},
	{ /* I2S SoC Master */
		.rx = GPIO29_SDATA_IN_I2S_MD,
		.tx = GPIO30_SDATA_OUT_I2S_MD,
		.clk = GPIO28_BITCLK_OUT_I2S_MD,
		.frm = GPIO31_SYNC_I2S_MD,
	},
};

static u32 get_sadiv(int rate)
{
	int i;
	for (i = 0; i < ARRAY_SIZE(pxa2xx_i2s); i++) {
		if (pxa2xx_i2s[i].rate == rate)
			break;
	}
	return pxa2xx_i2s[i].priv1;
};

static int pxa2xx_i2s_startup(snd_pcm_substream_t *substream)
{	
	down(&pxa_i2s_sem);
	if(!i2s_open) {
		SACR0 |= SACR0_RST;
		SACR0 = 0;
	}
	i2s_open++;
	up(&pxa_i2s_sem);
	return 0;
}

/* wait for I2S controller to be ready */
static int pxa_i2s_wait(void)
{
	int i;
	/* flush the Rx FIFO */
	for(i = 0; i < 16; i++)
		SADR;
	return 0;
}

static int pxa2xx_i2s_hw_params(snd_pcm_substream_t *substream,
				snd_pcm_hw_params_t *params)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	int master = 0;
	
	if(pcm_i->hw_runtime.hformat & SND_SOC_HWFMT_CBS_CFS)
		master = 1;
		
	pxa_gpio_mode(gpio_bus[master].rx);
	pxa_gpio_mode(gpio_bus[master].tx);
	pxa_gpio_mode(gpio_bus[master].frm);
	pxa_gpio_mode(gpio_bus[master].clk);
	pxa_set_cken(CKEN8_I2S, 1);
	pxa_i2s_wait();
	
	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
		if (params_channels(params) == 2)
			pcm_i->platform_data = &pxa2xx_i2s_pcm_stereo_out;
		else
			pcm_i->platform_data = &pxa2xx_i2s_pcm_mono_out;
	} else {
		if (params_channels(params) == 2)
			pcm_i->platform_data = &pxa2xx_i2s_pcm_stereo_in;
		else
			pcm_i->platform_data = &pxa2xx_i2s_pcm_mono_in;
	}
	
	/* is port used by another stream */
	if (SACR0 & SACR0_ENB) {
		if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
			SACR1 &= ~SACR1_DRPL;
			SAIMR |= SAIMR_TFS;
		} else {
			SACR1 &= ~SACR1_DREC;
			SAIMR |= SAIMR_RFS;
		}
	} else {
		
		SACR0 = 0;
		SACR1 = 0;
		SAIMR = SACR1_DRPL | SACR1_DREC;
		if (master)
			SACR0 |= SACR0_BCKD;
		
		SACR0 |= SACR0_RFTH(14) | SACR0_TFTH(1);
	
		if (pcm_i->hw_runtime.hformat & SND_SOC_HWFMT_LEFT_J)
			SACR1 |= SACR1_AMSL;
	
		if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
			SACR1 &= ~SACR1_DRPL;
			SAIMR |= SAIMR_TFS;
		} else {
			SACR1 &= ~SACR1_DREC;
			SAIMR |= SAIMR_RFS;
		}
	}
	
	SADIV = get_sadiv(pcm_i->hw_runtime.rate);
	SACR0 |= SACR0_ENB;
	
	return 0;
}

static void pxa2xx_i2s_shutdown(snd_pcm_substream_t *substream)
{
	down(&pxa_i2s_sem);
	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
		SACR1 |= SACR1_DRPL;
		SAIMR &= ~SAIMR_TFS;
	} else {
		SACR1 |= SACR1_DREC;
		SAIMR &= ~SAIMR_RFS;
	}

	if (SACR1 & (SACR1_DREC | SACR1_DRPL)) {
		pxa_i2s_wait();
		SACR0 &= ~SACR0_ENB;
		pxa_set_cken(CKEN8_I2S, 0);
	}
	i2s_open--;
	up(&pxa_i2s_sem);
}

#ifdef CONFIG_PM
#ifndef USE_OLD_PM_CODE
static int pxa2xx_i2s_suspend(struct platform_device *dev, struct snd_soc_pcm_interface *pcm_i)
{
    int ret = 0;
    // if(!pcm_i->active)   -njv
    //    return 0;

    ret = pxa2xx_ac97_do_suspend(card, SNDRV_CTL_POWER_D3cold);

    /* disable link and store registers */
    SACR0 &= ~SACR0_ENB;
    pxa_i2s_wait();
    pxa_i2s.sacr0 = SACR0;
    pxa_i2s.sacr1 = SACR1;
    pxa_i2s.saimr = SAIMR;
    pxa_i2s.sadiv = SADIV;
    pxa_set_cken(CKEN8_I2S, 0);
    return 0;
}

static int pxa2xx_i2s_resume(struct platform_device *pdev, struct snd_soc_pcm_interface *pcm_i)
{
    int ret = 0;
    //if(!pcm_i->active)  -njv
    //    return 0;

    pxa_set_cken(CKEN8_I2S, 1);
    pxa_i2s_wait();
   
    SACR0 = pxa_i2s.sacr0;
    SACR1 = pxa_i2s.sacr1;
    SAIMR = pxa_i2s.saimr;
    SADIV = pxa_i2s.sadiv;
    SACR0 |= SACR0_ENB;

    ret = pxa2xx_ac97_do_resume(card, SNDRV_CTL_POWER_D0);
    return 0;
}
#else
static int pxa2xx_i2s_suspend(struct snd_soc_pcm_interface *interface, pm_message_t state)
{
	int ret = 0;

	//if (interface->card && level == SUSPEND_DISABLE)
        //ret = pxa2xx_ac97_do_suspend(card, SNDRV_CTL_POWER_D3cold);

	return ret;
}

static int pxa2xx_i2s_resume(struct snd_soc_pcm_interface *interface)
{
	int ret = 0;

	//if (interface->card && level == RESUME_ENABLE)
		// ret = pxa2xx_ac97_do_resume(card, SNDRV_CTL_POWER_D0);

	return ret;
}
#endif // USE_OLD_PM_CODE
#else
#define pxa2xx_i2s_suspend	NULL
#define pxa2xx_i2s_resume	NULL
#endif

static struct snd_soc_pcm_interface pxa_i2s_interface[] = {
	{	.name = "pxa2xx-i2s",
		.id = 0,
		.type = SND_SOC_I2S,
		.suspend = pxa2xx_i2s_suspend,
		.resume = pxa2xx_i2s_resume,
		.playback = {
			.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_11025 | 
			SNDRV_PCM_RATE_16000 | SNDRV_PCM_RATE_22050 |
			SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 | SNDRV_PCM_RATE_96000,
			.formats = SNDRV_PCM_FMTBIT_S16_LE,
			.channels_min = 2,
			.channels_max = 2,},
		.capture = {
			.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_11025 | 
			SNDRV_PCM_RATE_16000 | SNDRV_PCM_RATE_22050 |
			SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 | SNDRV_PCM_RATE_96000,
			.formats = SNDRV_PCM_FMTBIT_S16_LE,
			.channels_min = 2,
			.channels_max = 2,},
		.ops = {
			.startup = pxa2xx_i2s_startup,
			.shutdown = pxa2xx_i2s_shutdown,
			.hw_params = pxa2xx_i2s_hw_params,},
		.hw = {
			.num_hmodes = ARRAY_SIZE(pxa2xx_i2s),
			.hmodes = &pxa2xx_i2s[0],},
	},
};

static int __init pxa2xx_soc_i2s_init(void)
{
	snd_soc_register_interface(pxa_i2s_interface);
	return 0;
}

static void __exit pxa2xx_soc_i2s_exit(void)
{
	snd_soc_unregister_interface(pxa_i2s_interface);
}

EXPORT_SYMBOL_GPL(pxa_i2s_interface);

module_init(pxa2xx_soc_i2s_init);
module_exit(pxa2xx_soc_i2s_exit);

/* Module information */ 
MODULE_AUTHOR("Liam Girdwood, liam.girdwood@wolfsonmicro.com, www.wolfsonmicro.com");
MODULE_DESCRIPTION("pxa2xx I2S SoC Interface");
MODULE_LICENSE("GPL");
