/*
 * soc-dpm.c  --  ALSA SoC Dynamic Power Management
 *
 * Copyright 2005 Wolfson Microelectronics PLC.
 * Author: Liam Girdwood
 *         liam.girdwood@wolfsonmicro.com or linux@wolfsonmicro.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  Revision history
 *    12th Aug 2005   Initial version.
 *    25th Oct 2005   Implemented path power domain.
 *
 *  Features:
 *    o Changes power status of internal codec blocks depending on the 
 *      dynamic configuration of codec internal audio paths and active
 *      DAC's/ADC's.
 * 
 *  Todo:
 *    o Platform power domain - could this be tied into machine layer ???
 *    o DPM power change sequencing - allow for configurable sequences
 *    o Support for analogue bias optimisation.
 *    o Support for reduced codec oversampling rates.
 *    o Support for reduced codec bias currents.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/pm.h>
#include <linux/bitops.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc-dpm.h>
#include <sound/initval.h>


static int dpm_status = 1;
module_param(dpm_status, int, 0);
MODULE_PARM_DESC(dpm_status, "enable DPM sysfs entries");

static struct snd_soc_dpm_widget *dpm_cnew_widget(const struct snd_soc_dpm_widget *_widget)
{
	struct snd_soc_dpm_widget* widget;
	widget = kmalloc(sizeof(struct snd_soc_dpm_widget), GFP_KERNEL);
	if (!widget)
		return NULL;
	
	memcpy(widget, _widget, sizeof(struct snd_soc_dpm_widget));
	return widget;
}

/* set up initial codec paths */  
static void dpm_set_path_status(struct snd_soc_dpm_widget *w, 
	struct snd_soc_dpm_path *p, int i)
{	
	switch (w->id){
	case snd_soc_dpm_mixer: {
		int val;
		int reg = w->kcontrols[i].private_value & 0xff;
		int shift = (w->kcontrols[i].private_value >> 8) & 0x0f;
		int mask = (w->kcontrols[i].private_value >> 16) & 0xff;
		int invert = (w->kcontrols[i].private_value >> 24) & 0x01;
		
		val = snd_soc_read(w->codec, reg);
		val = (reg >> shift) & mask;
	
		if((invert && !val) || (!invert && val))
			p->connect = 1;
		else
			p->connect = 0;
	}
	break;
	case snd_soc_dpm_mux: {
		struct soc_enum *e = (struct soc_enum *)w->kcontrols[i].private_value;
		int val, item, bitmask;
		
		for (bitmask = 1; bitmask < e->mask; bitmask <<= 1)
		;
		val = snd_soc_read(w->codec, e->reg);
		item = (val >> e->shift_l) & (bitmask - 1);
		
		p->connect = 0;
		for (i = 0; i < e->mask; i++) {
			if (!(strcmp(p->name, e->texts[i])) && item == i)
				p->connect = 1;
		}
	}
	break;
	/* does not effect routing - always connected */
	case snd_soc_dpm_volume:
	case snd_soc_dpm_output:
	case snd_soc_dpm_adc:
	case snd_soc_dpm_input:
	case snd_soc_dpm_dac:
	case snd_soc_dpm_clock:
	case snd_soc_dpm_vref:	
	case snd_soc_dpm_amp:
	case snd_soc_dpm_micbias:
		p->connect = 1;
	case snd_soc_dpm_mute:
	/* does effect routing - dynamically connected */
	case snd_soc_dpm_hp:
	case snd_soc_dpm_mic:			
		p->connect = 0;
	break;
	}
}	

static int dpm_connect_mux(struct snd_soc_codec *codec, struct snd_soc_dpm_widget *src, 
	struct snd_soc_dpm_widget *dest, struct snd_soc_dpm_path *path, 
	const char *control_name, const snd_kcontrol_new_t *kcontrol)
{
	struct soc_enum *e = (struct soc_enum *)kcontrol->private_value;
	int i;
	
	for (i = 0; i < e->mask; i++) {
		if (!(strcmp(control_name, e->texts[i]))) {
			list_add(&path->list, &codec->dpm_paths);
			list_add(&path->list_out, &dest->inputs);
			list_add(&path->list_in, &src->outputs);
			path->name = (char*)e->texts[i];
			dpm_set_path_status(dest, path, 0);
			return 0;
		}
	}

	return -ENODEV;
}

static int dpm_connect_mixer(struct snd_soc_codec *codec, struct snd_soc_dpm_widget *src, 
	struct snd_soc_dpm_widget *dest, struct snd_soc_dpm_path *path, const char *control_name)
{
	int i;
	
	/* search for mixer kcontrol */
	for (i = 0; i < dest->num_kcontrols; i++) {		
		if(!strcmp(control_name, dest->kcontrols[i].name)) {
			list_add(&path->list, &codec->dpm_paths);
			list_add(&path->list_out, &dest->inputs);
			list_add(&path->list_in, &src->outputs);
			path->name = dest->kcontrols[i].name;
			dpm_set_path_status(dest, path, i);
			return 0;
		}			
	}
	return -ENODEV;
}

int snd_soc_dpm_connect_input(struct snd_soc_codec *codec, const char *dest_name, 
	const char * control_name, const char *src_name)
{
	struct snd_soc_dpm_path *path;
	struct snd_soc_dpm_widget *src = NULL, *dest = NULL, *w;
	struct list_head *l;
	int ret = 0;

	/* find src and dest widgets */
	list_for_each(l, &codec->dpm_widgets)
	{
		w = list_entry(l, struct snd_soc_dpm_widget, list);
		
		if(!(strcmp(w->name, dest_name))) {
			dest = w;
			continue;
		}
		if(!(strcmp(w->name, src_name))) {
			src = w;
		}				
	}

	if(src == NULL || dest == NULL)
		return -ENODEV;
	
	path = kzalloc(sizeof(struct snd_soc_dpm_path), GFP_KERNEL);
	if (!path)
		return -ENOMEM;	

	path->in = src;
	path->out = dest;
	INIT_LIST_HEAD(&path->list);
	INIT_LIST_HEAD(&path->list_in);
	INIT_LIST_HEAD(&path->list_out);

	/* check for external widgets */
	if(dest->id == snd_soc_dpm_input) {	
		if (src->id == snd_soc_dpm_micbias || src->id == snd_soc_dpm_mic)
			dest->ext = 1;
	}
	if(src->id == snd_soc_dpm_output) {	
		if (dest->id == snd_soc_dpm_amp || dest->id == snd_soc_dpm_hp)
			src->ext = 1;
	}
	
	if(control_name == NULL) {
		list_add(&path->list, &codec->dpm_paths);
		list_add(&path->list_out, &dest->inputs);
		list_add(&path->list_in, &src->outputs);
		path->connect = 1;
		return 0;
	}

	switch(dest->id) {
	case snd_soc_dpm_adc:
	case snd_soc_dpm_dac:
	case snd_soc_dpm_volume:
	case snd_soc_dpm_input:
	case snd_soc_dpm_output:
	case snd_soc_dpm_clock:
	case snd_soc_dpm_vref:
	case snd_soc_dpm_micbias:
	case snd_soc_dpm_mute:
	case snd_soc_dpm_amp:
		list_add(&path->list, &codec->dpm_paths);
		list_add(&path->list_out, &dest->inputs);
		list_add(&path->list_in, &src->outputs);
		path->connect = 1;
		return 0;
	case snd_soc_dpm_mux:
		if((ret = dpm_connect_mux(codec, src, dest, path, control_name, &dest->kcontrols[0])) != 0)
			goto err;
		break;
	case snd_soc_dpm_mixer:
		if((ret = dpm_connect_mixer(codec, src, dest, path, control_name)) != 0)
			goto err;
		break;
	case snd_soc_dpm_hp:
	case snd_soc_dpm_mic:
		list_add(&path->list, &codec->dpm_paths);
		list_add(&path->list_out, &dest->inputs);
		list_add(&path->list_in, &src->outputs);
		path->connect = 0;
		return 0;
	}
	return 0;

err:
	printk(KERN_WARNING "soc: no dpm match for %s --> %s --> %s\n", src_name, 
		control_name, dest_name); 
	kfree(path);
	return ret;
}

static int dpm_update_bits(struct snd_soc_dpm_widget *widget, int power)
{
	int change;
	unsigned short old, new;
	struct snd_soc_codec *codec = widget->codec;
	
	if(widget->reg < 0 || widget->id == snd_soc_dpm_input ||
		widget->id == snd_soc_dpm_output)
		return 0;
	
	if(power)
		power = 1;
	
	old = snd_soc_read(codec, widget->reg);
	new = (old & ~(0x1 << widget->shift)) | (power << widget->shift);
	
	change = old != new;
	if (change) {
		snd_soc_write(codec, widget->reg, new);
		if (!widget->invert)
			widget->power = power;
		else
			widget->power = power ? 0:1;
	}

	return change;
}

/* ramps the volume up or down to minimise pops before or after a power event */
static int dpm_set_volume(struct snd_soc_dpm_widget *widget, int power)
{
	const snd_kcontrol_new_t *k = widget->kcontrols;
	
	if(widget->muted && !power)
		return 0;
	if(!widget->muted && power)
		return 0;
	
	if(widget->num_kcontrols && k) {
		int reg = k->private_value & 0xff;
		int shift = (k->private_value >> 8) & 0x0f;
		int mask = (k->private_value >> 16) & 0xff;
		int invert = (k->private_value >> 24) & 0x01;
		
		if (power) {
			int i;
			/* power up has happended, increase volume to last level */
			if(invert) {
				for (i = mask; i >= widget->value; i--)
					snd_soc_update_bits(widget->codec, reg, mask, i);
			} else {
				for (i = 0; i <= widget->value; i++)
					snd_soc_update_bits(widget->codec, reg, mask, i);
			}
			widget->muted = 0;
		} else {
			/* power down is about to occur, decrease volume to mute */
			int val = snd_soc_read(widget->codec, reg);
			int i = widget->value = (val >> shift) & mask;
			if(invert) {
				for (; i < mask; i++)
					snd_soc_update_bits(widget->codec, reg, mask, i);
			} else {
				for (; i > 0; i--)
					snd_soc_update_bits(widget->codec, reg, mask, i);
			}
			widget->muted = 1;
		}
	}
	return 0;
}

static int dpm_sync_mixer(struct snd_soc_codec *codec, struct snd_soc_dpm_widget *w)
{
	int i, ret = 0;
	char name[32];
	struct list_head *lp;
	struct snd_soc_dpm_path *path;
	
	/* add kcontrol */
	for (i = 0; i < w->num_kcontrols; i++) {
			
		/* match name */
		list_for_each(lp, &w->inputs)
		{
			path = list_entry(lp, struct snd_soc_dpm_path, list_out);
			
			/* mixer/mux paths name must match control name */
			if (path->name != (char*)w->kcontrols[i].name) 
				continue;
				
			/* add dpm control with long name */
			snprintf(name, 32, "%s %s", w->name, w->kcontrols[i].name);
			path->long_name = kstrdup (name, GFP_KERNEL);
			if(path->long_name == NULL)
				return -ENOMEM;
			
			path->kcontrol = snd_soc_cnew(&w->kcontrols[i], w, path->long_name);
			if ((ret = snd_ctl_add(codec->card, path->kcontrol)) < 0) {
				printk(KERN_ERR "soc: failed to add dpm kcontrol %s\n", path->long_name);
				kfree(path->long_name);
				path->long_name = NULL;
				return ret;
			}
		}
	}
	return ret;
}

static int dpm_sync_mux(struct snd_soc_codec *codec, struct snd_soc_dpm_widget *w)
{
	struct list_head *lp;
	struct snd_soc_dpm_path *path = NULL;
	snd_kcontrol_t *kcontrol;
	int ret = 0;
	
	if(!w->num_kcontrols) {
		printk(KERN_ERR "soc: mux %s has no controls\n", w->name);
		return -EINVAL;
	}
	
	kcontrol = snd_soc_cnew(&w->kcontrols[0], w, w->name);
	if ((ret = snd_ctl_add(codec->card, kcontrol)) < 0) {
		printk(KERN_ERR "soc: failed to add kcontrol %s\n", w->name);
		return ret;
	}

	list_for_each(lp, &w->inputs)
	{
		path = list_entry(lp, struct snd_soc_dpm_path, list_out);
		path->kcontrol = kcontrol;
	}
	
	return ret;
}

static int dpm_sync_volume(struct snd_soc_codec *codec, struct snd_soc_dpm_widget *w)
{
	snd_kcontrol_t *kcontrol;
	int ret = 0;
	
	if(!w->num_kcontrols)
		return -EINVAL;
	
	kcontrol = snd_soc_cnew(&w->kcontrols[0], w, w->name);
	if ((ret = snd_ctl_add(codec->card, kcontrol)) < 0) {
		printk(KERN_ERR "soc: failed to add kcontrol %s\n", w->name);
		return ret;
	}
	
	return ret;
}

/*
 * Reset 'walked' bit for each path
 */
static void dpm_clear_walk(struct snd_soc_codec *codec)
{
	struct snd_soc_dpm_path *p = NULL;
	struct list_head *l = NULL;
		
	list_for_each(l, &codec->dpm_paths)
	{
		p = list_entry(l, struct snd_soc_dpm_path, list);
		p->walked = 0;		
	}
}

/*
 * Check for a completed path to an active or physically connected output widget
 */
static int is_connected_output_ep(struct snd_soc_dpm_widget *widget)
{
	struct list_head *lp;
	struct snd_soc_dpm_path *path = NULL;
	int con = 0;

	if(widget->id == snd_soc_dpm_adc && widget->active)
		return 1;
	if(widget->id == snd_soc_dpm_output && widget->connected && !widget->ext)
		return 1;
	if(widget->id == snd_soc_dpm_hp && widget->connected)
		return 1;
	if(widget->id == snd_soc_dpm_amp && widget->connected)
		return 1;
	
	list_for_each(lp, &widget->outputs)
	{
		path = list_entry(lp, struct snd_soc_dpm_path, list_in);
		if(path->walked)
			continue;
	
		if (path->out && path->connect) {
			path->walked = 1;
			con += is_connected_output_ep(path->out);
		}
	}
	
	return con;
}

/*
 * Check for a completed path to an active or physically connected input widget
 */
static int is_connected_input_ep(struct snd_soc_dpm_widget *widget)
{
	struct list_head *lp;
	struct snd_soc_dpm_path *path = NULL;
	int con = 0;
	
	if(widget->id == snd_soc_dpm_dac && widget->active)
		return 1;
	if(widget->id == snd_soc_dpm_input && widget->connected && !widget->ext)
		return 1;
	if(widget->id == snd_soc_dpm_mic && widget->connected)
		return 1;
	
	list_for_each(lp, &widget->inputs)
	{
		path = list_entry(lp, struct snd_soc_dpm_path, list_out);
		if(path->walked)
			continue;
	
		if (path->in && path->connect) {
			path->walked = 1;
			con += is_connected_input_ep(path->in);
		}
	}
	
	return con;
}

static int dpm_power_widgets(struct snd_soc_codec *codec, int event)
{
	struct snd_soc_dpm_widget *w;
	struct list_head *l;
	int in, out, pm = 0, i, c = 1, *seq = NULL;
	int up_seq[] = {snd_soc_dpm_dac, snd_soc_dpm_mixer, snd_soc_dpm_volume, snd_soc_dpm_mute};
	int down_seq[] = {snd_soc_dpm_volume, snd_soc_dpm_mixer, snd_soc_dpm_dac, snd_soc_dpm_mute};

	if(event == SND_SOC_DPM_STREAM_START) {
		c = ARRAY_SIZE(up_seq);
		seq = up_seq;
	} else if (event == SND_SOC_DPM_STREAM_STOP) {
		c = ARRAY_SIZE(down_seq);
		seq = down_seq;
	}
	
	for(i = 0; i < c; i++){
		list_for_each(l, &codec->dpm_widgets)
		{
			w = list_entry(l, struct snd_soc_dpm_widget, list);
			
			if(seq && seq[i] && w->id != seq[i])
				continue;
			
			if(w->id == snd_soc_dpm_vref || w->id == snd_soc_dpm_clock ||
				w->id == snd_soc_dpm_mute)
				continue;
			
			if(w->id == snd_soc_dpm_adc && w->active) {
				in = is_connected_input_ep(w);
				dpm_clear_walk(w->codec);
				pm = (in != 0) ? 1 : 0;
				if(w->invert)
					pm = (pm ? 0:1);
				dpm_update_bits(w, pm);
				continue;
			}
			
			if(w->id == snd_soc_dpm_dac && w->active) {
				out = is_connected_output_ep(w);
				dpm_clear_walk(w->codec);
				pm = (out != 0) ? 1 : 0;
				if(w->invert)
					pm = (pm ? 0:1);
				dpm_update_bits(w, pm);
				continue;
			}
			
			if(w->id == snd_soc_dpm_volume) {
				int on;
				in = is_connected_input_ep(w);
				dpm_clear_walk(w->codec);
				out = is_connected_output_ep(w);
				dpm_clear_walk(w->codec);
				pm = on = (out != 0 && in != 0) ? 1 : 0;
				
				if(!on)
					dpm_set_volume(w, on); /* lower volume to reduce pops */
				
				if(w->invert)
					pm = (on ? 0:1);
				dpm_update_bits(w, pm);
				
				if(on) 
					dpm_set_volume(w, on); /* restore volume from zero */
				
				continue;
			}
			in = is_connected_input_ep(w);
			dpm_clear_walk(w->codec);
			out = is_connected_output_ep(w);
			dpm_clear_walk(w->codec);
			pm = (out != 0 && in != 0) ? 1 : 0;
			if(w->invert)
					pm = (pm ? 0:1);
			dpm_update_bits(w, pm);
			if(w->ext_event)
				w->ext_event(w->codec, pm);
			/*printk("widget %s\n in %d out %d pm %s ext %d con %d\n", 
				w->name, in, out, pm ? "On":"Off", w->ext, w->connected);*/	
		}
	}
	
	return 0;
}

static int dpm_mux_update_power(struct snd_soc_dpm_widget *widget, 
	snd_kcontrol_t *kcontrol, int mask, int val, struct soc_enum* e)
{
	struct list_head *l;
	struct snd_soc_dpm_path *path;
	
	if(widget->id != snd_soc_dpm_mux)
		return -ENODEV; 

	if (!snd_soc_test_bits(widget->codec, e->reg, mask, val))
		return 0;
	
	/* find dpm widget path assoc with kcontrol */
	list_for_each(l, &widget->codec->dpm_paths)
	{
		path = list_entry(l, struct snd_soc_dpm_path, list);
		if (path->kcontrol != kcontrol)
			continue;

		if(!path->name || ! e->texts[val])
			continue;

		/* we now need to match the string in the enum to the path */
		if (!(strcmp(path->name, e->texts[val])))
			path->connect = 1; /* new connection */
		else
			path->connect = 0; /* old connection must be powered down */	
	}
	
	dpm_power_widgets(widget->codec, 0);			
	return 0;
}

static int dpm_mixer_update_power(struct snd_soc_dpm_widget *widget, 
	snd_kcontrol_t *kcontrol, int reg, int val_mask, int val, int invert)
{
	struct list_head *l;
	struct snd_soc_dpm_path *path;

	if(widget->id != snd_soc_dpm_mixer)
		return -ENODEV; 
	
	if (!snd_soc_test_bits(widget->codec, reg, val_mask, val))
		return 0;
	
	/* find dpm widget path assoc with kcontrol */
	list_for_each(l, &widget->codec->dpm_paths)
	{
		path = list_entry(l, struct snd_soc_dpm_path, list);
		if (path->kcontrol != kcontrol)
			continue;
			
		/* found, now check type */
		if (val)
			path->connect = invert ? 0:1; /* new connection */
		else
			path->connect = invert ? 1:0; /* old connection must be powered down */
		break;
	}
	
	dpm_power_widgets(widget->codec, 0);
	return 0;
}

static ssize_t dpm_status_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct snd_soc_codec *codec = (struct snd_soc_codec*)dev->driver_data;
	struct snd_soc_dpm_widget *w = NULL;
	struct list_head *l = NULL;
	int count = 0; 

	list_for_each(l, &codec->dpm_widgets)
	{
		w = list_entry(l, struct snd_soc_dpm_widget, list);
		if (w->name && w->reg > 0 && 
			w->id != snd_soc_dpm_input && w->id != snd_soc_dpm_output)
			count += sprintf(buf + count, "%s: %s\n", w->name, w->power ? "On":"Off");
	}
	
	return count;
}

static DEVICE_ATTR(dpm, 0444, dpm_status_show, NULL);
		     
int snd_soc_dpm_sys_add(struct device *dev) 
{
	int ret = 0;

	if (dpm_status)
		ret = device_create_file(dev, &dev_attr_dpm);
	return ret;
}	     

void snd_soc_dpm_sys_remove(struct device *dev) 
{
	if (dpm_status)
		device_remove_file(dev, &dev_attr_dpm);
} 

int snd_soc_dpm_sync(struct snd_soc_codec *codec)
{
	struct snd_soc_dpm_widget *w = NULL;
	struct list_head *lw;
	
	list_for_each(lw, &codec->dpm_widgets)
	{
		w = list_entry(lw, struct snd_soc_dpm_widget, list);
		
		if(w->sync)
			continue;
		
		switch(w->id) {
		case snd_soc_dpm_mixer:
			dpm_sync_mixer(codec, w);
			break;
		case snd_soc_dpm_mux:
			dpm_sync_mux(codec, w);
			break;
		case snd_soc_dpm_adc:
		case snd_soc_dpm_dac:
		case snd_soc_dpm_volume:
			dpm_sync_volume(codec, w);
			break;
		case snd_soc_dpm_input:
		case snd_soc_dpm_output:
		case snd_soc_dpm_clock:
		case snd_soc_dpm_vref:
		case snd_soc_dpm_micbias:
		case snd_soc_dpm_mute:
		case snd_soc_dpm_amp:
		case snd_soc_dpm_hp:
		case snd_soc_dpm_mic:  				
			break;
		}
		w->sync = 1;
	}
	
	dpm_power_widgets(codec, 0);
	return 0;
}

int snd_soc_dpm_get_volsw(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol)
{
	struct snd_soc_dpm_widget *widget = snd_kcontrol_chip(kcontrol);
	int reg = kcontrol->private_value & 0xff;
	int shift = (kcontrol->private_value >> 8) & 0x0f;
	int rshift = (kcontrol->private_value >> 12) & 0x0f;
	int mask = (kcontrol->private_value >> 16) & 0xff;
	int invert = (kcontrol->private_value >> 24) & 0x01;

	if(widget->id == snd_soc_dpm_volume && !widget->power) {
		ucontrol->value.integer.value[0] = widget->value;
		return 0;
	}
	
	ucontrol->value.integer.value[0] = (snd_soc_read(widget->codec, reg) >> shift) & mask;
	if (shift != rshift)
		ucontrol->value.integer.value[1] = (snd_soc_read(widget->codec, reg) >> rshift) & mask;
	if (invert) {
		ucontrol->value.integer.value[0] = mask - ucontrol->value.integer.value[0];
		if (shift != rshift)
			ucontrol->value.integer.value[1] = mask - ucontrol->value.integer.value[1];
	}

	return 0;
}

int snd_soc_dpm_put_volsw(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol)
{
	struct snd_soc_dpm_widget *widget = snd_kcontrol_chip(kcontrol);
	int reg = kcontrol->private_value & 0xff;
	int shift = (kcontrol->private_value >> 8) & 0x0f;
	int rshift = (kcontrol->private_value >> 12) & 0x0f;
	int mask = (kcontrol->private_value >> 16) & 0xff;
	int invert = (kcontrol->private_value >> 24) & 0x01;
	unsigned short val, val2, val_mask;
	
	val = (ucontrol->value.integer.value[0] & mask);
	if (invert)
		val = mask - val;
	val_mask = mask << shift;
	val = val << shift;
	if (shift != rshift) {
		val2 = (ucontrol->value.integer.value[1] & mask);
		if (invert)
			val2 = mask - val2;
		val_mask |= mask << rshift;
		val |= val2 << rshift;
	}
	
	if(widget->id == snd_soc_dpm_volume && !widget->power) {
		widget->value = val;
		return 1;
	}
	
	dpm_mixer_update_power(widget, kcontrol, reg, val_mask, val, invert);
	return snd_soc_update_bits(widget->codec, reg, val_mask, val);
}

int snd_soc_dpm_get_enum_double(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol)
{
	struct snd_soc_dpm_widget *widget = snd_kcontrol_chip(kcontrol);
	struct soc_enum *e = (struct soc_enum *)kcontrol->private_value;
	unsigned short val, bitmask;
	
	for (bitmask = 1; bitmask < e->mask; bitmask <<= 1)
		;
	val = snd_soc_read(widget->codec, e->reg);
	ucontrol->value.enumerated.item[0] = (val >> e->shift_l) & (bitmask - 1);
	if (e->shift_l != e->shift_r)
		ucontrol->value.enumerated.item[1] = (val >> e->shift_r) & (bitmask - 1);

	return 0;
}

int snd_soc_dpm_put_enum_double(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol)
{
	struct snd_soc_dpm_widget *widget = snd_kcontrol_chip(kcontrol);
	struct soc_enum *e = (struct soc_enum *)kcontrol->private_value;
	unsigned short val, mux;
	unsigned short mask, bitmask;
	
	for (bitmask = 1; bitmask < e->mask; bitmask <<= 1)
		;
	if (ucontrol->value.enumerated.item[0] > e->mask - 1)
		return -EINVAL;
	mux = ucontrol->value.enumerated.item[0];
	val = mux << e->shift_l;
	mask = (bitmask - 1) << e->shift_l;
	if (e->shift_l != e->shift_r) {
		if (ucontrol->value.enumerated.item[1] > e->mask - 1)
			return -EINVAL;
		val |= ucontrol->value.enumerated.item[1] << e->shift_r;
		mask |= (bitmask - 1) << e->shift_r;
	}

	dpm_mux_update_power(widget, kcontrol, mask, mux, e);
	return snd_soc_update_bits(widget->codec, e->reg, mask, val);
}

int snd_soc_dpm_new_control(struct snd_soc_codec *codec, const struct snd_soc_dpm_widget *widget)
{
	struct snd_soc_dpm_widget *w;

	if ((w = dpm_cnew_widget(widget)) == NULL)
		return -ENOMEM;

	w->codec = codec;
	INIT_LIST_HEAD(&w->inputs);
	INIT_LIST_HEAD(&w->outputs);
	INIT_LIST_HEAD(&w->list);
	list_add(&w->list, &codec->dpm_widgets);
	
	/* machine layer set ups unconnected pins and insertions */
	w->connected = 1;
	return 0;
}

/* stream event notification */
int snd_soc_dpm_stream_event(struct snd_soc_codec *codec, 
	const char * stream, int event)
{
	struct snd_soc_dpm_widget *w;
	struct list_head *l = NULL;	
	
	list_for_each(l, &codec->dpm_widgets)
	{
		w = list_entry(l, struct snd_soc_dpm_widget, list);

		if (!w->sname)
			continue;
		
		if (strstr(w->sname, stream)) {
			switch(event) {
			case SND_SOC_DPM_STREAM_START:
				w->active = 1;
				break;
			case SND_SOC_DPM_STREAM_STOP:
			case SND_SOC_DPM_STREAM_SUSPEND:	
				w->active = 0;
				break;
			case SND_SOC_DPM_STREAM_PAUSE_PUSH:
				break;
			case SND_SOC_DPM_STREAM_PAUSE_RELEASE:
				break;
			}
		}					
	}
	
	dpm_power_widgets(codec, event);
	return 0;
}

/* mute/unmute codec DAC mute bits - used to reduce pops */
int snd_soc_dpm_codec_mute(struct snd_soc_codec *codec, int mute)
{
	struct snd_soc_dpm_widget *w;
	struct list_head *l = NULL;	

	list_for_each(l, &codec->dpm_widgets)
	{
		w = list_entry(l, struct snd_soc_dpm_widget, list);

		if(w->id == snd_soc_dpm_mute)
			dpm_update_bits(w, mute);
	}
	
	return 0;
}

/* handle codec level dpm events - i.e. clocks */ 
int snd_soc_dpm_codec_event(struct snd_soc_codec *codec, int event)
{
	struct snd_soc_dpm_widget *w;
	struct list_head *l = NULL;	
	int pm = 0, active = 0;

	if(event == SNDRV_CTL_POWER_D1) {
		if(codec->active_streams)
			return 0;
		pm = 1;
		active = 1;
	
	} else if(event == SNDRV_CTL_POWER_D3hot){
		if(!codec->active_streams)
			return 0;
		pm = 0;
		active = -1;
	}
		
	list_for_each(l, &codec->dpm_widgets)
	{
		int tpm = pm;
		w = list_entry(l, struct snd_soc_dpm_widget, list);
		if(w->id == snd_soc_dpm_vref || w->id == snd_soc_dpm_clock) {
			codec->active_streams += active;
			
			if(w->invert)
				tpm = (pm ? 0:1);
			
			dpm_update_bits(w, tpm);
			//printk("active %d name %s state %d\n", codec->active_streams, w->name, tpm);	
		}
	}

	return 0;
}

/* update codec dpm status with connected pins/inputs */
int snd_soc_dpm_set_connection(struct snd_soc_codec *codec, 
	const struct snd_soc_dpm_pin *pin)
{
	struct snd_soc_dpm_widget *w;
	struct list_head *l = NULL;	

	list_for_each(l, &codec->dpm_widgets)
	{
		w = list_entry(l, struct snd_soc_dpm_widget, list);
			
		if(!strcmp(w->name, pin->name)) {
			w->connected = pin->connected;
		}
	}
	
	return 0;	
}

EXPORT_SYMBOL_GPL(snd_soc_dpm_put_volsw);
EXPORT_SYMBOL_GPL(snd_soc_dpm_get_volsw);
EXPORT_SYMBOL_GPL(snd_soc_dpm_connect_input);
EXPORT_SYMBOL_GPL(snd_soc_dpm_new_control);
EXPORT_SYMBOL_GPL(snd_soc_dpm_sync);
EXPORT_SYMBOL_GPL(snd_soc_dpm_put_enum_double);
EXPORT_SYMBOL_GPL(snd_soc_dpm_get_enum_double);
EXPORT_SYMBOL_GPL(snd_soc_dpm_stream_event);
EXPORT_SYMBOL_GPL(snd_soc_dpm_codec_event);
EXPORT_SYMBOL_GPL(snd_soc_dpm_codec_mute);
EXPORT_SYMBOL_GPL(snd_soc_dpm_set_connection);

/* Module information */ 
MODULE_AUTHOR("Liam Girdwood, liam.girdwood@wolfsonmicro.com, www.wolfsonmicro.com");
MODULE_DESCRIPTION("ALSA SoC DPM");
MODULE_LICENSE("GPL");
