/*
 * bootblock.S - based on memsetup.S, but included inline with reset code
 *
 * Copyright (C) 2005, 2006 Lab126, Inc.  All rights reserved.
 * Nick Vaccaro <nvaccaro@lab126.com>
 * Jerry Katzung <katzung@lab126.com>
 *
 * Based on cpu/pxa/start.S and board/fiona/memsetup.S, but included inline
 * by cpu/pxa/start.S to guarantee that everything ends up in the first 1K
 * of the overall ROM image.
 *
 * Since the OneNAND flash parts preload only the first 1K of the bootloader,
 * it is important to pack as much of the initial boot code as possible into
 * that first 1K.  So this file is included into the reset handling code.
 *
 */

/* header files are included by cpu/pxa/start.S	*/


/* Define to support NOR.
 * Undefine to save space
 */
#undef CONFIG_NOR_SUPPORT

/* Define to enable GPIO5 debug signal code
 */
#undef USE_GPIO_DEBUG_SIGNAL 

/* Define to remove code to create space (NOT FOR REAL BUILDS, ONLY DEV !!)
 */
#ifdef USE_GPIO_DEBUG_SIGNAL
        #define STEAL_SPACE
#else
        #undef STEAL_SPACE
#endif

/* Define to reference BSS globals from inside the boot block.
 */
#undef BSS_IN_BOOTBLOCK

/* Pad bootblock to 1K so that build tool will squawk if we exceed 1K.
 */
#define BOOTBLOCK_PAD	0xEEEEEEEE, 0xEEEEEEEE, 0xEEEEEEEE

/* PXA255 register offsets */

#define PSSR_OFFSET	0x04
#define PSPR_OFFSET	0x08
#define PEDR_OFFSET	0x18
#define RCSR_OFFSET	0x30
#define OSMR3_OFFSET    0x0C
#define OSCR_OFFSET     0x10
#define OWER_OFFSET     0x18

/* OneNAND */

#define ONENAND_DATARAM0_BASE	0x00000400	/* addr of DataRAM0 buf	*/

#define ONENAND_REG_BASE	0x0001E000	/* assume OneNAND boot	*/
#define ONENAND_REG_BASE_CS2	0x0801E000

#define ON_MANU_ID_OFFSET	0x0000		/* manufacturer ID	*/
#define ON_DEV_ID_OFFSET	0x0002		/* device ID		*/

#define ON_START_ADDR_1_OFFSET	0x0200		/* start addr1 - FBA	*/
#define ON_START_ADDR_2_OFFSET	    0x02	/* start addr2 - DBS	*/
#define ON_START_ADDR_8_OFFSET	    0x0E	/* start addr8 - FPA/FSA */

#define ON_START_BUFFER_OFFSET	0x0400		/* start buffer - BSA/BSC */
#define ON_COMMAND_OFFSET	    0x40	/* command		*/
#define ON_CTRLR_STATUS_OFFSET	    0x80	/* controller status	*/
#define ON_INTERRUPT_OFFSET	    0x82	/* interrupt status	*/

#define ON_MANU_ID_EXPECTED	0x00EC
#define ON_DEV_ID_EXPECTED	0x0035

#define ON_PAGE_SIZE		2048		/* OneNAND page size	*/

/* Recovery mode entry point definitions */

#ifdef CONFIG_RECOVERY_MODE

#define ENTRY_PT_OFFSET			0x20

#define ENTRY_PT_CHECK_RECOVERY		0
#define ENTRY_PT_AFTER_CHECK_RECOVERY	1

#endif

	.macro CPWAIT
	mrc	p15,0,r0,c2,c0,0
	mov	r0,r0
	sub	pc,pc,#4
	.endm

/****************************************************************************
 *
 *	The 1K boot block starts here.
 *
 ****************************************************************************/

.globl _start
_start:
        b	reset
	ldr	pc, __undefined_instruction
	ldr	pc, __software_interrupt
	ldr	pc, __prefetch_abort
	ldr	pc, __data_abort
	ldr	pc, __not_used
	ldr	pc, __irq
	ldr	pc, __fiq

	b	image_entry_point

#ifndef STEAL_SPACE
__undefined_instruction:	.word _undefined_instruction
__software_interrupt:	        .word _software_interrupt
__prefetch_abort:	        .word _prefetch_abort
__data_abort:	        	.word _data_abort
__not_used:	        	.word _not_used
__irq:		        	.word _irq
__fiq:		        	.word _fiq
#else
__undefined_instruction:
__software_interrupt:
__prefetch_abort:
__data_abort:
__not_used:
__irq:
__fiq:		        	.word _data_abort
#endif

/*
 * Startup Code (reset vector)
 *
 * do important init only if we don't start from RAM!
 * - relocate armboot to ram
 * - setup stack
 * - jump to second stage
 */

_TEXT_BASE:	.word	TEXT_BASE

.globl _armboot_start
_armboot_start:	.word	_start

#ifdef BSS_IN_BOOTBLOCK
/*
 * These are defined in the board-specific linker script.
 */
.globl _bss_start
_bss_start:	.word	__bss_start

.globl _bss_end
_bss_end:	.word	_end
#endif

#ifdef CONFIG_USE_IRQ
/* IRQ stack memory (calculated at run-time) */
.globl IRQ_STACK_START
IRQ_STACK_START: .word	0x0badc0de

/* IRQ stack memory (calculated at run-time) */
.globl FIQ_STACK_START
FIQ_STACK_START: .word	0x0badc0de
#endif

/*
 * We only change the version number of the bootblock when the functionality
 * of the boot block (i.e., the first 1K of data in OneNAND) changes, which
 * isn't very often.
 *
 * The boot block version number is at offset 0x48 (72), and we're starting
 * this on 2006-08-17.  Notice that the the version number is the date
 * encoded in hexadecimal.
 *
 * One must fully power cycle the device in order for the boot block to be
 * be reloaded.
 */
_BB_VERSION:	.word	0x20080909

_ICMR:		.word	ICMR
_CCCR:		.word	CCCR
_CSPD:		.word	CFG_CPUSPEED
_0x2001:	.word	0x2001

/****************************************************************************/
/*									    */
/* the actual reset code						    */
/*									    */
/****************************************************************************/

reset:
	/* This code can be executed in two different ways - typically in
	 * a ROM because of an actual reset, but also from RAM through
	 * either JTAG or the bootloader.
	 *
	 * We need to limit the code in the bootblock only to the code
	 * needed to boot from ROM.  Everything else needs to be moved
	 * out of the bootblock.
	 */

	mrs	r0,cpsr		 	/* set the cpu to SVC32 mode	*/
	bic	r0,r0,#0x1f	 	/* (superviser mode, M=10011)	*/
	orr	r0,r0,#0x13
	msr	cpsr,r0

/****************************************************************************/
/*									    */
/* CPU_init_critical registers						    */
/*									    */
/* - setup important registers						    */
/* - setup memory timing						    */
/*									    */
/****************************************************************************/

cpu_init_crit:

	ldr	r0, _ICMR		/* mask all IRQ's */
	mov	r1, #0x00
	str	r1, [r0]

	/* set clock speed */
	ldr	r0, _CCCR
	ldr	r1, _CSPD	   	/* CFG_CPUSPEED */
	str	r1, [r0]
	mov	r0, #2
	mcr	p14, 0, r0, c6, c0, 0

	/*
	 * before relocating, we have to setup RAM timing
	 * because memory timing is board-dependent, you will
	 * find a memsetup.S in your board directory.
	 */
	bl	memsetup

	/* Memory interfaces are working. Disable MMU and enable I-cache.   */

	ldr	r0, _0x2001		/* enable access to all coproc.	 */
	mcr	p15, 0, r0, c15, c1, 0
	CPWAIT

	mcr	p15, 0, r0, c7, c10, 4  /* drain the write & fill buffers   */
	CPWAIT

	mcr	p15, 0, r0, c7, c7, 0   /* flush Icache, Dcache and BTB	 */
	CPWAIT

	mcr	p15, 0, r0, c8, c7, 0   /* flush instuction and data TLBs   */
	CPWAIT

	/* Enable the Icache	*/
/*
	mrc	 p15, 0, r0, c1, c0, 0
	orr	 r0, r0, #0x1800
	mcr	 p15, 0, r0, c1, c0, 0
	CPWAIT
*/


relocate:				/* relocate U-Boot to RAM	    */
	adrl	r0, _start		/* r0 <- current position of code   */
	ldr	r1, _TEXT_BASE		/* test if we run from flash or RAM */
	cmp	r0, r1			/* don't reloc during debug	    */
	beq	stack_setup

#ifndef BSS_IN_BOOTBLOCK
	mov	r2, #(N_BL*BL_SIZE)	/* size of all bootloader images   */
	add	r2, r0, r2		/* r2 <- source end address 	    */
	sub	r1, r1, #BL_SIZE	/* copy recovery image 256K below   */
#else
	ldr	r2, _armboot_start
	ldr	r3, _bss_start
	sub	r2, r3, r2		/* r2 <- size of armboot	*/
	add	r2, r0, r2		/* r2 <- source end address	*/
#endif

	cmp	r0, #CFG_DRAM_BASE	/* check if we're in SDRAM	*/
	bhi	booting_from_ram
#ifdef CONFIG_NOR_SUPPORT
	/* Runtime check to determine if we're booting from NAND or NOR.
	 * If a OneNAND is visible in the CS2 space, then we are booting
	 * from NOR.
	 */

	ldr	r4, unp_0x0801E000		/* Check for nand on CS2 */
	ldrh	r5, [r4, #ON_MANU_ID_OFFSET]
	cmp	r5, #ON_MANU_ID_EXPECTED
	bne	booting_from_onenand
	ldrh	r5, [r4, #ON_DEV_ID_OFFSET]
	cmp	r5, #ON_DEV_ID_EXPECTED
	bne	booting_from_nor
#endif

booting_from_onenand:
	/*
	 * Register usage:
	 * r0 = "virtual" source ptr	(starts at 0x00000000 or 0xA2000000)
	 * r1 = destination ptr		(_TEXT_BASE - typically 0xA3F00000)
	 * r2 = source end addr		(typically 0x00022xxx or 0xA2022xxx)
	 * r3 = DataRAM0 buffer ptr	(ranges from 0x0000-0x07FF)
	 * r4 = OneNAND register ptr	(0x0001E000)
	 */

copy_from_onenand:
unpack_loop:
	ldr	r4, unp_0x0001E200

	mov	r5, r0, LSL #5		/* convert addr to block number */
	mov	r5, r5, LSR #(17+5)
	strh	r5, [r4]		/* set FBA			*/

	mov	r5, r0, LSL #15		/* convert addr to page number	*/
	mov	r5, r5, LSR #(11+15)
	mov	r5, r5, LSL #2		/* FPA = page << 2, FSA = 0	*/
	strh	r5, [r4, #ON_START_ADDR_8_OFFSET]	/* set FPA/FSA	*/

	mov	r5, #0
	strh	r5, [r4, #ON_START_ADDR_2_OFFSET]	/* clear DDP	*/

	ldr	r4, unp_0x0001E400

	mov	r5, #0x0800
	strh	r5, [r4]

	mov	r5, #0
	strh	r5, [r4, #ON_INTERRUPT_OFFSET]	/* clear intr flags	*/

	strh	r5, [r4, #ON_COMMAND_OFFSET]	/* "Load" cmd = 0	*/

unp_wait_for_intr:
	ldrh	r5, [r4, #ON_INTERRUPT_OFFSET]
	tst	r5, #0x8000			 /* FIXME - needs a timeout */
	beq	unp_wait_for_intr

	/* Don't check for an error - what can we do?  Just plow ahead... */

	mov	r3, #0x00000400
	mov	r9, #ON_PAGE_SIZE
copy_page_loop:
	ldrh	r5, [r3]
	strh	r5, [r1]
	add	r3, r3, #2
	add	r1, r1, #2
	subs	r9, r9, #2
	bne	copy_page_loop

	add	r0, r0, #2048
	cmp	r0, r2
	ble	unpack_loop


	/* All of the bootloader code is copied into memory, so we
	 * can call outside of the bootblock for the first time
	 */
relocate_done:

#ifdef CONFIG_RECOVERY_MODE
	mov	r0, #ENTRY_PT_CHECK_RECOVERY
	ldr	r1, _TEXT_BASE		/* jumps into RAM copy of recovery  */
	sub	r1, r1, #BL_SIZE	/* code 256K below TEXT_BASE	    */
	add	r1, r1, #ENTRY_PT_OFFSET
	blx	r1
#endif

#ifdef CONFIG_NOR_SUPPORT
unp_0x0801E000: .word	0x0801E000
#endif

unp_0x0001E200: .word	0x0001E200
unp_0x0001E400: .word	0x0001E400




.globl	init_table, iGPSR0, iGPSR1, iGPCR0, _OSCR
init_table:
iGPSR0:		.word	GPSR0,	 CFG_GPSR0_VAL
iGPSR1:		.word	GPSR1,	 CFG_GPSR1_VAL
		.word	GPSR2,	 CFG_GPSR2_VAL
iGPCR0:		.word	GPCR0,	 CFG_GPCR0_VAL
		.word	GPCR1,	 CFG_GPCR1_VAL
		.word	GPCR2,	 CFG_GPCR2_VAL
iGPDR0:		.word	GPDR0,	 CFG_GPDR0_VAL
		.word	GPDR1,	 CFG_GPDR1_VAL
		.word	GPDR2,	 CFG_GPDR2_VAL
		.word	GAFR0_L, CFG_GAFR0_L_VAL
		.word	GAFR0_U, CFG_GAFR0_U_VAL
		.word	GAFR1_L, CFG_GAFR1_L_VAL
		.word	GAFR1_U, CFG_GAFR1_U_VAL
		.word	GAFR2_L, CFG_GAFR2_L_VAL
		.word	GAFR2_U, CFG_GAFR2_U_VAL
iPSSR:		.word	PSSR,	 CFG_PSSR_VAL
		.word	0,	 0

_OSCR:		.word	OSCR
_CFG_MSC0_VAL: 	.word	CFG_MSC0_VAL
_CFG_MSC1_VAL: 	.word	CFG_MSC1_VAL
_CFG_MSC2_VAL: 	.word	CFG_MSC2_VAL
_0xfff:		.word	0xfff
_0x00220022:	.word	0x00220022
_0xC020C020:	.word	0xC020C020


//
// Start of Memory Setup Code
//

// Define the following positions to make driving JTAG easier

.globl memsetup
.globl mem_init_complete
.globl start_mem_controller_init

memsetup:

	/*
	 * If this bootloader code is running at our blessed location in RAM,
	 * typically TEXT_BASE (0xA3F00000), then we return immediately
	 * because the GPIO's and memory are already setup.
	 */

	ldr	r0, _BLESSED_RAM_LOCATION
	cmp	pc, r0			/* running at blessed location? */
	bhs	end_of_memsetup		/* if so, then return		*/

	//
	// JTAG Support
	//
	// To have this bootloader behave as a JTAG-specific bootloader, all
	// that needs to be done is to load the bootloader image via JTAG to
	// a memory address that IS NOT >= BLESSED_RAM_LOCATION.
	//
	// This code will check the address it's running from, and if it's
	// not the actual blessed bootloader running from the blessed location
	// in RAM, it will then check to see if we're running from RAM. 
	// If we are running from RAM, but not at the blessed location, then
	// we must want JTAG support and we'll skip the appropriate stuff
	// for doing JTAG recovery.
	//
	// Okay, we're not running at the normal location for the bootloader -
	// let's see if we're running from RAM.  If we are, this might be a
	// JTAG request.
	//

#ifndef STEAL_SPACE
	cmp	pc, #CFG_DRAM_BASE	/* test if we're in SDRAM */
	bhs	init_gpios		/* if we are, might be JTAG - jump
					 * over wake check and init GPIO's */
#endif


#if defined(CONFIG_FIONA) && defined(CONFIG_POWER_SAVINGS)
    ldr     r1, _PMGR_BASE

    /* Check for BATT_FAULT first... */
    ldr     r0, [r1, #PSSR_OFFSET]
    ands    r0, r0, #PSSR_BFS
    bne     init_gpios	/* Branch if we DID BATT_FAULT */

    /* Check for wake from sleep */
    ldr     r0, [r1, #RCSR_OFFSET]
    cmp     r0, #RCSR_SMR
    bne     init_gpios	/* Branch if we DIDN'T wake from sleep */

        /*
         * We may be in a situation where we put ourselves to sleep
         * from the bootloader (power applied, but power switch OFF).
         * IF this is the case, treat this like a fresh boot by clearing
         * the "we were sleeping" bit in the RCSR.
         */
        ldr     r2, [r1, #PSPR_OFFSET]
        ldr     r0, _FAKE_SLEEP
        cmp     r2, r0
        bne     no_init_gpios

        mov     r0, #RCSR_SMR
        str     r0, [r1, #RCSR_OFFSET]

#endif	/* CONFIG_FIONA && CONFIG_POWER_SAVINGS */

init_gpios:

	/* First, setup GPIOs */

	adrl	r2, init_table
init_loop:
	ldr	r0, [r2]
	ldr	r1, [r2, #4]
	cmp	r0, #0
	beq	init_done
	str	r1, [r0]
	add	r2, r2, #8
	b	init_loop
init_done:

no_init_gpios:

	/* The code below uses sample code generated by the Intel PXA250 Memory
	 * Configuration Tool [http://appzone.intel.com/pcg/pxa250/memory/]
	 * which has then been modified manually while following the
	 * documentation in "Intel PXA255 Process Developer's Manual"
	 * [order #278693-002] 
	 * Chapter 6.11 - Hardware, Watchdog, or Sleep Reset Operation */

	/* Step 1: After hardware reset, complete a power-on wait of 200us
	 * to allow clocks to stabilize */

	ldr	r3, _OSCR		/* reset the OS Timer Count to zero */
	mov	r2, #0
	str	r2, [r3]
	ldr	r4, =0x300		/* 0x2E1 is about 200usec, so	*/
					/* 0x300 should be plenty	*/
10:
	ldr	r2, [r3]
	cmp	r4, r2
	bgt	10b

	/*
	 * If running a test image from MMC or JTAG recovery, skip memory
	 * controller initialization.  (Either the JTAG config script
	 * or the existing ROM image has already initialized the controller.)
	 */
start_mem_controller_init:

#ifndef STEAL_SPACE
	// If we're running from RAM, skip memory controller initialization.

	cmp	pc, #CFG_DRAM_BASE	/* are we in SDRAM?		*/
	bhs	mem_init_complete	/* if so, then we're done	*/
#endif

	/* Step 1a: Write MSC0, MSC1, MSC2 */
	/*	The MSCx registers should be read after they are written
	 *	with a new value before an access to the memory is attempted.
	 *	(See MSC section of manual.)				*/
	ldr	r3, _MEMC_BASE		/* Base of memory controller	*/
	ldr	r2, _CFG_MSC0_VAL	/* Generated value CFG_MSC0_VAL */
	str	r2, [r3, #MSC0_OFFSET]	/* Configure CS1 and CS0	*/
	ldr	r2, [r3, #MSC0_OFFSET]	/* Read it back after write	*/

	ldr	r2, _CFG_MSC1_VAL	/* Generated value CFG_MSC1_VAL	*/
	str	r2, [r3, #MSC1_OFFSET]	/* Configure CS3 and CS2	*/
	ldr	r2, [r3, #MSC1_OFFSET]	/* Read it back after write	*/

	ldr	r2, _CFG_MSC2_VAL	/* Generated value CFG_MSC2_VAL	*/
	str	r2, [r3, #MSC2_OFFSET]	/* Configure CS5 and CS4	*/
	ldr	r2, [r3, #MSC2_OFFSET]	/* Read it back after write	*/

	/* Step 1b: Write MECR,MCMEM0/1,MCATT0/1,MCIO0/1 (Skip on gumstix) */
	/* Step 1c: Write MDREFR:K0RUN and MDREFR:E0PIN.		   */
	/*	    Configure MDREFR:K0DB2.				   */
	/*	    Retain current values of MDREFR:APD and MDREFR:SLFRSH. */
	/*	    MDREFR:DRI must contain a valid value.		   */
	/*	    Deassert MDREFR:KxFREE. (Skip on gumstix)		   */

	ldr	r2, [r3, #MDREFR_OFFSET] /* read MDREFR value */
	str	r2, [r3, #MDREFR_OFFSET] /* configure K2DB1 and K2DB2 */

#if 0
	ldr	r1, =0xfffff000
	bic	r2, r2, r1		/* Retain APD,SLFRSH,K1DB2,K2DB2   */
	orr	r2, r2, #0x000F		/* SDRAM refresh interval (DRI) */
	str	r2, [r3, #MDREFR_OFFSET]
#endif

	ldr	r1, _0xfff
	bic	r2, r2, r1 /* Clear DRI -- INTEL TOOL LEAVES THIS STEP OUT!! */
	orr	r2, r2, #0x000F	/* set a valid SDRAM Refresh Interval (DRI) */
	str	r2, [r3, #MDREFR_OFFSET]

	/* Step 2: Configure Synchronous Static memory (Skip on gumstix) */

	mov	r1, #0
	str	r1, [r3, #SXCNFG_OFFSET]

	/* Step 4: In systems that contain SDRAM, transition the SDRAM
	 *	   controller through the following state sequence:
	 *		 a. self-refresh
	 *		 b. power-down
	 *		 c. PWRDNX
	 *		 d. NOP						*/

	/* change from "self-refresh and clock-stop" to "self-refresh" state */
	orr	r2, r2, #0x00010000	/* assert K1RUN for SDCLK1 */
	str	r2, [r3, #MDREFR_OFFSET]

	/* change from "self-refresh" to "Power-down" state		*/
	bic	r2, r2, #0x00400000	/* clear SLFRSH bit field */
	str	r2, [r3, #MDREFR_OFFSET]

	/* change from "Power-down" to "PWRDNX" state */
	orr	r2, r2, #0x00008000	/* set the E1PIN bit field */
	str	r2, [r3, #MDREFR_OFFSET]

	/* no action required to change from "PWRDNX" to "NOP" state */


	/* Step 4f: Write MDCNFG (with enable bits deasserted),
	 *	    MDCNFG:DE3:2,1:0 set to 0.
	 *	    Must not be enabled yet.				*/

	ldr	r2, =0x000001C8		/* Value from Intel tool */
	str	r2, [r3, #MDCNFG_OFFSET]

	/* Step 5: For systems that contain SDRAM, wait a specified NOP
	 * power-up waiting period required by the SDRAMs to ensure the SDRAMs
	 * receive a stable clock with a NOP condition */

	ldr	r1, _OSCR		/* reset the OS Timer Count to zero */
	mov	r2, #0 
	str	r2, [r1] 
	ldr	r2, =0x300		/* really 0x2E1 is about 200usec,
					 * so 0x300 should be plenty	    */
20:
	ldr	r0, [r1] 
	cmp	r2, r0
	bgt	20b 

	/* Step 6: Ensure the data cache is disabled -- should not be needed
	 * out of reset, but best to be safe*/

	mov	r0, #0x78		/* turn everything off */
	mcr	p15, 0, r0, c1, c0, 0	/* caches off, MMU off, etc. */

	CPWAIT				/* wait for co-processor */

	/* Step 7:
	 *	On a hardware reset in systems that contain SDRAM, trigger
	 *	the specified number (typically eight) of refresh cycles
	 *	by attempting non-burst read or write accesses to any
	 *	disabled SDRAM bank.
	 *	Each such access causes a simultaneous CBR refresh cycles
	 *	for all four banks, which causes a pass through the CBR
	 *	state and back to NOP.
	 *	On the first pass, the PALL state occurs before CBR state. */

	ldr	r1, =PHYS_SDRAM_1
	mov	r2, #9		/* now must do 9 (!) refresh or CBR commands
				 * before the first access		    */
CBR_refresh1:
	str	r2, [r1]
	subs	r2, r2, #1 
	bne	CBR_refresh1 

	/* Step 8: Can re-enable DCACHE if disabled above (Skip on fiona) */

	/* Step 9: Enable SDRAM partitions */

	ldr	r2, [r3, #MDCNFG_OFFSET] /* enable sdram		*/
	orr	r2, r2, #0x00000001	/* enable partition 0 only	*/
	str	r2, [r3, #MDCNFG_OFFSET]

	/* Step 10: In systems containing SDRAM, write the MDMRS register to
	 * trigger an MRS command to all enabled banks of SDRAM. For each SDRAM
	 * partition pair that has one or both partitions enabled, this forces a
	 * pass through the MRS state and back to NOP. The CAS latency must be
	 * the only variable option and is derived from the value programmed in
	 * the MDCNFG:MDTC0,2 fields. The burst type is programmed to sequential
	 * and the length is set to four. */

	ldr	r2, _0x00220022		/* writeable bits will be set to 0 */
	str	r2, [r3, #MDMRS_OFFSET]

	/* Check the physical board_id to see if this has the mobile SDRAM
	 * that we need to drive to half-drive strength.
	 * If board_id == BOARD_REV_ID_REV3, then we configure the SDRAM EMRS
	 * register to half drive strength, full array
	 */

	ldr	r1, =PHYS_BOARD_REV_ID  /* read board ID */
	ldr	r1, [r1]
	ands	r1, r1, #BOARD_REV_ID_MASK  /* clear all but board ID nybble */
	cmp	r1, #BOARD_REV_ID_REV3
	ldr	r2, _0xC020C020
	strge	r2, [r3, #MDMRSLP_OFFSET] /* only set MDMRSLP on REV3 boards */

	/* Step 11: (optional) Enable auto-power-down */

	ldr	r2, [r3, #MDREFR_OFFSET] 
	orr	r2, r2, #0x00100000		/* set the APD bit to	  */
	str	r2, [r3, #MDREFR_OFFSET]	/* enable auto-power-down */

mem_init_complete:

#if defined(CONFIG_FIONA) && defined(CONFIG_POWER_SAVINGS)
	ldr	r3, _PMGR_BASE

    /* Check for BATT_FAULT */
    ldr     r0, [r3, #PSSR_OFFSET]
    ands    r0, r0, #PSSR_BFS
    strne   r0, [r3, #PSSR_OFFSET]                /* clear bit if it was set... */
    bne     end_of_memsetup         /* Branch if we DID BATT_FAULT */

	/* If we're waking up, branch to resume code
	 * Don't resume on reset, though.
	 */
	ldr	r1, [r3, #RCSR_OFFSET]
	cmp	r1, #RCSR_SMR
	ldreq	r1, [r3, #PSPR_OFFSET]
	moveq	pc, r1		/* Jump to resume pc if we just woke up */
#endif /* CONFIG_FIONA */

end_of_memsetup:
	mov	pc, lr


_BLESSED_RAM_LOCATION:	.word	BLESSED_RAM_LOCATION /* typical RAM address */
_MEMC_BASE:	        .word	MEMC_BASE	/* Memory controller base address */
_PMGR_BASE:	        .word	PMCR		/* Power Manager base address	  */
_WAKE_PEDR:             .word   WAKE_PEDR_FIONA /* PEDR Wake Possibility */ 
_FAKE_SLEEP:            .word   NO_SLEEP_PSPR   /* PSPR value on fake sleep */
_OSMR_BASE:             .word   OSMR0           /* OS Timer Space */

/*
 * The BOOTBLOCK_PAD servers two purposes:
 *
 *	(1)  We use it to ensure that we haven't exceeded the boot block size (1024 bytes).
 *
 *	(2)  We use it to prevent the following two globals from causing us to think that the boot
 *	     block itself changed since they are independent of the boot block.
 */
.globl  _end_of_bootblock
_end_of_bootblock:	.word	.-_start, BOOTBLOCK_PAD
_end_of_bootblock_pad:	.fill	1024-., 1, 0xEE

#ifndef BSS_IN_BOOTBLOCK
/*
 * These are defined in the board-specific linker script.
 */
.globl _bss_start
_bss_start:	.word	__bss_start

.globl _bss_end
_bss_end:	.word	_end
#endif

/****************************************************************************
 *
 *	Code after this point is not available on OneNAND until after the
 *	image has been copied to RAM.
 *
 ****************************************************************************/


/* Copy loop for NOR flash AND images loaded into RAM (JTAG booting, for example).
 *
 * r0 - start address of source image
 * r1 - destination address for copy (256K below TEXT_BASE)
 * r2 - end address of source image (2 * 256K)
 */
booting_from_nor:
booting_from_ram:
copy_loop:
	ldmia	r0!, {r3-r10}	/* copy from source address [r0]	*/
	stmia	r1!, {r3-r10}	/* copy to   target address [r1]	*/
	cmp	r0, r2		/* until source end address [r2]	*/
	ble	copy_loop

	b	relocate_done


/****************************************************************************
 *
 *	After the recovery mode check, we return here.
 *
 ****************************************************************************/

after_check_recovery_mode:

        /* We didn't have space to setup the 32KHz clock in the early
         * bootloader, so we set it up here...
         */
        ldr     r2, =OSCC
        mov     r0, #CFG_OSCC_VAL
        str     r0, [r2]
        ldr     r2, =PCFR
        mov     r0, #CFG_PCFR_VAL
        str     r0, [r2]

#ifdef CONFIG_SHOW_BOOT_PROGRESS
        /* If we have a mismatch between normal and recovery bootloader,
         * AND the recovery bootloader is old and doesn't have the IOC uart
         * initialization code in it, then we would hang the first time we
         * try and talk to the IOC.  To "fix" this possible condition,
         * we re-initialize the UART here. -njv
         */
        bl      ioc_uart_initialized
        bne     uart_already_inited

        // Initialize the progress bar and indicate we're at position #1
        bl      init_btuart
        bl      progress_bar_init
        mov     r0, #BOOTBLOCK_PROGRESS_1
        bl      progress_bar
        mov     r0, #BOOTBLOCK_PROGRESS_2
        bl      progress_bar
        b       uart_already_inited

ioc_uart_initialized:
        ldr     r0, =0x40200004         /* BTIER */
        ldr     r0, [r0]
        tst     r0,#0x40
        mov     pc, lr
        
uart_already_inited:
#endif // CONFIG_SHOW_BOOT_PROGRESS

#ifdef CONFIG_SHOW_BOOT_PROGRESS
        /* Update progress bar */
        mov     r0, #BOOTBLOCK_PROGRESS_3
        bl      progress_bar
#endif

	/* Set up the stack						    */
stack_setup:
	ldr	r0, _TEXT_BASE		/* upper 128 KiB: relocated uboot   */
	sub	r0, r0, #CFG_MALLOC_LEN 	/* malloc area		    */
	sub	r0, r0, #CFG_GBL_DATA_SIZE	/* bdinfo						*/
#ifdef CONFIG_USE_IRQ
	sub	r0, r0, #(CONFIG_STACKSIZE_IRQ+CONFIG_STACKSIZE_FIQ)
#endif
	sub	sp, r0, #12		/* leave 3 words for abort-stack    */

clear_bss:
	ldr	r0, _bss_start		/* find start of bss segment	    */
	ldr	r1, _bss_end		/* stop here			    */
	mov	r2, #0x00000000		/* clear			    */

clbss_l:
	str	r2, [r0]		/* clear loop...		*/
	add	r0, r0, #4
	cmp	r0, r1
	bne	clbss_l

	/* set clock speed */
	ldr	r0, _CCCR
	ldr	r1, _CSPD	   	/* CFG_CPUSPEED */
	str	r1, [r0]
	mov	r0, #2
	mcr	p14, 0, r0, c6, c0, 0

#ifdef CONFIG_SHOW_BOOT_PROGRESS
        /* Update progress bar */
        mov     r0, #BOOTBLOCK_PROGRESS_4
        bl      progress_bar
#endif

	ldr	pc, _start_armboot

_start_armboot:	.word	start_armboot

/* vim: set ts=8 sw=8 tw=78: */
