/*
 *	isp1761.c - Philips ISP1761 USB host/device controller
 *
 *	Copyright (C) 2006 Lab126, Inc.  All rights reserved.
 */

#include <common.h>
#include <command.h>
#include "ub_boot_globals.h"

#define USB_CMD_ARGS		2	// usb [sleep|wake]



#define HC_REVISION_REG         0x00    /* 8-31Revision Register (read only) */
#define HC_CAPABILITY_REG       0x00    /* 0-7 Capability Register (read only) */

#define HC_SPARAMS_REG          0x04    /* Structural Parameters Register */
#define		PPC			0x00000010
#define		N_PORTS_MASK		0x0000000F

#define HC_CPARAMS_REG          0x08    /* Capability Parameters Register */

#define HC_SP_PORTROUTE1_REG    0x0C    /* Companion Port Route Register 1 */
#define HC_SP_PORTROUTE2_REG    0x10    /* Companion Port Route Register 1 */

#define HC_USBCMD_REG           0x20    /* USB Command Register */
#define		RS			0x00000001
#define		HCRESET			0x00000002
#define		LHCR			0x00000080

#define HC_USBSTS_REG           0x24    /* USB Status Register */
#define HC_INTERRUPT_REG_EHCI   0x28    /* INterrupt Enable Register */
#define HC_FRINDEX_REG          0x2C    /* Frame Index Register */

#define HC_CTRLDSSEGMENT_REG    0x30    /* 4G Segment Selector (64-bit only) */
#define HC_PERIODICBASE_REG     0x34    /* Periodic List Base Address */
#define HC_ASYNCLISTADDR_REG    0x38    /* Next Asynchronous List register */

#define HC_CONFIGFLAG_REG       0x60    /* Configured Flag  Register */
#define		CF			0x00000001  // Configure Flag

#define HC_PORTSC1_REG          0x64    /* Port Status Control for Port1 */
#define		PP			0x00001000  // Port Power
#define		PR			0x00000100  // Port Reset
#define		SUSP			0x00000080  // Suspend
#define		PED			0x00000004  // Port Enabled/Disabled
#define		ECSC			0x00000002  // Connect Status Change
#define		ECCS			0x00000001  // Current Connect Status

//ISO Transfer Registers
#define HC_ISO_PTD_DONEMAP_REG  0x130   /* ISO PTD Done Map Register */
#define HC_ISO_PTD_SKIPMAP_REG  0x134   /* ISO PTD Skip Map Register */
#define HC_ISO_PTD_LASTPTD_REG  0x138   /* ISO PTD Last PTD Register */

//INT Transfer Registers
#define HC_INT_PTD_DONEMAP_REG  0x140   /* INT PTD Done Map Register */
#define HC_INT_PTD_SKIPMAP_REG  0x144   /* INT PTD Skip Map Register */
#define HC_INT_PTD_LASTPTD_REG  0x148   /* INT PTD Last PTD Register  */

//ATL Transfer Registers
#define HC_ATL_PTD_DONEMAP_REG  0x150   /* ATL PTD Last PTD Register  */
#define HC_ATL_PTD_SKIPMAP_REG  0x154   /* ATL PTD Last PTD Register  */
#define HC_ATL_PTD_LASTPTD_REG  0x158   /* ATL PTD Last PTD Register  */

#define HC_ATL_PTDDONE_THRESHOLDCNT_REG     0x350   /* Revision (read only) */
#define HC_ATL_PTDDONE_THRESHOLDTIMEOUT_REG 0x351   /* Control Register */

//Device controller registers

#define DC_MODE_REG		0x20C	/* Mode register */
#define		DMACLKON		0x0200
#define		VBUSSTAT		0x0100
#define		CLKAON			0x0080
#define		SNDRSU			0x0040
#define		GOSUSP			0x0020
#define		SFRESET			0x0010
#define		GLINTENA		0x0008
#define		WKUPCS			0x0004

#define DC_SCRATCH_REG		0x278
#define		USB_RUNNING_SIGNATURE	0x13579BDF
#define		USB_SUSPEND_SIGNATURE	0x11235813

#define	DC_UNLOCK_DEVICE_REG	0x27C
#define		UNLOCK_VALUE		0xAA37

//General Purpose Registers

#define HC_HW_MODE_REG          0x300   /* H/W Mode Register  */
#define		COMN_IRQ		0x00000400
#define		DATA_BUS_WIDTH		0x00000100
#define		GLOBAL_INTR_EN		0x00000001

#define HC_CHIP_ID_REG          0x304   /* Chip ID Register */
#define HC_SCRATCH_REG          0x308   /* Scratch Register */
#define HC_SW_RESET_REG		0x30C   /* SW Reset Register */
#define		RESET_ALL		0x00000001
#define		RESET_HC		0x00000002

//Interrupt Registers
#define HC_INTERRUPT_REG        0x310   /* Interrupt Register */

#define HC_INTENABLE_REG        0x314   /* Interrupt enable Register */
#define		CLKREADY		0x00000040
#define		HCSUSP			0x00000020

#define HC_ISO_IRQ_MASK_OR_REG  0x318   /* ISO Mask OR Register */
#define HC_INT_IRQ_MASK_OR_REG  0x31C   /* INT Mask OR Register*/
#define HC_ATL_IRQ_MASK_OR_REG  0x320   /* ATL Mask OR Register */
#define HC_ISO_IRQ_MASK_AND_REG 0x324   /* ISO Mask AND Register */
#define HC_INT_IRQ_MASK_AND_REG 0x328   /* INT Mask AND Register */
#define HC_ATL_IRQ_MASK_AND_REG 0x32C   /* ATL Mask AND Register */

//RAM Registers
#define HC_DMACONFIG_REG        0x330   /* DMA Config Register */
#define HC_BUFFER_STATUS_REG    0x334   /* Buffer Status Register */
#define HC_ATL_DONE_TIMEOUT_REG 0x338   /* ATL Done Timeout Register */
#define HC_MEM_READ_REG         0x33C   /* Memory Register */

/*interrupt count and buffer status register*/

#define HC_INT_THRESHOLD_REG    0x340 

#define HC_POWER_DOWN_CTRL_REG	0x354

#define OTG_CONTROL_REG         0x374   /* ISP176x Port 1 Control Register */



#define DEVICE_DESCRIPTOR	1

#define PORT_SUSPEND		2
#define PORT_RESET		4
#define PORT_POWER		8
#define C_PORT_CONNECTION	16



static int rd32(int reg_offset)
{
	volatile int	*addr = (volatile int *) (0x0C000000 + reg_offset);

	return *addr;
}

static void wr32(int reg_offset, int value)
{
	volatile int	*addr = (volatile int *) (0x0C000000 + reg_offset);
	volatile int	v;

	* (volatile int *) addr = value;
	
	v = *(volatile int *) addr;
	if (v != value) {
//		printf("%s: write to 0x%08X of 0x%08X returned 0x%08X\n",
//				__FUNCTION__, addr, value, v);
	}
}

static void mdelay(unsigned long n)
{
	unsigned long msec = n;

	while (msec--)
		udelay(1000);
}


#if 0
static int wait_for_completion(int ptd_num)
{
	volatile int	donemap;
	int	count = 0;

	wr32(HC_BUFFER_STATUS_REG, 1);	// indicate there is a ptd to process

	do {
		donemap = rd32(HC_ATL_PTD_DONEMAP_REG);
		count++;
		if ((count & (0x1000 - 1)) == 0)
			printf(" count = %d\n", count);

	} while (!donemap);

#if 1
	printf("donemap = 0x%08X\n", donemap);
	wr32(HC_MEM_READ_REG, 0x00000C00);
	for (count = 0 ; count < 4 ; count++)
		printf("ptd[%d] = 0x%08X\n", count, rd32(0xC00));
#endif

	return SUCCESS;
}

static void dump(int addr, int len)
{
	volatile int	base = addr;

	wr32(0x33c, base);
	
	while (len) {
		printf("0x%08X: 0x%08X 0x%08X 0x%08X 0x%08X\n", addr,
				rd32(base), rd32(base), rd32(base), rd32(base));
		addr += 16;
		len  -= 16;
	}
}

static int setup_token(int addr, /* int ep, */ int setup0, int setup1)
{
	volatile int	*ptd	 = (int *) 0x0C000C00;
	volatile int	*payload = (int *) 0x0C001000;

	payload[0] = setup0;
	payload[1] = setup1;
	payload[2] = 0;
	payload[3] = 0;

	// DW1: S=0, EPType=0, Token=2, DevAddr=addr, EndPt[3:1]=0
	//...  = HS_BULK_DW1(S, EPType, Token, DeviceAddress, EndPt_3_1);
	//...  = HS_BULK_DW1(0, EPTYPE_CTRL, TOKEN_SETUP, addr, 0);
	ptd[1] = 0x00000800 | (addr << 3);
	ptd[2] = 0x10000000 | (((0x1000-0x0400)>>3) << 8);
		// RL=8,DataStartAddr=0x1000
	ptd[3] = 0x83800000;	// A=1, H=0, B=0, X=0, P=0, DT=1, Cerr=3,
				// NakCnt=0, NrBytesTransferred=0
	ptd[4] = 0;		// J=0, NextPTDPtr=0
	ptd[5] = 0;
	ptd[6] = 0;
	ptd[7] = 0;
	//...  = HS_BULK_DW0(EndPt, Mult, MaxPktLen, NrBytesToTransfer, V);
	//...  = HS_BULK_DW0(0, 1, 64, 8, 1);
	ptd[0] = 0x21000041;	// Mult=1, MaxPktLen=64, NrBytes=8, V=1

	// wait for transaction to complete
	return wait_for_completion(0);
}

static int in_token(int addr, /* int ep, */ char *buf, int len)
{
	volatile int	*ptd	 = (int *) 0x0C000C00;
	volatile int	*payload = (int *) 0x0C001000;
	int	i;
	int	retval;

	for (i = 0 ; i < 64 ; i++)
		*payload++ = 0xFFFFFFFF;
	
	ptd[1] = 0x00002400 | (addr << 3);
	ptd[2] = 0x10000000 | (((0x1000-0x0400)>>3) << 8);
	ptd[3] = 0x83800000;
	ptd[4] = 0;
	ptd[5] = 0;
	ptd[6] = 0;
	ptd[7] = 0;
	ptd[0] = 0x21000201;	// NrBytesToTransfer=64

	// wait for transaction to complete
	retval = wait_for_completion(0);

	if (buf)
		dump(0x1000, 64);

	return retval;
}

static int out_token(int addr, /* int ep, */ char *buf, int len)
{
	volatile int	*ptd	 = (int *) 0x0C000C00;
	volatile int	*payload = (int *) 0x0C001000;

	if (buf) {
		printf("Whoa - non-NULL out token!\n");
		payload[0] = 0;
	}	

	ptd[1] = 0x00002000 | (addr << 3);
	ptd[2] = 0x10000000 | (((0x1000-0x0400)>>3) << 8);
	ptd[3] = 0x83800000;
	ptd[4] = 0;
	ptd[5] = 0;
	ptd[6] = 0;
	ptd[7] = 0;
	ptd[0] = 0x21000201;	// NrBytesToTransfer=64

	// wait for transaction to complete
	wait_for_completion(0);
}

static int get_descriptor(int addr, int type, int index, int len,
				char *buf, int buflen)
{
	int	setup0 = (type << 24) | (index << 16) | 0x00000680;
	int	setup1 = (len << 16) | 0x00000000;
	
	if (type == 3)
		setup1 |= 0x00000409;
	
	setup_token(addr, setup0, setup1);

	in_token(addr, buf, len);

	out_token(addr, NULL, 0);

	return SUCCESS;
}

static int set_address(int addr, int new_addr)
{
	int	setup0 = 0x00000500 | (new_addr << 16);
	int	setup1 = 0x00000000;
	
	setup_token(addr, setup0, setup1);

	in_token(addr, NULL, 0);

	return SUCCESS;
}

static int set_configuration(int addr, int config_number)
{
	int	setup0 = 0x00000900;
	int	setup1 = 0x00000000;
	
	setup_token(addr, setup0, setup1);

	in_token(addr, NULL, 0);
	
	return SUCCESS;
}

static int set_hub_feature(int addr, int feature, int port)
{
	int	setup0 = 0x00000323 | (feature << 16);
	int	setup1 = 0x00000000 | (port);
	
	setup_token(addr, setup0, setup1);

	in_token(addr, NULL, 0);
	
	return SUCCESS;
}

static int clear_hub_feature(int addr, int feature, int port)
{
	int	setup0 = 0x00000123 | (feature << 16);
	int	setup1 = 0x00000000 | (port);

	setup_token(addr, setup0, setup1);

	in_token(addr, NULL, 0);

	return SUCCESS;
}


int do_get_descriptor(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	char	buf[80];
	int	result;

	result = get_descriptor(0, 1 /*device*/, 0, 18, buf, sizeof(buf));
	result = get_descriptor(0, 2 /*config*/, 0,  9, buf, sizeof(buf));
	result = get_descriptor(0, 3 /*string*/, 1, 64, buf, sizeof(buf));
	return SUCCESS;
}


U_BOOT_CMD(
    get_descriptor, 2, CMD_FLAG_NONE, do_get_descriptor,
    "get_descriptor - issue a get_descriptor request to address 'n'\n",
    "       - issue get_descriptor to address 'n'\n"
);


static int do_enumerate(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	char	buf[80];
	int	port;

	get_descriptor(0, DEVICE_DESCRIPTOR, 0, 18, buf, sizeof(buf)); 

	set_address(0, 2);

	set_configuration(2, 1);

	for (port = 1 ; port <= 3 ; port++) {
		set_hub_feature(2, PORT_POWER, port);
		clear_hub_feature(2, C_PORT_CONNECTION, port);
		set_hub_feature(2, PORT_RESET, port);
	}

	return SUCCESS;
}

U_BOOT_CMD(
    enumerate, 1, CMD_FLAG_NONE, do_enumerate,
    "enumerate - enumerate and configure the internal 1761 hub\n",
    "      - enumerate and configure the internal 1761 hub\n"
);
#endif


static int do_usb_init(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	volatile int	intr;
	volatile int	*msc1 = (volatile int *)(0x4800000C);
	int		scratch;
	int		tmp;
	int		wait_for_clkready = 0;

	*msc1 = 0x7ff47ff8;	// kick timings out to max value
	
	// if host controller scratch register is zero, wait for CLKREADY
	
	scratch = rd32(HC_SCRATCH_REG);

	if (scratch == USB_SUSPEND_SIGNATURE) {
		//printf("Host controller is already suspended.\n");
		return SUCCESS;
	} else if (scratch == USB_RUNNING_SIGNATURE) {
		//printf("Host controller is already running.\n");
		// just fall through and initialize
	} else if ((scratch == 0) && 
			((rd32(HC_HW_MODE_REG) & GLOBAL_INTR_EN) == 0)) {
		//printf("Host controller is in the reset state.\n");
		wait_for_clkready = 1;
		// just fall through and initialize
	} else {
		printf("Unexpected scratch value: 0x%08X\n", scratch);
		// just fall through and initialize
	}
	
	if (wait_for_clkready) {
		//printf("waiting for CLKREADY\n");
		do {
			intr = rd32(HC_INTERRUPT_REG);
		} while ((intr & CLKREADY) == 0);
	}

	wr32(HC_INTERRUPT_REG, intr);	// clear pending interrupts

	// Reset part, and wait for reset to complete
	wr32(HC_SW_RESET_REG, (RESET_HC | RESET_ALL));
	while (rd32(HC_SW_RESET_REG) & RESET_HC)
		;

	// Setup HW Mode Control register, reset ATX
	//printf("setup HW Mode Control, reset ATX\n");
	wr32(HC_HW_MODE_REG, 0x00000100);
	
	wr32(HC_HW_MODE_REG, 0x80000100);
	mdelay(15);
	wr32(HC_HW_MODE_REG, 0x00000100);
	
	// program Edge Intr Count (optional)
	//wr32(0x340, 0x0000FFFF);

	// program Port 1 Control to connect port 1 to host controller
	//wr32(0x374, 0x00800018);

	// program EHCI controller
	//printf("program EHCI controller\n");
	tmp = rd32(HC_USBCMD_REG);
	tmp |= HCRESET;
	wr32(HC_USBCMD_REG, tmp);		// host controller reset
	
	wr32(HC_CONFIGFLAG_REG, 0x00000001);	// All ports routed to EHCI
	
	wr32(HC_PORTSC1_REG, 0x00001000);	// Port power enable
	
	wr32(HC_PORTSC1_REG, 0x00001100);	// assert port reset
	mdelay(50);				// assert reset for 50ms
	wr32(HC_PORTSC1_REG, 0x00001000);	// remove reset
	// wait for PR (port reset) bit to be cleared by EHCI hardware
	while (rd32(HC_PORTSC1_REG) & 0x00000100)
		;
	//printf("starting up controller\n");
	tmp = rd32(HC_USBCMD_REG);
	tmp |= RS;
	wr32(HC_USBCMD_REG, tmp);	// USBCMD - R/S=1
	wr32(0x154, 0x00000000); // ATL PTD Skip Map
	wr32(0x158, 0x10000000); // ATL PTD Last PTD

	wr32(HC_SCRATCH_REG, USB_RUNNING_SIGNATURE);

	/* Initialize the OTG controller to device-only out port 1 */

	wr32(OTG_CONTROL_REG, 0x00000486);	// disable OTG
	
	/* Initialize the device controller portion of the 1761 */

	tmp = rd32(DC_SCRATCH_REG);
	mdelay(3*100);

	wr32(DC_UNLOCK_DEVICE_REG, 0xAA37);
	mdelay(2*100);
	
	wr32(DC_MODE_REG, SFRESET);
	mdelay(2*100);
	wr32(DC_MODE_REG, 0);
	mdelay(5*100);

	return SUCCESS;
}

U_BOOT_CMD(
	usb_init, 1, CMD_FLAG_NONE, do_usb_init,
	"usb_init  - init USB host controller\n",
	"      - init USB host controller\n"
);


static int do_host_suspend(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int	tmp;

#ifdef USB_SUSPEND_INDIVIDUAL_PORTS
	{
		int	port;
	
		for (port = 1 ; port <= 3 ; port++) {
			set_hub_feature(2, PORT_SUSPEND, port);
		}
	}
#endif

	wr32(HC_SCRATCH_REG, USB_SUSPEND_SIGNATURE);

	tmp = rd32(HC_USBCMD_REG);
	tmp &= ~RS;
	wr32(HC_USBCMD_REG, tmp);
	mdelay(100);
	
	wr32(HC_PORTSC1_REG, 0x00001084);
	mdelay(100);

#ifdef USB_USE_DEEP_SLEEP_MODE
	wr32(HC_POWER_DOWN_CTRL_REG, 0x000A1C5E);
#endif
	
	return SUCCESS;
}

U_BOOT_CMD(
    hsuspend, 1, CMD_FLAG_NONE, do_host_suspend,
    "hsuspend - suspend the 1761 USB host controller\n",
    "      - suspend the 1761 USB host controller\n"
);


static int usb_host_wake(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int	tmp;
	
	// Perform a dummy read to wake up the host controller
	// or Assert_HCSUSPEND_WAKEUP();
	
	tmp = rd32(HC_SCRATCH_REG);
	//printf("usb_host_wake: scratch = 0x%08X\n", tmp);

	while ((rd32(HC_INTERRUPT_REG) & CLKREADY) == 0)
		;

#ifdef USB_USE_DEEP_SLEEP_MODE
	wr32(HC_POWER_DOWN_CTRL_REG, 0x03E81BA0); // only if max power saved
#endif

	wr32(HC_PORTSC1_REG, 0x00001004);	// Remove the SUSP bit 

	mdelay(100);
	
	tmp = rd32(HC_USBCMD_REG);
	tmp |= RS;
	wr32(HC_USBCMD_REG, tmp);		// Set RunStop bit to '1'

	mdelay(100);

#ifdef USB_SUSPEND_INDIVIDUAL_PORTS
	{
		int	port;

		for (port = 1 ; port <= 3 ; port++) {
			clear_hub_feature(2, PORT_SUSPEND, port);
		}
	}
#endif

	wr32(HC_SCRATCH_REG, USB_RUNNING_SIGNATURE); // note that we're running

	return SUCCESS;
}

U_BOOT_CMD(
	hwake, 1, CMD_FLAG_NONE, usb_host_wake,
	"hwake   - wake USB host controller\n",
	"      - wake USB host controller\n"
);


#ifdef FIXME_OBSOLETE
static int usb_host_sleep(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int	temp;
	//
	// From Foxconn
	//

//	wr32(OTG_CONTROL_REG,0x00800018);  //## as a host control
//	wr32(HC_HW_MODE_REG,COMN_INT |  DATA_BUS_32BIT | GLOBAL_INT_EN  );
        wr32(HC_INTENABLE_REG, CLKREADY | HCSUSP );
//	wr32(HC_SPARAMS_REG, HC_SPARAMS_PPC | 1); //## read-only register
        /*put the host controller to ehci mode*/
        wr32(HC_CONFIGFLAG_REG, CF);
        wr32(HC_USBCMD_REG,0);
        wr32(HC_PORTSC1_REG, PP | SUSP | PED); //0X1084
        udelay(20000);
        wr32(HC_PORTSC1_REG, PP | PR);   /*Reset */
        wr32(HC_PORTSC1_REG, PP | SUSP | PED);   //0X1084
        wr32(HC_POWER_DOWN_CTRL_REG, 0x000a1f9e);
        temp=rd32(HC_INTERRUPT_REG);
        wr32(HC_INTERRUPT_REG, temp);     //clear interrupt source
        temp=rd32(HC_PORTSC1_REG);
        temp &= (~PED);
//	wr32(HC_PORTSC1_REG, temp);  // seems to make current draw increase
#if 0
	Create_PortSuspendTD(PayloadAddr, 0x00020323);
	Create_PortSuspendTD((PayloadAddr+4), 0x00020323);
	IssueSetupToken(PayloadAddr);
	CheckDoneBit(PayloadAddr);
	IssueINToken();
	CheckDoneBit(PayloadAddr);
	Write_USBCMD(0x00000000);
	Write_PORTSC1(0x00001084);
	Write_PowerDownCntrlReg(0x000A1C5E);
#endif
	return SUCCESS;
}
#endif

static int usb_device_sleep(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int	tmp = rd32(DC_MODE_REG);
	
	wr32(DC_SCRATCH_REG, USB_SUSPEND_SIGNATURE);
	wr32(DC_MODE_REG, tmp | GOSUSP | WKUPCS);
	wr32(DC_MODE_REG, tmp | WKUPCS);
	return SUCCESS;
}

static int usb_device_wake(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	volatile int	tmp;
#if 0
	volatile int	*gpsr0 = (volatile int *) 0x40E00018;
	volatile int	*gpcr0 = (volatile int *) 0x40E00024;

printf("wiggling DC_WAKEUP\n");
	tmp = 0x04000000;
	*gpcr0 = tmp;
	mdelay(2);
	*gpsr0 = tmp;
#endif

printf("waiting for CLKREADY\n");
	do {
		tmp = rd32(DC_MODE_REG);
		tmp = rd32(HC_INTERRUPT_REG) & CLKREADY;
	} while (tmp == 0);
	
	mdelay(3);			// wait for clocks to stabilize
	//printf("usb_wake_device: Mode reg = 0x%08X\n", tmp);
	wr32(DC_UNLOCK_DEVICE_REG, UNLOCK_VALUE);
	wr32(DC_SCRATCH_REG, USB_RUNNING_SIGNATURE);
	tmp = rd32(DC_SCRATCH_REG);
	printf("Mode reg = 0x%08X\n", rd32(DC_MODE_REG));

printf("soft resetting device controller\n");
	wr32(DC_MODE_REG, SFRESET);
	mdelay(2);
	wr32(DC_MODE_REG, 0);
	mdelay(5);
	
	return SUCCESS;
}

void usb_force_sleep(void)
{
	printf("usb_force_sleep: hw_flags.disable_pdc_suspend = %d\n",
			get_hw_flags()->disable_pdc_suspend);
	do_usb_init(NULL, 0, 0, NULL);
	do_host_suspend(NULL, 0, 0, NULL);
	if (DISABLE_PDC_SUSPEND())
		printf("pdc suspend disabled.\n");
	else {
		usb_device_wake(NULL, 0, 0, NULL);
		mdelay(100);
		usb_device_sleep(NULL, 0, 0, NULL);
	}
}

#ifdef FIXME_OBSOLETE
U_BOOT_CMD(
	hsleep, 1, CMD_FLAG_NONE, usb_host_sleep,
	"hsleep  - sleep USB host controller\n",
	"      - sleep USB host controller\n"
);
#endif

U_BOOT_CMD(
	dsleep, 1, CMD_FLAG_NONE, usb_device_sleep,
	"dsleep  - sleep USB device controller\n",
	"      - sleep USB device controller\n"
);

U_BOOT_CMD(
	dwake, 1, CMD_FLAG_NONE, usb_device_wake,
	"dwake   - wake USB device controller\n",
	"      - wake USB device controller\n"
);

#if 0

static int do_usbcmd_test(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	volatile int	tmp;
	volatile int	*gpsr0 = (volatile int *) 0x40E00018;
	volatile int	*gpcr0 = (volatile int *) 0x40E00024;

printf("wiggling HC_WAKEUP and DC_WAKEUP\n");
	tmp = 0x0C000000;
	*gpcr0 = tmp;
	mdelay(2);
	wr32(HC_USBCMD_REG, RS);
	mdelay(100);
	*gpsr0 = tmp;

	return SUCCESS;
}

U_BOOT_CMD(
	usbcmd, 1, CMD_FLAG_NONE, do_usbcmd_test,
	"usbcmd  - USBCMD register test\n",
	"      - USBCMD register test\n"
);


static int foxconn_suspend(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	wr32(HC_INTERRUPT_REG, rd32(HC_INTERRUPT_REG));	// clear interrupts
	
	wr32(OTG_CONTROL_REG, 0x00800018);  // use port 1 as host cntr (p.93)
	
	wr32(HC_CONFIGFLAG_REG, 1);

	wr32(HC_USBCMD_REG, 0);

	// enter suspend mode by setting SUSP bit
	wr32(HC_PORTSC1_REG, 0x1084);
	mdelay(20);
	wr32(HC_PORTSC1_REG, 0x1100);
	wr32(HC_PORTSC1_REG, 0x1084);

	// turn off internal 3.3V and 1.8V during suspend mode
	wr32(HC_POWER_DOWN_CTRL_REG, 0x000a1f9e);

	return SUCCESS;
}

U_BOOT_CMD(
	fsuspend, 1, CMD_FLAG_NONE, foxconn_suspend,
	"fsuspend - test Foxconn suspend code\n",
	"      - test Foxconn suspend code\n"
);

#define HC_WAKEUP_GPIO	(1 << 27)

static int foxconn_wake(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	volatile int	*gpsr0 = (volatile int *) 0x40E00018;
	volatile int	*gpcr0 = (volatile int *) 0x40E00024;

	wr32(HC_POWER_DOWN_CTRL_REG, 0x3e81ba0);

	*gpcr0 = HC_WAKEUP_GPIO;
	mdelay(1);
	*gpsr0 = HC_WAKEUP_GPIO;

	wr32(HC_PORTSC1_REG, 0x1004);

	wr32(HC_USBCMD_REG, RS | rd32(HC_USBCMD_REG));

	wr32(HC_CONFIGFLAG_REG, 0);
	
	wr32(OTG_CONTROL_REG, 0x00060401);

	return SUCCESS;
}

U_BOOT_CMD(
	fwake, 1, CMD_FLAG_NONE, foxconn_wake,
	"fwake  - test Foxconn wake code\n",
	"      - test Foxconn wake code\n"
);

#endif
