/*
 * (C) Copyright 2005-2006, Lab126, Inc.
 * Jerry Katzung, katzung@lab126.com
 * Mike Puckett, mikepuck@amazon.com
 *
 * Driver for NAND support, Rick Bronson
 * borrowed heavily from:
 * (c) 1999 Machine Vision Holdings, Inc.
 * (c) 1999, 2000 David Woodhouse <dwmw2@infradead.org>
 */

#include <common.h>
#include <command.h>
#include <malloc.h>
#include <asm/io.h>
#include <watchdog.h>

#ifdef CONFIG_SHOW_BOOT_PROGRESS
# include <status_led.h>
# define SHOW_BOOT_PROGRESS(arg)	show_boot_progress(arg)
#else
# define SHOW_BOOT_PROGRESS(arg)
#endif

#if (CONFIG_COMMANDS & CFG_CMD_NAND)

#ifdef CONFIG_AT91RM9200DK
#include <asm/arch/hardware.h>
#endif

#include <linux/mtd/nand.h>
#ifdef CONFIG_FLASH_SAMSUNG_ONENAND
#include <linux/mtd/onenand.h>
#include <asm/arch/pxa-regs.h>
#else
#include <linux/mtd/nand_ids.h>
#endif
#include <jffs2/jffs2.h>

#ifdef CONFIG_FIONA
#include "../board/fiona/ub_boot_globals.h"
#endif

#ifdef CONFIG_OMAP1510
void archflashwp(void *archdata, int wp);
#endif

#define ROUND_DOWN(value,boundary)      ((value) & (~((boundary)-1)))

/*
 * Definition of the out of band configuration structure
 */
struct nand_oob_config {
	int ecc_pos[6];		/* position of ECC bytes inside oob */
	int badblock_pos;	/* position of bad block flag inside oob -1 = inactive */
	int eccvalid_pos;	/* position of ECC valid flag inside oob -1 = inactive */
} oob_config = { {0}, 0, 0};

#undef	NAND_DEBUG
#undef	PSYCHO_DEBUG

/* ****************** WARNING *********************
 * When ALLOW_ERASE_BAD_DEBUG is non-zero, the erase command will
 * erase (or at least attempt to erase) blocks that are marked
 * bad.  This can be very handy if you are _sure_ that the block
 * is OK, say because you marked a good block bad to test bad
 * block handling and you are done testing, or if you have
 * accidentally marked blocks bad.
 *
 * Erasing factory marked bad blocks is a _bad_ idea.  If the
 * erase succeeds there is no reliable way to find them again,
 * and attempting to program or erase bad blocks can effect
 * the data in _other_ (good) blocks.
 */
#define ALLOW_ERASE_BAD_DEBUG 0

#ifndef CONFIG_FLASH_SAMSUNG_ONENAND
#define CONFIG_MTD_NAND_ECC  /* enable ECC */
#define CONFIG_MTD_NAND_ECC_JFFS2
#endif

// Support for nand_rw_dispatch().
//
// start, len, cmd, retlen, buf
//
#define NAND_RW(s, l, c, r, b) \
			nand_rw_dispatch((s), (l), (c), (r), (b), NANDRW_QUIET)

// start, len, cmd
//
#define NAND_ERASE(s, l, c) \
			nand_rw_dispatch((s), (l), (c), NULL, NULL, NANDRW_QUIET)

// device, start
//
#define NAND_DEV_START(d, s) \
			(nand_dev_desc[(d)].BANK_ADDR + (s))

// start
//
#define NAND_BLOCK(s) \
			((s)/nand_rw_get_device_block_size())
#define NAND_LOAD_PAGE(s) \
			(((s) & (nand_rw_get_device_block_size() - 1))/nand_rw_get_device_oobblock())
#define NAND_PROGRAM_PAGE(s) \
			(((s)/nand_rw_get_device_oobblock()) & (nand_rw_get_device_oobsize() - 1))

// Support for nand_mark_oob()
//
#define NAND_MARK_OOB_BAD	true
#define NAND_MARK_OOB_GOOD	false

/*
 * Function Prototypes
 */
static size_t nand_rw_get_device_oobblock(void);
static size_t nand_rw_get_device_oobsize(void);
static size_t nand_rw_get_device_start(void);

static int nand_rw(struct nand_chip* nand, int cmd,
			size_t start, size_t len,
			size_t* retlen, u_char* buf);
static int nand_erase(struct nand_chip* nand,
			size_t ofs, size_t len, int clean,
			bool verify);

static void nand_print(struct nand_chip *nand);
static void nand_print_bad(struct nand_chip *nand);

static int nand_read_ecc(struct nand_chip *nand, size_t start, size_t len,
			size_t *retlen, u_char *buf, u_char *ecc_code);
static int nand_write_ecc(struct nand_chip *nand, size_t to, size_t len,
			size_t *retlen, const u_char *buf, u_char *ecc_code);
static int nand_read_oob(struct nand_chip *nand, size_t ofs, size_t len,
			size_t *retlen, u_char *buf);
static int nand_write_oob(struct nand_chip *nand, size_t ofs, size_t len,
			size_t *retlen, const u_char *buf);
static int nand_mark_oob(struct nand_chip *nand, size_t ofs, bool mark_bad); 

static int NanD_WaitReady(struct nand_chip *nand, int ale_wait);

#ifdef CONFIG_MTD_NAND_ECC
static int  nand_correct_data (u_char *dat, u_char *read_ecc, u_char *calc_ecc);
static void nand_calculate_ecc (const u_char *dat, u_char *ecc_code);
#endif

#ifdef CONFIG_FLASH_SAMSUNG_ONENAND

static int nand_time(int num_mb, int max_mb);
static int onenand_enter_otp_mode(struct nand_chip *nand);
static int onenand_hot_reset(struct nand_chip *nand, ushort command);
static int onenand_lock_otp_forever(struct nand_chip *nand);

int nand_dump(struct nand_chip *nand, unsigned long ofs, unsigned long len);

#endif

struct nand_chip	nand_dev_desc[CFG_MAX_NAND_DEVICE] = {{0}};
extern ulong        bank_io_base[];

/*
 * FIXME:  We need to get rid of curr_device in favor of treating integrated
 *         and separate devices as though they were just a single device.
 * 
 * Current NAND device (bank).  Initialized in nand_probe().
 */
#define NO_CURR_DEVICE (-1)
static int curr_device = NO_CURR_DEVICE;

/*
 * FIXME:  Need to transition from knowing about individual banks and, instead,
 *         knowing about devices as a whole.  These get initialized during
 *         nand_probe().
 */
static int    nand_dev_base_bank  = 0;
static size_t nand_dev_block_size = 0;
static size_t nand_dev_oobblock   = 0;
static size_t nand_dev_oobsize	  = 0;
static size_t nand_dev_bb_size    = 0;
static size_t nand_dev_start      = 0;
static size_t nand_dev_len        = 0;

static bool last_erase_required_verify = false;

/* ------------------------------------------------------------------------- */

int do_nand (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
#ifdef CONFIG_FLASH_SAMSUNG_ONENAND
	int num_mb, max_mb;
#endif

    int rcode = 0;

    switch (argc) {
    case 0:
    case 1:
	goto do_nand_usage;
    
    case 2:
	if (strcmp(argv[1],"info") == 0) {
		int i;

		putc ('\n');

		for (i=0; i<CFG_MAX_NAND_DEVICE; ++i) {
			if(nand_dev_desc[i].ChipID == NAND_ChipID_UNKNOWN)
				continue; /* list only known devices */
			printf ("Device %d: ", i);
			nand_print(&nand_dev_desc[i]);
		}
		return 0;

	} else if (strcmp(argv[1],"device") == 0) {
		if ((NO_CURR_DEVICE == curr_device) || (CFG_MAX_NAND_DEVICE < curr_device)) {
			puts ("\nno devices available\n");
			return 1;
		}
		printf ("\nDevice %d: ", curr_device);
		nand_print(&nand_dev_desc[curr_device]);
		return 0;

	} else if (strcmp(argv[1],"bad") == 0) {
		if ((NO_CURR_DEVICE == curr_device) || (CFG_MAX_NAND_DEVICE < curr_device)) {
			puts ("\nno devices available\n");
			return 1;
		}
		printf ("\nDevice %d bad blocks:\n", curr_device);
		nand_print_bad(&nand_dev_desc[curr_device]);
		return 0;
	}
	goto do_nand_usage;
    
    case 3:
	if (strcmp(argv[1],"device") == 0) {
		int dev = (int)simple_strtoul(argv[2], NULL, 10);

		printf ("\nDevice %d: ", dev);
		if (dev >= CFG_MAX_NAND_DEVICE) {
			puts ("unknown device\n");
			return 1;
		}
		nand_print(&nand_dev_desc[dev]);
		/*nand_print (dev);*/

		if (nand_dev_desc[dev].ChipID == NAND_ChipID_UNKNOWN) {
			return 1;
		}

		curr_device = dev;

		puts ("... is now current device\n");

		return 0;		
	} 
#ifdef CONFIG_FLASH_SAMSUNG_ONENAND
		else if ((strncmp(argv[1], "dump", 4) == 0) &&
		   (strncmp(argv[2], "all", 3) == 0)) {
		rcode = nand_dump(&nand_dev_desc[0], 0,nand_dev_desc[0].totlen);
		if (rcode)
			printf("ERROR\n");
		rcode = nand_dump(&nand_dev_desc[1],0,nand_dev_desc[1].totlen);
		if (rcode)
			printf("ERROR\n");
			
		return rcode;
	}
		else if (strcmp(argv[1],"time") == 0) {
		
		// We're supposed to write & time <num_mb>s; argv[2]
		// holds <num_mb> (decimal) as a string.
		//
		num_mb = (int)simple_strtoul(argv[2], NULL, 10);
		
		return nand_time(num_mb, 0);
	}
#endif
#if 0
		/* This behavior is pretty dangerous: forget to enter an "offset" and
		   "length", and your entire flash gets blasted.  So we'll eliminate
		   this command option.  If you want to erase the entire device,
		   enter offset=0, and length=<device_length>.
		 */
		else if (strcmp(argv[1],"erase") == 0 && strcmp(argv[2], "clean") == 0) {
			struct nand_chip* nand = &nand_dev_desc[curr_device];
			ulong off = 0;
			ulong size = nand->totlen;
			int ret;
	
			printf ("\nNAND erase: device %d offset 0x%x, size 0x%x ... ",
				curr_device, off, size);
	
			ret = NAND_ERASE(NAND_DEV_START(curr_device, off), size, NANDRW_ERASE_CLEAN);
	
			printf("%s\n", ret ? "ERROR" : "OK");
	
			return ret;
	}
#endif
	goto do_nand_usage;
    
    default:
	/* at least 4 args */
	if (strncmp(argv[1], "read", 4) == 0 ||
	    strncmp(argv[1], "write", 5) == 0) {
		ulong addr = simple_strtoul(argv[2], NULL, 16);
		ulong off  = simple_strtoul(argv[3], NULL, 16);
		ulong size = simple_strtoul(argv[4], NULL, 16);
		int cmd    = (strncmp(argv[1], "read", 4) == 0) ?
				NANDRW_READ : NANDRW_WRITE;
		int ret, total;
		char* cmdtail = strchr(argv[1], '.');

		if (cmdtail && !strncmp(cmdtail, ".oob", 2)) {
			/* read out-of-band data */
			if (cmd & NANDRW_READ) {
				ret = nand_read_oob(nand_dev_desc + curr_device,
						    off, size, &total,
						    (u_char*)addr);
			}
			else {
				ret = nand_write_oob(nand_dev_desc + curr_device,
							off, size, &total,
						    (u_char*)addr);
			}
			return ret;
		}
		else if (cmdtail && !strncmp(cmdtail, ".jffs2", 2))
			cmd |= NANDRW_JFFS2;	/* skip bad blocks */
		else if (cmdtail && !strncmp(cmdtail, ".jffs2s", 2)) {
			cmd |= NANDRW_JFFS2;	/* skip bad blocks (on read too) */
			if (cmd & NANDRW_READ)
				cmd |= NANDRW_JFFS2_SKIP;	/* skip bad blocks (on read too) */
		}
#ifdef SXNI855T
		/* need ".e" same as ".j" for compatibility with older units */
		else if (cmdtail && !strcmp(cmdtail, ".e"))
			cmd |= NANDRW_JFFS2;	/* skip bad blocks */
#endif
#ifdef CFG_NAND_SKIP_BAD_DOT_I
		/* need ".i" same as ".jffs2s" for compatibility with older units (esd) */
		/* ".i" for image -> read skips bad block (no 0xff) */
		else if (cmdtail && !strcmp(cmdtail, ".i")) {
			cmd |= NANDRW_JFFS2;			/* skip bad blocks (on read too) */
			if (cmd & NANDRW_READ)
				cmd |= NANDRW_JFFS2_SKIP;	/* skip bad blocks (on read too) */
		}
#endif /* CFG_NAND_SKIP_BAD_DOT_I */
		else if (cmdtail) {
			goto do_nand_usage;
		}

		printf ("\nNAND %s: device %d offset 0x%x, size 0x%x ... ",
			(cmd & NANDRW_READ) ? "read" : "write",
			curr_device, off, size);

		ret = NAND_RW(NAND_DEV_START(curr_device, off), size, cmd,
			&total, (u_char*)addr);

		printf (" %d bytes %s: %s\n", total,
			(cmd & NANDRW_READ) ? "read" : "written",
			ret ? "ERROR" : "OK");

		return ret;
		}
		else if (strcmp(argv[1],"erase") == 0 &&
		   (argc == 4 || strcmp("clean", argv[2]) == 0)) {
		int clean = argc == 5;
		ulong off = simple_strtoul(argv[2 + clean], NULL, 16);
		ulong size = simple_strtoul(argv[3 + clean], NULL, 16);
		int ret;

		printf ("\nNAND erase: device %d offset 0x%x, size 0x%x ... ",
			curr_device, off, size);

		ret = NAND_ERASE(NAND_DEV_START(curr_device, off), size, (clean ? NANDRW_ERASE_CLEAN : NANDRW_ERASE_RAW));

		printf("%s\n", ret ? "ERROR" : "OK");

		return ret;
		}
		else if (strcmp(argv[1],"erase") == 0 &&
		   (argc == 4 || strcmp("verify", argv[2]) == 0)) {
		int verify = argc == 5;
		ulong off = simple_strtoul(argv[2 + verify], NULL, 16);
		ulong size = simple_strtoul(argv[3 + verify], NULL, 16);
		int ret;
		
		printf ("\nNAND erase: device %d offset 0x%x, size 0x%x ... ",
			curr_device, off, size);

		ret = NAND_ERASE(NAND_DEV_START(curr_device, off), size, (verify ? NANDRW_ERASE_VERIFY : NANDRW_ERASE_RAW));

		printf("%s\n", ret ? "ERROR" : "OK");

		return ret;
		}
#ifdef CONFIG_FLASH_SAMSUNG_ONENAND
		else if ((strncmp(argv[1], "dump", 4) == 0) && (argc == 4)) {
		ulong off  = simple_strtoul(argv[2], NULL, 16);
		ulong size = simple_strtoul(argv[3], NULL, 16);
	
		rcode = nand_dump(nand_dev_desc + curr_device, off, size);

		if (rcode)
			printf("ERROR\n");
		}
		else if ((strcmp(argv[1], "time") == 0) && (argc == 4)) {
	
		// We're supposed to write & time <num_mb>s for <max_mb>s; argv[2]
		// holds <num_mb> (decimal) as a string, and argv[3] holds <max_mb>
		// as a string.
		//
		num_mb = (int)simple_strtoul(argv[2], NULL, 10);
		max_mb = (int)simple_strtoul(argv[3], NULL, 10);
		
		rcode = nand_time(num_mb, max_mb);
		}
		else if ((strcmp(argv[1], "mark") == 0) && (argc == 4)) {
		ulong off = simple_strtoul(argv[3], NULL, 16);
		bool mark_bad;

		if (strcmp(argv[2], "bad") == 0)
			mark_bad = true;
		else if (strcmp(argv[2], "good") == 0)
			mark_bad = false;
		else
			goto do_nand_usage;

		rcode = nand_mark_oob(nand_dev_desc + curr_device, off, mark_bad);
		}
#endif
	else
		goto do_nand_usage;

	return rcode;
    }
    
// Save some space; do all usage printing in one spot.
//
do_nand_usage:
    printf("Usage:\n%s\n", cmdtp->usage);
    return 1;
}

U_BOOT_CMD(
	nand,	5,	1,	do_nand,
	"nand    - NAND sub-system\n",
	"info - show available NAND devices\n"
	"nand device [dev] - show or set current device\n"
	"nand read[.jffs2[s]] addr off size\n"
	"nand write[.jffs2] addr off size - read/write `size' bytes starting\n"
	"      at offset `off' to/from memory address `addr'\n"
#if 0
	"nand erase [clean] [off size] - erase `size' bytes from\n"
	"      offset `off' (entire device if not specified)\n"
#else
	"nand erase [clean | verify] off size - erase `size' bytes from offset `off'\n"
#endif
	"nand bad - show bad blocks\n"
	"nand read.oob  addr off size - read  out-of-band data\n"
	"nand write.oob addr off size - write out-of-band data\n"
#ifdef CONFIG_FLASH_SAMSUNG_ONENAND
	"nand dump off size - dump nand contents\n"
	"      (off/size must be multiples of page size)\n"
	"nand time num [max] - `num' is the number of MBs to write (destructive),\n"
	"      read, and time; limit device's length by `max' MBs if present\n"
	"nand mark [bad | good] off - mark block at offset `off' as bad or good;\n"
	"      note that (re-)marking a block good requires an erase\n"
#endif
);

int do_nandboot (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	char *boot_device = NULL;
	char *ep;
	int dev;
	ulong cnt;
	ulong addr;
	ulong offset = 0;
	image_header_t *hdr;
	int rcode = 0;
	switch (argc) {
	case 1:
		addr = CFG_LOAD_ADDR;
		boot_device = getenv ("bootdevice");
		break;
	case 2:
		addr = simple_strtoul(argv[1], NULL, 16);
		boot_device = getenv ("bootdevice");
		break;
	case 3:
		addr = simple_strtoul(argv[1], NULL, 16);
		boot_device = argv[2];
		break;
	case 4:
		addr = simple_strtoul(argv[1], NULL, 16);
		boot_device = argv[2];
		offset = simple_strtoul(argv[3], NULL, 16);
		break;
	default:
		printf ("Usage:\n%s\n", cmdtp->usage);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	if (!boot_device) {
		puts ("\n** No boot device **\n");
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	dev = simple_strtoul(boot_device, &ep, 16);

	if ((dev >= CFG_MAX_NAND_DEVICE) ||
	    (nand_dev_desc[dev].ChipID == NAND_ChipID_UNKNOWN)) {
		printf ("\n** Device %d not available\n", dev);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	printf ("\nLoading from device %d: %s at 0x%lx (offset 0x%lx)\n",
		dev, nand_dev_desc[dev].name, ONENAND_REGS_BASE(nand_dev_desc[dev].IO_ADDR),
		offset);

	if (NAND_RW(NAND_DEV_START(dev, offset), SECTORSIZE, NANDRW_READ, NULL,
			(u_char *)addr)) {
		printf ("** Read error on %d\n", dev);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	hdr = (image_header_t *)addr;

	if (ntohl(hdr->ih_magic) == IH_MAGIC) {

		print_image_hdr (hdr);

		cnt = (ntohl(hdr->ih_size) + sizeof(image_header_t));
		cnt -= SECTORSIZE;
	} else {
		printf ("\n** Bad Magic Number 0x%x **\n", hdr->ih_magic);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	if (NAND_RW(NAND_DEV_START(dev, offset + SECTORSIZE), cnt, NANDRW_READ, NULL,
			(u_char *)(addr+SECTORSIZE))) {
		printf ("** Read error on %d\n", dev);
		SHOW_BOOT_PROGRESS (-1);
		return 1;
	}

	/* Loading ok, update default load address */

	load_addr = addr;

	/* Check if we should attempt an auto-start */
	if (((ep = getenv("autostart")) != NULL) && (strcmp(ep,"yes") == 0)) {
		char *local_args[2];
		extern int do_bootm (cmd_tbl_t *, int, int, char *[]);

		local_args[0] = argv[0];
		local_args[1] = NULL;

		printf ("Automatic boot of image at addr 0x%08lx ...\n", addr);

		do_bootm (cmdtp, 0, 1, local_args);
		rcode = 1;
	}
	return rcode;
}

U_BOOT_CMD(
	nboot,	4,	1,	do_nandboot,
	"nboot   - boot from NAND device\n",
	"loadAddr dev\n"
);

/* returns 0 if block containing pos is OK:
 *		valid erase block and
 *		not marked bad, or no bad mark position is specified
 * returns 1 if marked bad or otherwise invalid
 */
static int check_block(struct nand_chip* nand, unsigned long pos)
{
	int badpos = oob_config.badblock_pos;
	
	if (pos >= nand->totlen)
		return 1;

	if (badpos < 0)
		return 0;	/* no way to check, assume OK */
	
#ifdef CONFIG_FLASH_SAMSUNG_ONENAND

	if ( onenand_block_is_bad(nand->IO_ADDR, NAND_BLOCK(pos)) )
		return 1;
	else
		return 0;

#else
{
	uint8_t oob_data;
	int retlen;

	int page0  = pos & (-nand->erasesize);
	int page1  = page0 + nand->oobblock;

	if (nand_read_oob(nand, page0 + badpos, 1, &retlen, &oob_data) ||
	    oob_data != 0xff ||
	    nand_read_oob(nand, page1 + badpos, 1, &retlen, &oob_data) ||
	    oob_data != 0xff)
		return 1;
	
	return 0;
}
#endif
}

#ifdef CONFIG_FLASH_SAMSUNG_ONENAND
static int check_block_verify(struct nand_chip* nand, unsigned long pos)
{
	if ( check_block(nand, pos) )
		if ( onenand_block_is_truly_bad(nand->IO_ADDR, NAND_BLOCK(pos)) )
			return 1;

	return 0;
}
#endif

/* print bad blocks in NAND flash */
static void nand_print_bad(struct nand_chip* nand)
{
	char how_bad[STRING_MAX];
	unsigned long pos;

	for (pos = 0; pos < nand->totlen; pos += nand->erasesize) {
		if (check_block(nand, pos)) {
#ifdef CONFIG_FLASH_SAMSUNG_ONENAND
			strcpy(how_bad, "soft");
			
			if (check_block_verify(nand, pos))
				strcpy(how_bad, "hard");
				
			printf(" 0x%8.8lx (block %4.4d) - %s failure\n", pos, NAND_BLOCK(pos), how_bad);
#else
			printf(" 0x%8.8lx (block %4.4d)\n", pos, NAND_BLOCK(pos));
#endif
		}
	}
	puts("\n");
}

// Entry:
//
// - nand_rw_dispatch(), attempts to make it so that, whether we have distinct
//   devices acting in concert, or we have fully integrated devices, we treat
//   them the same.
//
// So, unlike the functions it calls, nand_rw() and nand_erase(), nand_rw_dispatch()
// wants <start> to be absolute within the ENTIRE device space, not relative
// to a particular bank.  In other words, <start> is now always relative to
// the physical address of the first bank, and the caller should treat each "device"
// as though it were one contiguous address space.  Equally, <len> should be
// absolute.
//
// The macros NAND_RW(), NAND_ERASE(), and NAND_DEV_START() bridge the gap between
// calling nand_rw_dispatch() directly or not.
//
// Utils:
//
// - nand_rw_get_device_oobblock()   -- These should remain static.
// - nand_rw_get_device_oobsize()    -- |
// - nand_rw_get_device_start()      -- (Eliminate?  Assume logical zero?)
//
// - nand_rw_last_erase_required_verify()
// - nand_rw_get_device_block_size()
// - nand_rw_get_device_bb_size()
// - nand_rw_get_device_len()
//
static size_t nand_rw_get_device_oobblock(void)
{
	return ( nand_dev_oobblock );
}

static size_t nand_rw_get_device_oobsize(void)
{
	return ( nand_dev_oobsize );
}

static size_t nand_rw_get_device_start(void)
{
	return ( nand_dev_start );		
}

bool nand_rw_last_erase_required_verify(void)
{
	return ( last_erase_required_verify );
}

size_t nand_rw_get_device_block_size(void)
{
	return ( nand_dev_block_size );
}

size_t nand_rw_get_device_bb_size(void)
{
	return ( nand_dev_bb_size );
}

size_t nand_rw_get_device_len(void)
{
	return ( nand_dev_len );
}

int nand_rw_dispatch(size_t start, size_t len, int cmd,
	size_t *retlen, u_char *buf, bool verbose)
{
	int  result   = FAILURE;
	bool is_rw    = false,
	     is_erase = false,
	     verify   = false;
	int  clean    = (int)false;
	
	char string1[STRING_MAX],
		 string2[16];

	// Validate the incoming selector (cmd).
	//
	switch ( cmd )
	{
		case NANDRW_WRITE:
		case NANDRW_WRITE | NANDRW_JFFS2:
			if ( verbose )
			{
				strcpy(string1, "write");
				strcpy(string2, "written");
			}
		goto rw_common;
		
		case NANDRW_READ:
		case NANDRW_READ  | NANDRW_JFFS2:
		case NANDRW_READ  | NANDRW_JFFS2 | NANDRW_JFFS2_SKIP:
			if ( verbose )
			{
				strcpy(string1, "read_");
				strcpy(string2, "read");
			}

		rw_common:
			if ( retlen ) *retlen = 0;
			is_rw = true;
		break;
		
		case NANDRW_ERASE_VERIFY:
			verify = true;
		case NANDRW_ERASE_RAW:
			clean = (int)false;
		goto erase_common;
		
		case NANDRW_ERASE_CLEAN:
			clean = (int)true;
		
		erase_common:
			if ( verbose )
			{
				strcpy(string1, "erase");
				strcpy(string2, "erased");
			}
				
			is_erase = true;
		break;
	}

	// If selector (cmd) is valid, get to work.
	//
	if ( is_rw || is_erase )
	{
		unsigned long bank_start, bank_len;
		bool done = false;
		int bank;
		
		size_t cur_start = start, cur_len = len, cur_retlen = 0;
		uchar  *cur_buf = buf;
		
		// Print out header when <verbose>.
		//
		if ( verbose )
		{
			if ( verify )
			{
				for ( bank = nand_dev_base_bank; bank < CFG_MAX_NAND_DEVICE; bank++ )
				{
					onenand_display_bad_blocks(bank+1); putc('\n');
					nand_print_bad(&nand_dev_desc[bank]);
				}
			}
			
			if ( (is_rw && (cmd & NANDRW_JFFS2)) || (is_rw && (cmd & NANDRW_JFFS2_SKIP))
												 || (is_erase && clean) )
				strcat(string1, "_JFFS2");
			else
			    strcat(string1, "_RAW  ");
			
			printf("NAND %s:  0x%08X (start), 0x%08X (length)... ", string1,
				start, len);
		}

		// Walk all the banks, skipping any invalid ones, reducing
		// our workload as we go.
		//
		for ( bank = nand_dev_base_bank; (bank < CFG_MAX_NAND_DEVICE) && !done; bank++ )
		{
			if ( NAND_ChipID_UNKNOWN != nand_dev_desc[bank].ChipID )
			{
				bank_start  = nand_dev_desc[bank].BANK_ADDR;
				bank_len    = nand_dev_desc[bank].totlen;
				
				// Skip to the bank where we've been asked to start.
				//
				if ( (cur_start >= bank_start) && (cur_start < (bank_start + bank_len)) )
				{
					// The nand_rw() and nand_erase() functions want the start
					// to be relative to the bank's base address.
					//
					cur_start -= bank_start;
					
					// Chop up the data on bank bounds.  We're done when the
					// current length (offset from start, of course) is less
					// than the bank's length.
					//
					if ( (cur_start + cur_len) > bank_len )
						cur_len = bank_len - cur_start;
					else
						done = true;

					if ( is_rw )
						result = nand_rw(&nand_dev_desc[bank], cmd, cur_start,
									cur_len, &cur_retlen, cur_buf);
					else
						result = nand_erase(&nand_dev_desc[bank], cur_start,
									cur_len, clean, verify);
				
					// Normalize result.
					//
					result = result ? FAILURE : SUCCESS;
				
					// Don't continue if there was a failure.
					//
					if ( FAILURE == result )
						done = true;
					
					if ( is_rw && (SUCCESS == result) )
					{
						if ( retlen )
							*retlen += cur_retlen;
						
						if ( buf )
							cur_buf += cur_retlen;
					}
					
					// The next start is at the next bank's base address if we're
					// not done.  Similarly, the next length is what we haven't
					// yet processed.
					//
					if ( ((bank + 1) < CFG_MAX_NAND_DEVICE) && !done )
					{
						cur_start = nand_dev_desc[bank+1].BANK_ADDR;
						cur_len   = len - cur_len;
					}
				}
			}
		}
		
		// Print out trailer when <verbose>.
		//
		if ( verbose )
		{
			if ( SUCCESS == result )
			{
				if ( is_erase )
					cur_retlen = len;
					
				printf("OK, 0x%08X bytes %s\n", cur_retlen, string2);
			}
			else
				printf("ERROR\n");
				
			if ( verify )
			{
				for ( bank = nand_dev_base_bank; bank < CFG_MAX_NAND_DEVICE; bank++ )
				{
					onenand_display_bad_blocks(bank+1); putc('\n');
					nand_print_bad(&nand_dev_desc[bank]);
				}
			}
		}
	}

	return ( result );
}

/* cmd: 0: NANDRW_WRITE write, fail on bad block
 *      1: NANDRW_READ  read,  fail on bad block
 *      2: NANDRW_WRITE | NANDRW_JFFS2	write, skip bad blocks
 *      3: NANDRW_READ  | NANDRW_JFFS2	read, data all 0xff for bad blocks
 *      7: NANDRW_READ  | NANDRW_JFFS2 | NANDRW_JFFS2_SKIP read, skip bad blocks
 */
static int nand_rw(struct nand_chip* nand, int cmd,
	size_t start, size_t len,
	size_t *retlen, u_char *buf)
{
	int ret = 0, n, total = 0;
	char eccbuf[6];
	/* eblk (once set) is the start of the erase block containing the
	 * data being processed.
	 */
	unsigned long eblk = ~0;	/* force mismatch on first pass */
	unsigned long erasesize = nand->erasesize;

	while (len) {
		if ((start & (-erasesize)) != eblk) {
			/* have crossed into new erase block, deal with
			 * it if it is marked bad.
			 */
			eblk = start & (-erasesize); /* start of block */
			if (check_block(nand, eblk)) {
				if (cmd == (NANDRW_READ | NANDRW_JFFS2)) {
					while (len > 0 &&
					       start - eblk < erasesize) {
						*(buf++) = 0xff;
						++start;
						++total;
						--len;
					}
					continue;
				}
				else if (cmd == (NANDRW_READ | NANDRW_JFFS2 | NANDRW_JFFS2_SKIP)) {
					printf("%s: skipping bad block at 0x%x\n", __FUNCTION__, eblk);
					start += erasesize;
					continue;
				}
				else if (cmd == (NANDRW_WRITE | NANDRW_JFFS2)) {
					printf("%s: skipping bad block at 0x%x\n", __FUNCTION__, eblk);
					start += erasesize;
					continue;
				}
				else {
					ret = 1;
					break;
				}
			}
		}
		/* The ECC will not be calculated correctly if
		   less than 512 is written or read */
		/* Is request at least 512 bytes AND it starts on a proper boundry */
		if((start != ROUND_DOWN(start, 0x200)) || (len < 0x200))
			printf("Warning block writes should be at least 512 bytes and start on a 512 byte boundry\n");

		if (cmd & NANDRW_READ)
			ret = nand_read_ecc(nand, start,
					   min(len, eblk + erasesize - start),
					   &n, (u_char*)buf, eccbuf);
		else
			ret = nand_write_ecc(nand, start,
					    min(len, eblk + erasesize - start),
					    &n, (u_char*)buf, eccbuf);

		if (ret)
			break;

		start += n;
		buf   += n;
		total += n;
		len   -= n;
	}
	if (retlen)
		*retlen = total;

	return ret;
}

static void nand_print(struct nand_chip *nand)
{
	ulong regs_base = ONENAND_REGS_BASE(nand->IO_ADDR);
	
	if (nand->numchips > 1) {
		printf("%s at 0x%08lX,\n"
		       "\t  %d chips %s, size %d MB, \n"
		       "\t  total size %ld MB, sector size %ld kB\n",
		       nand->name, regs_base, nand->numchips,
		       nand->chips_name, 1 << (nand->chipshift - 20),
		       nand->totlen >> 20, nand->erasesize >> 10);
	}
	else {
		printf("%s at 0x%08lX (", nand->chips_name, regs_base);
		print_size(nand->totlen, ", ");
		print_size(nand->erasesize, " block)\n");
	}
}

/* ------------------------------------------------------------------------- */

#ifdef CONFIG_FLASH_SAMSUNG_ONENAND

static ushort onenand_get_boot_buffer_size(ulong io_base_addr)
{
	 // Return the bootblock (aka the "boot buffer") size in bytes rather
	 // than words.
	 //
	 onenand_regs_t *regs = ONENAND_REG_BASE(io_base_addr);
	 return ( regs->boot_buffer_size << 1 );
}

static void onenand_set_sys_config_1(ulong io_base_addr, ushort config)
{
	onenand_regs_t *regs;

	if ( PHYS_FLASH_1 == io_base_addr )
	{
		struct nand_chip *nand = &nand_dev_desc[0];
		io_base_addr = (ulong)nand->IO_ADDR;
	}

	regs = ONENAND_REG_BASE(io_base_addr);
	regs->sys_config_1 = config;
}

void onenand_clear_dbs(void)
{
	int clear_dbs = get_clear_dbs();
	
	switch ( clear_dbs )
	{
		case CLEAR_DBS_UNKNOWN:
		case CLEAR_DBS_TRUE:
		{
			struct nand_chip *nand = &nand_dev_desc[0];
			ONENAND_SET_DBS(nand->IO_ADDR);
		}
		break;
	}
}

bool onenand_get_sync_default(void)
{
	struct nand_chip *nand = &nand_dev_desc[0];
	return ( ONENAND_SAMSUNG_256MBx1(nand->mfr, nand->id) );
}

// SXCNFG
//
#define SXMEM_SET_FIXED_CLOCK	0x00000000	// Bit 14
#define SXMEM_SET_SYNC_FLASH	0x00002000	// Bits 12-13

#define SXMEM_SET_SYNC_LATENCY	0x000000E0	// Bits 5-7

#define SXMEM_3_CLOCK_DELAY		0x00000008	// Bits 2-4
#define SXMEM_4_CLOCK_DELAY		0x0000000C
#define SXMEM_5_CLOCK_DELAY		0x00000010
#define SXMEM_6_CLOCK_DELAY		0x00000014
#define SXMEM_7_CLOCK_DELAY		0x00000018

#define SXMEM_DISABLED			0x00000000	// Bits 0-1
#define SXMEM_ENABLED			0x00000001

// MSC0
//
#define RT_FLASH_NON_BURST		0x00000000	// Bits 0-2
#define RT_FLASH_BURST_OF_FOUR	0x00000002
#define RT_FLASH_BURST_OF_EIGHT	0x00000003

bool onenand_sync_reads = CONFIG_ONENAND_SYNC;

void onenand_async_set(ulong io_base_addr)
{
	sync_async_t *async_settings = get_async_settings();
	ushort flash_async;
	ulong  sxcnfg_async,
		   mdrefr_async,
		   msc0_async;
	
	if ( SYNC_SETTINGS_UNKNOWN == async_settings->initialized )
	{
		// Set register values.
		//
		flash_async =		ONENAND_RM_ASYNC		|
							ONENAND_BRL_04_CYCLES	|
							ONENAND_BL_CONTINUOUS	|
							ONENAND_ECC_ON 			|
							ONENAND_RDY_HIGH		|
							ONENAND_INT_HIGH		|
							ONENAND_IOBE_DISABLED	|
							ONENAND_RDY_ACTIVE_EARLY;	// 0x40C0
							
		sxcnfg_async =		SXMEM_DISABLED,				// 0x00000000		
		
		mdrefr_async =		MDREFR & ~(MDREFR_K0RUN | MDREFR_E0PIN);
		msc0_async	 =		MSC0   & ~(RT_FLASH_NON_BURST);
		
		// Cache them.
		//
		async_settings->initialized = SYNC_SETTINGS_INITED;
		
		async_settings->onenand_sys_config_1 = flash_async;
		async_settings->pxa_sxcnfg = sxcnfg_async;
		async_settings->pxa_mdrefr = mdrefr_async;
		async_settings->pxa_msc0   = msc0_async;
		
		set_async_settings(async_settings);
	}
	else
	{
		flash_async  = async_settings->onenand_sys_config_1;
		sxcnfg_async = async_settings->pxa_sxcnfg;
		mdrefr_async = async_settings->pxa_mdrefr;
		msc0_async   = async_settings->pxa_msc0;
	}

	onenand_set_sys_config_1(io_base_addr, flash_async);
	SXCNFG = sxcnfg_async;
	MDREFR = mdrefr_async;
	
	MSC0 = msc0_async;
	msc0_async = MSC0;
}

void onenand_sync_set(ulong io_base_addr)
{
	sync_async_t *sync_settings = get_sync_settings();

	if ( onenand_sync_reads )
	{
		ushort flash_sync;
		ulong  sxcnfg_sync,
			   mdrefr_sync,
			   msc0_sync;
		
		// Set register values.
		//
		if ( SYNC_SETTINGS_UNKNOWN == sync_settings->initialized )
		{
			flash_sync = 	ONENAND_RM_SYNC			|
							ONENAND_BRL_04_CYCLES	|
							ONENAND_BL_08_WORDS		|
							ONENAND_ECC_ON			|
							ONENAND_RDY_HIGH		|
							ONENAND_INT_HIGH		|
							ONENAND_IOBE_ENABLED	|
							ONENAND_RDY_ACTIVE_EARLY;	// 0xC4E0
		
			sxcnfg_sync =	SXMEM_SET_FIXED_CLOCK	|
							SXMEM_SET_SYNC_FLASH	|
							SXMEM_SET_SYNC_LATENCY	|
							SXMEM_4_CLOCK_DELAY		|
							SXMEM_ENABLED,				// 0x000020ED,
			
			mdrefr_sync =	MDREFR | (MDREFR_K0RUN | MDREFR_E0PIN);
			msc0_sync   =	MSC0   | (RT_FLASH_BURST_OF_EIGHT);

			// Cache them.
			//
			sync_settings->initialized = SYNC_SETTINGS_INITED;
			
			sync_settings->onenand_sys_config_1 = flash_sync;
			sync_settings->pxa_sxcnfg = sxcnfg_sync;
			sync_settings->pxa_mdrefr = mdrefr_sync;
			sync_settings->pxa_msc0   = msc0_sync;
			
			set_sync_settings(sync_settings);
		}
		else
		{
			flash_sync  = sync_settings->onenand_sys_config_1;
			sxcnfg_sync = sync_settings->pxa_sxcnfg;
			mdrefr_sync = sync_settings->pxa_mdrefr;
			msc0_sync   = sync_settings->pxa_msc0;
		}
		
		onenand_set_sys_config_1(io_base_addr, flash_sync);
		SXCNFG = sxcnfg_sync;
		MDREFR = mdrefr_sync;
		
		MSC0 = msc0_sync;
		msc0_sync = MSC0;
	}
	else
	{
		onenand_async_set(io_base_addr);
		
		if ( SYNC_SETTINGS_UNKNOWN == sync_settings->initialized )
		{
			sync_settings = get_async_settings();
			set_sync_settings(sync_settings);
		}
	}
}

void onenand_sync_init(void)
{
	struct nand_chip *nand = &nand_dev_desc[0];
	set_clear_dbs(CLEAR_DBS_UNKNOWN);
	onenand_clear_dbs();
	
	clear_async_sync_settings();
	onenand_async_set(nand->IO_ADDR);
	onenand_sync_set(nand->IO_ADDR);
}

static void onenand_clear_spare_area(ulong io_base_addr)
{
	ushort *spare_area = ONENAND_DATARAM0_ECC_BASE(io_base_addr);
	size_t oobsize = nand_rw_get_device_oobsize() >> 1;
	int i;
	
	// Clear the spare area with the erased marker to clear it.
	//
    for ( i = 0; i < oobsize; i++ )
		spare_area[i] = ONENAND_ERASED_MARKER;
}

bool onenand_block_is_bad(ulong io_base_addr, ulong block)
{
	bool block_is_bad, done = false;
	onenand_spare_t *spare_area;
	ushort page = 0;
	
	// As long as the BI (bad block information) on ONENAND_BADBLOCK_PAGES
	// of a <block> is erased, the <block> is not considered to be bad.
	//
	do
	{
		block_is_bad = true;
		
		if ( SUCCESS == onenand_load_spare(io_base_addr, block, page) )
		{
			spare_area = (onenand_spare_t *)ONENAND_DATARAM0_ECC_BASE(io_base_addr);
			
			if ( ONENAND_ERASED_MARKER == spare_area->bi )
				block_is_bad = false;
		}
	
		page++;

		// We're done if <block_is_bad> or ONENAND_BADBLOCK_PAGES has been exceeded.
		//
		done = block_is_bad || (ONENAND_BADBLOCK_PAGES < page);
	}
	while ( !done );
	
	return ( block_is_bad );
}

bool onenand_block_is_truly_bad(ulong io_base_addr, ulong block)
{
	bool block_is_truly_bad = false;

	// A block isn't truly bad unless it is first just bad.
	//
	if ( onenand_block_is_bad(io_base_addr, block) )
	{
		size_t	oobblock  = nand_rw_get_device_oobblock(),
				oobsize   = nand_rw_get_device_oobsize();
		
		bool	good;
		
		int		num_pages = ONENAND_BADBLOCK_PAGES,
				page, i;
		
		ushort	*data;
		
		oobsize  /= 2; // Ensure word (16-bit) read/writes.
		oobblock /= 2;
		
		for ( page = 0; (page < num_pages) && !block_is_truly_bad; page++ )
		{		
			// Point to the oob & page buffers and zero them out.
			//
			data = ONENAND_DATARAM0_ECC_BASE(io_base_addr);
			
			for ( i = 0; i < oobsize; i++ )
				data[i] = 0;
				
			data = ONENAND_DATARAM0_BASE(io_base_addr);
			
			for ( i = 0; i < oobblock; i++ )
				data[i] = 0;
				
			// Load the page into its respective buffers.
			//
			onenand_load_page(io_base_addr, block, page);
			
			// Point back to the oob buffer.  If ALL of its bytes are still
			// zero, the block is truly bad
			//
			data = ONENAND_DATARAM0_ECC_BASE(io_base_addr);
			good = false;
			
			for ( i = 0; (i < oobsize) && !good; i++ )
				good = (0 != data[i]) ? true : false;
				
			// Point back to the page buffer.  If ALL of its bytes are still
			// zero, the block is truly bad.
			//
			if ( !good )
			{
				data = ONENAND_DATARAM0_BASE(io_base_addr);

				for ( i = 0; (i < oobblock) && !good; i++ )
					good = (0 != data[i]) ? true : false;
			}
			
			// A block is truly bad when our attempt to read its bytes resulted
			// in no data coming back from the device (i.e., all the data in the
			// buffers is still zero).
			//
			block_is_truly_bad = !good;
		}
	}
	
	return ( block_is_truly_bad );
}

int onenand_wait_for_int(ulong io_base_addr, ushort int_type, bool check_status)
{
    onenand_regs_t *regs = ONENAND_REG_BASE(io_base_addr);
    ushort pend_int = (ONENAND_INT | int_type);
   	int result = SUCCESS;

    // Wait for pending interrupt bit(s) in ISR to transition.
    //
    while ( pend_int != (regs->interrupt & pend_int) )
    	;
    	
    // When checking status, normalize the <result> on FAILURE.
    //
    if ( check_status )
    	if ( ONENAND_ERROR(regs->controller_status) )
    		result = FAILURE;

    return ( result );
}

static int onenand_issue_command(ulong io_base_addr, ushort command, bool check_status)
{
	onenand_regs_t *regs = ONENAND_REG_BASE(io_base_addr);
	ushort int_type = ONENAND_INT;
	
	// Set the read mode, either synchronous or asynchronous, as appropriate.
	//
	onenand_sync_set(io_base_addr);

	// Clear ISR, determine ISR mask based on <command>, and issue <command>.
	//
	regs->interrupt = ONENAND_CLEAR_ISR;
	regs->command   = command;

	switch ( command )
	{
		case ONENAND_COMMAND_LOAD:
		case ONENAND_COMMAND_LOAD_SPARE:
			int_type = ONENAND_RI;
		break;
		
		case ONENAND_COMMAND_PROGRAM:
		case ONENAND_COMMAND_PROGRAM_SPARE:
		case ONENAND_COMMAND_PROGRAM_COPY:
			int_type = ONENAND_WI;
		break;
		
		case ONENAND_COMMAND_ERASE_VERIFY:
		case ONENAND_COMMAND_ERASE:
		case ONENAND_COMMAND_ERASE_BLOCKS:
		case ONENAND_COMMAND_ERASE_SUSPEND:
		case ONENAND_COMMAND_ERASE_RESUME:
			int_type = ONENAND_EI;
		break;
		
		case ONENAND_COMMAND_RESET_FLASH:
		case ONENAND_COMMAND_RESET_NAND:
			int_type = ONENAND_RSTI;
		break;
		
		case ONENAND_COMMAND_UNLOCK:
		case ONENAND_COMMAND_LOCK:
		case ONENAND_COMMAND_LOCK_TIGHT:
		case ONENAND_COMMAND_OTP_ACCESS:
		break;
	}
	
	// Wait for command's interrupt to clear, check and return
	// status if requested.
	//
	return ( onenand_wait_for_int(io_base_addr, int_type, check_status) );
}

static void onenand_set_addresses(ulong io_base_addr, ulong block, ushort page,
								  ushort sector, bool fba_or_sba)
{
	// Use of the FBA or SBA, it seems, is mutually exclusive.  Set one or the
	// other, but not both.
	//
	switch ( fba_or_sba )
	{
		// Set FBA, respecting DFS as appropriate.
		//
		case ONENAND_USE_FBA:
			ONENAND_SET_FBA(io_base_addr, block);
		break;
		
		// Set SBA, respecting DFS as appropriate.
		//
		case ONENAND_USE_SBA:
			ONENAND_SET_SBA(io_base_addr, block);
		break;
	}

	// Set FPA/FSA of Flash.
	//
	ONENAND_SET_FXA(io_base_addr, page, sector);

	// Set BSA/BSC of DataRAM.
	//
	ONENAND_SET_BSA(io_base_addr, sector, (0 != sector) ? ONENAND_STBUF_BSC_1_SECTOR
														: ONENAND_STBUF_BSC_4_SECTORS);

	// Select DataRAM for DDP.
	//
	ONENAND_SET_DBS(io_base_addr);
}

int onenand_set_protection_state(ulong io_base_addr, ulong block, ushort command)
{
	// Set SBA, respecting DFS as appropriate.
	//
	onenand_set_addresses(io_base_addr, block, ONENAND_FPA_ZERO, ONENAND_FSA_ZERO, ONENAND_USE_SBA);
	
	// Issue protection <command>:  ONENAND_COMMAND_LOCK,
	//                              ONENAND_COMMAND_UNLOCK, or
	//                              ONENAND_COMMAND_LOCK_TIGHT
	//
	// Then, follow the command protocol, noting that setting of the
	// protection states doesn't require a status check.
	//
	return ( onenand_issue_command(io_base_addr, command, ONENAND_NO_STATUS) );
}

int onenand_get_protection_state(ulong io_base_addr, ulong block)
{
	onenand_regs_t *regs = ONENAND_REG_BASE(io_base_addr);
	int protection_state = ONENAND_LOCKED;
	ushort wp_status;

	// Set FBA, respecting DFS as appropriate.
	//
	onenand_set_addresses(io_base_addr, block, ONENAND_FPA_ZERO, ONENAND_FSA_ZERO, ONENAND_USE_FBA);
	
	// Get protection status of specified <block>.
	//
	wp_status = regs->write_protection_status;
	
	// Normalize status to ONENAND_LOCKED or ONENAND_UNLOCKED.
	//
	switch ( wp_status )
	{
		case ONENAND_WP_LOCKED:
		case ONENAND_WP_LOCKED_TIGHT:
			// Assume locked; set above.
		break;
	
		case ONENAND_WP_UNLOCKED:
			protection_state = ONENAND_UNLOCKED;
		break;
		
		default:
			printf("Unexpected OneNAND WP state: %d\n", wp_status);
		break;
	}

	return ( protection_state );
}

int onenand_load_spare(ulong io_base_addr, ulong block, ushort page)
{
	// Set addresses (DFS/FBA, FPA/FSA, BSA/BSC, and DataRAM as appropropriate).
	//
	onenand_set_addresses(io_base_addr, block, page, ONENAND_FSA_ZERO, ONENAND_USE_FBA);

	// Issue the load-spare command, check its status, and return the result.
	//
	return ( onenand_issue_command(io_base_addr, ONENAND_COMMAND_LOAD_SPARE, ONENAND_CHECK_STATUS) );
}

int onenand_write_spare(ulong io_base_addr, ulong block, ushort page)
{
	// Set addresses (DFS/FBA, FPA/FSA, BSA/BSC, and DataRAM as appropropriate).
	//
	onenand_set_addresses(io_base_addr, block, page, ONENAND_FSA_ZERO, ONENAND_USE_FBA);

	// Issue the load-spare command, check its status, and return the result.
	//
	return ( onenand_issue_command(io_base_addr, ONENAND_COMMAND_PROGRAM_SPARE, ONENAND_CHECK_STATUS) );
}

int onenand_load_page(ulong io_base_addr, ulong block, ushort page)
{
	int result = SUCCESS;

	// Set addresses (DFS/FBA, FPA/FSA, BSA/BSC, and DataRAM as appropropriate).
	//
	onenand_set_addresses(io_base_addr, block, page, ONENAND_FSA_ZERO, ONENAND_USE_FBA);

	// Issue the load command and check its status.
	//
	if ( FAILURE == onenand_issue_command(io_base_addr, ONENAND_COMMAND_LOAD, ONENAND_CHECK_STATUS) )
	{
#ifdef NAND_DEBUG
		printf("onenand_load_page failed: block %d, page %d\n", block, page); /* map_out() */
#endif
		result = FAILURE;
	}

	return ( result );
}

int onenand_write_page(ulong io_base_addr, ulong block, ushort page)
{
	int result = SUCCESS;
	
	// Set addresses (DFS/FBA, FPA/FSA, BSA/BSC, and DataRAM as appropropriate).
	//
	onenand_set_addresses(io_base_addr, block, page, ONENAND_FSA_ZERO, ONENAND_USE_FBA);
	
	// Issue the program command and check its status.
	//
	if ( FAILURE == onenand_issue_command(io_base_addr, ONENAND_COMMAND_PROGRAM, ONENAND_CHECK_STATUS) )
	{
#ifdef NAND_DEBUG
		printf("onenand_write_page failed: block %d, page %d\n", block, page); /* map_out() */
#endif		
		result = FAILURE;
	}
	
	return ( result );
}

static int onenand_clean_block(ulong io_base_addr, ulong block, bool verify)
{
	ushort *spare_area = ONENAND_DATARAM0_ECC_BASE(io_base_addr);
	size_t oobsize = nand_rw_get_device_oobsize() >> 1;
	int i, result;

    // First, make sure the spare area is clear.
    //
    onenand_clear_spare_area(io_base_addr);
		
	// Next, apply the JFFS2 cleanmarker format to the spare buffer.
	// 
	//   JFFS2_MAGIC_BITMASK 0x1985 (16 bits)
	//   JFFS2_NODETYPE_CLEANMARKER 0x2003 (16 bits)
	//   length value 0x00000008 (32 bits)
	//
	spare_area[1] = 0x1985;		// offset 2, 3
	spare_area[2] = 0xFF03;		// 4

	// Set addresses (DFS/FBA, FPA/FSA, BSA/BSC, and DataRAM as appropropriate).
	//
	onenand_set_addresses(io_base_addr, block, ONENAND_FPA_ZERO, ONENAND_FSA_ZERO, ONENAND_USE_FBA);
	
	// Issue the program-spare command and check its status.
	//
	result = onenand_issue_command(io_base_addr, ONENAND_COMMAND_PROGRAM_SPARE, ONENAND_CHECK_STATUS);
	
	if ( (SUCCESS == result) && verify )
	{
		// Now, zero out the spare (oob) area to verify it.
		//
		for ( i = 0; i < oobsize; i++ )
			spare_area[i] = 0;
			
		onenand_load_page(io_base_addr, block, ONENAND_FPA_ZERO);
		
		for ( i = 0; (i < oobsize) && (SUCCESS == result); i++ )
		{
			switch ( i )
			{
				// Check for the cleanmarker.
				//
				case 1:
					result = 0x1985 == spare_area[i] ? SUCCESS : FAILURE;
				break;
				
				case 2:
					result = 0xFF03 == spare_area[i] ? SUCCESS : FAILURE;
				break;
				
				// Check for the cleanmarker's ECC.
				//
				case 5:
					result = 0x30FF == spare_area[i] ? SUCCESS : FAILURE;
				break;
				
				// Everything else should be 0xFFFFs.
				//
				default:
					result = 0xFFFF == spare_area[i] ? SUCCESS : FAILURE;
				break;
			}
		}
	}

#ifdef NAND_DEBUG
	if ( FAILURE == result )
		printf("onenand_clean_block failed: block %d\n", block); /* map_out() */
#endif

	return ( result );
}

#define GET_RTC() ((ulong)RCNR)

static char  spinner[]  = { '|', '/', '-', '\\' };
static ulong spin_count = 0,
             spin_rtc   = 0;

static int onenand_verify_erase(ulong io_base_addr, ulong block, bool verify)
{
	int result;
	
	// Set FBA, respecting DFS as appropriate.
	//
	onenand_set_addresses(io_base_addr, block, ONENAND_FPA_ZERO, ONENAND_FSA_ZERO, ONENAND_USE_FBA);
	
	// Issue the erase command.
	//
	result = onenand_issue_command(io_base_addr, ONENAND_COMMAND_ERASE, ONENAND_CHECK_STATUS);
	
	// If asked, verify that all of the pages in this block have been erased.
	//
	if ( verify )
	{
		size_t	blocksize = nand_rw_get_device_block_size(),
				oobblock  = nand_rw_get_device_oobblock(),
				oobsize   = nand_rw_get_device_oobsize();
		
		int		num_pages = blocksize/oobblock,
				page, i;
		
		ushort	*data;
		
		bool	spinner_started = false;
		
		oobsize  /= 2; // Ensure word (16-bit) read/writes.
		oobblock /= 2;
		
		for ( page = 0; (page < num_pages) && (SUCCESS == result); page++ )
		{
			// About once a second, update spinner.
			//
			if ( (0 == page) || (GET_RTC() != spin_rtc) )
			{
				printf("\b%c", spinner[spin_count++ % sizeof(spinner)]);
				spinner_started = true;
				spin_rtc = GET_RTC();
			}
			
			// Point to the oob & page buffers and zero them out.
			//
			data = ONENAND_DATARAM0_ECC_BASE(io_base_addr);
			
			for ( i = 0; i < oobsize; i++ )
				data[i] = 0;
				
			data = ONENAND_DATARAM0_BASE(io_base_addr);
			
			for ( i = 0; i < oobblock; i++ )
				data[i] = 0;
				
			// Load the page into its respective buffers.
			//
			onenand_load_page(io_base_addr, block, page);
			
			// Point back to the oob buffer.  If any of its bytes is not
			// the ONENAND_ERASED_MARKER, the erase failed.
			//
			data = ONENAND_DATARAM0_ECC_BASE(io_base_addr);
			
			for ( i = 0; (i < oobsize) && (SUCCESS == result); i++ )
				result = (ONENAND_ERASED_MARKER == data[i]) ? SUCCESS : FAILURE;
				
			// Point back to the page buffer.  If any of its bytes is not
			// the ONENAND_ERASED_MARKER, the erase failed.
			//
			if ( SUCCESS == result )
			{
				data = ONENAND_DATARAM0_BASE(io_base_addr);
				
				for ( i = 0; (i < oobblock) && (SUCCESS == result); i++ )
					result = (ONENAND_ERASED_MARKER == data[i]) ? SUCCESS : FAILURE;
			}
		}
	
		// Clear spinner if necessary.
		//
		if ( spinner_started )
			puts("\b ");
	}
	
	return ( result );
}

int onenand_erase_block(ulong io_base_addr, ulong block, bool clean)
{
	int erase_attempts = 0,
		result;
		
	bool verify = false;

	// Only go into verification mode if the erase fails.
	//
	do
	{
		result = onenand_verify_erase(io_base_addr, block, verify);
		
		if ( FAILURE == result )
		{
#ifdef NAND_DEBUG
			if ( 0 == erase_attempts )
				putc('\n');
			printf("onenand_erase_block failed: trying again\n");
#endif
			erase_attempts++;
			verify = true;
		}
		
	}
	while ( (FAILURE == result) && (ONENAND_MAX_ERASE_ATTEMPTS > erase_attempts) );
	
	if ( SUCCESS == result )
	{
		// If the erase succeeded and we've been asked to write out the
		// JFFS2 clean marker, then do that now.  Also, if we had to
		// verify the erase, verify the write.
		//
		if ( clean )
			result = onenand_clean_block(io_base_addr, block, verify);
	}
	else
	{
#ifdef NAND_DEBUG
		printf("onenand_erase_block failed: block %d\n", block); /* map_out() */
#endif		
	}
	
	// Remember whether we had to verify this erase or not.
	//
	last_erase_required_verify = verify;
	
	return ( result );
}

static int onenand_program_page(struct nand_chip *nand, size_t ofs)
{
    ushort	*data_area = ONENAND_DATARAM0_BASE(nand->IO_ADDR);
    ushort	*w_src;
    int		i, cnt;

    // First, make sure the spare area is clear.
    //
    onenand_clear_spare_area(nand->IO_ADDR);

    // Next, copy the page data into device's buffer.
    //
    w_src = (ushort *)nand->data_buf;
    cnt = nand->oobblock >> 1;
    for (i = 0; i < cnt; i++)
		data_area[i] = w_src[i];

	// Write the page from the buffer into flash per standard protocol.
	//
	return ( onenand_write_page(nand->IO_ADDR, NAND_BLOCK(ofs), NAND_PROGRAM_PAGE(ofs)) );
}

static int nand_read_ecc(struct nand_chip *nand, size_t start, size_t len,
	size_t * retlen, u_char *buf, u_char *ecc_code)
{
	int	col;
	int	ecc_status = 0;
	int	pagesize = nand->oobblock;
	u_char *dataram0 = (u_char *) ONENAND_DATARAM0_BASE(nand->IO_ADDR);

	*retlen = 0;		// in case of failure

	/* Do not allow reads past end of device */
	if ((start + len) > nand->totlen) {
		printf("%s: Attempted read beyond end of device ", __FUNCTION__);
		printf("(start=0x%x, len=0x%x, totlen=0x%x)\n", start, len, nand->totlen);
		return -1;
	}

	col = start & (pagesize - 1);	// raw starting column

	/* Loop until all data is read */
	while (*retlen < len) {
		onenand_load_page(nand->IO_ADDR, NAND_BLOCK(start), NAND_LOAD_PAGE(start));

		if ((*retlen + (pagesize - col)) >= len) {
			memcpy(buf + *retlen, dataram0 + col, len - *retlen);
			*retlen = len;
			continue;	// we're done
		} else {
			memcpy(buf + *retlen, dataram0 + col, pagesize - col);
			*retlen += pagesize - col;
			start	+= pagesize - col;
		}
		col = 0;  // subsequent reads align to page boundary
	}

	return ecc_status ? -2 : 0;
}

static int nand_write_ecc (struct nand_chip* nand, size_t to, size_t len,
			   size_t * retlen, const u_char * buf, u_char * ecc_code)
{
	int	col;
	int	pagesize = nand->oobblock;
	int	i, page, cnt, err, ret = 0;

	*retlen = 0;		// in case of failure

#if 1	// FIXME
	if ((to | len) & (pagesize-1)) {
	    printf("%s: temporarily require len/offset to be page aligned\n",
			__FUNCTION__);
	    return -5;
	}
#endif

	/* Do not allow writes past end of device */
	if ((to + len) > nand->totlen) {
		printf("%s: Attempted write past end of device ", __FUNCTION__);
		printf("(to=0x%x), len=0x%x, totlen=0x%x)\n",
			to, len, nand->totlen);
		return -1;
	}

	page = ((int) to) >> nand->page_shift;	// shift to get page

	col = to & (pagesize - 1);		// raw starting column

	/* FIXME Check if device is write-protected ??? */

	/* Loop until all data is read */
	while (*retlen < len) {
	    if ((col + len) >= pagesize) {
	    	for (i = 0 ; i < col ; i++)	// preserve existing data
		    nand->data_buf[i] = 0xFF;
		for (i = col, cnt = 0 ; i < pagesize ; i++, cnt++)
		    nand->data_buf[i] = buf[(*retlen + cnt)];
	    } else {
	    	for (i = 0 ; i < col ; i++)
		    nand->data_buf[i] = 0xFF;
	    	for (i = pagesize - ((len - *retlen)+col) ; i < pagesize ; i++)
		    nand->data_buf[i] = 0xFF;
		for (i = col, cnt = 0 ; cnt < (len - *retlen); i++, cnt++)
		    nand->data_buf[i] = buf[(*retlen + cnt)];
	    }

	    err = onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(to + *retlen), ONENAND_COMMAND_UNLOCK);
	    if (err) {
		printf("%s: failed to unprotect block at 0x%x\n",
			__FUNCTION__, to + *retlen);
	    }

	    ret = onenand_program_page(nand, to + *retlen);
	
	    err = onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(to + *retlen), ONENAND_COMMAND_LOCK);
	    if (err) {
		printf("%s: failed to protect block at 0x%x\n",
			__FUNCTION__, to + *retlen);
	    }

	    if (ret)
		goto out;

	    col = 0;		// Next transfer starts on page boundary

	    *retlen += cnt;

	    page++;
	}
	*retlen = len;
out:
	return ret;
}

static int nand_read_oob(struct nand_chip* nand, size_t ofs, size_t len,
			 size_t *retlen, u_char *buf)
{
	int result = FAILURE;
	*retlen = 0;

	// Load the spare data from the requested block.
	//
	if ( FAILURE == onenand_load_spare(nand->IO_ADDR, NAND_BLOCK(ofs), ONENAND_FPA_ZERO) )
		printf("nand_read_oob: failed to load spare data!!!\n");
	else
	{
		ushort	*spare_area = ONENAND_DATARAM0_ECC_BASE(nand->IO_ADDR);
		u_char	*src = (u_char *)spare_area;
		
		// Copy the data from spare area into the caller's buffer.
		//
		src += ofs & (nand->oobblock - 1);
		memcpy(buf, src, len);
		*retlen = len;
		
		result = SUCCESS;
	}

	return ( result );
}

static int nand_write_oob(struct nand_chip *nand, size_t ofs, size_t len,
		  size_t *retlen, const u_char *buf)
{
	ushort	*spare_area = ONENAND_DATARAM0_ECC_BASE(nand->IO_ADDR);
	int		i, result = FAILURE;
	ushort	*word_ptr;
	u_char	*dst;
	char	spare_img[nand->oobsize];
	
	// Assume failure.
	//
	*retlen = 0;

	// Set the <spare_image> to all 0xFFs and then fill it in with caller's
	// data.
	//
	memset(spare_img, 0xFF, nand->oobsize);
	
	dst  = (u_char *)spare_img;
	dst += ofs & (nand->oobblock - 1);
	memcpy(dst, buf, len);

	word_ptr = (ushort *)spare_img;

	// Now, fill the spare_area with caller's data.
	//
	for (i = 0 ; i < nand->oobsize/2 ; i++)
		spare_area[i] = word_ptr[i];

	// Write the spare data to the requested block.
	//
	if ( SUCCESS == onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(ofs), ONENAND_COMMAND_UNLOCK) )
	{
		if ( SUCCESS == onenand_write_spare(nand->IO_ADDR, NAND_BLOCK(ofs), ONENAND_FPA_ZERO ) )
		{
			*retlen = len;
			result = SUCCESS;
		}
		
		onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(ofs), ONENAND_COMMAND_LOCK);
	}

	if ( FAILURE == result )
		printf("nand_write_oob: failed to write spare data!!!\n");

	return ( result );
}

static int nand_mark_oob(struct nand_chip *nand, size_t ofs, bool mark_bad)
{
	onenand_spare_t spare_buffer;
	int result = FAILURE;
	size_t retlen = 0;
	
	// Get the specified oob's current data.
	//
	if ( SUCCESS == nand_read_oob(nand, ofs, sizeof(spare_buffer), &retlen, (char *)&spare_buffer) )
	{
		// Mark it either bad or good if we can.
		//
		if ( mark_bad )
		{
			// If it's not already bad, mark it bad in a way that we
			// can recover from.  Otherwise, just leave it alone.
			//
			// Note:  We're just looking at the bad block marker on
			//		  page zero
			//
			if ( ONENAND_ERASED_MARKER == spare_buffer.bi )
			{
				spare_buffer.bi = ONENAND_FAKE_BAD_MARKER;
				result = nand_write_oob(nand, ofs, sizeof(spare_buffer), &retlen, (char *)&spare_buffer);
			}
		}
		else
		{
			// If it's marked bad by us, recover it by erasing the block.
			//
			if ( ONENAND_FAKE_BAD_MARKER == spare_buffer.bi )
			{
				if ( SUCCESS == onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(ofs), ONENAND_COMMAND_UNLOCK) )
					result = onenand_erase_block(nand->IO_ADDR, NAND_BLOCK(ofs), ONENAND_ERASE_RAW);

				onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(ofs), ONENAND_COMMAND_LOCK);
			}
		}
	}
	
	return ( result );
}

static void check_block_erase(struct nand_chip *nand, size_t ofs, int clean, bool verify)
{
	if (ALLOW_ERASE_BAD_DEBUG || (verify && !check_block_verify(nand, ofs)) || !check_block(nand, ofs)) {
		int err = onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(ofs), ONENAND_COMMAND_UNLOCK);
		if (err) {
			printf("%s: failed to unprotect block at 0x%x\n",
				__FUNCTION__, ofs);
		}

		err = onenand_erase_block(nand->IO_ADDR, NAND_BLOCK(ofs), clean ? ONENAND_ERASE_JFFS2 : ONENAND_ERASE_RAW);
		if (err) {
			printf("%s: failed to erase block at 0x%x\n",
				__FUNCTION__, ofs);
		}
		
		err = onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(ofs), ONENAND_COMMAND_LOCK);
		if (err) {
			printf("%s: failed to protect block at 0x%x\n",
				__FUNCTION__, ofs);
		}
	}
}

static int nand_erase(struct nand_chip *nand, size_t ofs, size_t len, int clean, bool verify)
{
	if (ofs & (nand->erasesize-1) || len & (nand->erasesize-1)) {
		printf("Offset/length must be block aligned (0x%08X)\n",
			nand->erasesize);
		return 1;
	}

	while (len) {
	    check_block_erase(nand, ofs, clean, verify);

	    ofs += nand->erasesize;
	    len -= nand->erasesize;

        // Don't let code walk off of onenand space in case where it 
        // can't find enough good blocks !! it will wrap to 0, and 
        // then start writing over bootloaders !!
        if ((ofs >= nand->totlen) && len) {
            printf("ERROR: ofs = 0x%x, len = 0x%x, erasesize = 0x%x\n",ofs,len,nand->erasesize);
            return 1;
        }
	}

	return 0;
}

static int nand_onenand_scan(struct nand_chip *nand)
{
	onenand_regs_t	*regs = ONENAND_REG_BASE(nand->IO_ADDR);
	unsigned short	manu_id, dev_id;

	// FIXME:  By the time this code runs, we've already run flash_init()
	// in flash.c, and it has performed the same check.  So, we should be
	// able to skip it here (we'd need a way for flash.c to cache this
	// information, perhaps in the onenand_info[] array).
	//
	regs->manufacturer_id = 0xFFFF;
	
	manu_id = regs->manufacturer_id;
	dev_id  = regs->device_id;

	if ( ONENAND_SAMSUNG_128MBx2(manu_id, dev_id) ||
		 ONENAND_SAMSUNG_256MBx1(manu_id, dev_id) )
	{
		nand->page_shift	= 9;
		nand->oobblock		= 2048;
		nand->oobsize		= 64;
		nand->data_buf		= malloc(nand->oobblock + nand->oobsize);
		
		nand->data_cache	= NULL;
		nand->chipshift		= 27;
		nand->erasesize		= PHYS_FLASH_SECT_SIZE;
		nand->mfr			= manu_id;
		nand->id			= dev_id;
		nand->name			= NULL;
		nand->numchips		= 1;
		nand->page256		= 0;
		nand->pageadrlen	= (4-1);
		nand->eccsize		= 3;
	
		switch ( dev_id )
		{
			case SAMSUNG_DEVICE_ID_128MBx2:
				nand->chips_name = "Samsung 256MB OneNAND (stacked)";
			break;
			
			case SAMSUNG_DEVICE_ID_256MBx1:
				nand->chips_name = "Samsung 256MB OneNAND (monolithic)";
			break;
		}

		// We're treating the 256MB parts as though they
		// were two 128MB parts.  This way, the low-level
		// code is the same in both places.  We bottleneck the
		// particulars (bank crossing and such) higher up.
		//
		nand->totlen = PHYS_FLASH_SECT_SIZE * CFG_MAX_FLASH_SECT;
		
		// It's unlikely that we couldn't get memory, so we
		// put it at the end.
		//
		if (!nand->data_buf)
			puts ("Cannot allocate memory for buffer.\n");
		else
			nand->ChipID = dev_id & 0xFF; // SUCCESS!
	}
	else
		printf("Unknown flash chip found: manufacturer = 0x%02X, "
			   "device = 0x%02X\n", manu_id, dev_id);
	
	return ( (nand->ChipID != NAND_ChipID_UNKNOWN) ? SUCCESS : FAILURE );
}

unsigned long nand_probe(unsigned long physaddr)
{
	int bank;
	
	oob_config.badblock_pos =  0;
	oob_config.eccvalid_pos = -1;

	// Go build up our nand_dev_desc array.
	//
	for ( bank = 0; bank < CFG_MAX_NAND_DEVICE; bank++ )
	{
		// Assume our scan will fail, but prep the IO_ADDR so that we
		// can at least perform the scan; ChipID will get filled out
		// appropriately if the scan succeeds.
		//
		nand_dev_desc[bank].ChipID     = NAND_ChipID_UNKNOWN;
		nand_dev_desc[bank].IO_ADDR    = bank_io_base[bank];
		nand_dev_desc[bank].cache_page = -1;
		
		if ( SUCCESS == nand_onenand_scan(&nand_dev_desc[bank]) )
		{
			// Except for bank 0, the (logical) base address of THIS
			// bank is the current length.  Otherwise, it's <physaddr>.
			//
			if ( bank > 0 )
				nand_dev_desc[bank].BANK_ADDR = nand_dev_len;
			else
				nand_dev_desc[bank].BANK_ADDR = physaddr;

			// Update our total length.
			//
			nand_dev_len += nand_dev_desc[bank].totlen;
			
			// Set curr_device to our first successful bank scanned.
			//
			if ( (NO_CURR_DEVICE == curr_device) || (CFG_MAX_NAND_DEVICE < curr_device) )
			{
				curr_device = bank;

				// For stitching, cache which bank is at the lowest base
				// address and remember which bank that is.
				//
				// FIXME?  We assume that various things in this bank are
				//         representative of all of the banks.
				//
				nand_dev_block_size = nand_dev_desc[bank].erasesize;
				nand_dev_oobblock   = nand_dev_desc[bank].oobblock;
				nand_dev_oobsize    = nand_dev_desc[bank].oobsize;
				nand_dev_start      = nand_dev_desc[bank].BANK_ADDR;
				nand_dev_base_bank  = bank;

				nand_dev_bb_size    = onenand_get_boot_buffer_size(nand_dev_desc[bank].IO_ADDR);
			}
			
			// FIXME:  Why is this here?  It does nothing.
			//
			NanD_WaitReady(&nand_dev_desc[bank], 0);
		}
	}
	
	// Initialize the sync-read setting.
	//
	onenand_sync_init();
	
	// Done.
	//
	return ( nand_dev_len );
}

static int NanD_WaitReady(struct nand_chip *nand, int ale_wait)
{
	return 0;
}

/* Dump all the OneNAND data - formatted as follows:

# 0x00000000   0 0  <== offset, block number, page number
# FFFF851903FF6789 34FFFFFFFFFFFFFFFFFF 851903FF6789 34FFFFFFFFFFFFFF	<== oob info for page
# FFFF851903FF6789 34FFFFFFFFFFFFFFFFFF 851903FF6789 34FFFFFFFFFFFFFF
00000000: FFFF851903FF6789 34FFFFFFFFFFFFFF FFFF851903FF6789 34FFFFFFFFFFFFFF
00000020: 1234123412341234 1234123412341234 1234123412341234 1234123412341234
...
...

 */

int nand_dump(struct nand_chip *nand, unsigned long ofs, unsigned long len)
{
	unsigned long	erasesize, pagesize;
	char			*p;
	int			    i;

	if ((ofs | len) & (nand->oobblock - 1)) {
		printf("dump: offset/length must be multiples of page size(0x%X).\n",
			nand->oobblock);
		return 1;
	}
	if ((ofs + len) > nand->totlen) {
		printf("dump: offset + length (0x%08X) out of range.\n",
			ofs + len);
		return 1;
	}

	erasesize  = nand->erasesize;
	pagesize   = nand->oobblock;

	while (len) {

		// dump header

		printf("# 0x%08X %4d %2d\n", ofs, ofs/erasesize,
				(ofs & (erasesize - 1))/pagesize);

		// load the next page
		onenand_load_page(nand->IO_ADDR, NAND_BLOCK(ofs), NAND_LOAD_PAGE(ofs));

		// dump oob info
		p = (char *) ONENAND_DATARAM0_ECC_BASE(nand->IO_ADDR);
		
		for (i = 0 ; i < nand->oobsize ; i++) {
			if ((i % 32) == 0)
				printf("#");
			if ((i % 8) == 0)
				printf(" ");
			printf("%02X", p[i]);
			if ((i % 32) == 31)
				printf("\n");
		}

		// dump page data

		p = (char *) ONENAND_DATARAM0_BASE(nand->IO_ADDR);
		
		for (i = 0 ; i < nand->oobblock ; i++) {
			if ((i % 32) == 0)
				printf("%08X:", ofs + i);
			if ((i % 8) == 0)
				printf(" ");
			printf("%02X", p[i]);
			if ((i % 32) == 31)
				printf("\n");
		}

		// update ptrs
		
		ofs += nand->oobblock;
		len -= nand->oobblock;
	}

	return 0;
}

// Entry:
//
// - nand_time(), attempts to time writes to <num_mb>s of flash on the NAND
//   device specified by <which_dev>.
//
// Utils:  
//
// - nand_time_timer_to_secs(), converts OSCR ticks to seconds and deciseconds
//   or to milliseconds.
//
// - nand_time_diff_ticks(), returns the delta between two OSCR ticks, fixing
//   the delta if the ticks have wrapped (with the timings we're doing, we assume
//   that the wrapping won't have occurred more than once).
//
// - nand_time_get_ticks(), returns the raw OSCR (OS Timer Count Regiser)
//   ticks.
// 
// - nand_time_prime_timer(), nand_start_timer(), nand_stop_timer(), support
//   timing operations in ticks.
//
// - nand_time_fill_buffer(), fills our write buffer, which is one block size,
//   with known (computed) data.
//
// - nand_time_check_buffer(), checks our buffer after a read to see whether
//   it has the data we expect in it.
//
// - nand_time_print_results(), prints out the results (times) for each of the
//   write, read, and erase operations.
//
#define NAND_TIME_SHIFT_MB			20			// 2^20 is 1MB.

#define NAND_TIME_END_TICKS			0xFFFFFFFF	// End of unsigned long ticks.
#define NAND_TIME_SECS				CFG_HZ		// 3.6864MHz clock (~19:25).

#define NAND_TIME_DECISECS			(NAND_TIME_SECS / 100)
#define NAND_TIME_MINS				(NAND_TIME_SECS * 60)
#define NAND_TIME_ROLL_TICKS		(NAND_TIME_MINS * 5)
#define NAND_TIME_ROLL_SECS			(60 * 5)

#define NAND_TIME_BLOCKS_PER_ROW	64			// 64 blocks per row.
#define NAND_TIME_BLOCKS_PER_SPACE	8   		//  8 blocks per section.

#define NAND_TIME_BUFFER_OKAY		0			// False is okay.
#define NAND_TIME_BUFFER_BAD		1			// True is bad.

struct nand_timer_t
{
	unsigned long	count,						// Count of NAND_TIME_ROLL_TICKS.
					ticks,						// Current elapsed ticks.
					start_ticks,				// Start, or re-start, ticks.
					stop_ticks;					// Stop ticks.
};
typedef struct nand_timer_t nand_timer_t;

static void nand_time_timer_to_secs(nand_timer_t  *timer,
									unsigned long *secs,
									unsigned long *decisecs,
									unsigned long *millisecs)
{
	// Clear all the return values first for good measure.
	//
	*secs = *decisecs = *millisecs = 0;
	
	if ( timer )
	{
		// Use seconds if we're in the seconds range; otherwise,
		// use milliseconds.
		//		
		if ( timer->count || (NAND_TIME_SECS <= timer->ticks) )
		{
			// Compute seconds to the 100ths.  (FIXME:  Should round; we're just trunc'ing.)
			//
			*secs	   = timer->ticks/NAND_TIME_SECS;
			*decisecs  = timer->ticks/NAND_TIME_DECISECS - (*secs * 100);
			
			if ( timer->count )
				*secs += timer->count * NAND_TIME_ROLL_SECS;
		}
		else
			*millisecs = (timer->ticks * 1000)/NAND_TIME_SECS;
	}
}

static unsigned long nand_time_diff_ticks(unsigned long start_ticks,
										  unsigned long end_ticks)
{
	unsigned long diff_ticks;
	
	// Fix the difference if the clock has wrapped (i.e., <end_ticks> is
	// is less than <start_ticks>).  We assume, of course, that multiple
	// wraps have not occurred; and we mitigate multiple wraps from occurring
	// by making NAND_TIME_ROLL_TICKS relatively small.
	//
	if ( end_ticks < start_ticks )
		diff_ticks = (NAND_TIME_END_TICKS - start_ticks) + end_ticks;
	else
		diff_ticks = end_ticks - start_ticks;
	
	return ( diff_ticks );
}

static unsigned long nand_time_get_ticks(void)
{
	// Return raw OSCR ticks.
	//
	return ( (unsigned long)OSCR );
}

static void nand_time_prime_timer(nand_timer_t *timer)
{
	// Prime the timer.
	//
	if ( timer )
		timer->count = timer->ticks = timer->start_ticks = timer->stop_ticks = 0;
}

static void nand_time_start_timer(nand_timer_t *timer)
{
	// Start (or restart) the timer.
	//
	if ( timer )
		timer->start_ticks = nand_time_get_ticks();
}

static void nand_time_stop_timer(nand_timer_t *timer)
{
	// Stop the timer and update things as appropriate.
	//
	if ( timer )
	{
		timer->stop_ticks = nand_time_get_ticks();
		
		timer->ticks += nand_time_diff_ticks(timer->start_ticks, timer->stop_ticks);
		
		if ( NAND_TIME_ROLL_TICKS < timer->ticks )
		{
			timer->ticks -= NAND_TIME_ROLL_TICKS;
			timer->count++;
		}
	}
}

static unsigned long *nand_time_fill_buffer(int block_num, int block_size)
{
	unsigned long *mem_start = (unsigned long *)CFG_MEMTEST_START;
	int i, write_count = block_size/sizeof(unsigned long);
	
	for (i = 0; i < write_count; i += 2 )
	{
		mem_start[i+0] = block_num;
		mem_start[i+1] = (unsigned long)i;
	}
	
	return ( mem_start );
}

static int nand_time_check_buffer(int block_num, int block_size,
								  unsigned long *buffer_start)
{
	int read_count = block_size/sizeof(unsigned long),
		i, buffer_status = NAND_TIME_BUFFER_OKAY;
	
	for (i = 0; (i < read_count) && (NAND_TIME_BUFFER_OKAY == buffer_status); i += 2)
	{
		if ( (buffer_start[i] != block_num) || (buffer_start[i+1] != (unsigned long)i) )
			buffer_status = NAND_TIME_BUFFER_BAD;
	}
	
	return ( buffer_status );
}

static void nand_time_print_results(nand_timer_t *which_timer, char *which_string, 
									unsigned long mb)
{
	unsigned long secs, decisecs, millisecs, mbps, decimbps;
	
	printf("%s time:  ", which_string);
	
	nand_time_timer_to_secs(which_timer, &secs, &decisecs, &millisecs);
	
	// If everything is zero, we'll end up printing "0.00 s", and
	// that's okay.
	//
	if ( millisecs )
		printf("%d ms", millisecs);
	else
	{
		// FIXME:  Convert to mm:ss.dd when seconds are in the minute
		//         range.
		//
		printf("%d.%02d s", secs, decisecs);
		
		if ( secs || decisecs )
		{
			// Compute MB/s to 100ths.  (FIXME:  Should round; we're just trunc'ing.)
			//
			decisecs = (secs * 100) + decisecs;
			
			mbps	 = (mb * 100)/decisecs;
			decimbps = (mb * 10000)/decisecs - (mbps * 100);
			
			printf(", %d.%02d MB/s", mbps, decimbps);
		}
	}
	
	printf("\n");
}

static int nand_time(int num_mb, int max_mb)
{
	// Assume failure.
	//
	int result = FAILURE;

	// Note:  nand_time() is the test bed for device/bank independence in that
	//        it no longer uses the nand_chip structure directly.  Rather, it
	//        just calls nand_rw_dispatch() and its utility routines and lets
	//        them handle all the gory details.
	//
	//        Actually, since nand_time()'s nature is destructive, it does
	//        handle one gory detail itself:  It protects the root image area.
	//
	size_t device_start = nand_rw_get_device_start(),
	       device_len   = nand_rw_get_device_len();
	       	
	// Convert to bytes since everything else is in bytes.
	//
	int num_b = num_mb << NAND_TIME_SHIFT_MB,
	    max_b = max_mb << NAND_TIME_SHIFT_MB;

	// Only limit the device's length it makes sense to do so.
	//
	if ( max_b && (device_len > max_b) )
		device_len = max_b;
	
	if ( (0 < num_b) && (device_len >= num_b) )
	{
		nand_timer_t erase_timer, write_timer, read_timer;
			
		u_char	*buffer_start;
		size_t	actual_bytes_buffered;
		
		int	block_size = nand_rw_get_device_block_size(),
			read_write_status,
			blocks_to_write,
			total_blocks,
			block_count,
			block_num;
		
		// We have 0..<total_blocks>-1 we can write to.
		//
		total_blocks = device_len/block_size;
		
		// We have <blocks_to_write> blocks to write.
		//
		blocks_to_write = num_b/block_size;
		
		// Protect the root image area since this routine does destructive
		// writes.
		//
		// FIXME?  Should we read a block into RAM, save it, do our timed
		//         write, and then restore the block we just timed?
		//
		if ( 0 == device_start )
		{
			int num_blocks_to_protect = FLASH_ROOTIMAGE_BASE/block_size;
			
			if ( blocks_to_write > (total_blocks - num_blocks_to_protect) )
				blocks_to_write -= num_blocks_to_protect;
		}

		// Start writing at the highest block possible to avoid overwriting
		// anything of interest.  But print the header and erase the blocks
		// we'll be writing to first.
		//
		printf("Blocks total:  %d (", total_blocks);
			print_size(device_len, "); ");
		printf("blocks to write & time:  %d (", blocks_to_write);
			print_size(num_b, ").\n");
		
		// Prime all of our timers.
		//
		nand_time_prime_timer(&erase_timer);
		nand_time_prime_timer(&write_timer);
		nand_time_prime_timer(&read_timer);
		
		nand_time_start_timer(&erase_timer);
			nand_rw_dispatch(((total_blocks - blocks_to_write) * block_size),
				(blocks_to_write * block_size), NANDRW_ERASE_RAW, NULL, NULL,
				NANDRW_QUIET);
		nand_time_stop_timer(&erase_timer);
		
		for ( block_num = (total_blocks - blocks_to_write), block_count = 0;
			  block_num < total_blocks;
			  block_num++, block_count++ )
		{
			// Write out the block number at the beginning of each row,
			// and space each row out neatly.
			//
			if ( 0 == (block_count % NAND_TIME_BLOCKS_PER_ROW) )
				printf("%4d:", block_num);
			if ( 0 == (block_count % NAND_TIME_BLOCKS_PER_SPACE) )
				putc(' ');
		
			// Fill our buffer, write it to flash, and indicate if it failed
			// by writing a 'W' (bad write).
			//
			buffer_start = (u_char *)nand_time_fill_buffer(block_num, block_size);
			
			nand_time_start_timer(&write_timer);			
				read_write_status = nand_rw_dispatch((block_num * block_size),
					block_size, NANDRW_WRITE, &actual_bytes_buffered, buffer_start,
					NANDRW_QUIET);
			nand_time_stop_timer(&write_timer);
			
			if ( FAILURE == read_write_status )
				putc('W');
			else
			{
				// Clear our buffer, read the flash back into it, and indicate
				// if it failed by writing an 'R' (bad read).
				//
				memset(buffer_start, 0, block_size);
				actual_bytes_buffered = 0;
				
				nand_time_start_timer(&read_timer);
					read_write_status = nand_rw_dispatch((block_num * block_size),
						block_size, NANDRW_READ, &actual_bytes_buffered, buffer_start,
						NANDRW_QUIET);
				nand_time_stop_timer(&read_timer);
				
				if ( FAILURE == read_write_status )
					putc('R');
				else
				{
					// Write an 'I' (incomplete) if our buffer wasn't filled.
					//
					if ( actual_bytes_buffered != block_size )
						putc('I');
					else
					{
						// Write a 'V' (verify failed) if our buffer doesn't
						// have the data in it that it should.
						//
						if ( nand_time_check_buffer(block_num, block_size,
								(unsigned long *)buffer_start) )
						{
							putc('V');
						}
						else
						{
							// Otherwise, write a '.' to say that all is well.
							//
							putc('.');
						}
					}
				}
			}

			// Either we're at the end of a particular row or we're done
			// entirely.  Either way, end the row with a newline.
			//
			if ( ((NAND_TIME_BLOCKS_PER_ROW - 1) == (block_count % NAND_TIME_BLOCKS_PER_ROW)) ||
				 ((blocks_to_write - 1) == block_count) )
				putc('\n');
		}

		// Finish up by printing out the results.
		//
		nand_time_print_results(&write_timer, "Write", num_mb);
		nand_time_print_results(&read_timer,  "Read ", num_mb);
		nand_time_print_results(&erase_timer, "Erase", num_mb);

		// Say that, at the very least, we were able to get in here.
		//
		result = SUCCESS;
	}
    
    return ( result );
}

int do_otp (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
    struct nand_chip	*nand = &nand_dev_desc[curr_device];
    int					rcode = 1;	// parse error result

    switch (argc) {
    case 3:
    	if (strncmp(argv[1], "lock", 4) == 0 &&
	    strncmp(argv[2], "forever", 7) == 0) {
		rcode = onenand_lock_otp_forever(nand);
		if (rcode) {
			printf("ERROR\n");
		}
	}
	break;
    case 4:
	if (strncmp(argv[1], "dump", 4) == 0) {
		ulong offset = simple_strtoul(argv[2], NULL, 16);
		ulong length = simple_strtoul(argv[3], NULL, 16);

		// check for alignment and "offset+length" inside OTP range

		if ((offset | length) & (nand->oobblock - 1)) {
			printf("otp: offset/length must be multiples of "
				"page size (0x%X).\n", nand->oobblock);
			return 3;
		}

		if ((offset + length) > nand->erasesize) {
			printf("otp: offset + length (0x%08X) out of range.\n",
					offset+length);
			return 3;
		}

		rcode = onenand_enter_otp_mode(nand);
		if (rcode) {
			printf("ERROR entering OneNAND OTP mode!\n");
			return 4;
		}

		rcode = nand_dump(nand, offset, length);

		if (rcode)
			printf("ERROR\n");

		if (onenand_hot_reset(nand, ONENAND_COMMAND_RESET_FLASH))
			printf("ERROR performing OneNAND hot reset!\n");
	}
	break;
    case 5:
    	if (strncmp(argv[1], "read", 4) == 0 ||
	    strncmp(argv[1], "write", 5) == 0) {
		ulong addr = simple_strtoul(argv[2], NULL, 16);
		ulong off  = simple_strtoul(argv[3], NULL, 16);
		ulong size = simple_strtoul(argv[4], NULL, 16);
		int cmd    = (strncmp(argv[1], "read", 4) == 0) ?
				NANDRW_READ : NANDRW_WRITE;
		int ret, total;

		// check for alignment and "offset+length" inside OTP range

		if ((off | size) & (nand->oobblock - 1)) {
			printf("otp: offset/length must be multiples of "
				"page size (0x%X).\n", nand->oobblock);
			return 3;
		}

		if ((off + size) > nand->erasesize) {
			printf("otp: offset + length (0x%08X) out of range.\n",
				off+size);
			return 3;
		}

		rcode = onenand_enter_otp_mode(nand);
		if (rcode) {
			printf("ERROR entering OneNAND OTP mode!\n");
			return 4;
		}

		printf ("\notp %s: device %d offset 0x%x, size 0x%x ... ",
			(cmd & NANDRW_READ) ? "read" : "write",
			curr_device, off, size);

		ret = NAND_RW(NAND_DEV_START(curr_device, off), size, cmd, &total, (u_char*)addr);

		printf (" %d bytes %s: %s\n", total,
			(cmd & NANDRW_READ) ? "read" : "written",
			ret ? "ERROR" : "OK");

		if (onenand_hot_reset(nand, ONENAND_COMMAND_RESET_FLASH))
			printf("ERROR performing OneNAND hot reset!\n");

		return ret;
	}
	break;
    default:
	printf ("Usage:\n%s\n", cmdtp->usage);
	rcode = 1;
	break;
    }
    return rcode;
}

U_BOOT_CMD(
	otp,	5,	0,	do_otp,
	"otp     - dump/read/write OneNAND OTP\n",
	"read  addr offset length\n"
	"otp write addr offset length - read/write `length' bytes starting\n"
	"    at offset `offset' to/from memory address `addr'\n"
	"otp dump offset length - dump OTP contents\n"
	"    (offset/length must be multiples of page size)\n"
	"otp lock forever - lock OTP contents forever (can never be altered)\n"
);

static bool onenand_otp_locked(struct nand_chip *nand)
{
	onenand_regs_t *regs = ONENAND_REG_BASE(nand->IO_ADDR);
	onenand_hot_reset(nand, ONENAND_COMMAND_RESET_FLASH);
	return ( ONENAND_OTPL(regs->controller_status) ? true : false );
}

bool serial_number_locked(void)
{
	struct nand_chip *nand = &nand_dev_desc[0];
	return ( onenand_otp_locked(nand) );
}

int onenand_enter_otp_mode(struct nand_chip *nand)
{
	// Set FBA, respecting DFS as appropriate.
	//
	onenand_set_addresses(nand->IO_ADDR, ONENAND_FBA_ZERO, ONENAND_FPA_ZERO, ONENAND_FSA_ZERO, ONENAND_USE_FBA);
	
	// Issue the OTP-access command.
	//
	return ( onenand_issue_command(nand->IO_ADDR, ONENAND_COMMAND_OTP_ACCESS, ONENAND_CHECK_STATUS) );
}

int onenand_hot_reset(struct nand_chip *nand, ushort command)
{
	// Set FBA, respecting DFS as appropriate.
	//
	onenand_set_addresses(nand->IO_ADDR, ONENAND_FBA_ZERO, ONENAND_FPA_ZERO, ONENAND_FSA_ZERO, ONENAND_USE_FBA);
	
	// Issue the OTP-reset command.
	//
	return ( onenand_issue_command(nand->IO_ADDR, command, ONENAND_CHECK_STATUS) );
}

static int onenand_is_sector_empty(ushort *w_addr)
{
	ushort value = *w_addr;
	int	i, cnt = 512 / sizeof(ushort);

	for (i = 0 ; i < cnt ; i++)
		value &= w_addr[i];

	if (value != ONENAND_ERASED_MARKER)
		return 0;
	else
		return 1;
}

static int onenand_program_otp(struct nand_chip *nand, size_t ofs, ushort command)
{
    ushort	*data_area = (ONENAND_COMMAND_PROGRAM == command) ? ONENAND_DATARAM0_BASE(nand->IO_ADDR) :
    															ONENAND_DATARAM0_ECC_BASE(nand->IO_ADDR);
    ushort	*w_src;
    ushort	block  = ofs / nand->erasesize;
    ushort	page   = (ofs / nand->oobblock) & (64-1);
    ushort	sector = (ofs / 512) & 0x0003;
    int		i, cnt;

    // First, make sure the spare area is clear.
    //
    onenand_clear_spare_area(nand->IO_ADDR);

    // Data to be programmed is located in <nand->data_buf>.
    //
    w_src = (ushort *)nand->data_buf;
    cnt = nand->oobblock >> 1;
    for (i = 0 ; i < cnt ; i++)
		data_area[i] = w_src[i];

	// Set addresses (DFS/FBA, FPA/FSA, BSA/BSC, and DataRAM as appropropriate).
	//
	onenand_set_addresses(nand->IO_ADDR, block, page, sector, ONENAND_USE_FBA);

	// Issue <command: ONENAND_COMMAND_PROGRAM | ONENAND_COMMAND_PROGRAM_SPARE> and check its status.
	//
    return ( onenand_issue_command(nand->IO_ADDR, command, ONENAND_CHECK_STATUS) );
}

static int onenand_lock_otp_forever(struct nand_chip *nand)
{
	int	err, rcode = 0;
	
	if (!onenand_otp_locked(nand)) {
		memset(nand->data_buf, 0xFF, nand->oobblock);
		nand->data_buf[14] = 0xCC;
		nand->data_buf[15] = 0xCC;
		
		err = onenand_enter_otp_mode(nand);
		if (err) {
			printf("ERROR entering OneNAND OTP mode!\n");
			rcode = 4;
			goto exit;
		}
	
		err = onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(0), ONENAND_COMMAND_UNLOCK);
		if (err) {
			printf("ERROR unlocking OTP block!\n");
			rcode = 8;
			goto exit_hot_reset;
		}
	
		rcode = onenand_program_otp(nand, 0, ONENAND_COMMAND_PROGRAM_SPARE);
		if (rcode)
			printf("ERROR writing OTP lock pattern!\n");
	
		exit_hot_reset:
		if (onenand_hot_reset(nand, ONENAND_COMMAND_RESET_NAND))
			printf("ERROR performing OneNAND hot reset!\n");
	}

exit:
	if (rcode == 0) {
		printf("OTP permanently locked");
		
		if (!onenand_otp_locked(nand))
			printf("; power cycle device for locking to take effect.\n");
		else
			printf(".\n");
	}
	
	return rcode;
}

int lock_otp(void)
{
	struct nand_chip *nand = &nand_dev_desc[0];
	return ( onenand_lock_otp_forever(nand) );
}

static int page_onenand_get_serial_number(ushort page, char *buf, int len)
{
	struct nand_chip *nand = &nand_dev_desc[0];
	ushort *dataram0 = ONENAND_DATARAM0_BASE(nand->IO_ADDR);
	char *cp = (char *)dataram0;
	int	i, offset, rcode = 0;

	memset(buf, 0, len);

	if (onenand_enter_otp_mode(nand)) {
		printf("ERROR entering OneNAND OTP mode!\n");
		return 4;
	}

	onenand_load_page(nand->IO_ADDR, NAND_BLOCK(0), page);

	if (onenand_hot_reset(nand, ONENAND_COMMAND_RESET_FLASH))
		printf("ERROR performing OneNAND hot reset!\n");

	/* Find last sector with data in it. */

	offset = -1;
	
	for (i = 0 ; i < 2048 ; i += 512) {
		if (onenand_is_sector_empty((ushort *) &cp[i]) == 0) {
			offset = i;
		}
	}

	if (offset != -1) {
		int limit = offset + 512;
		i = offset;
		offset = -1;
		for (; (i < limit) && (offset == -1) ; i++) {
			if (strncmp(&cp[i], "serial#=", 8) == 0) {
				offset = i + 8;
			}
		}
	}

	if (offset != -1) {
		strcpy(buf, &cp[offset]);
		printf("serial#=%s", buf);
		
		if (onenand_otp_locked(nand))
			puts(" (locked)\n");
		else
			puts("\n");
		
	} else {
#ifdef NAND_DEBUG
		printf("Serial number not found in page %d.\n", page);
#endif	
		rcode = 1;
	}

	return rcode;
}

int onenand_get_serial_number(char *buf, int len)
{
	int result = FAILURE;
	ushort page;
	
	for ( page = ONENAND_NUM_OTP_USER_PAGES; (page > 0) && (FAILURE == result); page-- )
		result = (SUCCESS == page_onenand_get_serial_number((page-1), buf, len)) ? SUCCESS : FAILURE;

	return ( result );
}

static int page_onenand_set_serial_number(ushort page, char *buf, int len)
{
	struct nand_chip *nand = &nand_dev_desc[0];
	char tmpbuf[80];
	int rcode = 0;

	if (!onenand_otp_locked(nand)) {
		ushort	*dataram0 = ONENAND_DATARAM0_BASE(nand->IO_ADDR);
		char	*cp = (char *)dataram0;
		int		i, err, offset;
	
		if (!buf || len >= 80) {
			printf("ERROR bad address or length\n");
			printf("buf=0x%x, len=%d\n", buf, len);
			rcode = 3;
			goto exit;
		}
	
		err = onenand_enter_otp_mode(nand);
		if (err) {
			printf("ERROR entering OneNAND OTP mode!\n");
			rcode = 4;
			goto exit;
		}
	
		onenand_load_page(nand->IO_ADDR, NAND_BLOCK(0), page);
	
		/* Find the first 512-byte sector that hasn't been programmed. */
	
		offset = -1;
	
		for (i = 0 ; (i < 2048) && (offset == -1) ; i += 512) {
			if (onenand_is_sector_empty((ushort *) &cp[i])) {
				offset = i;
			}
		}
	
		if (offset == -1) {
#ifdef NAND_DEBUG
			printf("All sectors in page %d are already written.\n", page);
#endif
			rcode = 7;
			goto exit_hot_reset;
		}
	
		err = onenand_set_protection_state(nand->IO_ADDR, NAND_BLOCK(0), ONENAND_COMMAND_UNLOCK);
		if (err) {
			printf("ERROR unlocking OTP block!\n");
			rcode = 8;
			goto exit_hot_reset;
		}
	
		memset(tmpbuf, 0, sizeof(tmpbuf));
		sprintf(tmpbuf, "serial#=%s", buf);
	
		memset(nand->data_buf, 0xFF, nand->oobblock);
		memcpy(&nand->data_buf[offset], tmpbuf, strlen(tmpbuf) + 1);
	
		rcode = onenand_program_otp(nand, ((nand->oobblock * page) + offset), ONENAND_COMMAND_PROGRAM);
	
		exit_hot_reset:
		if (onenand_hot_reset(nand, ONENAND_COMMAND_RESET_FLASH))
			printf("ERROR performing OneNAND hot reset!\n");
	}
	
exit:
	if (rcode == 0) {
		extern void set_serial_num(char *serial_num);
		extern void serial_number_init(void);
		
		set_serial_num("");		// Clear cached serial number.
		serial_number_init();	// Re-init with new one.
	}

	return rcode;
}

int onenand_set_serial_number(char *buf, int len)
{
	int result = FAILURE;
	ushort page;
	
	for ( page = 0; (page < ONENAND_NUM_OTP_USER_PAGES) && (FAILURE == result); page++ )
		result = (SUCCESS == page_onenand_set_serial_number(page, buf, len)) ? SUCCESS : FAILURE;
	
	return ( result );
}

int do_serial_number(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int	rcode = 0;
	char buf[80];

	if (argc == 1) {		// display serial number
		onenand_get_serial_number(buf, 80);
	} else if (argc == 2) {	// set serial number
		onenand_get_serial_number(buf, 80);
		if (0 != strcmp(buf, argv[1]))
			rcode = onenand_set_serial_number(argv[1], strlen(argv[1]));
	} else {
		printf("Usage:\n%s\n", cmdtp->usage);
		rcode = 1;
	}
	return(rcode);
}

U_BOOT_CMD(
	serial,	3,	0,	do_serial_number,
	"serial  - set/display board serial number in OTP\n",
	"string - display/set serial number to string\n"
);

/* ------------------------------------------------------------------------- */

#else /* (not) CONFIG_FLASH_SAMSUNG_ONENAND */

static int NanD_WaitReady(struct nand_chip *nand, int ale_wait)
{
	/* This is inline, to optimise the common case, where it's ready instantly */
	int ret = 0;

#ifdef NAND_NO_RB	/* in config file, shorter delays currently wrap accesses */
	if(ale_wait)
		NAND_WAIT_READY(nand);	/* do the worst case 25us wait */
	else
		udelay(10);
#else	/* has functional r/b signal */
	NAND_WAIT_READY(nand);
#endif
	return ret;
}

/* NanD_Command: Send a flash command to the flash chip */

static inline int NanD_Command(struct nand_chip *nand, unsigned char command)
{
	unsigned long nandptr = nand->IO_ADDR;

	/* Assert the CLE (Command Latch Enable) line to the flash chip */
	NAND_CTL_SETCLE(nandptr);

	/* Send the command */
	WRITE_NAND_COMMAND(command, nandptr);

	/* Lower the CLE line */
	NAND_CTL_CLRCLE(nandptr);

#ifdef NAND_NO_RB
	if(command == NAND_CMD_RESET){
		u_char ret_val;
		NanD_Command(nand, NAND_CMD_STATUS);
		do{
			ret_val = READ_NAND(nandptr);/* wait till ready */
		} while((ret_val & 0x40) != 0x40);
	}
#endif
	return NanD_WaitReady(nand, 0);
}

/* NanD_Address: Set the current address for the flash chip */

static int NanD_Address(struct nand_chip *nand, int numbytes, unsigned long ofs)
{
	unsigned long nandptr;
	int i;

	nandptr = nand->IO_ADDR;

	/* Assert the ALE (Address Latch Enable) line to the flash chip */
	NAND_CTL_SETALE(nandptr);

	/* Send the address */
	/* Devices with 256-byte page are addressed as:
	 * Column (bits 0-7), Page (bits 8-15, 16-23, 24-31)
	 * there is no device on the market with page256
	 * and more than 24 bits.
	 * Devices with 512-byte page are addressed as:
	 * Column (bits 0-7), Page (bits 9-16, 17-24, 25-31)
	 * 25-31 is sent only if the chip support it.
	 * bit 8 changes the read command to be sent
	 * (NAND_CMD_READ0 or NAND_CMD_READ1).
	 */

	if (numbytes == ADDR_COLUMN || numbytes == ADDR_COLUMN_PAGE)
		WRITE_NAND_ADDRESS(ofs, nandptr);

	ofs = ofs >> nand->page_shift;

	if (numbytes == ADDR_PAGE || numbytes == ADDR_COLUMN_PAGE)
		for (i = 0; i < nand->pageadrlen; i++, ofs = ofs >> 8)
			WRITE_NAND_ADDRESS(ofs, nandptr);

	/* Lower the ALE line */
	NAND_CTL_CLRALE(nandptr);

	/* Wait for the chip to respond */
	return NanD_WaitReady(nand, 1);
}

/* NanD_SelectChip: Select a given flash chip within the current floor */

static inline int NanD_SelectChip(struct nand_chip *nand, int chip)
{
	/* Wait for it to be ready */
	return NanD_WaitReady(nand, 0);
}

/* NanD_IdentChip: Identify a given NAND chip given {floor,chip} */

static int NanD_IdentChip(struct nand_chip *nand, int floor, int chip)
{
	int mfr, id, i;

	NAND_ENABLE_CE(nand);  /* set pin low */
	/* Reset the chip */
	if (NanD_Command(nand, NAND_CMD_RESET)) {
#ifdef NAND_DEBUG
		printf("NanD_Command (reset) for %d,%d returned true\n",
		       floor, chip);
#endif
		NAND_DISABLE_CE(nand);  /* set pin high */
		return 0;
	}

	/* Read the NAND chip ID: 1. Send ReadID command */
	if (NanD_Command(nand, NAND_CMD_READID)) {
#ifdef NAND_DEBUG
		printf("NanD_Command (ReadID) for %d,%d returned true\n",
		       floor, chip);
#endif
		NAND_DISABLE_CE(nand);  /* set pin high */
		return 0;
	}

	/* Read the NAND chip ID: 2. Send address byte zero */
	NanD_Address(nand, ADDR_COLUMN, 0);

	/* Read the manufacturer and device id codes from the device */

	mfr = READ_NAND(nand->IO_ADDR);

	id = READ_NAND(nand->IO_ADDR);

	NAND_DISABLE_CE(nand);  /* set pin high */
	/* No response - return failure */
	if (mfr == 0xff || mfr == 0) {
#ifdef NAND_DEBUG
		printf("NanD_Command (ReadID) got %d %d\n", mfr, id);
#endif
		return 0;
	}

	/* Check it's the same as the first chip we identified.
	 * M-Systems say that any given nand_chip device should only
	 * contain _one_ type of flash part, although that's not a
	 * hardware restriction. */
	if (nand->mfr) {
		if (nand->mfr == mfr && nand->id == id)
			return 1;	/* This is another the same the first */
		else
			printf("Flash chip at floor %d, chip %d is different:\n",
			       floor, chip);
	}

	/* Print and store the manufacturer and ID codes. */
	for (i = 0; nand_flash_ids[i].name != NULL; i++) {
		if (mfr == nand_flash_ids[i].manufacture_id &&
		    id == nand_flash_ids[i].model_id) {
#ifdef NAND_DEBUG
			printf("Flash chip found:\n\t Manufacturer ID: 0x%2.2X, "
			       "Chip ID: 0x%2.2X (%s)\n", mfr, id,
			       nand_flash_ids[i].name);
#endif
			if (!nand->mfr) {
				nand->mfr = mfr;
				nand->id = id;
				nand->chipshift =
				    nand_flash_ids[i].chipshift;
				nand->page256 = nand_flash_ids[i].page256;
				nand->eccsize = 256;
				if (nand->page256) {
					nand->oobblock = 256;
					nand->oobsize = 8;
					nand->page_shift = 8;
				} else {
					nand->oobblock = 512;
					nand->oobsize = 16;
					nand->page_shift = 9;
				}
				nand->pageadrlen =
				    nand_flash_ids[i].pageadrlen;
				nand->erasesize =
				    nand_flash_ids[i].erasesize;
				nand->chips_name =
				    nand_flash_ids[i].name;
				return 1;
			}
			return 0;
		}
	}

#ifdef NAND_DEBUG
	/* We haven't fully identified the chip. Print as much as we know. */
	printf("Unknown flash chip found: %2.2X %2.2X\n",
	       id, mfr);
#endif

	return 0;
}

/* NanD_ScanChips: Find all NAND chips present in a nand_chip, and identify them */

static void NanD_ScanChips(struct nand_chip *nand)
{
	int floor, chip;
	int numchips[NAND_MAX_FLOORS];
	int maxchips = NAND_MAX_CHIPS;
	int ret = 1;

	nand->numchips = 0;
	nand->mfr = 0;
	nand->id = 0;


	/* For each floor, find the number of valid chips it contains */
	for (floor = 0; floor < NAND_MAX_FLOORS; floor++) {
		ret = 1;
		numchips[floor] = 0;
		for (chip = 0; chip < maxchips && ret != 0; chip++) {

			ret = NanD_IdentChip(nand, floor, chip);
			if (ret) {
				numchips[floor]++;
				nand->numchips++;
			}
		}
	}

	/* If there are none at all that we recognise, bail */
	if (!nand->numchips) {
#ifdef NAND_DEBUG
		puts ("No NAND flash chips recognised.\n");
#endif
		return;
	}

	/* Allocate an array to hold the information for each chip */
	nand->chips = malloc(sizeof(struct Nand) * nand->numchips);
	if (!nand->chips) {
		puts ("No memory for allocating chip info structures\n");
		return;
	}

	ret = 0;

	/* Fill out the chip array with {floor, chipno} for each
	 * detected chip in the device. */
	for (floor = 0; floor < NAND_MAX_FLOORS; floor++) {
		for (chip = 0; chip < numchips[floor]; chip++) {
			nand->chips[ret].floor = floor;
			nand->chips[ret].chip = chip;
			nand->chips[ret].curadr = 0;
			nand->chips[ret].curmode = 0x50;
			ret++;
		}
	}

	/* Calculate and print the total size of the device */
	nand->totlen = nand->numchips * (1 << nand->chipshift);

#ifdef NAND_DEBUG
	printf("%d flash chips found. Total nand_chip size: %ld MB\n",
	       nand->numchips, nand->totlen >> 20);
#endif
}

/* we need to be fast here, 1 us per read translates to 1 second per meg */
static void NanD_ReadBuf(struct nand_chip *nand, u_char *data_buf, int cntr)
{
	unsigned long nandptr = nand->IO_ADDR;

	while (cntr >= 16) {
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		*data_buf++ = READ_NAND(nandptr);
		cntr -= 16;
	}

	while (cntr > 0) {
		*data_buf++ = READ_NAND(nandptr);
		cntr--;
	}
}

/*
 * NAND read with ECC
 */
static int nand_read_ecc(struct nand_chip *nand, size_t start, size_t len,
		 size_t * retlen, u_char *buf, u_char *ecc_code)
{
	int col, page;
	int ecc_status = 0;
#ifdef CONFIG_MTD_NAND_ECC
	int j;
	int ecc_failed = 0;
	u_char *data_poi;
	u_char ecc_calc[6];
#endif

	/* Do not allow reads past end of device */
	if ((start + len) > nand->totlen) {
		printf ("%s: Attempt read beyond end of device %x %x %x\n", __FUNCTION__, (uint) start, (uint) len, (uint) nand->totlen);
		*retlen = 0;
		return -1;
	}

	/* First we calculate the starting page */
	/*page = shr(start, nand->page_shift);*/
	page = start >> nand->page_shift;

	/* Get raw starting column */
	col = start & (nand->oobblock - 1);

	/* Initialize return value */
	*retlen = 0;

	/* Select the NAND device */
	NAND_ENABLE_CE(nand);  /* set pin low */

	/* Loop until all data read */
	while (*retlen < len) {


#ifdef CONFIG_MTD_NAND_ECC

		/* Do we have this page in cache ? */
		if (nand->cache_page == page)
			goto readdata;
		/* Send the read command */
		NanD_Command(nand, NAND_CMD_READ0);
		NanD_Address(nand, ADDR_COLUMN_PAGE, (page << nand->page_shift) + col);
		/* Read in a page + oob data */
		NanD_ReadBuf(nand, nand->data_buf, nand->oobblock + nand->oobsize);

		/* copy data into cache, for read out of cache and if ecc fails */
		if (nand->data_cache)
			memcpy (nand->data_cache, nand->data_buf, nand->oobblock + nand->oobsize);

		/* Pick the ECC bytes out of the oob data */
		for (j = 0; j < 6; j++)
			ecc_code[j] = nand->data_buf[(nand->oobblock + oob_config.ecc_pos[j])];

		/* Calculate the ECC and verify it */
		/* If block was not written with ECC, skip ECC */
		if (oob_config.eccvalid_pos != -1 &&
		    (nand->data_buf[nand->oobblock + oob_config.eccvalid_pos] & 0x0f) != 0x0f) {

			nand_calculate_ecc (&nand->data_buf[0], &ecc_calc[0]);
			switch (nand_correct_data (&nand->data_buf[0], &ecc_code[0], &ecc_calc[0])) {
			case -1:
				printf ("%s: Failed ECC read, page 0x%08x\n", __FUNCTION__, page);
				ecc_failed++;
				break;
			case 1:
			case 2:	/* transfer ECC corrected data to cache */
				if (nand->data_cache)
					memcpy (nand->data_cache, nand->data_buf, 256);
				break;
			}
		}

		if (oob_config.eccvalid_pos != -1 &&
		    nand->oobblock == 512 && (nand->data_buf[nand->oobblock + oob_config.eccvalid_pos] & 0xf0) != 0xf0) {

			nand_calculate_ecc (&nand->data_buf[256], &ecc_calc[3]);
			switch (nand_correct_data (&nand->data_buf[256], &ecc_code[3], &ecc_calc[3])) {
			case -1:
				printf ("%s: Failed ECC read, page 0x%08x\n", __FUNCTION__, page);
				ecc_failed++;
				break;
			case 1:
			case 2:	/* transfer ECC corrected data to cache */
				if (nand->data_cache)
					memcpy (&nand->data_cache[256], &nand->data_buf[256], 256);
				break;
			}
		}
readdata:
		/* Read the data from ECC data buffer into return buffer */
		data_poi = (nand->data_cache) ? nand->data_cache : nand->data_buf;
		data_poi += col;
		if ((*retlen + (nand->oobblock - col)) >= len) {
			memcpy (buf + *retlen, data_poi, len - *retlen);
			*retlen = len;
		} else {
			memcpy (buf + *retlen, data_poi,  nand->oobblock - col);
			*retlen += nand->oobblock - col;
		}
		/* Set cache page address, invalidate, if ecc_failed */
		nand->cache_page = (nand->data_cache && !ecc_failed) ? page : -1;

		ecc_status += ecc_failed;
		ecc_failed = 0;

#else
		/* Send the read command */
		NanD_Command(nand, NAND_CMD_READ0);
		NanD_Address(nand, ADDR_COLUMN_PAGE, (page << nand->page_shift) + col);
		/* Read the data directly into the return buffer */
		if ((*retlen + (nand->oobblock - col)) >= len) {
			NanD_ReadBuf(nand, buf + *retlen, len - *retlen);
			*retlen = len;
			/* We're done */
			continue;
		} else {
			NanD_ReadBuf(nand, buf + *retlen, nand->oobblock - col);
			*retlen += nand->oobblock - col;
			}
#endif
		/* For subsequent reads align to page boundary. */
		col = 0;
		/* Increment page address */
		page++;
	}

	/* De-select the NAND device */
	NAND_DISABLE_CE(nand);  /* set pin high */

	/*
	 * Return success, if no ECC failures, else -EIO
	 * fs driver will take care of that, because
	 * retlen == desired len and result == -EIO
	 */
	return ecc_status ? -1 : 0;
}

/*
 *	Nand_page_program function is used for write and writev !
 */
static int nand_write_page (struct nand_chip *nand,
			    int page, int col, int last, u_char * ecc_code)
{

	int i;
	unsigned long nandptr = nand->IO_ADDR;
#ifdef CONFIG_MTD_NAND_ECC
#ifdef CONFIG_MTD_NAND_VERIFY_WRITE
	int ecc_bytes = (nand->oobblock == 512) ? 6 : 3;
#endif
#endif
	/* pad oob area */
	for (i = nand->oobblock; i < nand->oobblock + nand->oobsize; i++)
		nand->data_buf[i] = 0xff;

#ifdef CONFIG_MTD_NAND_ECC
	/* Zero out the ECC array */
	for (i = 0; i < 6; i++)
		ecc_code[i] = 0x00;

	/* Read back previous written data, if col > 0 */
	if (col) {
		NanD_Command(nand, NAND_CMD_READ0);
		NanD_Address(nand, ADDR_COLUMN_PAGE, (page << nand->page_shift) + col);
		for (i = 0; i < col; i++)
			nand->data_buf[i] = READ_NAND (nandptr);
	}

	/* Calculate and write the ECC if we have enough data */
	if ((col < nand->eccsize) && (last >= nand->eccsize)) {
		nand_calculate_ecc (&nand->data_buf[0], &(ecc_code[0]));
		for (i = 0; i < 3; i++)
			nand->data_buf[(nand->oobblock + oob_config.ecc_pos[i])] = ecc_code[i];
		if (oob_config.eccvalid_pos != -1)
			nand->data_buf[nand->oobblock + oob_config.eccvalid_pos] = 0xf0;
	}

	/* Calculate and write the second ECC if we have enough data */
	if ((nand->oobblock == 512) && (last == nand->oobblock)) {
		nand_calculate_ecc (&nand->data_buf[256], &(ecc_code[3]));
		for (i = 3; i < 6; i++)
			nand->data_buf[(nand->oobblock + oob_config.ecc_pos[i])] = ecc_code[i];
		if (oob_config.eccvalid_pos != -1)
			nand->data_buf[nand->oobblock + oob_config.eccvalid_pos] &= 0x0f;
	}
#endif
	/* Prepad for partial page programming !!! */
	for (i = 0; i < col; i++)
		nand->data_buf[i] = 0xff;

	/* Postpad for partial page programming !!! oob is already padded */
	for (i = last; i < nand->oobblock; i++)
		nand->data_buf[i] = 0xff;

	/* Send command to begin auto page programming */
	NanD_Command(nand, NAND_CMD_READ0);
	NanD_Command(nand, NAND_CMD_SEQIN);
	NanD_Address(nand, ADDR_COLUMN_PAGE, (page << nand->page_shift) + col);

	/* Write out complete page of data */
	for (i = 0; i < (nand->oobblock + nand->oobsize); i++)
		WRITE_NAND(nand->data_buf[i], nand->IO_ADDR);

	/* Send command to actually program the data */
	NanD_Command(nand, NAND_CMD_PAGEPROG);
	NanD_Command(nand, NAND_CMD_STATUS);
#ifdef NAND_NO_RB
	{ u_char ret_val;

	  do{
		ret_val = READ_NAND(nandptr);	/* wait till ready */
	  } while((ret_val & 0x40) != 0x40);
	}
#endif
	/* See if device thinks it succeeded */
	if (READ_NAND(nand->IO_ADDR) & 0x01) {
		printf ("%s: Failed write, page 0x%08x, ", __FUNCTION__, page);
		return -1;
	}

#ifdef CONFIG_MTD_NAND_VERIFY_WRITE
	/*
	 * The NAND device assumes that it is always writing to
	 * a cleanly erased page. Hence, it performs its internal
	 * write verification only on bits that transitioned from
	 * 1 to 0. The device does NOT verify the whole page on a
	 * byte by byte basis. It is possible that the page was
	 * not completely erased or the page is becoming unusable
	 * due to wear. The read with ECC would catch the error
	 * later when the ECC page check fails, but we would rather
	 * catch it early in the page write stage. Better to write
	 * no data than invalid data.
	 */

	/* Send command to read back the page */
	if (col < nand->eccsize)
		NanD_Command(nand, NAND_CMD_READ0);
	else
		NanD_Command(nand, NAND_CMD_READ1);
	NanD_Address(nand, ADDR_COLUMN_PAGE, (page << nand->page_shift) + col);

	/* Loop through and verify the data */
	for (i = col; i < last; i++) {
		if (nand->data_buf[i] != readb (nand->IO_ADDR)) {
			printf ("%s: Failed write verify, page 0x%08x ", __FUNCTION__, page);
			return -1;
		}
	}

#ifdef CONFIG_MTD_NAND_ECC
	/*
	 * We also want to check that the ECC bytes wrote
	 * correctly for the same reasons stated above.
	 */
	NanD_Command(nand, NAND_CMD_READOOB);
	NanD_Address(nand, ADDR_COLUMN_PAGE, (page << nand->page_shift) + col);
	for (i = 0; i < nand->oobsize; i++)
		nand->data_buf[i] = readb (nand->IO_ADDR);
	for (i = 0; i < ecc_bytes; i++) {
		if ((nand->data_buf[(oob_config.ecc_pos[i])] != ecc_code[i]) && ecc_code[i]) {
			printf ("%s: Failed ECC write "
			       "verify, page 0x%08x, " "%6i bytes were succesful\n", __FUNCTION__, page, i);
			return -1;
		}
	}
#endif
#endif
	return 0;
}

static int nand_write_ecc (struct nand_chip* nand, size_t to, size_t len,
			   size_t * retlen, const u_char * buf, u_char * ecc_code)
{
	int i, page, col, cnt, ret = 0;

	/* Do not allow write past end of device */
	if ((to + len) > nand->totlen) {
		printf ("%s: Attempt to write past end of page\n", __FUNCTION__);
		return -1;
	}

	/* Shift to get page */
	page = ((int) to) >> nand->page_shift;

	/* Get the starting column */
	col = to & (nand->oobblock - 1);

	/* Initialize return length value */
	*retlen = 0;

	/* Select the NAND device */
#ifdef CONFIG_OMAP1510
	archflashwp(0,0);
#endif
    	NAND_ENABLE_CE(nand);  /* set pin low */

	/* Check the WP bit */
	NanD_Command(nand, NAND_CMD_STATUS);
	if (!(READ_NAND(nand->IO_ADDR) & 0x80)) {
		printf ("%s: Device is write protected!!!\n", __FUNCTION__);
		ret = -1;
		goto out;
	}

	/* Loop until all data is written */
	while (*retlen < len) {
		/* Invalidate cache, if we write to this page */
		if (nand->cache_page == page)
			nand->cache_page = -1;

		/* Write data into buffer */
		if ((col + len) >= nand->oobblock)
			for (i = col, cnt = 0; i < nand->oobblock; i++, cnt++)
				nand->data_buf[i] = buf[(*retlen + cnt)];
		else
			for (i = col, cnt = 0; cnt < (len - *retlen); i++, cnt++)
				nand->data_buf[i] = buf[(*retlen + cnt)];
		/* We use the same function for write and writev !) */
		ret = nand_write_page (nand, page, col, i, ecc_code);
		if (ret)
			goto out;

		/* Next data start at page boundary */
		col = 0;

		/* Update written bytes count */
		*retlen += cnt;

		/* Increment page address */
		page++;
	}

	/* Return happy */
	*retlen = len;

out:
	/* De-select the NAND device */
	NAND_DISABLE_CE(nand);  /* set pin high */
#ifdef CONFIG_OMAP1510
    	archflashwp(0,1);
#endif
	return ret;
}

/* read from the 16 bytes of oob data that correspond to a 512 byte
 * page or 2 256-byte pages.
 */
static int nand_read_oob(struct nand_chip* nand, size_t ofs, size_t len,
			 size_t * retlen, u_char * buf)
{
	int len256 = 0;
	struct Nand *mychip;
	int ret = 0;

	mychip = &nand->chips[ofs >> nand->chipshift];

	/* update address for 2M x 8bit devices. OOB starts on the second */
	/* page to maintain compatibility with nand_read_ecc. */
	if (nand->page256) {
		if (!(ofs & 0x8))
			ofs += 0x100;
		else
			ofs -= 0x8;
	}

	NAND_ENABLE_CE(nand);  /* set pin low */
	NanD_Command(nand, NAND_CMD_READOOB);
	NanD_Address(nand, ADDR_COLUMN_PAGE, ofs);

	/* treat crossing 8-byte OOB data for 2M x 8bit devices */
	/* Note: datasheet says it should automaticaly wrap to the */
	/*       next OOB block, but it didn't work here. mf.      */
	if (nand->page256 && ofs + len > (ofs | 0x7) + 1) {
		len256 = (ofs | 0x7) + 1 - ofs;
		NanD_ReadBuf(nand, buf, len256);

		NanD_Command(nand, NAND_CMD_READOOB);
		NanD_Address(nand, ADDR_COLUMN_PAGE, ofs & (~0x1ff));
	}

	NanD_ReadBuf(nand, &buf[len256], len - len256);

	*retlen = len;
	/* Reading the full OOB data drops us off of the end of the page,
	 * causing the flash device to go into busy mode, so we need
	 * to wait until ready 11.4.1 and Toshiba TC58256FT nands */

	ret = NanD_WaitReady(nand, 1);
	NAND_DISABLE_CE(nand);  /* set pin high */

	return ret;

}

/* write to the 16 bytes of oob data that correspond to a 512 byte
 * page or 2 256-byte pages.
 */
static int nand_write_oob(struct nand_chip* nand, size_t ofs, size_t len,
		  size_t * retlen, const u_char * buf)
{
	int len256 = 0;
	int i;
	unsigned long nandptr = nand->IO_ADDR;

#ifdef PSYCHO_DEBUG
	printf("nand_write_oob(%lx, %d): %2.2X %2.2X %2.2X %2.2X ... %2.2X %2.2X .. %2.2X %2.2X\n",
	       (long)ofs, len, buf[0], buf[1], buf[2], buf[3],
	       buf[8], buf[9], buf[14],buf[15]);
#endif

	NAND_ENABLE_CE(nand);  /* set pin low to enable chip */

	/* Reset the chip */
	NanD_Command(nand, NAND_CMD_RESET);

	/* issue the Read2 command to set the pointer to the Spare Data Area. */
	NanD_Command(nand, NAND_CMD_READOOB);
	NanD_Address(nand, ADDR_COLUMN_PAGE, ofs);

	/* update address for 2M x 8bit devices. OOB starts on the second */
	/* page to maintain compatibility with nand_read_ecc. */
	if (nand->page256) {
		if (!(ofs & 0x8))
			ofs += 0x100;
		else
			ofs -= 0x8;
	}

	/* issue the Serial Data In command to initial the Page Program process */
	NanD_Command(nand, NAND_CMD_SEQIN);
	NanD_Address(nand, ADDR_COLUMN_PAGE, ofs);

	/* treat crossing 8-byte OOB data for 2M x 8bit devices */
	/* Note: datasheet says it should automaticaly wrap to the */
	/*       next OOB block, but it didn't work here. mf.      */
	if (nand->page256 && ofs + len > (ofs | 0x7) + 1) {
		len256 = (ofs | 0x7) + 1 - ofs;
		for (i = 0; i < len256; i++)
			WRITE_NAND(buf[i], nandptr);

		NanD_Command(nand, NAND_CMD_PAGEPROG);
		NanD_Command(nand, NAND_CMD_STATUS);
#ifdef NAND_NO_RB
   		{ u_char ret_val;
    		  do{
		  	ret_val = READ_NAND(nandptr); /* wait till ready */
    		  }while((ret_val & 0x40) != 0x40);
		}
#endif
		if (READ_NAND(nandptr) & 1) {
			puts ("Error programming oob data\n");
			/* There was an error */
			NAND_DISABLE_CE(nand);  /* set pin high */
			*retlen = 0;
			return -1;
		}
		NanD_Command(nand, NAND_CMD_SEQIN);
		NanD_Address(nand, ADDR_COLUMN_PAGE, ofs & (~0x1ff));
	}

	for (i = len256; i < len; i++)
		WRITE_NAND(buf[i], nandptr);

	NanD_Command(nand, NAND_CMD_PAGEPROG);
	NanD_Command(nand, NAND_CMD_STATUS);
#ifdef NAND_NO_RB
	{ u_char ret_val;
	  do{
		ret_val = READ_NAND(nandptr); /* wait till ready */
	  } while((ret_val & 0x40) != 0x40);
	}
#endif
	if (READ_NAND(nandptr) & 1) {
		puts ("Error programming oob data\n");
		/* There was an error */
		NAND_DISABLE_CE(nand);  /* set pin high */
		*retlen = 0;
		return -1;
	}

	NAND_DISABLE_CE(nand);  /* set pin high */
	*retlen = len;
	return 0;

}

int nand_erase(struct nand_chip* nand, size_t ofs, size_t len, int clean)
{
	/* This is defined as a structure so it will work on any system
	 * using native endian jffs2 (the default).
	 */
	static struct jffs2_unknown_node clean_marker = {
		JFFS2_MAGIC_BITMASK,
		JFFS2_NODETYPE_CLEANMARKER,
		8		/* 8 bytes in this node */
	};
	unsigned long nandptr;
	struct Nand *mychip;
	int ret = 0;

	if (ofs & (nand->erasesize-1) || len & (nand->erasesize-1)) {
		printf ("Offset and size must be sector aligned, erasesize = %d\n",
			(int) nand->erasesize);
		return -1;
	}

	nandptr = nand->IO_ADDR;

	/* Select the NAND device */
#ifdef CONFIG_OMAP1510
	archflashwp(0,0);
#endif
    NAND_ENABLE_CE(nand);  /* set pin low */

	/* Check the WP bit */
	NanD_Command(nand, NAND_CMD_STATUS);
	if (!(READ_NAND(nand->IO_ADDR) & 0x80)) {
		printf ("nand_write_ecc: Device is write protected!!!\n");
		ret = -1;
		goto out;
	}

	/* Check the WP bit */
	NanD_Command(nand, NAND_CMD_STATUS);
	if (!(READ_NAND(nand->IO_ADDR) & 0x80)) {
		printf ("%s: Device is write protected!!!\n", __FUNCTION__);
		ret = -1;
		goto out;
	}

	/* FIXME: Do nand in the background. Use timers or schedule_task() */
	while(len) {
		/*mychip = &nand->chips[shr(ofs, nand->chipshift)];*/
		mychip = &nand->chips[ofs >> nand->chipshift];

		/* always check for bad block first, genuine bad blocks
		 * should _never_  be erased.
		 */
		if (ALLOW_ERASE_BAD_DEBUG || !check_block(nand, ofs)) {
			/* Select the NAND device */
			NAND_ENABLE_CE(nand);  /* set pin low */

			NanD_Command(nand, NAND_CMD_ERASE1);
			NanD_Address(nand, ADDR_PAGE, ofs);
			NanD_Command(nand, NAND_CMD_ERASE2);

			NanD_Command(nand, NAND_CMD_STATUS);

#ifdef NAND_NO_RB
			{ u_char ret_val;
			  do{
				ret_val = READ_NAND(nandptr); /* wait till ready */
			  } while((ret_val & 0x40) != 0x40);
			}
#endif
			if (READ_NAND(nandptr) & 1) {
				printf ("%s: Error erasing at 0x%lx\n",
					__FUNCTION__, (long)ofs);
				/* There was an error */
				ret = -1;
				goto out;
			}
			if (clean) {
				int n;	/* return value not used */
				int p, l;

				/* clean marker position and size depend
				 * on the page size, since 256 byte pages
				 * only have 8 bytes of oob data
				 */
				if (nand->page256) {
					p = NAND_JFFS2_OOB8_FSDAPOS;
					l = NAND_JFFS2_OOB8_FSDALEN;
				}
				else {
					p = NAND_JFFS2_OOB16_FSDAPOS;
					l = NAND_JFFS2_OOB16_FSDALEN;
				}

				ret = nand_write_oob(nand, ofs + p, l, &n,
						     (u_char *)&clean_marker);
				/* quit here if write failed */
				if (ret)
					goto out;
			}
		}
		ofs += nand->erasesize;
		len -= nand->erasesize;
	}

out:
	/* De-select the NAND device */
	NAND_DISABLE_CE(nand);  /* set pin high */
#ifdef CONFIG_OMAP1510
    	archflashwp(0,1);
#endif
	return ret;
}

unsigned long nand_probe(unsigned long physadr)
{
	struct nand_chip *nand = NULL;
	int i = 0, ChipID = 1;

#ifdef CONFIG_MTD_NAND_ECC_JFFS2
	oob_config.ecc_pos[0] = NAND_JFFS2_OOB_ECCPOS0;
	oob_config.ecc_pos[1] = NAND_JFFS2_OOB_ECCPOS1;
	oob_config.ecc_pos[2] = NAND_JFFS2_OOB_ECCPOS2;
	oob_config.ecc_pos[3] = NAND_JFFS2_OOB_ECCPOS3;
	oob_config.ecc_pos[4] = NAND_JFFS2_OOB_ECCPOS4;
	oob_config.ecc_pos[5] = NAND_JFFS2_OOB_ECCPOS5;
	oob_config.eccvalid_pos = 4;
#else
	oob_config.ecc_pos[0] = NAND_NOOB_ECCPOS0;
	oob_config.ecc_pos[1] = NAND_NOOB_ECCPOS1;
	oob_config.ecc_pos[2] = NAND_NOOB_ECCPOS2;
	oob_config.ecc_pos[3] = NAND_NOOB_ECCPOS3;
	oob_config.ecc_pos[4] = NAND_NOOB_ECCPOS4;
	oob_config.ecc_pos[5] = NAND_NOOB_ECCPOS5;
	oob_config.eccvalid_pos = NAND_NOOB_ECCVPOS;
#endif
	oob_config.badblock_pos = 5;

	for (i=0; i<CFG_MAX_NAND_DEVICE; i++) {
		if (nand_dev_desc[i].ChipID == NAND_ChipID_UNKNOWN) {
			nand = &nand_dev_desc[i];
			break;
		}
	}
	if (!nand)
		return (0);

	memset((char *)nand, 0, sizeof(struct nand_chip));

	nand->IO_ADDR = physadr;
	nand->cache_page = -1;  /* init the cache page */
	NanD_ScanChips(nand);

	if (nand->totlen == 0) {
		/* no chips found, clean up and quit */
		memset((char *)nand, 0, sizeof(struct nand_chip));
		nand->ChipID = NAND_ChipID_UNKNOWN;
		return (0);
	}

	nand->ChipID = ChipID;
	if ((NO_CURR_DEVICE == curr_device) || (CFG_MAX_NAND_DEVICE < curr_device))
		curr_device = i;

	nand->data_buf = malloc (nand->oobblock + nand->oobsize);
	if (!nand->data_buf) {
		puts ("Cannot allocate memory for data structures.\n");
		return (0);
	}

	return (nand->totlen);
}

#ifdef CONFIG_MTD_NAND_ECC
/*
 * Pre-calculated 256-way 1 byte column parity
 */
static const u_char nand_ecc_precalc_table[] = {
	0x00, 0x55, 0x56, 0x03, 0x59, 0x0c, 0x0f, 0x5a, 0x5a, 0x0f, 0x0c, 0x59, 0x03, 0x56, 0x55, 0x00,
	0x65, 0x30, 0x33, 0x66, 0x3c, 0x69, 0x6a, 0x3f, 0x3f, 0x6a, 0x69, 0x3c, 0x66, 0x33, 0x30, 0x65,
	0x66, 0x33, 0x30, 0x65, 0x3f, 0x6a, 0x69, 0x3c, 0x3c, 0x69, 0x6a, 0x3f, 0x65, 0x30, 0x33, 0x66,
	0x03, 0x56, 0x55, 0x00, 0x5a, 0x0f, 0x0c, 0x59, 0x59, 0x0c, 0x0f, 0x5a, 0x00, 0x55, 0x56, 0x03,
	0x69, 0x3c, 0x3f, 0x6a, 0x30, 0x65, 0x66, 0x33, 0x33, 0x66, 0x65, 0x30, 0x6a, 0x3f, 0x3c, 0x69,
	0x0c, 0x59, 0x5a, 0x0f, 0x55, 0x00, 0x03, 0x56, 0x56, 0x03, 0x00, 0x55, 0x0f, 0x5a, 0x59, 0x0c,
	0x0f, 0x5a, 0x59, 0x0c, 0x56, 0x03, 0x00, 0x55, 0x55, 0x00, 0x03, 0x56, 0x0c, 0x59, 0x5a, 0x0f,
	0x6a, 0x3f, 0x3c, 0x69, 0x33, 0x66, 0x65, 0x30, 0x30, 0x65, 0x66, 0x33, 0x69, 0x3c, 0x3f, 0x6a,
	0x6a, 0x3f, 0x3c, 0x69, 0x33, 0x66, 0x65, 0x30, 0x30, 0x65, 0x66, 0x33, 0x69, 0x3c, 0x3f, 0x6a,
	0x0f, 0x5a, 0x59, 0x0c, 0x56, 0x03, 0x00, 0x55, 0x55, 0x00, 0x03, 0x56, 0x0c, 0x59, 0x5a, 0x0f,
	0x0c, 0x59, 0x5a, 0x0f, 0x55, 0x00, 0x03, 0x56, 0x56, 0x03, 0x00, 0x55, 0x0f, 0x5a, 0x59, 0x0c,
	0x69, 0x3c, 0x3f, 0x6a, 0x30, 0x65, 0x66, 0x33, 0x33, 0x66, 0x65, 0x30, 0x6a, 0x3f, 0x3c, 0x69,
	0x03, 0x56, 0x55, 0x00, 0x5a, 0x0f, 0x0c, 0x59, 0x59, 0x0c, 0x0f, 0x5a, 0x00, 0x55, 0x56, 0x03,
	0x66, 0x33, 0x30, 0x65, 0x3f, 0x6a, 0x69, 0x3c, 0x3c, 0x69, 0x6a, 0x3f, 0x65, 0x30, 0x33, 0x66,
	0x65, 0x30, 0x33, 0x66, 0x3c, 0x69, 0x6a, 0x3f, 0x3f, 0x6a, 0x69, 0x3c, 0x66, 0x33, 0x30, 0x65,
	0x00, 0x55, 0x56, 0x03, 0x59, 0x0c, 0x0f, 0x5a, 0x5a, 0x0f, 0x0c, 0x59, 0x03, 0x56, 0x55, 0x00
};


/*
 * Creates non-inverted ECC code from line parity
 */
static void nand_trans_result(u_char reg2, u_char reg3,
	u_char *ecc_code)
{
	u_char a, b, i, tmp1, tmp2;

	/* Initialize variables */
	a = b = 0x80;
	tmp1 = tmp2 = 0;

	/* Calculate first ECC byte */
	for (i = 0; i < 4; i++) {
		if (reg3 & a)		/* LP15,13,11,9 --> ecc_code[0] */
			tmp1 |= b;
		b >>= 1;
		if (reg2 & a)		/* LP14,12,10,8 --> ecc_code[0] */
			tmp1 |= b;
		b >>= 1;
		a >>= 1;
	}

	/* Calculate second ECC byte */
	b = 0x80;
	for (i = 0; i < 4; i++) {
		if (reg3 & a)		/* LP7,5,3,1 --> ecc_code[1] */
			tmp2 |= b;
		b >>= 1;
		if (reg2 & a)		/* LP6,4,2,0 --> ecc_code[1] */
			tmp2 |= b;
		b >>= 1;
		a >>= 1;
	}

	/* Store two of the ECC bytes */
	ecc_code[0] = tmp1;
	ecc_code[1] = tmp2;
}

/*
 * Calculate 3 byte ECC code for 256 byte block
 */
static void nand_calculate_ecc (const u_char *dat, u_char *ecc_code)
{
	u_char idx, reg1, reg3;
	int j;

	/* Initialize variables */
	reg1 = reg3 = 0;
	ecc_code[0] = ecc_code[1] = ecc_code[2] = 0;

	/* Build up column parity */
	for(j = 0; j < 256; j++) {

		/* Get CP0 - CP5 from table */
		idx = nand_ecc_precalc_table[dat[j]];
		reg1 ^= idx;

		/* All bit XOR = 1 ? */
		if (idx & 0x40) {
			reg3 ^= (u_char) j;
		}
	}

	/* Create non-inverted ECC code from line parity */
	nand_trans_result((reg1 & 0x40) ? ~reg3 : reg3, reg3, ecc_code);

	/* Calculate final ECC code */
	ecc_code[0] = ~ecc_code[0];
	ecc_code[1] = ~ecc_code[1];
	ecc_code[2] = ((~reg1) << 2) | 0x03;
}

/*
 * Detect and correct a 1 bit error for 256 byte block
 */
static int nand_correct_data (u_char *dat, u_char *read_ecc, u_char *calc_ecc)
{
	u_char a, b, c, d1, d2, d3, add, bit, i;

	/* Do error detection */
	d1 = calc_ecc[0] ^ read_ecc[0];
	d2 = calc_ecc[1] ^ read_ecc[1];
	d3 = calc_ecc[2] ^ read_ecc[2];

	if ((d1 | d2 | d3) == 0) {
		/* No errors */
		return 0;
	}
	else {
		a = (d1 ^ (d1 >> 1)) & 0x55;
		b = (d2 ^ (d2 >> 1)) & 0x55;
		c = (d3 ^ (d3 >> 1)) & 0x54;

		/* Found and will correct single bit error in the data */
		if ((a == 0x55) && (b == 0x55) && (c == 0x54)) {
			c = 0x80;
			add = 0;
			a = 0x80;
			for (i=0; i<4; i++) {
				if (d1 & c)
					add |= a;
				c >>= 2;
				a >>= 1;
			}
			c = 0x80;
			for (i=0; i<4; i++) {
				if (d2 & c)
					add |= a;
				c >>= 2;
				a >>= 1;
			}
			bit = 0;
			b = 0x04;
			c = 0x80;
			for (i=0; i<3; i++) {
				if (d3 & c)
					bit |= b;
				c >>= 2;
				b >>= 1;
			}
			b = 0x01;
			a = dat[add];
			a ^= (b << bit);
			dat[add] = a;
			return 1;
		}
		else {
			i = 0;
			while (d1) {
				if (d1 & 0x01)
					++i;
				d1 >>= 1;
			}
			while (d2) {
				if (d2 & 0x01)
					++i;
				d2 >>= 1;
			}
			while (d3) {
				if (d3 & 0x01)
					++i;
				d3 >>= 1;
			}
			if (i == 1) {
				/* ECC Code Error Correction */
				read_ecc[0] = calc_ecc[0];
				read_ecc[1] = calc_ecc[1];
				read_ecc[2] = calc_ecc[2];
				return 2;
			}
			else {
				/* Uncorrectable Error */
				return -1;
			}
		}
	}

	/* Should never happen */
	return -1;
}

#endif

#endif	/* CONFIG_FLASH_SAMSUNG_ONENAND */

/* ------------------------------------------------------------------------- */

#ifdef CONFIG_JFFS2_NAND

int read_jffs2_nand(size_t start, size_t len,
		    size_t * retlen, u_char * buf, int nanddev)
{
#ifndef CONFIG_FIONA
	return NAND_RW(NAND_DEV_START(nandev, start), len, (NANDRW_READ | NANDRW_JFFS2),
			retlen, buf);
#else
	int fiona_read_jffs2_nand(size_t,size_t,size_t*,u_char*,int);
	return fiona_read_jffs2_nand(start, len, retlen, buf, nanddev);
#endif
}

#endif /* CONFIG_JFFS2_NAND */

#endif /* (CONFIG_COMMANDS & CFG_CMD_NAND) */

/* vim: set ts=4 sw=4 tw=80: */
