/*
 * Low-level PXA250/210 sleep/wakeUp support
 *
 * Initial SA1110 code:
 * Copyright (c) 2001 Cliff Brake <cbrake@accelent.com>
 *
 * Adapted for PXA by Nicolas Pitre:
 * Copyright (c) 2002 Monta Vista Software, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License.
 */

#include <configs/fiona.h>
#include <asm/arch/pxa-regs.h>

/*
 * pxa_cpu_suspend()
 *
 * Forces CPU into sleep state
 */

#define CCCR_N_MASK 0x0380      /* Run Mode Frequency to Turbo Mode Frequency Multiplier */
#define CCCR_M_MASK 0x0060      /* Memory Frequency to Run Mode Frequency Multiplier */
#define CCCR_L_MASK 0x001f      /* Crystal Frequency to Memory Frequency Multiplier */
#define MDREFR_SLFRSH   (1 << 22)   /* SDRAM Self-Refresh Control/Status */
#define UNCACHED_PHYS_0     0xff000000
#define UNCACHED_ADDR       UNCACHED_PHYS_0

#if 0
#define MDREFR  0x48000004
#define CCCR    0x41300000
#define PWER    0x40f0000c
#define PRER    0x40f00010
#define PFER    0x40f00014
#endif

/*
 * This macro cleans the entire dcache using line allocate.
 * The main loop has been unrolled to reduce loop overhead.
 * rd and rs are two scratch registers.
 */
#define VM_EXEC             0x00000004
#define CACHESIZE           32768
#define CACHELINESIZE       32
#define CLEAN_ADDR          0xfffe0000

    .macro  clean_d_cache, rd, rs
    ldr \rs, =clean_addr
    ldr \rd, [\rs]
    eor \rd, \rd, #CACHESIZE
    str \rd, [\rs]
    add \rs, \rd, #CACHESIZE
1:  mcr p15, 0, \rd, c7, c2, 5      @ allocate D cache line
    add \rd, \rd, #CACHELINESIZE
    mcr p15, 0, \rd, c7, c2, 5      @ allocate D cache line
    add \rd, \rd, #CACHELINESIZE
    mcr p15, 0, \rd, c7, c2, 5      @ allocate D cache line
    add \rd, \rd, #CACHELINESIZE
    mcr p15, 0, \rd, c7, c2, 5      @ allocate D cache line
    add \rd, \rd, #CACHELINESIZE
    teq \rd, \rs
    bne 1b
    .endm

    .data
clean_addr: .word   CLEAN_ADDR

    .text

/*
 *  flush_kern_cache_all()
 *
 *  Clean and invalidate the entire cache.
 */

.globl xscale_flush_kern_cache_all
xscale_flush_kern_cache_all:
    mov r2, #VM_EXEC
    mov ip, #0
__flush_whole_cache:
    clean_d_cache r0, r1
    tst r2, #VM_EXEC
    mcrne   p15, 0, ip, c7, c5, 0       @ Invalidate I cache & BTB
    mcrne   p15, 0, ip, c7, c10, 4      @ Drain Write (& Fill) Buffer
    mov pc, lr

.globl  pxa_cpu_suspend
pxa_cpu_suspend:

	mra	r2, r3, acc0
	stmfd	sp!, {r2 - r12, lr}		@ save registers on stack

	@ get coprocessor registers
	mrc	p15, 0, r4, c15, c1, 0		@ CP access reg
	mrc	p15, 0, r5, c13, c0, 0		@ PID
	mrc 	p15, 0, r6, c3, c0, 0		@ domain ID
	mrc 	p15, 0, r7, c2, c0, 0		@ translation table base addr
	mrc	p15, 0, r8, c1, c1, 0           @ auxiliary control reg
	mrc 	p15, 0, r9, c1, c0, 0		@ control reg

	@ store them plus current virtual stack ptr on stack
	mov	r10, sp
	stmfd	sp!, {r4 - r10}

#if 0
	@ preserve phys address of stack
	mov	r0, sp
	bl	sleep_phys_sp
	ldr	r1, =sleep_save_sp
	str	r0, [r1]
#endif

	@ clean data cache
	bl	xscale_flush_kern_cache_all

	@ Put the processor to sleep
	@ (also workaround for sighting 28071)

	@ prepare value for sleep mode
	mov	r1, #3				@ sleep mode

#if 0
	@ prepare to put SDRAM into self-refresh manually
	ldr	r4, =MDREFR
	ldr	r5, [r4]
	orr	r5, r5, #MDREFR_SLFRSH
#endif

	@ prepare pointer to physical address 0 (virtual mapping in generic.c)
	mov	r2, #UNCACHED_PHYS_0

	@ Intel PXA255 Specification Update notes problems
	@ about suspending with PXBus operating above 133MHz
	@ (see Errata 31, GPIO output signals, ... unpredictable in sleep
	@
	@ We keep the change-down close to the actual suspend on SDRAM
	@ as possible to eliminate messing about with the refresh clock
	@ as the system will restore with the original speed settings
	@
	@ Ben Dooks, 13-Sep-2004

    ldr r6, =PWER
    mov r8, #3
    str r8, [r6]
    ldr r6, =PRER
    mov r8, #2
    str r8, [r6]
    ldr r6, =PFER
    mov r8, #3
    str r8, [r6]
    ldr r6, =PGSR0
    ldr r8, =CFG_PGSR0_FIONA
    str r8, [r6]
    ldr r6, =PGSR1
    ldr r8, =CFG_PGSR1_FIONA
    str r8, [r6]
    ldr r6, =PGSR2
    ldr r8, =CFG_PGSR2_FIONA
    str r8, [r6]
    ldr r6, =PSPR
    ldr r8, =NO_SLEEP_PSPR
    str r8, [r6]

	ldr	r6, =CCCR
	ldr	r8, [r6]		@ keep original value for resume

	@ ensure x1 for run and turbo mode with memory clock
	bic	r7, r8, #CCCR_M_MASK | CCCR_N_MASK
	orr	r7, r7, #(1<<5) | (2<<7)

	@ check that the memory frequency is within limits
	and	r14, r7, #CCCR_L_MASK
	teq	r14, #1
	bicne	r7, r7, #CCCR_L_MASK
	orrne	r7, r7, #1			@@ 99.53MHz

	@ get ready for the change
	@ note, since we are making turbo=run, do not remove the turbo
	@ as this may cause non-turbo mode on resume
	mrc	p14, 0, r0, c6, c0, 0
	bic	r0, r0, #2			@ clear change bit
	mcr	p14, 0, r0, c6, c0, 0
	orr	r0, r0, #2			@ initiate change bit

	@ align execution to a cache line
	b	1f

	.ltorg
	.align	5
1:

	@ All needed values are now in registers.
	@ These last instructions should be in cache

	@ initiate the frequency change...
	str	r7, [r6]
	mcr	p14, 0, r0, c6, c0, 0

	@ restore the original cpu speed value for resume
	str	r8, [r6]

#if 0
	@ put SDRAM into self-refresh
	str	r5, [r4]
#endif

	@ force address lines low by reading at physical address 0
	ldr	r3, [r2]

	@ enter sleep mode
	mcr	p14, 0, r1, c7, c0, 0

20:	b	20b				@ loop waiting for sleep
