/******************************************************************
 *
 *  File:   fpow.c
 *
 *  Author: Nick Vaccaro <nvaccaro@lab126.com>
 *
 *  Date:   05/05/06
 *
 * Copyright 2006, Lab126, Inc.  All rights reserved.
 *
 *  Description:
 *      Core driver for the Fiona Power Management System
 *
 ******************************************************************/

#include <linux/autoconf.h>
#include <stdarg.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/errno.h>
#include <asm/string.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/proc_fs.h>
#include <linux/device.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/suspend.h>
#include <linux/ctype.h>
#include <linux/cpufreq.h>
#include <linux/syscalls.h>
#include <asm/segment.h>

#include <asm/arch/fpow.h>
#include <asm/arch/fiona.h>
#include <asm/arch/fiona_eventmap.h>
#include <asm/arch/fiona_keymap.h>
#include <asm/arch/boot_globals.h>
#include <linux/fiona_power.h>
#include <linux/fiona_events.h>
#include <linux/fb.h>
#include <linux/delay.h>
#include <linux/einkfb.h>
#include <linux/ioc.h>
#include <asm/arch/fiona_sniffer.h>

extern char *fiona_event_keycode_to_str(int keycode);
extern int  ac_adapter_installed(void);
extern int  fiona_get_current_draw(void);
extern void fiona_clear_unstall_event_queue(void);
extern void print_queue(char *str, int forcePrint);
extern int  remove_events_in_queue(int event);
extern void put_at_front_of_queue(int element);
extern void wake_on_wan_switch(int enable);
extern int  isEventQueueStalled(void);
extern void setEventQueueStalled(int stall);
extern void unstallEventQueue(void);
extern int  ok_to_unstall_queue;
extern void ioc_reset_voltage_averages(void);
extern int  get_last_passed_event(void);
extern asmlinkage int hlog(const char *fmt, ...);

// Force video ON when we wake from sleep....
#undef FORCE_VIDEO_ON_UPON_WAKE

// Test WD
#undef RUN_WD_TEST

// Debug IOC transactions on wake
#undef DEBUG_IOC_WAKE_TRANSACTIONS


#define VIDEO_CLEAR_DISABLE_MODE    FPOW_MODE_OFF_CLEAR_SCREEN
#define PNLCD_CLEAR_DISABLE_MODE    FPOW_MODE_OFF
#define VIDEO_ENABLE_MODE           FPOW_MODE_ON
#define PNLCD_ENABLE_MODE           FPOW_MODE_ON

// IOC Access Stuff
extern struct file_operations ioc_fops;

#ifdef DEBUG_IOC_WAKE_TRANSACTIONS
extern int g_do_core_debug;
#endif

#define ioc_sys_ioctl(x,y,z) ioc_fops.ioctl((struct inode*)0xDEADBEEF,(struct file*)x,y,(unsigned long)z)



// Define to enable /proc/fpow/event_test for testing event queue
// lookahead feature
#define ENABLE_EVENT_TEST_PROC


//
// More descriptive local labels
//
#define FPOW_STATE_WAKE_ON_UPDATES  FPOW_STATE_OFF


//
// FPOW Driver Version
//
#define FPOW_DRIVER_VERSION "v0.54"

// Define to enable the proc entry
#define ENABLE_SNIFF_DEBUG_PROC


#undef FPOW_REGISTER_DEBUG
#ifdef FPOW_REGISTER_DEBUG
  #define RDEBUG(x...)		powdebug(x)
#else
  #define RDEBUG(x...)		do {} while (0)
#endif

//
// Define to debug semaphore use around HW accesses
//
#undef LOCK_DEBUG
#ifdef LOCK_DEBUG
  #define lockdebug(x...)		printk(x)
#else
  #define lockdebug(x...)		do {} while (0)
#endif

#define power_switch()  ((GPLR(g_fiona_hw.sleep_switch_gpio) & (1<<g_fiona_hw.sleep_switch_gpio)) != g_fiona_hw.sleep_switch_polarity)

#define wan_switch()    ((GPLR(GPIO14_FIONA_P3_WAN_EN) & GPIO_bit(GPIO14_FIONA_P3_WAN_EN)) ? 1 : 0)

typedef struct FPOWGlobals {
    // Our signature, used for structure verification
    int             signature;

    // user_count holds number of clients that have /dev/fpow open
    int             user_count;

    // system_state - Holds current System Power State
    int             system_state;

    // system_state - Holds current System Power State
    spinlock_t      state_lock;

    // State Translation Table to translate a  system state into power states
    FPOW_POWER_MODE power_state_xlate[FPOW_TOTAL_NUM_STATES][FPOW_TOTAL_NUM_CLASSES];

    // registered_component_count - number of currently registered components
    unsigned long   registered_component_count;

    struct fasync_struct    *rx_fasync;

    // Rx Client Holding Queue
    wait_queue_head_t       rx_client_q; 

    // timer structure for watermark checking
    struct timer_list watermark_timer;

}FPOWGlobals;


// List for registered power components
static LIST_HEAD(g_fpow_components_list);
static DECLARE_MUTEX(g_fpow_components_lock);

// List for pending power state transition requests
static LIST_HEAD(g_fpow_pending_mode_list);
static DECLARE_MUTEX(g_fpow_pending_mode_lock);

// Semaphore for driver control (only 1 fpow request at a time)
DECLARE_MUTEX(fpow_lock);


////////////////////////////////////////////////////////
// Globals
///////////////////////////////////////////////////////
static struct FPOWGlobals    globs;
struct class_simple          *g_fpow_class;
static struct proc_dir_entry *proc_fpow_parent = NULL;
static struct proc_dir_entry *proc_fpow_components_parent = NULL;
static struct proc_dir_entry *proc_fpow_watermarks_parent = NULL;
static struct proc_dir_entry *proc_fpow_wan_watermark = NULL;
static struct proc_dir_entry *proc_fpow_low_watermark = NULL;
static struct proc_dir_entry *proc_fpow_critical_watermark = NULL;
static struct proc_dir_entry *proc_fpow_reg_list = NULL;
static struct proc_dir_entry *proc_fpow_event = NULL;
static struct proc_dir_entry *proc_fpow_event_raw = NULL;
static struct proc_dir_entry *proc_fpow_debug = NULL;
static struct proc_dir_entry *proc_fpow_vdebug = NULL;
static struct proc_dir_entry *proc_fpow_vtbl_debug = NULL;
static struct proc_dir_entry *proc_fpow_event_debug = NULL;
static struct proc_dir_entry *proc_fpow_uc_parent = NULL;
static struct proc_dir_entry *proc_fpow_uc_power = NULL;
static struct proc_dir_entry *proc_fpow_uc_ac = NULL;
static struct proc_dir_entry *proc_fpow_uc_wan = NULL;
static struct proc_dir_entry *proc_fpow_uc_hp = NULL;
static struct proc_dir_entry *proc_fpow_system_state = NULL;
static struct proc_dir_entry *proc_fpow_version = NULL;

// Queue Sniffer Query Support
static struct proc_dir_entry *proc_fpow_acieiq = NULL;  // AC_INSERT Event In Queue
static struct proc_dir_entry *proc_fpow_peiq = NULL;    // Power On Event In Queue
static struct proc_dir_entry *proc_fpow_iweiq = NULL;   // Idle Wake Event In Queue
static struct proc_dir_entry *proc_fpow_poiq = NULL;    // Power Off Event In Queue

// Queue Sniffer Control Support
static struct proc_dir_entry *proc_fpow_seq = NULL;     // Stall Event Queue
static struct proc_dir_entry *proc_fpow_useq = NULL;    // UnStall Event Queue
static struct proc_dir_entry *proc_fpow_cueq = NULL;    // Clear and Unstall Event Queue

// Queue Sniffer Correction Routines
static struct proc_dir_entry *proc_fpow_fuseq = NULL;   // Fix and UnStall Event Queue
static struct proc_dir_entry *proc_fpow_fueqpl = NULL;  // Fix and UnStall Event Queue for Power Lockout
static struct proc_dir_entry *proc_fpow_iwcueq = NULL;  // Idle Wake Correct and Unstall Event Queue

// Queue Sniffer Logging Support
#ifdef ENABLE_SNIFF_DEBUG_PROC
static struct proc_dir_entry *proc_fpow_sniff_debug = NULL; // Queue sniff'er logging
#endif

static struct proc_dir_entry *proc_fpow_wake = NULL;        // Return Wake Reason....
static struct proc_dir_entry *proc_fpow_last_event = NULL;  // Return Last Event read from kbd driver

static struct proc_dir_entry *proc_fpow_cache = NULL;
static struct proc_dir_entry *proc_fpow_dialog_up = NULL;

static int g_fpow_pnlcd_disable_pending = 0;
static int g_fpow_pnlcd_enable_pending = 0;
static int g_fpow_eink_disable_pending = 0;
static int g_fpow_eink_enable_pending = 0;

static int  g_wan_watermark = 0;
static int  g_low_watermark = 0;
static int  g_critical_watermark = 0;

#ifdef ENABLE_EVENT_TEST_PROC
static struct proc_dir_entry *proc_fpow_event_test = NULL;
static int g_fiona_insert_event = 0;
int g_ps_ignore_next_wake = 0;
#endif

#ifdef CONFIG_FIONA_PM_CPU
// Static global holding current FPOW power state of the driver
static FPOW_POWER_MODE  g_cpu_power_mode = FPOW_MODE_400_MHZ;

// Static global holding fpow device entry returned from fpow_register()
static struct fpow_component *g_cpu_fpow_component_ptr = NULL;
#endif

static FPOW_WAKE_SOURCE g_fpow_wake_source = WAKE_INVALID;

#ifdef CONFIG_FIONA_VOLTAGE_DEBUG
int g_fiona_voltage_debug = 1;
#else
int g_fiona_voltage_debug = 0;
#endif // CONFIG_FIONA_VOLTAGE_DEBUG
EXPORT_SYMBOL(g_fiona_voltage_debug);


// Set g_fiona_vtable_debug to 1 to dump voltage table to log on each voltage read
int g_fiona_vtable_debug = 0;
EXPORT_SYMBOL(g_fiona_vtable_debug);


#ifdef CONFIG_FIONA_PM_DEBUG
int g_fiona_power_debug = 1;
#else
int g_fiona_power_debug = 0;
#endif // CONFIG_FIONA_PM_DEBUG
EXPORT_SYMBOL(g_fiona_power_debug);

#ifdef CONFIG_FIONA_SNIFF_DEBUG
int g_fiona_sniff_debug = 1;
#else
int g_fiona_sniff_debug = 0;
#endif // CONFIG_FIONA_PM_DEBUG
EXPORT_SYMBOL(g_fiona_sniff_debug);

#ifdef CONFIG_FIONA_EVENT_DEBUG
int g_fiona_event_debug = 1;
#else
int g_fiona_event_debug = 0;
#endif // CONFIG_FIONA_PM_DEBUG
EXPORT_SYMBOL(g_fiona_event_debug);

int g_threads_are_stopped = 0;

//////////////////////////////
// External Structs and Prototypes
//////////////////////////////
extern int fiona_suspend_system(void);
extern int ac_adapter_status(void);

//////////////////////////////
// Local Prototypes
//////////////////////////////
static int  FPOW_Open(struct inode* inNode, struct file* inFile);
static int  FPOW_Close(struct inode* inNode, struct file* inFile);
static int  FPOW_Ioctl(struct inode *inNode, struct file *inFile, unsigned int inCommand, unsigned long inArg);
static int  FPOW_Close(struct inode* inNode, struct file* inFile);
static int  FPOWDriver_Remove(struct device *dev);
static void FPOWDriver_Platform_Release(struct device *device);
int         fasync_fpow(int fd, struct file *filp, int on);
static void do_display_fast_screen(int *screen);
static void do_display_fast_screen_and_sleep(int *screen);
static int  find_fix_idle_wake_event_in_event_queue(void);
static char *get_expected_results(int state);
void        set_remembered_wake_source(int event);

//////////////////////////////
// Power Management Support
//////////////////////////////
#define FPOWDriver_Suspend   NULL
#define FPOWDriver_Resume    NULL
//static int  FPOWDriver_Suspend(struct device *dev, u32 state, u32 level);
//static int  FPOWDriver_Resume(struct device *dev, u32 level);

static FPOW_ERR fiona_power_state(FPOW_POWER_STATE state);
static void fpow_print_registered_component_list(void);
char * 	fpow_passed_event_to_str(int event);
char * 	fpow_mode_to_str(FPOW_POWER_MODE mode);
u32		fpow_mode_to_syspm_level(FPOW_POWER_MODE mode);
char *	fpow_state_to_str(FPOW_POWER_STATE state);
char *	fpow_class_to_str(FPOW_POWER_CLASS class);
char *  fpow_wake_src_to_str(FPOW_WAKE_SOURCE src);
int		fpow_send_cmd(struct fpow_component *dev, FPOW_POWER_MODE mode);
static  FPOW_WAKE_SOURCE determine_ioc_wake_reason(void);
static void post_wake_event(void);
static int fpow_send_to_class(FPOW_POWER_MODE mode, FPOW_POWER_CLASS class);
static void do_init_start_watermark_timer(void);

#ifdef CONFIG_FIONA_PM_CPU
static int cpu_pm_init(void);
#endif
// Exported File Operators
//  Init and Destroy Routines
//////////////////////////////

static struct device_driver fpow_driver = {
    .name = "fpow",
    .bus = &platform_bus_type,
    .remove = FPOWDriver_Remove,
    .suspend = FPOWDriver_Suspend,
    .resume = FPOWDriver_Resume,
};

static struct platform_device fpow_device = {
    .name = "fpow",
    .id = 0,
    .dev = {
        .release = FPOWDriver_Platform_Release,
    }
};

struct file_operations fpow_fops = {
        owner:          THIS_MODULE,
        open:           FPOW_Open,
        ioctl:          FPOW_Ioctl,
        release:        FPOW_Close,
        fasync:         fasync_fpow
};

static struct miscdevice g_fpow_misc_device = {
    minor:      FPOW_MINOR,
    name:       FPOW_DRIVER_NAME,
    fops:       &fpow_fops,
};


// /proc interface
static int fpow_proc_init(void);
static void fpow_cleanup_procfs(void);
static void fpowSetWanWatermark(int value);
static void fpowSetLowWatermark(int value);
static void fpowSetCriticalWatermark(int value);
int fpowGetWanWatermark(void);
int fpowGetLowWatermark(void);
int fpowGetCriticalWatermark(void);


static void 
init_power_xlate_table(void)
{
    int state;
    struct FPOWGlobals *g = &globs;

    for (state = FPOW_STATE_FIRST; state <= FPOW_STATE_LAST; state++)  {
        switch(state)  {
		    case FPOW_STATE_A:  // Active / A1
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_400_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_SLEEP;
                break;

		    case FPOW_STATE_B:  // Active / A2
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_400_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_SLEEP;
                break;

		    case FPOW_STATE_C:  // Active / B1
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_400_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_ON;
                break;

		    case FPOW_STATE_D:  // Active / B2
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_400_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_ON;
                break;

		    case FPOW_STATE_E:  // Passive / PA1
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_SLEEP;
                break;

		    case FPOW_STATE_F:  // Passive / PA2
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_SLEEP;
                break;

		    case FPOW_STATE_G:  // Passive / PB1
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_ON;
                break;

		    case FPOW_STATE_H:  // Passive / PB2
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_ON;
                break;

		    case FPOW_STATE_I:  // Rest / RA1 KH
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_J:  // USB Wake Temporary Mode w/out WAN
            case FPOW_STATE_YY: // USB Faded Wake Temporary Mode w/out WAN
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_400_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_K:  // USB Wake Temporary Mode w WAN
            case FPOW_STATE_XX: // USB Faded Wake Temporary Mode wt WAN
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_400_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_L:  // Rest - RA2 KH
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_M:  // Rest - Off
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_N:  // First temporary state out-of-off-sleep....
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_WW:  // First temporary state out-of-idle-sleep....
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_O:  // Password w/ WAN w/o Audio
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_SLEEP;
                break;

		    case FPOW_STATE_P:  // Password w/o WAN w/o Audio
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_SLEEP;
                break;

		    case FPOW_STATE_Q:  // Password w WAN w Audio
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_ON;
                break;

		    case FPOW_STATE_R:  // Password w/o WAN w Audio
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_ON;
                break;

		    case FPOW_STATE_S:  // USB Mounted w/out WAN
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_400_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

            // Can't call or get to any of these states via SW.
            // They're all reached via HW (usually running out of juice)
		    case FPOW_STATE_T:          // OFF - No Power At All
		    case FPOW_STATE_U:          // DEAD
                break;

		    case FPOW_STATE_V:  // Power Lockout
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_ZZ: // Rest TPH Wake State from Z
		    case FPOW_STATE_W:  // Passive / TPH
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_200_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_X:  // USB Mounted w/ WAN
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_400_MHZ;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_Y:  // Rest State w/out WAN w/out screen content clear
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    case FPOW_STATE_Z:  // Rest State w/ WAN w/out screen content clear
                g->power_state_xlate[state][FPOW_CLASS_CPU] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_DISPLAY] = FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_WAN] =           FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_USB_HOST] =      FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_USB_DEVICE] =    FPOW_MODE_SLEEP;
                g->power_state_xlate[state][FPOW_CLASS_MMC] =           FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_IOC] =           FPOW_MODE_OFF;
                g->power_state_xlate[state][FPOW_CLASS_KBD_BUTTONS] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_SCROLLWHEEL] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_VIDEO_PNLCD] =   FPOW_MODE_ON;
                g->power_state_xlate[state][FPOW_CLASS_AUDIO] =         FPOW_MODE_OFF;
                break;

		    default:
			    break;
	    }
    }
}


//////////////////////////////
// Core Driver Routines
//////////////////////////////
static int
FPOW_Open(struct inode* inNode, struct file* inFile)
{
    // Inc user count
    globs.user_count++;

    // Return success
    return 0;
}

static int
FPOW_Close(struct inode* inNode, struct file* inFile)
{
    // Decrement user count
    globs.user_count--;

    // Return success
    return 0;
}

char *
fpow_event_to_str(int event)
{
    switch(event)  {
        case NO_WAN_BATTERY_EVENT:
            return("BATTERY_NO_WAN");
        case LOW_BATTERY_EVENT:
            return("BATTERY_LOW");
        case CRITICAL_BATTERY_EVENT:
            return("BATTERY_CRITICAL");
        case WAN_SWITCH_ON_EVENT:
            return("WAN_SWITCH_ON");
        case WAN_SWITCH_OFF_EVENT:
            return("WAN_SWITCH_OFF");
        case AC_PLUG_INSERTED_EVENT:
            return("AC_INSERTED");
        case AC_PLUG_REMOVED_EVENT:
            return("AC_REMOVED");
        case USB_PLUG_INSERTED_EVENT:
            return("USB_PLUG_INSERTED");
        case USB_PLUG_REMOVED_EVENT:
            return("USB_PLUG_REMOVED");
        case HEADPHONES_INSERTED_EVENT:
            return("HEADPHONES_INSERTED");
        case HEADPHONES_REMOVED_EVENT:
            return("HEADPHONES_REMOVED");
        case WAN_RING_EVENT:
            return("WAN_RING");
        case KEYLOCK_EVENT:
            return("KEYLOCK_WAKE");
        case AUDIO_ON_EVENT:
            return("AUDIO_ON");
        case AUDIO_OFF_EVENT:
            return("AUDIO_OFF");
        case POWER_SWITCH_ON_EVENT:
            return("POWER_SWITCH_ON");
        case POWER_SWITCH_OFF_EVENT:
            return("POWER_SWITCH_OFF");
        case MMC_EVENT:
            return("MMC");
        case VALID_PASSWORD_EVENT:
            return("VALID_PASSWORD");
        case INVALID_PASSWORD_EVENT:
            return("INVALID_PASSWORD");
        case TIMEOUT_EVENT:
            return("T1");
        case TIMEOUT2_EVENT:
            return("T2");
        case TIMEOUT3_EVENT:
            return("T3");
        case TIMEOUT_USB_WAKE_EVENT:
            return("TUSB");
        case WATERMARK_CHECK_EVENT:
            return("VC");
        case PRIME_EVENT:
            return("PRIME");
        case TPH_COMPLETE_EVENT:
            return("TPH_COMP");
        case USB_WAKE_EVENT:
            return("USB_WAKE");
        case WAN_SWITCH_SWEEPER_EVENT:
            return("WAN_SWEEP");
        case TWO_FINGER_SWEEPER_EVENT:
            return("AF_SWEEP");
        default:
            return("KBS");
    }
}

char *
fpow_passed_event_to_str(int event)
{
    switch(event)  {
        case EVENT_BATTERY_NO_WAN:
            return("BATTERY_NO_WAN");
        case EVENT_BATTERY_LOW:
            return("BATTERY_LOW");
        case EVENT_BATTERY_CRITICAL:
            return("BATTERY_CRITICAL");
        case EVENT_WAN_SWITCH_ON:
            return("WAN_SWITCH_ON");
        case EVENT_WAN_SWITCH_OFF:
            return("WAN_SWITCH_OFF");
        case EVENT_AC_INSERTED:
            return("AC_INSERTED");
        case EVENT_AC_REMOVED:
            return("AC_REMOVED");
        case EVENT_USB_PLUG_INSERTED:
            return("USB_PLUG_INSERTED");
        case EVENT_USB_PLUG_REMOVED:
            return("USB_PLUG_REMOVED");
        case EVENT_HEADPHONES_INSERTED:
            return("HEADPHONES_INSERTED");
        case EVENT_HEADPHONES_REMOVED:
            return("HEADPHONES_REMOVED");
        case EVENT_WAN_RING:
            return("WAN_RING");
        case EVENT_KEY_WAKE:
            return("KEY_WAKE");
        case EVENT_AUDIO_ON:
            return("AUDIO_ON");
        case EVENT_AUDIO_OFF:
            return("AUDIO_OFF");
        case EVENT_POWER_SWITCH_ON:
            return("POWER_SWITCH_ON");
        case EVENT_POWER_SWITCH_OFF:
            return("POWER_SWITCH_OFF");
        case EVENT_MMC:
            return("MMC");
        case EVENT_VALID_PASSWORD:
            return("VALID_PASSWORD");
        case EVENT_INVALID_PASSWORD:
            return("INVALID_PASSWORD");
        case EVENT_TIMEOUT_1:
            return("T1");
        case EVENT_TIMEOUT_2:
            return("T2");
        case EVENT_TIMEOUT_3:
            return("T3");
        case EVENT_TIMEOUT_USB_WAKE:
            return("TUSB");
        case EVENT_WATERMARK_CHECK:
            return("VC");
        case EVENT_PRIME_PUMP:
            return("PRIME");
        case EVENT_TPH_COMPLETE:
            return("TPH_COMP");
        case EVENT_USB_WAKE:
            return("USB_WAKE");
        case EVENT_WAN_SWITCH_SWEEPER:
            return("WAN_SWEEP");
        case EVENT_TWO_FINGER_SWEEPER:
            return("AF_SWEEP");
        default:
            return("KBS");
    }
}

static char *
fpow_ioctl_cmd_to_str(unsigned int inCommand)  {
    switch(inCommand)  {
        case FPOW_GET_STATE_IOCTL:
            return("FPOW_GET_STATE_IOCTL");
        case FPOW_SET_STATE_IOCTL:
            return("FPOW_GET_STATE_IOCTL");
        case FPOW_GET_WAKE_SRC_IOCTL:
            return("FPOW_GET_WAKE_SRC_IOCTL");
        case FPOW_CLEAR_DISABLE_VIDEO_IOCTL:
            return("FPOW_CLEAR_DISABLE_VIDEO_IOCTL");
        case FPOW_ENABLE_VIDEO_IOCTL:
            return("FPOW_ENABLE_VIDEO_IOCTL");
        case FPOW_DISPLAY_FAST_SCREEN_IOCTL:
            return("FPOW_DISPLAY_FAST_SCREEN_IOCTL");
        case FPOW_DISPLAY_FAST_SCREEN_SLEEP_IOCTL:
            return("FPOW_DISPLAY_FAST_SCREEN_SLEEP_IOCTL");
        default:
            return("UNKNOWN STATE !!");
    }
}

int fpow_get_eink_enable_pending(void) {
    return(g_fpow_eink_enable_pending);
} 
int fpow_get_eink_disable_pending(void) {
    return(g_fpow_eink_disable_pending);
} 
int fpow_get_pnlcd_enable_pending(void) {
    return(g_fpow_pnlcd_enable_pending);
} 
int fpow_get_pnlcd_disable_pending(void) {
    return(g_fpow_pnlcd_disable_pending);
} 

void fpow_clear_pnlcd_enable_pending(void) {
    g_fpow_pnlcd_enable_pending = 0;
}
void fpow_clear_pnlcd_disable_pending(void) {
    g_fpow_pnlcd_disable_pending = 0;
}
void fpow_clear_eink_enable_pending(void) {
    g_fpow_eink_enable_pending = 0;
}
void fpow_clear_eink_disable_pending(void) {
    g_fpow_eink_disable_pending = 0;
}

void fpow_set_pnlcd_enable_pending(void) {
    g_fpow_pnlcd_enable_pending = 1;
    fpow_clear_pnlcd_disable_pending();
}

void fpow_set_pnlcd_disable_pending(void) {
    g_fpow_pnlcd_disable_pending = 1;
    fpow_clear_pnlcd_enable_pending();
}

void fpow_set_eink_enable_pending(void) {
    g_fpow_eink_enable_pending = 1;
    fpow_clear_eink_disable_pending();
}

void fpow_set_eink_disable_pending(void) {
    g_fpow_eink_disable_pending = 1;
    fpow_clear_eink_enable_pending();
}

EXPORT_SYMBOL(fpow_get_eink_disable_pending);
EXPORT_SYMBOL(fpow_set_eink_disable_pending);
EXPORT_SYMBOL(fpow_clear_eink_disable_pending);
EXPORT_SYMBOL(fpow_clear_eink_enable_pending);

int __fpow_enable_video(int showScreen) {
    int retVal = 0;
    int retVal2 = 0;
    int screen = splash_screen_powering_on; // "Kindle is Powering On"

    // Send an ON command to the PNLCD video component
    powdebug("_fpow_enable_video() calling PNLCD_ENABLE_MODE\n");
    retVal2 = fpow_send_to_class(PNLCD_ENABLE_MODE, FPOW_CLASS_VIDEO_PNLCD);
    if (retVal2 != FPOW_NO_ERR)  {
        fpow_set_pnlcd_enable_pending();
        if (retVal == -FPOW_NO_MATCHES_FOUND_ERR) {
            // No PNLCD handler registered yet, deal with it directly
            powdebug("No pnlcd power component registered yet, set enable pending\n");
        }
        else
            printk("Power transition failed err=%d for pnlcd video display class\n",retVal2);
    }
    else  {
        // Clear any pending PNLCD ENABLE requests, we just enabled it
        powdebug("_fpow_enable_video() enabled PNLCD.\n");
        fpow_clear_pnlcd_enable_pending();
    }


    // Send an ON command to the main video component
    powdebug("__fpow_enable_video(%d) calling VIDEO_ENABLE_MODE\n",showScreen);
    retVal = fpow_send_to_class(VIDEO_ENABLE_MODE, FPOW_CLASS_VIDEO_DISPLAY);
    if (retVal != FPOW_NO_ERR)  {
        fpow_set_eink_enable_pending();
        if (retVal == -FPOW_NO_MATCHES_FOUND_ERR) {
            // No EINK handler registered yet, deal with it directly
            powdebug("No video power component registered yet, set enable pending\n");
        }
        else  
            printk("Power transition failed err=%d for video display class\n",retVal);
    }
    else  {
        fpow_clear_eink_enable_pending();
        // Draw the "Your Kindle is Powering Up" banner
        if (showScreen)
            do_display_fast_screen(&screen);
    }

    return((retVal | retVal2));
}

int fpow_enable_video(int showScreen) {
    int retVal = 0;

    powdebug("fpow_enable_video(%d)\n",showScreen);
    lockdebug("fpow_enable_video(%d, pid=%d) getting lock...\n",showScreen,(int)sys_getpid());
    if (down_interruptible(&fpow_lock)) {
        lockdebug("fpow_enable_video(%d, pid=%d) COULD NOT get lock!\n",showScreen,(int)sys_getpid());
        return(-ERESTARTSYS);
    }
    lockdebug("fpow_enable_video(pid=%d) got lock...\n",(int)sys_getpid());

    retVal = __fpow_enable_video(showScreen);

    up(&fpow_lock);
    lockdebug("fpow_enable_video(pid=%d) released lock...\n",(int)sys_getpid());

    return(retVal);
}


int __fpow_clear_disable_video(void) {
    int retVal1 = 0;
    int retVal2 = 0;

    // Clear and disable PNLCD video component
    powdebug("__fpow_clear_disable_video() calling PNLCD_CLEAR_DISABLE_MODE \n");
    retVal2 = fpow_send_to_class(PNLCD_CLEAR_DISABLE_MODE, FPOW_CLASS_VIDEO_PNLCD);
    if (retVal2 != FPOW_NO_ERR)  {
        fpow_set_pnlcd_disable_pending();
        if (retVal2 == -FPOW_NO_MATCHES_FOUND_ERR) {
            // No PNLCD handler registered yet, deal with it directly
            powdebug("No pnlcd power component registered yet, set enable pending\n");
        }
        else
            printk("Power transition failed err=%d for pnlcd video display class\n",retVal2);
    }
    else  {
        // Clear any pending PNLCD ENABLE requests, we just enabled it
        powdebug("_fpow_enable_video() enabled PNLCD.\n");
        fpow_clear_pnlcd_disable_pending();
    }

    // Clear and disable EINK display
    powdebug("__fpow_clear_disable_video() calling eink_sys_ioctl(OFF_CLEAR_SCREEN)...\n");
    retVal1 = eink_sys_ioctl(FBIO_EINK_OFF_CLEAR_SCREEN, NULL);
    powdebug("__fpow_clear_disable_video() back from eink_sys_ioctl(OFF_CLEAR_SCREEN)...\n");

    return((retVal1|retVal2));
}

int fpow_clear_disable_video(void) {
    int retVal = 0;

    lockdebug("fpow_clear_disable_video(pid=%d) getting lock...\n",(int)sys_getpid());
    if (down_interruptible(&fpow_lock)) {
        lockdebug("fpow_clear_disable_video(pid=%d) COULD NOT get lock!\n",(int)sys_getpid());
        return(-ERESTARTSYS);
    }
    lockdebug("fpow_clear_disable_video(pid=%d) got lock...\n",(int)sys_getpid());

    retVal = __fpow_clear_disable_video();

    up(&fpow_lock);
    lockdebug("fpow_clear_disable_video(pid=%d) released lock...\n",(int)sys_getpid());

    return(retVal);
}

static void do_display_fast_screen(int *screen)  {
    powdebug("do_display_fast_screen(%d) start...\n",*screen);
    eink_sys_ioctl(FBIO_EINK_SPLASH_SCREEN, *screen);
    powdebug("do_display_fast_screen() end.\n");
}


static void do_display_fast_screen_and_sleep(int *screen)  {
    powdebug("do_display_fast_screen_and_sleep(%d) start...\n",*screen);
    eink_sys_ioctl(FBIO_EINK_SPLASH_SCREEN_SLEEP, *screen);
    powdebug("do_display_fast_screen_and_sleep() end.\n");
}


static int
FPOW_Ioctl(struct inode *inNode, struct file *inFile,
            unsigned int inCommand, unsigned long inArg)
{
    int                         retVal = 0;
    FPOW_POWER_STATE_REC        stateRec;
    FPOW_WAKE_SRC_REC           wakeRec;

    powdebug("FPOW_Ioctl( %s )\n", fpow_ioctl_cmd_to_str(inCommand));

    if (_IOC_TYPE(inCommand) != FPOW_IOCTL_MAGIC_NUMBER) return -EINVAL;
    if (_IOC_NR(inCommand) > FPOW_IOCTLLast) return -EINVAL;

    lockdebug("FPOW_Ioctl( %s,pid=%d ) getting lock...\n", fpow_ioctl_cmd_to_str(inCommand),(int)sys_getpid());
    if (down_interruptible(&fpow_lock)) {
        lockdebug("FPOW_Ioctl( %s,pid=%d ) COULD NOT get lock!\n", fpow_ioctl_cmd_to_str(inCommand),(int)sys_getpid());
        retVal = -ERESTARTSYS;
        goto exit;
    }
    lockdebug("FPOW_Ioctl( %s,pid=%d ) got lock...\n", fpow_ioctl_cmd_to_str(inCommand),(int)sys_getpid());

    // Get user data
    switch(inCommand)  {
        case FPOW_GET_STATE_IOCTL:
            stateRec.state = globs.system_state;
            powdebug("FPOW_GET_STATE_IOCTL returning system_state = %d\n",stateRec.state);
            copy_to_user((void*) inArg, &stateRec, sizeof(stateRec));
            break;

        case FPOW_SET_STATE_IOCTL:
            // If we were called from inside, don't copy from user space
            if (inNode == (struct inode *)0xDEADBEEF)
                memcpy(&stateRec,(void*)inArg,sizeof(stateRec));
            else  {
                if (copy_from_user(&stateRec,(void *)inArg, sizeof(stateRec)))  {
                    powdebug("FPOW_IOCTL() failed copy_from_user !! Kernel Space ??? \n");
                    goto errOut;
                }
            }

            // Range-check the state
            if (!((stateRec.state >= FPOW_STATE_FIRST) && (stateRec.state <= FPOW_STATE_LAST))) {
                powdebug("State passed to FPOW_Ioctl(FPOW_SET_STATE_IOCTL) is out of range (%d is NOT >= %d AND <= %d)!\n",stateRec.state,FPOW_STATE_FIRST,FPOW_STATE_LAST);
                goto errOut;
            }

            // Transition power states
            retVal = fiona_power_state(stateRec.state);
            extpowdebug("FPOW_SET_STATE_IOCTL() set system_state to %s\n",fpow_state_to_str(globs.system_state));
            break;


        case FPOW_ENABLE_VIDEO_IOCTL:
            powdebug("FPOW_IOCTL(FPOW_ENABLE_VIDEO_IOCTL) calling fpow_enable_video(%d)\n",(int)inArg);
            retVal = __fpow_enable_video(inArg);
            break;

        case FPOW_CLEAR_DISABLE_VIDEO_IOCTL:
            retVal = __fpow_clear_disable_video();
            break;


        case FPOW_DISPLAY_FAST_SCREEN_IOCTL:
            // Send an DISPLAY_SCREEN command to the video driver
            powdebug("FPOW_IOCTL(FPOW_DISPLAY_FAST_SCREEN_IOCTL) calling do_display_fast_screen(%d)\n",*(int*)inArg);
            do_display_fast_screen((int *) inArg);
            break;

        case FPOW_DISPLAY_FAST_SCREEN_SLEEP_IOCTL:
            // Send an DISPLAY_SCREEN command to the video driver
            powdebug("FPOW_IOCTL(FPOW_DISPLAY_FAST_SCREEN_SLEEP_IOCTL) calling do_display_fast_screen_and_sleep(%d)\n",*(int*)inArg);
            do_display_fast_screen_and_sleep((int *) inArg);
            break;

        case FPOW_GET_WAKE_SRC_IOCTL:
            // If it was the IOC, go ask it now which reason it woke for...
            if (g_fpow_wake_source == IOC_WAKE)  {
                g_fpow_wake_source = determine_ioc_wake_reason();
                printk("FPOW_GET_WAKE_SRC_IOCTL() back from determine_ioc_wake_reason(), got %d\n",g_fpow_wake_source);
            }
            else
                printk("FPOW_GET_WAKE_SRC_IOCTL() returning known wake source %s (%d)\n",fpow_wake_src_to_str(g_fpow_wake_source),g_fpow_wake_source);

            wakeRec.source = g_fpow_wake_source;
            copy_to_user((void*) inArg, &wakeRec, sizeof(wakeRec));
            break;

        default:
            printk("FPOW_IOCTL() - unknown command 0x%x\n",inCommand);
            goto errOut;
    }
exit:
    up(&fpow_lock);
    lockdebug("FPOW_Ioctl( %s,pid=%d ) released lock...\n", fpow_ioctl_cmd_to_str(inCommand),(int)sys_getpid());

    extpowdebug("FPOW_Ioctl( %s ) returning %d\n", fpow_ioctl_cmd_to_str(inCommand),retVal);
    return(retVal);

errOut:
    up(&fpow_lock);
    lockdebug("FPOW_Ioctl( %s,pid=%d ) released lock...\n", fpow_ioctl_cmd_to_str(inCommand),(int)sys_getpid());
    powdebug("FPOW_Ioctl( %s ) returning %d\n", fpow_ioctl_cmd_to_str(inCommand),-EFAULT);
    return(-EFAULT);
}


static void
FPOWDriver_Platform_Release(
    struct device *device)
{
    // this is called when the reference count goes to zero
}



int fasync_fpow(int fd, struct file *filp, int on)
{
#if 0
    int retval;

    retval = fasync_helper(fd, filp, on, &g_hal_globs->rx_q.fasync);
    if( retval < 0)
        return retval;
#endif
    return 0;
}

static int
invalid_fpow_state(FPOW_POWER_STATE state)
{
    return(!((state >= FPOW_STATE_FIRST) && (state <= FPOW_STATE_LAST)));
}

static int
__init FPOWDriver_Init(void)
{
    int retVal = 0;

    // Initialize our globals...
    if (globs.signature != FPOW_GLOBALS_SIGNATURE)  {
        memclr(&globs,sizeof(globs));
        globs.signature = FPOW_GLOBALS_SIGNATURE;
        spin_lock_init(&globs.state_lock);
        globs.system_state = FPOW_STARTING_STATE;
        globs.rx_client_q.lock = SPIN_LOCK_UNLOCKED; 
        globs.rx_client_q.task_list.next = &globs.rx_client_q.task_list;
        globs.rx_client_q.task_list.prev = &globs.rx_client_q.task_list;
        g_fpow_components_list.next = &g_fpow_components_list;
        g_fpow_components_list.prev = &g_fpow_components_list;

        // Initialize the Power State Translation Table
        init_power_xlate_table();
    }

    // Initialize watermark variables
    fpowSetWanWatermark(g_fiona_hw.wan_watermark);
    fpowSetLowWatermark(g_fiona_hw.low_watermark);
    fpowSetCriticalWatermark(g_fiona_hw.critical_watermark);

    // Setup our own timer to watch watermarks when IOC is sleeping
    do_init_start_watermark_timer();

    // Initialize /proc interface
    fpow_proc_init();

    // Register our driver
    retVal = misc_register(&g_fpow_misc_device);
    if (retVal < 0)  {
        printk("***** FPOWDriver_Init failed to register misc driver !!\n");
        misc_deregister(&g_fpow_misc_device);

        // We couldn't register the driver, but we're going to leave
        // the /proc interface in place as it should still work
        // fpow_cleanup_procfs();
        goto exit;
    }

    retVal = driver_register(&fpow_driver);
    if (!retVal)  {
        retVal = platform_device_register(&fpow_device);
        if (retVal)  {
            printk("***** FPOWDriver_Init failed to platform_device_register!!\n");
            driver_unregister(&fpow_driver);
            goto exit;
        }
    }
    else {
        printk("***** FPOWDriver_Init failed to driver_register!!\n");
        goto exit;
    }

#ifdef CONFIG_FIONA_PM_CPU
    // Initialize the CPU client for FPOW
    cpu_pm_init();
#endif

    printk("FPOW: Fiona Power Management Driver %s\n",FPOW_DRIVER_VERSION);
exit:
    return(retVal);
}

static int 
FPOWDriver_Remove(struct device *dev)
{
    fpow_cleanup_procfs();
    return(0);
}

static void
__exit FPOWDriver_Exit(void)
{
    fpow_cleanup_procfs();
    misc_deregister(&g_fpow_misc_device);
    devfs_remove(FPOW_DRIVER_NAME);
    unregister_chrdev(FPOW_MAJOR,FPOW_DRIVER_NAME);
    class_simple_destroy(g_fpow_class);
}


////////////////////////////////////////////////////////
//
// /proc Interface Routines
//
////////////////////////////////////////////////////////


/******************************************************************************
 *
 * /proc/fpow/iweiq Support
 *
 * 'iweiq' is short for "Idle Wake Event In Queue", meaning that a wake event
 * for an idle sleep state is pending in the queue.
 *
 * Reading from /proc/fpow/iweiq will return :
 *  1 - Idle Wake Event is in the event queue
 *  0 - No Idle Wake events are in the event queue at this time
 *
 * Writing to /proc/fpow/iweiq is not supported.
 ******************************************************************************/
static int proc_read_fpow_iweiq(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;
    int wake_event_in_queue = 0;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    if (count < 2)
        return -EINVAL;

    // If there's a wake event in the queue, indicate such
    if (INVALID_EVENT != idle_wake_event_in_event_queue()) 
        wake_event_in_queue = 1;

    len = sprintf(page, "%d\n",wake_event_in_queue);

    *eof = 1;
    return len;
}

/******************************************************************************
 *
 * /proc/fpow/poiq Support
 *
 * 'poiq' is short for "Power Off Event In Queue", meaning that a Power Off event
 * is pending in the queue.
 *
 * Reading from /proc/fpow/poiq will return :
 *  1 - Power Off Event is in the event queue
 *  0 - No Power Off events are in the event queue at this time
 *
 * Writing to /proc/fpow/poiq is not supported.
 ******************************************************************************/
static int proc_read_fpow_poiq(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;
    int found = 0;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    if (count < 2)
        return -EINVAL;

    found = find_event_in_queue(POWER_SWITCH_OFF_EVENT);
    len = sprintf(page, "%d\n",found);

    *eof = 1;
    return len;
}

/******************************************************************************
 *
 * /proc/fpow/acieiq Support
 *
 * 'acieiq' is short for "AC Insert Event In Queue", meaning that an AC Inserted
 *  event is pending in the event queue.
 *
 * Reading from /proc/fpow/acieiq will return :
 *  1 - AC Insert Event is in the event queue
 *  0 - No AC Insert events are in the event queue at this time
 *
 * Writing to /proc/fpow/acieiq is not supported.
 ******************************************************************************/
static int proc_read_fpow_acieiq(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;
    int acInQ = 0;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    if (count < 2)
        return -EINVAL;

    // If there's a power on in the queue, fix the queue up and check again to
    // assure it's still there before returning true....
    acInQ = find_event_in_queue(AC_PLUG_INSERTED_EVENT);
    hlog("proc_read_fpow_acieiq() %s AC_INSERT event in the q\n",(acInQ?"saw an":"didn't see an"));
    len = sprintf(page, "%d\n",acInQ);

    *eof = 1;
    return len;
}

/******************************************************************************
 *
 * /proc/fpow/peiq Support
 *
 * 'peiq' is short for "Power Event In Queue", meaning that a Power On event
 * is pending in the queue.
 *
 * Reading from /proc/fpow/peiq will return :
 *  1 - Power On Event is in the event queue
 *  0 - No Power On events are in the event queue at this time
 *
 * Writing to /proc/fpow/peiq is not supported.
 ******************************************************************************/
static int proc_read_fpow_peiq(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;
    int onInQ = 0;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    if (count < 2)
        return -EINVAL;

    // If there's a power on in the queue, fix the queue up and check again to
    // assure it's still there before returning true....
    if (find_event_in_queue(POWER_SWITCH_ON_EVENT))  {
        hlog("proc_read_fpow_peiq() saw ON in q, calling fixEventQueue\n");
        onInQ = fixEventQueue(POWER_SWITCH_ON_EVENT);
        hlog("proc_read_fpow_peiq(): %sPOWER_ON in queue after calling fixEventQueue\n",(onInQ?"":"No "));
    }
    len = sprintf(page, "%d\n",onInQ);

    *eof = 1;
    return len;
}

/******************************************************************************
 *
 * /proc/fpow/last_event Support
 *
 * Reading from /proc/fpow/last_event will return the last event that was read
 *  from the keyboard driver.
 *
 * Writing to /proc/fpow/last_event is unsupported.
 *
 ******************************************************************************/
static int proc_read_last_event(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    if (count < 2)
        return -EINVAL;

    len = sprintf(page, "%d\n",get_last_passed_event());

    *eof = 1;
    return len;
}


/******************************************************************************
 *
 * /proc/fpow/wake Support
 *
 * 'wake' is short for "wake reason".
 *
 * Reading from /proc/fpow/wake will return the last event that woke the device
 *
 * Writing to it is unsupported.
 *
 ******************************************************************************/
static int proc_read_wake_reason(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    if (count < 2)
        return -EINVAL;

    len = sprintf(page, "%d\n",fiona_get_wake_source());

    *eof = 1;
    return len;
}


/******************************************************************************
 *
 * /proc/fpow/seq Support
 *
 * 'seq' is short for "Stall Event Queue".
 *  Writing a 1 to seq will stall the event queue
 *  Writing a 0 has no effect.
 *
 * Reading from /proc/fpow/seq will return if event queue is currently stalled.
 *
 ******************************************************************************/
static int proc_read_fpow_event_queue_stalled(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    if (count < 2)
        return -EINVAL;

    len = sprintf(page, "%d\n",isEventQueueStalled());

    *eof = 1;
    return len;
}


/******************************************************************************
 *
 * /proc/fpow/dialog_up Support
 *
 * Writing a value to /proc/fpow/dialog_up simply remembers it.
 * Reading from /proc/fpow/dialog_up will return last value written to 
 *  /proc/fpow/dialog_up
 *
 *  This is used by Java to indicate to user space when a dialog is up.
 *  This is highly useful for automated scripts.
 *
 ******************************************************************************/
static int g_fpow_dialog_up = 0;
static int proc_read_fpow_dialog_up(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len = 0;
    if (count < 2)
        return -EINVAL;

    // Skip second round calls
    if (off)
        goto exit;

    len = sprintf(page, "%d\n",g_fpow_dialog_up);
exit:
    *eof = 1;
    return len;
}

/******************************************************************************/

static int proc_write_fpow_dialog_up(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[8];
    int rc = 0;
    int value = 0;
    char *new_line = NULL;

    if (count > 7)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

    if (copy_from_user(valueStr, buffer, count)) {
        return -EFAULT;
    }

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value < 0) || (value > 1))
        goto usage;

    printk("FPOW: dialog_up now set to %d.\n",value);
    g_fpow_dialog_up = value;

    return count;

usage:
    printk("You must write a value between 0 and 1 to /proc/fpow/dialog_up (saw %d)\n",value);
    return count;   // Don't error or we'll get called again
}


/******************************************************************************
 *
 * /proc/fpow/cache Support
 *
 * Writing a value to /proc/fpow/cache simply remembers it.
 * Reading from /proc/fpow/cache will return last value written to /proc/fpow/cache
 *
 ******************************************************************************/
static int g_fpow_cached_value = 0;
static int proc_read_fpow_cache(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len = 0;
    if (count < 2)
        return -EINVAL;

    // Skip second round calls
    if (off)
        goto exit;

    len = sprintf(page, "%d\n",g_fpow_cached_value);
exit:
    *eof = 1;
    return len;
}

/******************************************************************************/

static int proc_write_fpow_cache(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[8];
    int rc = 0;
    int value = 0;
    char *new_line = NULL;

    if (count > 7)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

    if (copy_from_user(valueStr, buffer, count)) {
        return -EFAULT;
    }

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value < 0) || (value > 10))
        goto usage;

    printk("FPOW: Cached away value of %d for debug purposed.\n",value);
    g_fpow_cached_value = value;

    return count;

usage:
    printk("You must write a value between 0 and 10 to /proc/fpow/cache (saw %d)\n",value);
    return count;   // Don't error or we'll get called again
}


/******************************************************************************
 *
 * /proc/fpow/iwcueq Support
 *
 * 'iwcueq' is short for "Idle Wake Correct and Unstall Event Queue".
 *  Writing a 1 to iwcueq will fix and unstall the event queue
 *  Writing a 0 has no effect.
 *
 * Reading from /proc/fpow/iwcueq will return if event queue is currently stalled.
 *
 ******************************************************************************/
static int proc_write_fpow_iwcueq(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[10];
    int rc = 0;
    int value = 0;
    char *new_line = NULL;
    int wakeEventFoundInQ = INVALID_EVENT;

    // If too much data for our buffer, show usage
    if (count >= 10)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, 10)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if (value != 1)
        goto usage;

    if (isEventQueueStalled()) {
        wakeEventFoundInQ = find_fix_idle_wake_event_in_event_queue();
        if (INVALID_EVENT != wakeEventFoundInQ)  {
            hlog("FPOW iwcueq found wake event %s (%d) in queue\n",fpow_passed_event_to_str(wakeEventFoundInQ),wakeEventFoundInQ);
            switch(wakeEventFoundInQ)  {
                case ALT_FONT_KEY_COMBINATION:
                    // Case has been corrected, just unstall and go
                    unstallEventQueue();
                    break;
                case WAN_SWITCH_ON_EVENT:
                case WAN_SWITCH_OFF_EVENT:
                    fixUnstallEventQueue(wakeEventFoundInQ);
                    print_queue("FPOW found wake event WAN SWITCH in queue, Done fixing Q.\n",0);
                    break;
                default:
                    hlog("proc_write_fpow_iwcueq() ignoring wake event %s\n",fpow_passed_event_to_str(wakeEventFoundInQ));
                    break;
            }
        }
    }
        
	return count;

usage:
    printk("You must write a 1 to iwcueq\n");
    return count;
}



/******************************************************************************
 *
 * /proc/fpow/fueqpl Support
 *
 * 'fueqpl' is short for "Fix and UnStall Event Queue for Power Lockout".
 *  Writing a 1 to fueqpl will fix and unstall the event queue for power lockout mode
 *  Writing a 0 has no effect.
 *
 * Reading from /proc/fpow/fueqpl will return if event queue is currently stalled.
 *
 ******************************************************************************/
static int proc_write_fpow_fueqpl(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[10];
    int rc = 0;
    int value;
    char *new_line = NULL;

    // If too much data for our buffer, show usage
    if (count >= 10)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, 10)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if (value != 1)
        goto usage;

    if (isEventQueueStalled()) {
	    printk("Event queue stalled, fix unstall for power lockout from JAVA request...\n");
        fixUnstallEventQueue(CRITICAL_BATTERY_EVENT);
    }
	return count;

usage:
    printk("You must write a 1 to fueqpl\n");
    return count;
}


/******************************************************************************
 *
 * /proc/fpow/fuseq Support
 *
 * 'useq' is short for "Fix and UnStall Event Queue".
 *  Writing a 1 to fuseq will fix and unstall the event queue
 *  Writing a 0 has no effect.
 *
 * Reading from /proc/fpow/fuseq will return if event queue is currently stalled.
 *
 ******************************************************************************/
static int proc_write_fpow_fuseq(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[10];
    int rc = 0;
    int value;
    char *new_line = NULL;

    // If too much data for our buffer, show usage
    if (count >= 10)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, 10)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 1) && (value != 0))
        goto usage;

    if (isEventQueueStalled()) {
	    powdebug("Event queue stalled, fixunstall from JAVA request for POWER_%s...\n",(value ? "OFF":"ON"));
        if (value == 0)
            fixUnstallEventQueue(POWER_SWITCH_OFF_EVENT);
        else
            fixUnstallEventQueue(POWER_SWITCH_ON_EVENT);
    }
	return count;

usage:
    printk("You must write a 0 (fix for Power Switch OFF) or a 1 (Power Switch On) to fuseq\n");
    return count;
}


/******************************************************************************
 *
 * /proc/fpow/useq Support
 *
 * 'useq' is short for "UnStall Event Queue".
 *  Writing a 1 to useq will unstall the event queue
 *  Writing a 0 has no effect.
 *
 * Reading from /proc/fpow/useq will return if event queue is currently stalled.
 *
 ******************************************************************************/
static int proc_write_fpow_useq(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[10];
    int rc = 0;
    int value;
    char *new_line = NULL;

    // If too much data for our buffer, show usage
    if (count >= 10)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, 10)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if (value != 1)
        goto usage;

    if (isEventQueueStalled()) {
	    powdebug("Event queue stalled, unstalling event queue...\n");
        unstallEventQueue();
    }
	return count;

usage:
    printk("You must write a 1 to useq\n");
    return count;
}



static int proc_write_fpow_cueq(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[10];
    int rc = 0;
    int value;
    char *new_line = NULL;

    // If too much data for our buffer, show usage
    if (count >= 10)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, 10)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if (value != 1)
        goto usage;

	powdebug("Clearing and unstalling event queue from cueq proc...\n");
    fiona_clear_unstall_event_queue();
	return count;

usage:
    printk("You must write a 1 to seq\n");
    return count;
}

static int proc_write_fpow_seq(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[10];
    int rc = 0;
    int value;
    char *new_line = NULL;

    // If too much data for our buffer, show usage
    if (count >= 10)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, 10)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if (value != 1)
        goto usage;

	powdebug("Stalling event queue...\n");
    setEventQueueStalled(1);
	return count;

usage:
    printk("You must write a 1 to seq\n");
    return count;
}

/******************************************************************************
 *
 * /proc/fpow/version Support
 *
 * Reading from /proc/fpow/version will return the version of the FPOW driver
 * Writing anything to /proc/fpow/version is invalid and not supported.
 *
 ******************************************************************************/
static int proc_read_fpow_version(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;

    if (count < 2)
        return -EINVAL;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    len = sprintf(page, "%s\n",FPOW_DRIVER_VERSION);
    *eof = 1;
    return len;
}

/******************************************************************************/

static int proc_write_fpow_version(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    printk("Writing to /proc/fpow/version is not supported.\n");
    return count;
}


/******************************************************************************
 *
 * /proc/fpow/watermarks/wan Support
 *
 ******************************************************************************/

#define watermark_wan       1    // BAT_LEVEL_WAN
#define watermark_low       2    // BAT_LEVEL_LOW
#define watermark_critical  3    // BAT_LEVEL_CRIT

static void save_battery_watermarks(int which_watermark, int watermark)
{
    bat_level_t battery_watermarks;
    
    // Get the currently saved battery watermarks and, by default,
    // say that we want to update the environment.
    //
    memcpy(&battery_watermarks, get_battery_watermarks(), sizeof(bat_level_t));
    battery_watermarks.update_env = 1;
    
    // Change the requested one.
    // 
    switch ( which_watermark )
    {
        case watermark_wan:
            battery_watermarks.stage[BAT_LEVEL_UPDT] = (unsigned short)CALC_WATERMARK_UPDATE(watermark);
            battery_watermarks.stage[BAT_LEVEL_WAN]  = (unsigned short)watermark;
        break;
        
        case watermark_low:
            battery_watermarks.stage[BAT_LEVEL_LOW]  = (unsigned short)watermark;
        break;
        
        case watermark_critical:
            battery_watermarks.stage[BAT_LEVEL_CRIT] = (unsigned short)watermark;
        break;
        
        // Huh?  Unrecognized case.  Don't update the
        // environment.
        //
        default:
            battery_watermarks.update_env = 0;
        break;
    }
    
    // Save the new watermarks out if we should.
    //
    if ( battery_watermarks.update_env )
        set_battery_watermarks(&battery_watermarks);
}

static int proc_read_fpow_wan_watermark(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
	int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

	len = sprintf(page, "%d\n",fpowGetWanWatermark());
    *eof = 1;
	return len;
}


static int proc_write_fpow_wan_watermark(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[10];
    int rc = 0;
    int value;
    char *new_line = NULL;
    
    // If too much data for our buffer, show usage
    if (count >= 10)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, 10)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if (!((value >= 0) && (value <= 4200)))
        goto usage;

	printk("Set wan watermark to be %d\n", value);
	if (value != fpowGetWanWatermark()) {
	    save_battery_watermarks(watermark_wan, value);
        fpowSetWanWatermark(value);
    }
	return count;

usage:
    printk("You must write a value between 0 and 4200\n");
    return count;
}


/******************************************************************************
 *
 * /proc/fpow/watermarks/low Support
 *
 ******************************************************************************/
static int proc_read_fpow_low_watermark(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
	int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

	len = sprintf(page, "%d\n",fpowGetLowWatermark());
    *eof = 1;
	return len;
}


static int proc_write_fpow_low_watermark(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[10];
    int rc = 0;
    int value = 0;
    char *new_line = NULL;

    // If too much data for our buffer, show usage
    if (count >= 10)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, 10)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);

    if (!((value >= 0) && (value <= 4200)))
        goto usage;

	printk("Set low watermark to be %d\n", value);
    if (value != fpowGetLowWatermark()) {
        save_battery_watermarks(watermark_low, value);
        fpowSetLowWatermark(value);
    }
	return count;

usage:
    printk("You must write a value between 0 and 4200 (saw %d)\n",value);
    return count;
}


/******************************************************************************
 *
 * /proc/fpow/watermarks/critical Support
 *
 ******************************************************************************/
static int proc_read_fpow_critical_watermark(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
	int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

	len = sprintf(page, "%d\n",fpowGetCriticalWatermark());
    *eof = 1;
	return len;
}


static int proc_write_fpow_critical_watermark(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[10];
    int rc = 0;
    int value;
    char *new_line = NULL;

    // If too much data for our buffer, show usage
    if (count >= 10)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, 10)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if (!((value >= 0) && (value <= 4200)))
        goto usage;

	printk("Set critical watermark to be %d\n", value);
    if (value != fpowGetCriticalWatermark()) {
	    save_battery_watermarks(watermark_critical, value);
        fpowSetCriticalWatermark(value);
    }
	return count;

usage:
    printk("You must write a value between 0 and 4200\n");
    return count;
}

static void fpowSetWanWatermark(int value) {
    g_wan_watermark = value;
}
static void fpowSetLowWatermark(int value) {
    g_low_watermark = value;
}
static void fpowSetCriticalWatermark(int value) {
    g_critical_watermark = value;
}
int fpowGetWanWatermark(void) {
    return(g_wan_watermark);
}
int fpowGetLowWatermark(void) {
    return(g_low_watermark);
}
int fpowGetCriticalWatermark(void) {
    return(g_critical_watermark);
}

/******************************************************************************
 *
 * Timer support for watching battery level
 *
 ******************************************************************************/
#define WATERMARK_TIMEOUT   (30)    // Seconds in RCNR

static void disable_wakeup_timer(void) {
    // Clear RTC as a valid wake source
    g_fiona_hw.pwer &= ~PWER_RTC;
}

static void enable_wakeup_timer(void) {
    // Make RTC a valid wake source
    g_fiona_hw.pwer |= PWER_RTC;
}


void startWatermarkTimer(int timeout)
{
    printk("startWatermarkTimer() - Initialized timer to %d = RCNR\n\n",RCNR+timeout);
    RTAR = (RCNR + timeout);
    RTSR = ((RTSR & RTSR_RESPECT) | RTSR_ALE);
}

void deleteWatermarkTimer(void)
{
    RTSR = ((RTSR & RTSR_RESPECT) & ~RTSR_ALE);
    g_fiona_hw.pwer &= ~PWER_RTC;
    RTAR = 0;
}

static void do_init_start_watermark_timer(void)
{
    enable_wakeup_timer();
    //startWatermarkTimer(WATERMARK_TIMEOUT);
}


/******************************************************************************
 *
 * /proc/fpow/uc/ac_adapter Support
 *
 * Reading from /proc/fpow/uc/ac_adapter will return the state of the AC
 * adapter, with 1 being "IN" and 0 being "OUT".
 *
 *  NOTE:  This will query the IOC directly to get the answer (not rely on the local
 *      knowledge of the adapter status...
 *
 ******************************************************************************/
static int proc_read_fpow_uc_ac_adapter(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

	len = sprintf(page, "%d\n",ac_adapter_status());   // Check with IOC again for status
    *eof = 1;
	return len;
}

/******************************************************************************
 *
 * /proc/fpow/uc/power_switch Support
 *
 * Reading from /proc/fpow/uc/power_switch will return the state of the power
 * switch, with 1 being "ON" and 0 being "OFF".
 *
 ******************************************************************************/
static int proc_read_fpow_uc_power(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;
    int state = 1;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

    // If it's EVT3 or older, always return power switch in ON position...
    if (get_fiona_board_revision() >= FIONA_BOARD_DVT1)
        state = ((GPLR(g_fiona_hw.sleep_switch_gpio) & (1<<g_fiona_hw.sleep_switch_gpio)) == g_fiona_hw.sleep_switch_polarity) ? 0 : 1;

	len = sprintf(page, "%d\n",state);
    *eof = 1;
	return len;
}


/******************************************************************************
 *
 * /proc/fpow/uc/wan_switch Support
 *
 * Reading from /proc/fpow/uc/wan_switch will return the state of the WAN switch.
 * Writing anything to /proc/fpow/uc/wan_switch is invalid and not supported.
 *
 ******************************************************************************/
static int proc_read_fpow_uc_wan(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;
    int wan_enabled = 0;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

    wan_enabled = wan_switch();
	len = sprintf(page, "%d\n",wan_enabled);
    *eof = 1;
	return len;
}


/******************************************************************************
 *
 * /proc/fpow/uc/headphones Support
 *
 * Reading from /proc/fpow/uc/headphones will return the state of the headphones
 *  (1 means inserted, 0 means removed)
 * Writing anything to /proc/fpow/uc/headphones is invalid and not supported.
 *
 ******************************************************************************/
static int proc_read_fpow_uc_hp(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;
    int hp = 0;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

    hp = ((GPLR(GPIO_FIONA_HPDET_EVT3) & GPIO_bit(GPIO_FIONA_HPDET_EVT3)) ? 0 : 1);

	len = sprintf(page, "%d\n",hp);
    *eof = 1;
	return len;
}


/******************************************************************************
 *
 * /proc/fpow/components/xxx Support
 *
 * Reading from /proc/fpow/components/xxx will return the current power mode of 
 * 	the xxx component.
 *
 * Writing a valid FPOW_POWER_MODe to /proc/fpow/components/xxx will send that
 *	mode change request to said device driver component.
 ******************************************************************************/
static int proc_read_fpow_component_mode(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;
	struct fpow_component *dev = NULL;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if ((count < 2) || (data == NULL))  {
		len = -EINVAL;
        goto exit;
    }

	dev = (struct fpow_component*) data;
	len = sprintf(page, "%d\n",(*dev->getmode)((struct device *)dev->private));
    *eof = 1;

exit:
    if (len < 0) len = 0;
    if (len > count) len = count;
	return len;
}

/******************************************************************************/

static int proc_write_fpow_component_mode(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    int		rc = 0;
    int		mode;
    char	valueStr[2];
	struct	fpow_component *dev = NULL;
    char *new_line = NULL;

    if ((count != 2) || (data == NULL))
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&mode);
    if (!((mode >= FPOW_MODE_FIRST) && (mode <= FPOW_MODE_LAST))) {
        goto usage;
    }

    // First, get the lock
    lockdebug("proc_write_fpow_component_mode( pid=%d ) getting lock...\n", (int)sys_getpid());
    if (down_interruptible(&fpow_lock)) {
        lockdebug("proc_write_fpow_component_mode( pid=%d ) COULD NOT GET lock!\n", (int)sys_getpid());
        return(-ERESTARTSYS);
    }
    lockdebug("proc_write_fpow_component_mode( pid=%d ) got lock...\n", (int)sys_getpid());

	dev = (struct fpow_component*) data;
	powdebug("Sending %s mode command to %s component.\n",fpow_mode_to_str(mode),dev->name);
	rc = fpow_send_cmd(dev, mode);
	powdebug("%s result from sending %s mode command to the %s component.\n", \
		(rc == 0) ? "COMPLETE" : "NOT_COMPLETE", \
		fpow_mode_to_str(mode),dev->name);

    up(&fpow_lock);
    lockdebug("proc_write_fpow_component_mode( pid=%d ) released lock...\n", (int)sys_getpid());
	return count;

usage:
    printk("You must write a valid mode value to proc/fpow/component/%s.\n",dev->name); 
    printk("Valid mode values range between %d and %d inclusive.\n",FPOW_MODE_FIRST,FPOW_MODE_LAST);
    return 0;
}



#ifdef ENABLE_EVENT_TEST_PROC
/******************************************************************************
 *
 * /proc/fpow/event_test Support
 *
 * Reading from /proc/fpow/event_test will return the state of the 
 *  test flag g_fiona_insert_event, which indicates whether FPOW
 *  will inject a test event into the queue just prior to processing
 *  a sleep request to test the "user queue lookahead" feature for cancelling
 *  opposing user events.
 *  
 *  Writing to /proc/fpow/event_test supports the following values:
 *      1 - inject a POWER_ON event once and then reset itself.
 *      2 - inject a POWER_OFF event once and then reset itself
 *      3 - inject a WAN_SWITCH_OFF event once and then reset itself
 *      4 - inject a WAN_SWITCH_ON event once and then reset itself
 *      5 - inject an ALT+FONT key event sequence once and then reset itself
 *      6 - inject an AC_IN event once and then reset itself
 *
 * Writing "0" to /proc/fpow/event_test will keep FPOW from injecting any events.
 *
 ******************************************************************************/
static int proc_read_fpow_event_test(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

	len = sprintf(page, "%d\n",g_fiona_insert_event);
    *eof = 1;
	return len;
}

/******************************************************************************/

static int proc_write_fpow_event_test(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[256];
    int rc = 0;
    int value = 0;
    char *new_line = NULL;

    if (count > 3)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);

    if ((value < 0) || (value > 10))
        goto usage;

    g_fiona_insert_event = value;
    if (g_fiona_insert_event != 0)
        printk("Preparing Test Case %d\n%s",g_fiona_insert_event,get_expected_results(g_fiona_insert_event));

	return count;

usage:
    printk("You must write a value between 0 and 10 to /proc/fpow/event_test (value was %d)\n",value);
    return count;
}
#endif // ENABLE_EVENT_TEST_PROC


int get_fiona_power_state(void)
{
    return(globs.system_state);
}

/******************************************************************************
 *
 * /proc/fpow/system_state Support
 *
 * Reading from /proc/fpow/system_state will return the power state of the system.
 *  
 * Writing a valid FPOW_POWER_STATE to /proc/fpow/system_state will set the system
 *	to the appropriate power state. 
 *
 ******************************************************************************/
static int proc_read_fpow_system_state(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;
    int state;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

    state = get_fiona_power_state();
	len = sprintf(page, "%d\n",state);
    printk("FPOW State: %s\n",fpow_state_to_str(state));
    *eof = 1;
	return len;
}

/******************************************************************************/

static int proc_write_fpow_system_state(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char    valueStr[2];
    char    theChar;
    int     rc = 0;
    int     state;
    char *new_line = NULL;

    if (count != 2)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&state);
    if (invalid_fpow_state(state)) {
        rc = sscanf(valueStr,"%c",&theChar);
        theChar = toupper(theChar);
        state = theChar - 'A';
        if (invalid_fpow_state(state))
            goto usage;
    }

    // First, get the lock
    lockdebug("proc_write_fpow_system_state( pid=%d ) getting lock...\n", (int)sys_getpid());
    if (down_interruptible(&fpow_lock)) {
        lockdebug("proc_write_fpow_system_state( pid=%d ) COULD NOT GET lock!\n", (int)sys_getpid());
        return(-ERESTARTSYS);
    }
    lockdebug("proc_write_fpow_system_state( pid=%d ) got lock...\n", (int)sys_getpid());

    // Transition power states
	powdebug("\nSetting Fiona Power Mangement system state to %s...\n",fpow_state_to_str(state));
    rc = fiona_power_state(state);
	powdebug("Set Fiona Power Mangement system state to %s\n\n",fpow_state_to_str(globs.system_state));
   
    up(&fpow_lock);
    lockdebug("proc_write_fpow_system_state( pid=%d ) released lock...\n", (int)sys_getpid());

	return count;

usage:
    printk("You must write a value between '%d' and '%d' to /proc/fpow/system_state\n",FPOW_STATE_FIRST,FPOW_STATE_LAST);
	return count;   // Return count so we don't get an error callback
}

/******************************************************************************
 *
 * /proc/fpow/event_raw Support
 *
 * Reading from /proc/fpow/event_raw is not supported.
 *  
 * Writing a value to event_raw will write that value to the event queue.
 *  This should be used for raw character / button / key entry.
 *
 ******************************************************************************/
static int proc_write_fpow_event_raw(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[12];
    int rc = 0;
    int value;
    char *new_line = NULL;

    if (count >= 12) {
        printk("proc_write_fpow_event_raw(): BAILING out - count = %d\n",(int)count);
        goto usage;
    }

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%x",&value);

    // If it's a valid event, send it...
    if (value) {
	    poweventdebug("Inserting event %x into Fiona event queue.\n",value);
        handle_fiona_event(value);
    }
    else {
        printk("proc_write_fpow_event_raw(): Unknown event %d (0x%x)\n",value,value);
    }

	return count;

usage:
    printk("You must write a hex value between '0' and '0xff' inclusive to /proc/fpow/event_raw\n");
    return count;
}

/******************************************************************************
 *
 * /proc/fpow/event Support
 *
 * Reading from /proc/fpow/event is not supported.
 *  
 * Writing a value of 0-20 to /proc/fpow/event will insert that event into the
 *  Fiona event queue.
 *
 ******************************************************************************/
static int proc_write_fpow_event(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[12];
    int rc = 0;
    int ec = 0;
    int value = 0;
    char *new_line = NULL;

    if (count >= 12) {
        printk("proc_write_fpow_event(): BAILING out - count = %d\n",(int)count);
        goto usage;
    }

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);

    if ((value < 0) || (value >= FPOW_NUM_OF_EVENTS))
        goto usage;

    switch(value)  {
        case POWER_SWITCH_ON_EVENT:
            ec = POWER_SW_ON_EVENTCODE;
            break;
        case POWER_SWITCH_OFF_EVENT:
            ec = POWER_SW_OFF_EVENTCODE;
            break;
        case WAN_SWITCH_ON_EVENT:
            ec = WAN_SW_ON_EVENTCODE;
            break;
        case WAN_SWITCH_OFF_EVENT:
            ec = WAN_SW_OFF_EVENTCODE;
            break;
        case WAN_SWITCH_SWEEPER_EVENT:
            ec = WAN_SW_SWEEPER_EVENTCODE;
            break;
        case TWO_FINGER_SWEEPER_EVENT:
            ec = TWO_FINGER_SWEEPER_EVENTCODE;
            break;
        case TIMEOUT_USB_WAKE_EVENT:
            ec = TIMEOUT_USB_WAKE_EVENTCODE;
            break;
        case BUTTON_PRESS_EVENT:
            break;
        case SCROLLWHEEL_EVENT:
            break;
        case KEY_PRESS_EVENT:
            break;
        case MMC_EVENT:
            ec = MMC_EVENTCODE;
            break;
        case AC_PLUG_INSERTED_EVENT:
            ec = AC_IN_EVENTCODE;
            break;
        case AC_PLUG_REMOVED_EVENT:
            ec = AC_OUT_EVENTCODE;
            break;
        case USB_WAKE_EVENT:
            ec = USB_WAKE_EVENTCODE;
            break;
        case USB_PLUG_INSERTED_EVENT:
            ec = USB_IN_EVENTCODE;
            break;
        case USB_PLUG_REMOVED_EVENT:
            ec = USB_OUT_EVENTCODE;
            break;
        case HEADPHONES_INSERTED_EVENT:
            ec = HEADPHONES_IN_EVENTCODE;
            break;
        case HEADPHONES_REMOVED_EVENT:
            ec = HEADPHONES_OUT_EVENTCODE;
            break;
        case TIMEOUT_EVENT:
            ec = TIMEOUT1_EVENTCODE;
            break;
        case LOW_BATTERY_EVENT:
            ec = BATTERY_LOW_EVENTCODE;           
            break;
        case CRITICAL_BATTERY_EVENT:
            ec = BATTERY_CRITICAL_EVENTCODE;
            break;
        case OUT_OF_POWER_EVENT:
            break;
        case WAN_RING_EVENT:
            ec = WAN_RING_EVENTCODE;
            break;
        case AUDIO_ON_EVENT:
            ec = AUDIO_ON_EVENTCODE;
            break;
        case AUDIO_OFF_EVENT:
            ec = AUDIO_OFF_EVENTCODE;
            break;
        case TIMEOUT2_EVENT:
            ec = TIMEOUT2_EVENTCODE;
            break;
        case TIMEOUT3_EVENT:
            ec = TIMEOUT3_EVENTCODE;
            break;
        case VALID_PASSWORD_EVENT:
            ec = VALID_PASSWORD_EVENTCODE;
            break;
        case NO_WAN_BATTERY_EVENT:
            ec = BATTERY_NO_WAN_EVENTCODE;
            break;
        case INVALID_PASSWORD_EVENT:
            ec = INVALID_PASSWORD_EVENTCODE;
            break;
        case WATERMARK_CHECK_EVENT:
            ec = WATERMARK_CHECK_EVENTCODE;
            break;
        case TPH_COMPLETE_EVENT:
            ec = TPH_COMPLETE_EVENTCODE;
            break;
        case PRIME_EVENT:
            ec = PRIME_PUMP_EVENTCODE;
            break;
        default:
            break;
    }

    // If it's a valid event, send it...
    if (ec) {
	    poweventdebug("Inserting event %s into Fiona event queue.\n",fiona_event_keycode_to_str(ec));
        handle_fiona_event(ec);
    }
    else {
        printk("proc_write_fpow_event(): Unknown event %d (0x%x)\n",value,value);
    }

	return count;

usage:
    printk("You must write a value between '0' and '%d' inclusive to /proc/fpow/event (saw %d)\n",FPOW_NUM_OF_EVENTS-1,value);
    return count;
}


static int find_fix_idle_wake_event_in_event_queue(void) {
    int wakeEventFoundInQ = INVALID_EVENT;
    int done = 0;
    do {
        wakeEventFoundInQ = idle_wake_event_in_event_queue();
        if (INVALID_EVENT != wakeEventFoundInQ)  {
            hlog("FPOW found wake event %s in queue, dealing with it\n",fpow_wake_src_to_str(wakeEventFoundInQ));
            print_queue("FPOW found wake event in queue, dealing with it\n",0);
            switch(wakeEventFoundInQ) {
                case WAN_ON_WAKE:
                case WAN_OFF_WAKE:
                    hlog("FPOW wake event WAN_SWITCH in queue, removing paired events...\n");
                    // If we find and remove a pair, check again....
                    // If we can remove paired events and still leave one, do so..

                    // njv 03/27/07 12:44 am commented out following 2 lines
                    // if (remove_paired_wan_events_in_queue())
                    //    continue;
                    break;

                case ALT_FONT_KEY_COMBINATION:
                    hlog("FPOW wake event ALT_FONT_KEY_COMBINATION in queue, fixing...\n");
                    remove_alt_font_keys_from_event_queue();
                    // Add in ALT_FONT event after removing all of them...
                    break;

                default:
                    break;
            }
            done = 1;
            hlog("FPOW done dealing with wake event %s in queue\n",fpow_wake_src_to_str(wakeEventFoundInQ));
            print_queue("FPOW done dealing with wake event in queue\n",0);
        }
        else
            done = 1;
    } while (0 == done);
    return(wakeEventFoundInQ);
}

#ifdef ENABLE_SNIFF_DEBUG_PROC

/******************************************************************************
 *
 * /proc/fpow/sniff_debug Support
 *
 * Reading from /proc/fpow/sniff_debug will return the state of the 
 *  g_fiona_sniffer_debug flag, which turns on or off logging for the queue sniffer.
 *  
 * Writing "1" to /proc/fpow/sniff_debug will enable the logging
 * Writing "0" to /proc/fpow/sniff_debug will disable the logging
 *
 ******************************************************************************/
static int proc_read_fpow_sniff_debug(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

	len = sprintf(page, "%d\n",g_fiona_sniff_debug);
    *eof = 1;
	return len;
}

/******************************************************************************/

static int proc_write_fpow_sniff_debug(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char *new_line = NULL;

    if (count != 2)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 0) && (value != 1))
        goto usage;

    if (value) {
	    printk("Enabled Fiona Queue Sniffer logging\n");
        g_fiona_sniff_debug = 1;
    }
    else {
	    printk("Disabled Fiona Queue Sniffer logging\n");
        g_fiona_sniff_debug = 0;
    }

	return count;

usage:
    printk("You must write a '1' or a '0' to /proc/fpow/event_debug\n");
    return count;
}

#endif  // ENABLE_SNIFF_DEBUG_PROC

/******************************************************************************
 *
 * /proc/fpow/event_debug Support
 *
 * Reading from /proc/fpow/event_debug will return the state of the 
 *  g_fiona_event_debug flag, which turns on or off core Fiona Power 
 *  Management I/O logging.
 *  
 * Writing "1" to /proc/fpow/event_debug will enable FPOW core I/O logging
 * Writing "0" to /proc/fpow/event_debug will disable FPOW core I/O logging
 *
 ******************************************************************************/
static int proc_read_fpow_event_debug(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

	len = sprintf(page, "%d\n",g_fiona_event_debug);
    *eof = 1;
	return len;
}

/******************************************************************************/

static int proc_write_fpow_event_debug(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char *new_line = NULL;

    if (count != 2)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 0) && (value != 1))
        goto usage;

    if (value) {
	    printk("Enabled Fiona Event management core logging\n");
        g_fiona_event_debug = 1;
    }
    else {
	    printk("Disabled Fiona Event management core logging\n");
        g_fiona_event_debug = 0;
    }

	return count;

usage:
    printk("You must write a '1' or a '0' to /proc/fpow/event_debug\n");
    return count;
}


/******************************************************************************
 *
 * /proc/fpow/debug Support
 *
 * Reading from /proc/fpow/debug will return the state of the g_fiona_power_debug flag,
 *  which turns on or off core Fiona Power Management I/O logging.
 *  
 * Writing "1" to /proc/fpow/debug will enable FPOW core I/O logging
 * Writing "0" to /proc/fpow/debug will disable FPOW core I/O logging
 *
 ******************************************************************************/
static int proc_read_fpow_debug(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

	len = sprintf(page, "%d\n",g_fiona_power_debug);
    *eof = 1;
	return len;
}

/******************************************************************************/

static int proc_write_fpow_debug(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char *new_line = NULL;

    if (count != 2)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 0) && (value != 1))
        goto usage;

    if (value) {
	    printk("Enabled Fiona power mangement core logging\n");
        g_fiona_power_debug = 1;
    }
    else {
	    printk("Disabled Fiona power mangement core logging\n");
        g_fiona_power_debug = 0;
    }

	return count;

usage:
    printk("You must write a '1' or a '0' to /proc/fpow/debug\n");
    return count;
}


/******************************************************************************
 *
 * /proc/fpow/reg_list Support
 *
 * Reading from /proc/fpow/reg_list will print the list of registered
 * components.
 *  
 * Writing to /proc/fpow/reg_list is not supported.
 *
 ******************************************************************************/
static int proc_read_fpow_reg_list(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

	if (count < 2)
		return -EINVAL;

	// Print out the registration list...
	fpow_print_registered_component_list();

	*eof = 1;
	len = 0;
	return len;
}

/******************************************************************************
 *
 * /proc/fpow/vtbl_debug Support
 *
 * Reading from /proc/fpow/vtbl_debug will return the state of the
 *  g_fiona_vtable_debug flag, which enables/disables voltage sampler logging
 *  on every voltage sample taken.
 *
 * Writing "1" to /proc/fpow/vtbl_debug will dump table every time voltage is sampled
 * Writing "0" to /proc/fpow/vtbl_debug will disable dumping table on every sample
 *
 ******************************************************************************/
static int proc_read_fpow_vtbl_debug(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    if (count < 2)
        return -EINVAL;

    len = sprintf(page, "%d\n",g_fiona_vtable_debug);
    *eof = 1;
    return len;
}

/******************************************************************************/

static int proc_write_fpow_vtbl_debug(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char *new_line = NULL;

    if (count > sizeof(valueStr))
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

    if (copy_from_user(valueStr, buffer, count)) {
        return -EFAULT;
    }

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 0) && (value != 1))
        goto usage;

    if (value) {
        printk("Enabled voltage table dumping on every sample taken\n");
        g_fiona_vtable_debug = 1;
    }
    else {
        printk("Disabled voltage table dumping on every sample taken\n");
        g_fiona_vtable_debug = 0;
    }
    return count;

usage:
    printk("You must write a '1' or a '0' to /proc/fpow/vtbl_debug (saw %s)\n",valueStr);
    return count;
}



/******************************************************************************
 *
 * /proc/fpow/vdebug Support
 *
 * Reading from /proc/fpow/volt_debug will return the state of the
 *  g_fiona_voltage_debug flag, which turns on or off the voltage sampler logging
 *
 * Writing "1" to /proc/fpow/vdebug will enable voltage sampler logging
 * Writing "0" to /proc/fpow/vdebug will disable voltage sampler logging
 *
 ******************************************************************************/
static int proc_read_fpow_vdebug(char *page, char **start,
                off_t off, int count,
                int *eof, void *data)
{
    int len;

    // If it's the "second" call, ignore it...
    if (off > 0) {
        *eof = 1;
        return(0);
    }

    if (count < 2)
        return -EINVAL;

    len = sprintf(page, "%d\n",g_fiona_voltage_debug);
    *eof = 1;
    return len;
}

/******************************************************************************/

static int proc_write_fpow_vdebug(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char *new_line = NULL;

    if (count != 2)
        goto usage;

    // Clear the storage for string to assure null-termination
    memset(valueStr,0,sizeof(valueStr));

    if (copy_from_user(valueStr, buffer, count)) {
        return -EFAULT;
    }

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 0) && (value != 1))
        goto usage;

    if (value) {
        printk("Enabled voltage sampling logging\n");
        g_fiona_voltage_debug = 1;
    }
    else {
        printk("Disabled voltage sampling logging\n");
        g_fiona_voltage_debug = 0;
    }
    return count;

usage:
    printk("You must write a '1' or a '0' to /proc/fpow/vdebug\n");
    return count;
}


/******************************************************************************/

static int proc_write_fpow_reg_list(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    printk("Writing to /proc/fpow/reg_list is not supported.\n");
    return count;
}




/******************************************************************************/

static int fpow_proc_init(void)
{
	int rc = 0;

    // 
	// Create /proc/fpow directory
    //
    proc_fpow_parent = create_proc_entry(FPOW_PROC_PARENT_NAME, S_IFDIR | S_IRUGO | S_IXUGO, NULL);
    if (!proc_fpow_parent) {
		rc = -ENOMEM;
		goto exit;
    }

    // 
    // Create /proc/fpow/reg_list entry 
    //
	proc_fpow_reg_list = create_proc_entry(FPOW_PROC_REG_LIST_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_reg_list) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_reg_list->owner	= THIS_MODULE;
	proc_fpow_reg_list->data	= NULL;
	proc_fpow_reg_list->read_proc	= proc_read_fpow_reg_list;
	proc_fpow_reg_list->write_proc	= proc_write_fpow_reg_list;

    // 
    // Create /proc/fpow/uc directory entry 
    //
	proc_fpow_uc_parent = create_proc_entry(FPOW_PROC_UC_PARENT_NAME, S_IFDIR | S_IRUGO | S_IXUGO, proc_fpow_parent);
	if (!proc_fpow_uc_parent) {
		rc = -ENOMEM;
		goto exit;
	}

    // 
    // Create /proc/fpow/uc/ac_adapter directory entry 
    //
	proc_fpow_uc_ac = create_proc_entry(FPOW_PROC_UC_AC_ADAPTER_ENTRY_NAME, 0666, proc_fpow_uc_parent);
	if (!proc_fpow_uc_ac) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_uc_ac->owner	= THIS_MODULE;
	proc_fpow_uc_ac->data	= NULL;
	proc_fpow_uc_ac->read_proc	= proc_read_fpow_uc_ac_adapter;
	proc_fpow_uc_ac->write_proc	= NULL;

    // 
    // Create /proc/fpow/uc/power_switch directory entry 
    //
	proc_fpow_uc_power = create_proc_entry(FPOW_PROC_UC_POWER_ENTRY_NAME, 0666, proc_fpow_uc_parent);
	if (!proc_fpow_uc_power) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_uc_power->owner	= THIS_MODULE;
	proc_fpow_uc_power->data	= NULL;
	proc_fpow_uc_power->read_proc	= proc_read_fpow_uc_power;
	proc_fpow_uc_power->write_proc	= NULL;

    // 
    // Create /proc/fpow/uc/wan_switch directory entry 
    //
	proc_fpow_uc_wan = create_proc_entry(FPOW_PROC_UC_WAN_ENTRY_NAME, 0666, proc_fpow_uc_parent);
	if (!proc_fpow_uc_wan) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_uc_wan->owner	= THIS_MODULE;
	proc_fpow_uc_wan->data	= NULL;
	proc_fpow_uc_wan->read_proc	= proc_read_fpow_uc_wan;
	proc_fpow_uc_wan->write_proc	= NULL;

    // 
    // Create /proc/fpow/uc/headphones directory entry 
    //
	proc_fpow_uc_hp = create_proc_entry(FPOW_PROC_UC_HP_ENTRY_NAME, 0666, proc_fpow_uc_parent);
	if (!proc_fpow_uc_hp) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_uc_hp->owner	= THIS_MODULE;
	proc_fpow_uc_hp->data	= NULL;
	proc_fpow_uc_hp->read_proc	= proc_read_fpow_uc_hp;
	proc_fpow_uc_hp->write_proc	= NULL;

    // 
    // Create /proc/fpow/event entry 
    //
	proc_fpow_event = create_proc_entry(FPOW_PROC_EVENT_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_event ) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_event->owner	= THIS_MODULE;
	proc_fpow_event->data	= NULL;
	proc_fpow_event->read_proc	= NULL;
	proc_fpow_event->write_proc	= proc_write_fpow_event;

    // 
    // Create /proc/fpow/event_raw entry 
    //
	proc_fpow_event_raw = create_proc_entry(FPOW_PROC_EVENT_RAW_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_event_raw ) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_event_raw->owner	= THIS_MODULE;
	proc_fpow_event_raw->data	= NULL;
	proc_fpow_event_raw->read_proc	= NULL;
	proc_fpow_event_raw->write_proc	= proc_write_fpow_event_raw;

    // 
    // Create /proc/fpow/debug entry 
    //
	proc_fpow_debug = create_proc_entry(FPOW_PROC_DEBUG_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_debug) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_debug->owner	= THIS_MODULE;
	proc_fpow_debug->data	= NULL;
	proc_fpow_debug->read_proc	= proc_read_fpow_debug;
	proc_fpow_debug->write_proc	= proc_write_fpow_debug;

    //
    // Create /proc/fpow/volt_debug entry
    //
    proc_fpow_vdebug = create_proc_entry(FPOW_PROC_VDEBUG_ENTRY_NAME, 0666, proc_fpow_parent);
    if (!proc_fpow_vdebug) {
        rc = -ENOMEM;
        goto exit;
    }
    proc_fpow_vdebug->owner = THIS_MODULE;
    proc_fpow_vdebug->data  = NULL;
    proc_fpow_vdebug->read_proc = proc_read_fpow_vdebug;
    proc_fpow_vdebug->write_proc    = proc_write_fpow_vdebug;

    //
    // Create /proc/fpow/vtbl_debug entry
    //
    proc_fpow_vtbl_debug = create_proc_entry(FPOW_PROC_VTBL_DEBUG_ENTRY_NAME, 0666, proc_fpow_parent);
    if (!proc_fpow_vtbl_debug) {
        rc = -ENOMEM;
        goto exit;
    }
    proc_fpow_vtbl_debug->owner = THIS_MODULE;
    proc_fpow_vtbl_debug->data  = NULL;
    proc_fpow_vtbl_debug->read_proc = proc_read_fpow_vtbl_debug;
    proc_fpow_vtbl_debug->write_proc    = proc_write_fpow_vtbl_debug;

#ifdef ENABLE_SNIFF_DEBUG_PROC
    // 
    // Create /proc/fpow/event_debug entry 
    //
	proc_fpow_sniff_debug = create_proc_entry(FPOW_PROC_SNIFF_DEBUG_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_sniff_debug) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_sniff_debug->owner	= THIS_MODULE;
	proc_fpow_sniff_debug->data	= NULL;
	proc_fpow_sniff_debug->read_proc	= proc_read_fpow_sniff_debug;
	proc_fpow_sniff_debug->write_proc	= proc_write_fpow_sniff_debug;
#endif

    // 
    // Create /proc/fpow/event_debug entry 
    //
	proc_fpow_event_debug = create_proc_entry(FPOW_PROC_EVENT_DEBUG_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_event_debug) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_event_debug->owner	= THIS_MODULE;
	proc_fpow_event_debug->data	= NULL;
	proc_fpow_event_debug->read_proc	= proc_read_fpow_event_debug;
	proc_fpow_event_debug->write_proc	= proc_write_fpow_event_debug;

#ifdef ENABLE_EVENT_TEST_PROC
    // 
    // Create /proc/fpow/event_test entry 
    //
	proc_fpow_event_test = create_proc_entry(FPOW_PROC_EVENT_TEST_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_event_test) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_event_test->owner	= THIS_MODULE;
	proc_fpow_event_test->data	= NULL;
	proc_fpow_event_test->read_proc	= proc_read_fpow_event_test;
	proc_fpow_event_test->write_proc	= proc_write_fpow_event_test;
#endif
		
    // 
    // Create /proc/fpow/system_state entry 
    //
	proc_fpow_system_state = create_proc_entry(FPOW_PROC_SYSTEM_STATE_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_system_state) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_system_state->owner	= THIS_MODULE;
	proc_fpow_system_state->data	= NULL;
	proc_fpow_system_state->read_proc	= proc_read_fpow_system_state;
	proc_fpow_system_state->write_proc	= proc_write_fpow_system_state;


    // 
    // Create /proc/fpow/iweiq entry 
    //
	proc_fpow_iweiq = create_proc_entry(FPOW_PROC_IWEIQ_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_iweiq) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_iweiq->owner	= THIS_MODULE;
	proc_fpow_iweiq->data	= NULL;
	proc_fpow_iweiq->read_proc	= proc_read_fpow_iweiq;
	proc_fpow_iweiq->write_proc	= NULL;


    // 
    // Create /proc/fpow/poiq entry 
    //
	proc_fpow_poiq = create_proc_entry(FPOW_PROC_POIQ_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_poiq) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_poiq->owner	= THIS_MODULE;
	proc_fpow_poiq->data	= NULL;
	proc_fpow_poiq->read_proc	= proc_read_fpow_poiq;
	proc_fpow_poiq->write_proc	= NULL;


    //
    // Create /proc/fpow/acieiq entry 
    //
	proc_fpow_acieiq = create_proc_entry(FPOW_PROC_ACIEIQ_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_acieiq) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_acieiq->owner	= THIS_MODULE;
	proc_fpow_acieiq->data	= NULL;
	proc_fpow_acieiq->read_proc	= proc_read_fpow_acieiq;
	proc_fpow_acieiq->write_proc	= NULL;


    // 
    //
    // Create /proc/fpow/peiq entry 
    //
	proc_fpow_peiq = create_proc_entry(FPOW_PROC_PEIQ_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_peiq) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_peiq->owner	= THIS_MODULE;
	proc_fpow_peiq->data	= NULL;
	proc_fpow_peiq->read_proc	= proc_read_fpow_peiq;
	proc_fpow_peiq->write_proc	= NULL;


    // 
    // Create /proc/fpow/cueq entry 
    //
	proc_fpow_cueq = create_proc_entry(FPOW_PROC_CUEQ_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_cueq ) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_cueq->owner	= THIS_MODULE;
	proc_fpow_cueq->data	= NULL;
	proc_fpow_cueq->read_proc	= NULL;
	proc_fpow_cueq->write_proc	= proc_write_fpow_cueq;


    // 
    // Create /proc/fpow/seq entry 
    //
	proc_fpow_seq = create_proc_entry(FPOW_PROC_SEQ_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_seq) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_seq->owner	= THIS_MODULE;
	proc_fpow_seq->data	= NULL;
	proc_fpow_seq->read_proc	= proc_read_fpow_event_queue_stalled;
	proc_fpow_seq->write_proc	= proc_write_fpow_seq;


    //
    // Create /proc/fpow/dialog_up entry
    //
    proc_fpow_dialog_up = create_proc_entry(FPOW_PROC_DIALOG_UP_ENTRY_NAME, 0666, proc_fpow_parent);
    if (!proc_fpow_dialog_up) {
        rc = -ENOMEM;
        goto exit;
    }
    proc_fpow_dialog_up->owner   = THIS_MODULE;
    proc_fpow_dialog_up->data    = NULL;
    proc_fpow_dialog_up->read_proc   = proc_read_fpow_dialog_up;
    proc_fpow_dialog_up->write_proc = proc_write_fpow_dialog_up;



    //
    // Create /proc/fpow/cache entry
    //
    proc_fpow_cache = create_proc_entry(FPOW_PROC_CACHE_ENTRY_NAME, 0666, proc_fpow_parent);
    if (!proc_fpow_cache) {
        rc = -ENOMEM;
        goto exit;
    }
    proc_fpow_cache->owner   = THIS_MODULE;
    proc_fpow_cache->data    = NULL;
    proc_fpow_cache->read_proc   = proc_read_fpow_cache;
    proc_fpow_cache->write_proc = proc_write_fpow_cache;



    // 
    // Create /proc/fpow/iwcueq entry 
    //
	proc_fpow_iwcueq = create_proc_entry(FPOW_PROC_IWCUEQ_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_iwcueq ) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_iwcueq->owner	= THIS_MODULE;
	proc_fpow_iwcueq->data	= NULL;
	proc_fpow_iwcueq->read_proc	= proc_read_fpow_event_queue_stalled;
	proc_fpow_iwcueq->write_proc = proc_write_fpow_iwcueq;


    // 
    // Create /proc/fpow/last_event entry 
    //
	proc_fpow_last_event = create_proc_entry(FPOW_PROC_LAST_EVENT_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_last_event) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_last_event->owner	= THIS_MODULE;
	proc_fpow_last_event->data	= NULL;
	proc_fpow_last_event->read_proc	= proc_read_last_event;
	proc_fpow_last_event->write_proc	= NULL;

    // 
    // Create /proc/fpow/wake entry 
    //
	proc_fpow_wake = create_proc_entry(FPOW_PROC_WAKE_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_wake) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_wake->owner	= THIS_MODULE;
	proc_fpow_wake->data	= NULL;
	proc_fpow_wake->read_proc	= proc_read_wake_reason;
	proc_fpow_wake->write_proc	= NULL;


    // 
    // Create /proc/fpow/fueqpl entry 
    //
	proc_fpow_fueqpl = create_proc_entry(FPOW_PROC_FUEQPL_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_fueqpl) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_fueqpl->owner	= THIS_MODULE;
	proc_fpow_fueqpl->data	= NULL;
	proc_fpow_fueqpl->read_proc	= proc_read_fpow_event_queue_stalled;
	proc_fpow_fueqpl->write_proc	= proc_write_fpow_fueqpl;


    // 
    // Create /proc/fpow/fuseq entry 
    //
	proc_fpow_fuseq = create_proc_entry(FPOW_PROC_FUSEQ_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_fuseq) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_fuseq->owner	= THIS_MODULE;
	proc_fpow_fuseq->data	= NULL;
	proc_fpow_fuseq->read_proc	= proc_read_fpow_event_queue_stalled;
	proc_fpow_fuseq->write_proc	= proc_write_fpow_fuseq;


    // 
    // Create /proc/fpow/proc_fpow_useq entry 
    //
	proc_fpow_useq = create_proc_entry(FPOW_PROC_USEQ_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_useq) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_useq->owner	= THIS_MODULE;
	proc_fpow_useq->data	= NULL;
	proc_fpow_useq->read_proc	= proc_read_fpow_event_queue_stalled;
	proc_fpow_useq->write_proc	= proc_write_fpow_useq;


    // 
    // Create /proc/fpow/version entry 
    //
	proc_fpow_version = create_proc_entry(FPOW_PROC_VERSION_ENTRY_NAME, 0666, proc_fpow_parent);
	if (!proc_fpow_version) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_version->owner	= THIS_MODULE;
	proc_fpow_version->data	= NULL;
	proc_fpow_version->read_proc	= proc_read_fpow_version;
	proc_fpow_version->write_proc	= proc_write_fpow_version;

		
    // 
    // Create /proc/fpow/watermarks directory
	//
    if (proc_fpow_watermarks_parent == NULL)
        proc_fpow_watermarks_parent = create_proc_entry(FPOW_PROC_WATERMARKS_PARENT_NAME, S_IFDIR | S_IRUGO | S_IXUGO, proc_fpow_parent);
    if (!proc_fpow_watermarks_parent) {
		rc = -ENOMEM;
		goto exit;
    }

    // 
    // Create /proc/fpow/watermarks/wan
    //
    if (proc_fpow_wan_watermark == NULL)
	    proc_fpow_wan_watermark = create_proc_entry(FPOW_PROC_WAN_WATERMARK_NAME, 0666, proc_fpow_watermarks_parent);
	if (!proc_fpow_wan_watermark ) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_wan_watermark->owner	= THIS_MODULE;
	proc_fpow_wan_watermark->data	= NULL;
	proc_fpow_wan_watermark->read_proc	= proc_read_fpow_wan_watermark;
	proc_fpow_wan_watermark->write_proc	= proc_write_fpow_wan_watermark;

    // 
    // Create /proc/fpow/watermarks/low
    //
    if (proc_fpow_low_watermark == NULL)
	    proc_fpow_low_watermark = create_proc_entry(FPOW_PROC_LOW_WATERMARK_NAME, 0666, proc_fpow_watermarks_parent);
	if (!proc_fpow_low_watermark) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_low_watermark->owner	= THIS_MODULE;
	proc_fpow_low_watermark->data	= NULL;
	proc_fpow_low_watermark->read_proc	= proc_read_fpow_low_watermark;
	proc_fpow_low_watermark->write_proc	= proc_write_fpow_low_watermark;


    // 
    // Create /proc/fpow/watermarks/critical
    //
    if (proc_fpow_critical_watermark == NULL)
	    proc_fpow_critical_watermark = create_proc_entry(FPOW_PROC_CRITICAL_WATERMARK_NAME, 0666, proc_fpow_watermarks_parent);
	if (!proc_fpow_critical_watermark) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_fpow_critical_watermark->owner	= THIS_MODULE;
	proc_fpow_critical_watermark->data	= NULL;
	proc_fpow_critical_watermark->read_proc	= proc_read_fpow_critical_watermark;
	proc_fpow_critical_watermark->write_proc	= proc_write_fpow_critical_watermark;

exit:
	return(rc);
}


static void fpow_cleanup_procfs(void)
{
    if (proc_fpow_reg_list) {
	    remove_proc_entry(FPOW_PROC_REG_LIST_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_reg_list= NULL;
    }
    if (proc_fpow_vdebug) {
        remove_proc_entry(FPOW_PROC_VDEBUG_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_vdebug = NULL;
    }
    if (proc_fpow_vtbl_debug) {
        remove_proc_entry(FPOW_PROC_VTBL_DEBUG_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_vtbl_debug = NULL;
    }
    if (proc_fpow_debug) {
	    remove_proc_entry(FPOW_PROC_DEBUG_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_debug = NULL;
    }
#ifdef ENABLE_SNIFF_DEBUG_PROC
    if (proc_fpow_sniff_debug) {
	    remove_proc_entry(FPOW_PROC_SNIFF_DEBUG_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_sniff_debug = NULL;
    }
#endif
    if (proc_fpow_event_debug) {
	    remove_proc_entry(FPOW_PROC_EVENT_DEBUG_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_event_debug = NULL;
    }
    if (proc_fpow_event) {
	    remove_proc_entry(FPOW_PROC_EVENT_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_event = NULL;
    }
    if (proc_fpow_event_raw) {
	    remove_proc_entry(FPOW_PROC_EVENT_RAW_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_event_raw = NULL;
    }
#ifdef ENABLE_EVENT_TEST_PROC
    if (proc_fpow_event_test) {
	    remove_proc_entry(FPOW_PROC_EVENT_TEST_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_event_test = NULL;
    }
#endif  // ENABLE_EVENT_TEST_PROC
    if (proc_fpow_acieiq) {
	    remove_proc_entry(FPOW_PROC_ACIEIQ_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_acieiq = NULL;
    }
    if (proc_fpow_peiq) {
	    remove_proc_entry(FPOW_PROC_PEIQ_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_peiq = NULL;
    }
    if (proc_fpow_iweiq)  {
	    remove_proc_entry(FPOW_PROC_IWEIQ_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_iweiq = NULL;
    }
    if (proc_fpow_poiq) {
	    remove_proc_entry(FPOW_PROC_POIQ_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_poiq = NULL;
    }
    if (proc_fpow_cueq) {
	    remove_proc_entry(FPOW_PROC_CUEQ_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_cueq = NULL;
    }
    if (proc_fpow_seq) {
	    remove_proc_entry(FPOW_PROC_SEQ_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_seq = NULL;
    }
    if (proc_fpow_dialog_up)  {
        remove_proc_entry(FPOW_PROC_DIALOG_UP_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_dialog_up = NULL;
    }
    if (proc_fpow_cache) {
	    remove_proc_entry(FPOW_PROC_CACHE_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_cache = NULL;
    }
    if (proc_fpow_iwcueq) {
	    remove_proc_entry(FPOW_PROC_IWCUEQ_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_iwcueq = NULL;
    }
    if (proc_fpow_last_event) {
	    remove_proc_entry(FPOW_PROC_LAST_EVENT_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_last_event = NULL;
    }
    if (proc_fpow_wake) {
	    remove_proc_entry(FPOW_PROC_WAKE_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_wake = NULL;
    }
    if (proc_fpow_fueqpl) {
	    remove_proc_entry(FPOW_PROC_FUEQPL_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_fueqpl = NULL;
    }
    if (proc_fpow_fuseq) {
	    remove_proc_entry(FPOW_PROC_FUSEQ_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_fuseq = NULL;
    }
    if (proc_fpow_useq) {
	    remove_proc_entry(FPOW_PROC_USEQ_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_useq = NULL;
    }
    if (proc_fpow_version) {
	    remove_proc_entry(FPOW_PROC_VERSION_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_version = NULL;
    }
    if (proc_fpow_system_state) {
	    remove_proc_entry(FPOW_PROC_SYSTEM_STATE_ENTRY_NAME, proc_fpow_parent);
        proc_fpow_system_state = NULL;
    }
    if (proc_fpow_uc_ac)  {
	    remove_proc_entry(FPOW_PROC_UC_AC_ADAPTER_ENTRY_NAME, proc_fpow_uc_parent);
        proc_fpow_uc_ac = NULL;
    }
    if (proc_fpow_uc_wan)  {
	    remove_proc_entry(FPOW_PROC_UC_WAN_ENTRY_NAME, proc_fpow_uc_parent);
        proc_fpow_uc_wan = NULL;
    }
    if (proc_fpow_uc_power)  {
	    remove_proc_entry(FPOW_PROC_UC_POWER_ENTRY_NAME, proc_fpow_uc_parent);
        proc_fpow_uc_power = NULL;
    }
    if (proc_fpow_uc_hp)  {
	    remove_proc_entry(FPOW_PROC_UC_HP_ENTRY_NAME, proc_fpow_uc_parent);
        proc_fpow_uc_hp = NULL;
    }
    if (proc_fpow_uc_parent)  {
	    remove_proc_entry(FPOW_PROC_UC_PARENT_NAME, proc_fpow_parent);
        proc_fpow_uc_parent = NULL;
    }
    if (proc_fpow_parent)  {
	    remove_proc_entry(FPOW_PROC_PARENT_NAME, NULL);
        proc_fpow_parent = NULL;
    }

    // 
    // Remove /proc/fpow/watermarks/* entries
    //
    if (proc_fpow_low_watermark)  {
	    remove_proc_entry(FPOW_PROC_LOW_WATERMARK_NAME, proc_fpow_watermarks_parent);
        proc_fpow_low_watermark = NULL;
    }
    if (proc_fpow_critical_watermark)  {
	    remove_proc_entry(FPOW_PROC_CRITICAL_WATERMARK_NAME, proc_fpow_watermarks_parent);
        proc_fpow_critical_watermark = NULL;
    }
    if (proc_fpow_wan_watermark)  {
	    remove_proc_entry(FPOW_PROC_WAN_WATERMARK_NAME, proc_fpow_watermarks_parent);
        proc_fpow_wan_watermark = NULL;
    }
    if (proc_fpow_watermarks_parent)  {
	    remove_proc_entry(FPOW_PROC_WATERMARKS_PARENT_NAME, NULL);
        proc_fpow_watermarks_parent = NULL;
    }
}


///////////////////////////////////////////////////////////////////////////////////
//
// Queueing / Searching Routines
//
///////////////////////////////////////////////////////////////////////////////////

#ifdef I_LIKE_COMPILER_WARNINGS_FOR_UNCALLED_CODE

/*********************************************************
 * Routine:     fpow_find_component_by_class
 * Purpose:     Search the registered Fiona Power components
 *		list (g_fpow_components_list) looking for
 *              a device having the name of dev_name. 
 * Paramters:   dev_class - device class of device to find
 *		last - if NULL, this routine returns first occurance
 *			of a component registered as type dev_class
 *			If NOT NULL, this routine returns the next
 *			occurance of a component registered as type
 *			dev_class, or NULL if none left.
 * Returns:     Pointer to fpow_component entry in list
 *              0 - No (more) device of class dev_class found in list.
 * Assumptions: Lock is held for g_fpow_components_list
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
static struct fpow_component *fpow_find_component_by_class(int dev_class, struct fpow_component *last)
{
    struct list_head *entry = g_fpow_components_list.next;
    struct fpow_component *dev = NULL;
    int next = 0;	// If set to 1, just return next match

    RDEBUG("fpow_find_component_by_class(0x%x, 0x%x)\n", dev_class, (int)last);

    while (entry != &g_fpow_components_list) {
        dev = list_entry(entry, struct fpow_component, entry);

		// If next is 0, we're still looking for our "left off here" spot
		if ((next == 0) && (dev != last))
			continue;
		next = 1;

    	RDEBUG("fpow_find_component_by_class(0x%x) comparing against %s (class = 0x%x)...\n", dev_class,dev->name,dev->device_class);
        if (dev->device_class == dev_class) {
	    RDEBUG("fpow_find_component_by_class(0x%x) found device %s of class 0x%x.\n", dev_class, dev->name,dev->device_class);
            return dev;
        }
        entry = entry->next;
    }
    RDEBUG("fpow_find_component_by_class(0x%x) FAILED to find device !\n", dev_class);
    return 0;
}


/*********************************************************
 * Routine:     fpow_find_component_by_name
 * Purpose:     Search the registered Fiona Power components
 *		list (g_fpow_components_list) looking for
 *              a device having name of dev_name. 
 * Paramters:   dev_name - name of device to find
 * Returns:     Pointer to fpow_component entry in list
 *              0 - No device named dev_name found in list.
 * Assumptions: Lock is held for g_fpow_components_list
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
static struct fpow_component *fpow_find_component_by_name(const char *dev_name)
{
    struct list_head *entry = g_fpow_components_list.next;
    struct fpow_component *dev = NULL;

    RDEBUG("fpow_find_component_by_name(%s)\n", dev_name);

    while (entry != &g_fpow_components_list) {
        dev = list_entry(entry, struct fpow_component , entry);

    	RDEBUG("fpow_find_component_by_name(%s) comparing against %s...\n", dev_name,dev->name);
        if (0 == strcmp(dev->name, dev_name)) {
	    RDEBUG("fpow_find_component_by_name(%s) found device.\n", dev_name);
            return dev;
        }
        entry = entry->next;
    }
    RDEBUG("fpow_find_component_by_name(%s) FAILED to find device !\n", dev_name);
    return 0;
}

#endif  // Remove compile warning


/*********************************************************
 * Routine:     fpow_print_registered_component_list
 * Purpose:     Prints out the list of registered components.
 * Paramters:   none
 * Returns:     None
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
static void fpow_print_registered_component_list(void)
{
    struct list_head *entry = g_fpow_components_list.next;
    int x = 0;

    while (entry != &g_fpow_components_list) {
        struct fpow_component *dev = list_entry(entry, struct fpow_component , entry);
	x++;
        if (entry != &g_fpow_components_list) {
	        printk( "  Component Name   : %s\n"
		        "  Device Class     : 0x%x\n"
		        "  Supported Modes  : 0x%x\n"
		        "  Get Mode Vector  : 0x%x\n"
		        "  Set Mode Vector  : 0x%x\n"
		        "  Private          : 0x%x\n\n",
		        dev->name,dev->device_class,dev->supported_modes, \
		        (int)dev->getmode,(int)dev->setmode,(int) dev->private);
        }
        entry = entry->next;
    }

    printk("Total Components Registered : %d\n",x);
}


/*********************************************************
 * Routine:     fpow_remove_registered_component
 * Purpose:     Remove the device belonging to the passed 
 *		fpow_component structure from the list of
 *		registered power components (g_fpow_components_list).
 * Paramters:   dev - fpow_dev struct * belonging to registered
 *			component (i.e. pointer returned
 *			from original fpow_register() call)
 * Returns:     None
 * NOTE:	The dev parameter pointer is no longer valid
 *		upon return from this call as the space held 
 *		by the structure pointed to by that pointer
 *		will be freed up / released.
 * Assumptions: Caller has lock held for g_fpow_components_list
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
static void fpow_remove_registered_component(struct fpow_component *dev)
{
        RDEBUG("fpow_remove_registered_component(%s)\n", dev->name);
	list_del(&dev->entry);
}


/*********************************************************
 * Routine:     fpow_add_registered_component
 * Purpose:     Add the device to the sorted g_fpow_components_list.
 *              The list is sorted by priority order, where
 *              the priority is defined by the device_class field
 *              of the dev input parameter.  
 *              Ordering is from largest (first) to smallest (last).
 * Paramters:   dev_name - Device entry to add to sorted list.
 * Returns:     None
 * Assumptions: Caller has lock held for g_fpow_components_list
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
static void fpow_add_registered_component(struct fpow_component *dev)
{
    struct fpow_component *thisDev = NULL;
    struct list_head    *entry = g_fpow_components_list.next;
    int                 thisClass = dev->device_class;

    RDEBUG("fpow_add_registered_component(%s)\n", dev->name);

    while (entry != &g_fpow_components_list) {
        thisDev = list_entry(entry, struct fpow_component , entry);

        // Sort by device_class
        //
        RDEBUG("fpow_add_registered_component(%s): Comparing %d against %s (%d)\n", \
                dev->name, thisClass, thisDev->name,thisDev->device_class);

        // If our class is greater than current class,
        //  insert before this device to maintain ordering
        if (thisClass > thisDev->device_class) {
            RDEBUG("fpow_add_registered_component(%s): Inserting %s (%d) before %s (%d)\n", \
            	dev->name,dev->name,dev->device_class,thisDev->name,thisDev->device_class);
            __list_add(&dev->entry, entry->prev, entry);
            return;
        }
        entry = entry->next;
    }

    // We must be last if we're here, so get on the end of the queue
    //
    RDEBUG("fpow_add_registered_component(%s): Inserting %s (%d) onto end of queue.\n", \
                    dev->name,dev->name,dev->device_class);
    list_add_tail(&dev->entry, &g_fpow_components_list);
}


/////////////////////////////////////////////////////////////
//
// Watermark Proc Routines
//
/////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////
//
// Register / Unregister Routines
//
/////////////////////////////////////////////////////////////

int fpow_add_component_proc_entry(struct fpow_component *dev)
{
	int	rc = 0;

    // 
    // Create /proc/fpow/components directory
	//
    if (proc_fpow_components_parent == NULL)
        proc_fpow_components_parent = create_proc_entry(FPOW_PROC_COMPONENTS_PARENT_NAME, S_IFDIR | S_IRUGO | S_IXUGO, proc_fpow_parent);
    if (!proc_fpow_components_parent) {
		rc = -ENOMEM;
		goto exit;
    }

    // 
    // Create /proc/fpow/components/dev->name entry
    //
	dev->proc_fpow_component_entry = create_proc_entry(dev->name, 0666, proc_fpow_components_parent);
	if (!dev->proc_fpow_component_entry) {
		rc = -ENOMEM;
		goto exit;
	}
	dev->proc_fpow_component_entry->owner	= THIS_MODULE;
	dev->proc_fpow_component_entry->data	= (void*)dev;
	dev->proc_fpow_component_entry->read_proc	= proc_read_fpow_component_mode;
	dev->proc_fpow_component_entry->write_proc	= proc_write_fpow_component_mode;

	/* All's good */
	RDEBUG("/proc/fpow/%s initialized.\n",dev->name);
exit:
	return(rc);
}

void fpow_remove_component_proc_entry(struct fpow_component *dev)
{
    // 
    // Remove /proc/fpow/components/dev->name entry
    //
    if (dev->proc_fpow_component_entry)  {
	    remove_proc_entry(dev->name, proc_fpow_components_parent);
        dev->proc_fpow_component_entry = NULL;
    }
}


/*********************************************************
 * Routine:     fpow_register
 * Purpose:     Register a component with the Fiona Power Manager.
 * Paramters:   reg_form - Pointer to a registration information 
 *			record containing infomation on the component
 *			to register.
 * Returns:     *fpow_component - Pointer to the newly registered
 *			component's power information.  Keep this pointer 
 *			around as caller will use it in subsequent fpow
 *			interaction (i.e. most power routines require this
 *			pointer as a parameter)
 *		NULL	- Error, couldn't register component
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
struct fpow_component * fpow_register(struct fpow_registration_rec *reg_form)
{
	struct fpow_component *dev = NULL;

	powdebug("fpow_register(%s)\n",reg_form->name);

	// First, verify field validity
	if (((reg_form->getmode == NULL) || (reg_form->setmode == NULL)) ||
		(!(reg_form->device_class >= FPOW_CLASS_FIRST) && (reg_form->device_class <= FPOW_CLASS_LAST)))  {
		printk("INVALID registration form for %s\n",reg_form->name);
		return(NULL);
    	}

	// Create a registration list element for this component
	dev = kmalloc(sizeof(struct fpow_component ), GFP_KERNEL);
	if (dev == NULL)  {
		printk("Couldn't get space for registration list entry for %s.\n",reg_form->name);
		return(NULL);
	}

	// Transfer information
	dev->setmode            = reg_form->setmode;
	dev->getmode		    = reg_form->getmode;
	dev->device_class	    = reg_form->device_class;
	dev->supported_modes	= reg_form->supported_modes;
	dev->private		    = reg_form->private;
	strcpy(dev->name, reg_form->name);

	// Add the new registrant to our registration list
	down(&g_fpow_components_lock);
	fpow_add_registered_component(dev);
	globs.registered_component_count++;
	up(&g_fpow_components_lock);
	
	// Add new registrant to the proc entries
	fpow_add_component_proc_entry(dev);

	// FIXME - this would be the place to place the initial call
	// to this component to get it into the current, proper power
	// state...

	// Return the pointer to the registered component record
	return(dev);
}

EXPORT_SYMBOL(fpow_register);


/*********************************************************
 * Routine:     fpow_unregister
 * Purpose:     Unregister a component with the Fiona Power Manager.
 * Paramters:   dev - Pointer to the device's fpow_component structure.
 * Returns:     None
 * NOTE:	*fpow_component structure passed to this routine
 *			will NOT be valid upon return as the space
 *			is freed up and released (device has been
 *			unregistered).
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
void fpow_unregister(struct fpow_component *dev)
{
	RDEBUG("fpow_unregister(%s)\n", dev->name);
	if (NULL == dev)
		return;

	// Remove the registrant from our registration list
	down(&g_fpow_components_lock);
	fpow_remove_registered_component(dev);
	globs.registered_component_count--;
    fpow_remove_component_proc_entry(dev);
	up(&g_fpow_components_lock);
}

EXPORT_SYMBOL(fpow_unregister);


/////////////////////////////////////////////////////////////
//
// Callback Routines
//
/////////////////////////////////////////////////////////////

char * fpow_class_to_str(FPOW_POWER_CLASS class)
{
	switch(class)  {
		case FPOW_CLASS_CPU:
			return("FPOW_CLASS_CPU");
		case FPOW_CLASS_VIDEO_DISPLAY:
			return("FPOW_CLASS_VIDEO_DISPLAY");
		case FPOW_CLASS_WAN:
			return("FPOW_CLASS_WAN");
		case FPOW_CLASS_USB_HOST:
			return("FPOW_CLASS_USB_HOST");
		case FPOW_CLASS_USB_DEVICE:
			return("FPOW_CLASS_USB_DEVICE");
        case FPOW_CLASS_MMC:
            return("FPOW_CLASS_MMC");
        case FPOW_CLASS_IOC:
            return("FPOW_CLASS_IOC");
        case FPOW_CLASS_KBD_BUTTONS:
            return("FPOW_CLASS_KBD_BUTTONS");
        case FPOW_CLASS_SCROLLWHEEL:
            return("FPOW_CLASS_SCROLLWHEEL");
        case FPOW_CLASS_VIDEO_PNLCD:
            return("FPOW_CLASS_VIDEO_PNLCD");
        case FPOW_CLASS_AUDIO:
            return("FPOW_CLASS_AUDIO");
        case FPOW_CLASS_INVALID:
            return("FPOW_CLASS_INVALID");
		default:
			return("UNDEFINED");
	}
}

char * fpow_state_to_str_verbose(FPOW_POWER_STATE state)
{
	switch(state)  {
		case FPOW_STATE_A:
			return("FPOW_STATE_A (Active - A1)");
		case FPOW_STATE_B:
			return("FPOW_STATE_B (Active - A2)");
		case FPOW_STATE_C:
			return("FPOW_STATE_C (Active - B1)");
		case FPOW_STATE_D:
			return("FPOW_STATE_D (Active - B2)");
		case FPOW_STATE_E:
			return("FPOW_STATE_E (Passive - PA1)");
		case FPOW_STATE_F:
			return("FPOW_STATE_F (Passive - PA2)");
		case FPOW_STATE_G:
			return("FPOW_STATE_G (Passive - PB1)");
		case FPOW_STATE_H:
			return("FPOW_STATE_H (Passive - PB2)");
		case FPOW_STATE_I:
			return("FPOW_STATE_I (Rest - RA1 w/o WAN)");
		case FPOW_STATE_J:
			return("FPOW_STATE_J (USB Wake - w/o WAN)");
		case FPOW_STATE_K:
			return("FPOW_STATE_K (USB Wake - w WAN)");
		case FPOW_STATE_L:
			return("FPOW_STATE_L (Rest - RA2 w/ WAN)");
		case FPOW_STATE_M:
			return("FPOW_STATE_M (Rest - Off)");
		case FPOW_STATE_N:
			return("FPOW_STATE_N (temp just-woke-from-OFF-sleep state)");
		case FPOW_STATE_O:
			return("FPOW_STATE_O (Password Entry - w/ WAN w/o Audio)");
		case FPOW_STATE_P:
			return("FPOW_STATE_P (Password Entry - w/o WAN w/o Audio)");
		case FPOW_STATE_Q:
			return("FPOW_STATE_Q (Password Entry - w/ WAN w/ Audio)");
		case FPOW_STATE_R:
			return("FPOW_STATE_R (Password w/o WAN w/ Audio)");
		case FPOW_STATE_S:
			return("FPOW_STATE_S (USB Mounted w/out WAN)");
		case FPOW_STATE_T:
			return("FPOW_STATE_T (OFF - No Power At All)");
		case FPOW_STATE_U:
			return("FPOW_STATE_U (DEAD)");
		case FPOW_STATE_V:
			return("FPOW_STATE_V (Power Lockout)");
		case FPOW_STATE_W:
			return("FPOW_STATE_W (Passive - TPH)");
		case FPOW_STATE_X:
			return("FPOW_STATE_X (USB Mounted w/ WAN)");
		case FPOW_STATE_Y:
			return("FPOW_STATE_Y (Faded idle sleep w/o WAN)");
		case FPOW_STATE_Z:
			return("FPOW_STATE_Z (Faded idle sleep w/ WAN)");
		case FPOW_STATE_WW:
			return("FPOW_STATE_WW (temp just-woke-from-idle-sleep state)");
		case FPOW_STATE_XX:
			return("FPOW_STATE_XX (USB Faded Wake - w WAN)");
		case FPOW_STATE_YY:
			return("FPOW_STATE_YY (USB Faded Wake - w/o WAN)");
		case FPOW_STATE_ZZ:
			return("FPOW_STATE_ZZ (TPH Faded Wake)");
		default:
			return("UNDEFINED");
	}
}
char * fpow_state_to_str(FPOW_POWER_STATE state)
{
	switch(state)  {
		case FPOW_STATE_A:
			return("FPOW_STATE_A");
		case FPOW_STATE_B:
			return("FPOW_STATE_B");
		case FPOW_STATE_C:
			return("FPOW_STATE_C");
		case FPOW_STATE_D:
			return("FPOW_STATE_D");
		case FPOW_STATE_E:
			return("FPOW_STATE_E");
		case FPOW_STATE_F:
			return("FPOW_STATE_F");
		case FPOW_STATE_G:
			return("FPOW_STATE_G");
		case FPOW_STATE_H:
			return("FPOW_STATE_H");
		case FPOW_STATE_I:
			return("FPOW_STATE_I");
		case FPOW_STATE_J:
			return("FPOW_STATE_J");
		case FPOW_STATE_K:
			return("FPOW_STATE_K");
		case FPOW_STATE_L:
			return("FPOW_STATE_L");
		case FPOW_STATE_M:
			return("FPOW_STATE_M");
		case FPOW_STATE_N:
			return("FPOW_STATE_N");
		case FPOW_STATE_O:
			return("FPOW_STATE_O");
		case FPOW_STATE_P:
			return("FPOW_STATE_P");
		case FPOW_STATE_Q:
			return("FPOW_STATE_Q");
		case FPOW_STATE_R:
			return("FPOW_STATE_R");
		case FPOW_STATE_S:
			return("FPOW_STATE_S");
		case FPOW_STATE_T:
			return("FPOW_STATE_T");
		case FPOW_STATE_U:
			return("FPOW_STATE_U");
		case FPOW_STATE_V:
			return("FPOW_STATE_V");
		case FPOW_STATE_W:
			return("FPOW_STATE_W");
		case FPOW_STATE_X:
			return("FPOW_STATE_X");
		case FPOW_STATE_Y:
			return("FPOW_STATE_Y");
		case FPOW_STATE_Z:
			return("FPOW_STATE_Z");
		case FPOW_STATE_WW:
			return("FPOW_STATE_WW");
		case FPOW_STATE_XX:
			return("FPOW_STATE_XX");
		case FPOW_STATE_YY:
			return("FPOW_STATE_YY");
		case FPOW_STATE_ZZ:
			return("FPOW_STATE_ZZ");
		default:
			return("UNDEFINED");
	}
}
EXPORT_SYMBOL(fpow_state_to_str);

char * fpow_mode_to_str(FPOW_POWER_MODE mode)
{
	switch(mode)  {
		case FPOW_MODE_ON:
			return("FPOW_MODE_ON");
		case FPOW_MODE_SLEEP:
			return("FPOW_MODE_SLEEP");
		case FPOW_MODE_OFF:
			return("FPOW_MODE_OFF");
        case FPOW_MODE_200_MHZ:
			return("FPOW_MODE_200_MHZ");
        case FPOW_MODE_400_MHZ:
			return("FPOW_MODE_400_MHZ");
		case FPOW_MODE_INVALID:
			return("FPOW_MODE_INVALID");
		default:
			return("FPOW_MODE_UNDEFINED");
	}
}
EXPORT_SYMBOL(fpow_mode_to_str);

char *  fpow_wake_src_to_str(FPOW_WAKE_SOURCE src)
{
    switch(src)  {
        case USB_WAKE:
            return("USB_WAKE");
        case WAN_RING_WAKE:
            return("WAN_RING_WAKE");
        case POWER_ON_WAKE:
            return("POWER_ON_WAKE");
        case POWER_OFF_WAKE:
            return("POWER_OFF_WAKE");
        case WAN_ON_WAKE:
            return("WAN_ON_WAKE");
        case WAN_OFF_WAKE:
            return("WAN_OFF_WAKE");
        case KEY_PRESS_WAKE:
            return("KEY_PRESS_WAKE");
        case AC_PLUG_INSERTED_WAKE:
            return("AC_PLUG_INSERTED_WAKE");
        case LOW_BATTERY_WAKE:
            return("LOW_BATTERY_WAKE");
        case PXA_UART_WAKE:
            return("PXA_UART_WAKE");
        case IOC_WAKE:
            return("IOC_WAKE");
        case WATERMARK_WAKE:
            return("WATERMARK_WAKE");
        case ALT_FONT_KEY_COMBINATION:
            return("ALT_FONT_KEY_COMBINATION");
        case WAKE_INVALID:
        default:
            return("WAKE_INVALID");
    }
}

u32 fpow_mode_to_syspm_level(FPOW_POWER_MODE mode)
{
    switch (mode)  {
    	case FPOW_MODE_ON:
        	return(RESUME_ENABLE);
        case FPOW_MODE_SLEEP:
            return(SUSPEND_DISABLE);
        case FPOW_MODE_OFF:
            return(SUSPEND_POWER_DOWN);
        default:
    		return(FPOW_MODE_INVALID);
    }
}
                                                                                 
static int fpow_component_supports_mode(struct fpow_component *dev, FPOW_POWER_MODE mode)
{
	switch(mode)  {
		case FPOW_MODE_ON:
			return(dev->supported_modes & FPOW_MODE_ON_SUPPORTED);
		case FPOW_MODE_400_MHZ:
			return(dev->supported_modes & FPOW_MODE_400_MHZ_SUPPORTED);
		case FPOW_MODE_200_MHZ:
			return(dev->supported_modes & FPOW_MODE_200_MHZ_SUPPORTED);
		case FPOW_MODE_SLEEP:
			return(dev->supported_modes & FPOW_MODE_SLEEP_SUPPORTED);
		case FPOW_MODE_OFF:
			return(dev->supported_modes & FPOW_MODE_OFF_SUPPORTED);
		default:
			return(0);
	}
}


/*********************************************************
 * Routine:     fpow_send_cmd
 *
 * Purpose:     Send a power mode change request to specified
 *		component
 *
 * Parameters:  dev - component to send mode change request to
 *		mode - FPOW_POWER_MODE for component to switch to
 *
 * Returns: 
 *  FPOW_NO_ERR                     - Transition completed.
 *  -FPOW_COMPONENT_TRANSITION_ERR  - A component couldn't transition 
 *  -FPOW_NOT_ENOUGH_POWER_ERR      - Not enough power 
 *  -EINVAL                         - Invalid mode
 *
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
int fpow_send_cmd(struct fpow_component *dev, FPOW_POWER_MODE mode)
{
	int	complete = FPOW_NO_ERR;
	int	state = 0;	// For now, all state's are 0

	extpowdebug("fpow_send_cmd('%s', %s)\n",\
			dev->name,fpow_mode_to_str(mode));

	switch(mode)  {
		case FPOW_MODE_ON:
		case FPOW_MODE_SLEEP:
		case FPOW_MODE_OFF:
		case FPOW_MODE_400_MHZ:
		case FPOW_MODE_200_MHZ:
			complete = (*dev->setmode)((struct device*)dev->private, state, mode);
			break;
		default:
			printk("Unknown mode 0x%x passed to fpow_send_cmd.\n",mode);
			return(-EINVAL);
	}

	// Return complete result
	return(complete);
}



/*********************************************************
 * Routine:     fpow_send_to_class
 * Purpose:     Send a power mode change request to all components
 *		who registered for interest in that event AND are of
 *      device_class = class.
 * Paramters:   mode - Power mode for component to transition to.
 *              class - Device class type to send command to
 * Returns:
 *  FPOW_NO_ERR                     - Transition completed.
 *  -FPOW_COMPONENT_TRANSITION_ERR  - A component couldn't transition 
 *  -FPOW_NOT_ENOUGH_POWER_ERR      - Not enough power 
 *  -FPOW_NO_MATCHES_FOUND_ERR      - No one was registered for event
 *  -EINVAL                         - Invalid mode
 *
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
static int fpow_send_to_class(FPOW_POWER_MODE mode, FPOW_POWER_CLASS class)
{
	int	                    rc = -FPOW_NO_MATCHES_FOUND_ERR;
	struct fpow_component 	*dev = NULL;
	struct list_head	    *entry;

	extpowdebug("fpow_send_to_class(%s,%s)\n",fpow_mode_to_str(mode),fpow_class_to_str(class));
	down(&g_fpow_components_lock);
    entry = g_fpow_components_list.next;

	// For each item, if component is interested in this event, send it... 
	while (entry != &g_fpow_components_list) {
		dev = list_entry(entry, struct fpow_component, entry);
        if ((dev->device_class == class) && (fpow_component_supports_mode(dev,mode))) {
			rc = fpow_send_cmd(dev,mode);
            if (rc != FPOW_NO_ERR)
                break;
        }
		entry = entry->next;
	}
	up(&g_fpow_components_lock);
	return(rc);
}


/*********************************************************************
 *
 * Fiona Power State Machine Engine
 *
 *********************************************************************/
int
fiona_sleep_fpow_components(void)
{
    int rc = 0;

    // Going to Sleep here
    powdebug("fiona_sleep_fpow_components() entered.  Going to sleep...\n");
    rc = fiona_power_state(FPOW_STATE_OFF);
    if (rc)  {
        powdebug("fiona_sleep_fpow_components(): fiona_power_state(OFF) " \
                 "returned error %d\n",rc);
    } 
    powdebug("fiona_sleep_entire_system(): Back from sleeping....\n");
    return(rc);
}

int
fiona_wake_fpow_components(FPOW_POWER_STATE state)
{
    int rc = 0;

    // System's awake, time to wake the rest of the system...
    powdebug("fiona_wake_fpow_components() entered.  Waking components...\n");
    rc = fiona_power_state(state);
    if (rc)  {
        powdebug("fiona_wake_fpow_components(): fiona_power_state(%s) " \
                 "returned error %d\n",fpow_state_to_str(state),rc);
    } 
    return(rc);

}

static int is_WAN_state(FPOW_POWER_STATE state)
{
    FPOW_POWER_MODE mode = globs.power_state_xlate[state][FPOW_CLASS_WAN];
    return(mode != FPOW_MODE_OFF);
}


static int is_power_lockout_state(FPOW_POWER_STATE state)
{
    return(state == FPOW_STATE_V);
}


static int is_processor_off_state(FPOW_POWER_STATE state)
{
    return(state == FPOW_STATE_M);
}

int is_temp_wake_state(FPOW_POWER_STATE state)
{
    switch(state)  {
        case FPOW_OFF_WAKE_STATE:
        case FPOW_IDLE_WAKE_STATE:
            return(1);
        default:
            return(0);
    }
}

static int
is_processor_sleep_state(FPOW_POWER_STATE state)
{
    FPOW_POWER_MODE mode = globs.power_state_xlate[state][FPOW_CLASS_CPU];
    return((mode == FPOW_MODE_SLEEP) || (mode == FPOW_MODE_OFF));
}


static int
is_wakeup_timer_enabled_state(FPOW_POWER_STATE state)
{
    switch(state)  {
        case FPOW_STATE_I:
        case FPOW_STATE_L:
        case FPOW_STATE_Y:
        case FPOW_STATE_Z:
        case FPOW_STATE_XX:
        case FPOW_STATE_YY:
        case FPOW_STATE_ZZ:
            return(1);
        default:
            return(0);
    }
}


/***************************************************************************
 * Routine:     fpow_video_updated
 * Purpose:     Called by the EINK driver upon finishing a screen update/draw.
 *              This allows us to determine when to start respecting ALT+FONT
 *              events again by sending the TWO_FINGER_SWEEPER event to Java
 *              Power Management.
 * Paramters:   none
 * Returns:     none
 ***************************************************************************/
void fpow_video_updated(void)
{
    handle_fiona_event(TWO_FINGER_SWEEPER_EVENTCODE);
    powdebug("fpow_video_updated(): ALT_FONT sweeper sent.\n");
}
EXPORT_SYMBOL(fpow_video_updated);


static void post_wake_event(void)
{
    int event = 0;

    if (g_fpow_wake_source == IOC_WAKE) {
        g_fpow_wake_source = determine_ioc_wake_reason();
        powdebug("post_wake_event() got IOC wake reason %s (%d)...\n",fpow_wake_src_to_str(g_fpow_wake_source),g_fpow_wake_source);
    }

    // If it's one of our events that doesn't get posted elsewhere,
    //  post it from here
    powdebug("post_wake_event() checking wake %s event...\n",fpow_wake_src_to_str(g_fpow_wake_source));
    switch(g_fpow_wake_source)  {

        // These need to get handled
        case KEY_PRESS_WAKE:
            event = KEY_WAKE_EVENTCODE;
            break;

        case AC_PLUG_INSERTED_WAKE:
            event = AC_IN_EVENTCODE;
            break;

        case USB_WAKE:
            event = USB_WAKE_EVENTCODE;
            break;

        case LOW_BATTERY_WAKE:
            event = BATTERY_LOW_EVENTCODE;
            break;

        case WAN_ON_WAKE:
            event = WAN_SW_ON_EVENTCODE;
            break;

        case WAN_OFF_WAKE:
            event = WAN_SW_OFF_EVENTCODE;
            break;

        case WATERMARK_WAKE:
            event = WATERMARK_CHECK_EVENTCODE;
            break;

        // These get handled elsewhere
        case WAN_RING_WAKE:
        case POWER_ON_WAKE:
        case POWER_OFF_WAKE:
        case PXA_UART_WAKE:
        case IOC_WAKE:
        default:
            powdebug("post_wake_event() ignored %s event\n",fpow_wake_src_to_str(g_fpow_wake_source));
            break;
    }
    if (event)  {
        hlog("post_wake_event sending wake event %s as event %s\n",fpow_wake_src_to_str(g_fpow_wake_source),fiona_event_keycode_to_str(event));
        handle_fiona_event(event);
    }
    else
        hlog("post_wake_event not posting wake event %s (posted elsewhere)\n",fpow_wake_src_to_str(g_fpow_wake_source));
}


static FPOW_WAKE_SOURCE determine_ioc_wake_reason(void)
{
    IOC_WAKE_CAUSE_REC  ioc_wake_rec;
    int err = 0;

    // FIXME -njv Fill this out, have to deal with fact that IOC driver is not
    //  up yet, so how should we communicate with IOC (defer communication, 
    //  access directly, have framework access later?
    memclr(&ioc_wake_rec,sizeof(ioc_wake_rec));
    err = ioc_sys_ioctl(0, IOC_QUERY_IOC_WAKE_CAUSE_IOCTL, (unsigned long)&ioc_wake_rec);
    if (err)  {
        printk("Received error %d from ioc_sys_ioctl(IOC_QUERY_IOC_WAKE_CAUSE_IOCTL)\n",err);
        ioc_wake_rec.wake_source = WAKE_INVALID;
    }
    return(ioc_wake_rec.wake_source);
}


void set_remembered_wake_source(int event)
{
    powdebug("set_remembered_wake_source(%s = %d)\n",fpow_wake_src_to_str(event),event);
    g_fpow_wake_source = event;
}


/***************************************************************************
 * Routine:     fiona_get_wake_source
 * Purpose:     Returns the last wake source that woke the device
 * Paramters:   none
 * Returns:     FPOW_WAKE_SOURCE (defined in fiona_power.h)
 *              One of:
 *                  WAKE_INVALID    (device hasn't slept/woke yet)
 *                  WAN_RING_WAKE
 *                  POWER_ON_WAKE
 *                  POWER_OFF_WAKE
 *                  WAN_ON_WAKE
 *                  WAN_OFF_WAKE
 *                  KEY_PRESS_WAKE
 *                  AC_PLUG_INSERTED_WAKE
 *                  LOW_BATTERY_WAKE
 *                  WATERMARK_WAKE
 *                  USB_WAKE
 ***************************************************************************/
int fiona_get_wake_source(void)
{
    return(g_fpow_wake_source);
}


static void remember_wake_source(void)
{
    int source = PEDR;
    FPOW_WAKE_SOURCE reason = WATERMARK_WAKE;

    /*
     * Possible wake sources are:
     *  1) Power Switch On/Off
     *  2) WAN Switch On/Off
     *  3) IOC
     *  4) WAN Ring
     *  5) USB Plug insert/remove
     */

    //
    // NOTE:
    // There's a bug here.  If user toggles fast enough such
    // that the switch changes between system wakeup and this
    // call, we'll return the wrong reason because we check the
    // status of the switch real-time in remember_wake_source() instead of
    // having cached off the initial wake case in the interrupt handler
    // and reading that here (that would be one possible fix, but 
    // the method described below is a step better and does not require
    // this mechanism to change....
    //
    // We're leaving this as-is, and are going to instead turn
    // this into a feature by the table change for power tables v.57.
    // This table change simply enables the "Power Switch On" event
    // to be handled by STATE_I and STATE_L where it was ignored
    // before.  If the opposite polarity transition was missed, by
    // using the current setting, we get to bypass the Power Manager 
    // potentially starting the first Power Off event only to get
    // far enough to see the Power On and then reverse engines.  This
    // would take a bit of time.  By instead enabling Power Switch On
    // as a transitional event for STATE_I and STATE_L, we get a plus
    // effect when an event falls within this window as it gives us
    // an automatic skip of the "user changed mind, reverse engines"
    // process, thus saving user time.  This is especially true for
    // WAN switch transitions as those are much more costly in terms
    // of time (30-60 seconds) and in power consumption. -njv 11/16/07
    // 

    if (source & WAN_WAKE_PEDR_BIT)
        reason = WAN_RING_WAKE;
    else if (source & POWER_SWITCH_PEDR_BIT)
        reason = power_switch() ? POWER_ON_WAKE : POWER_OFF_WAKE;
    else if (source & WAN_SWITCH_PEDR_BIT)
        reason = wan_switch() ? WAN_ON_WAKE : WAN_OFF_WAKE;
    else if (source & USB_WAKE_PEDR_BIT)
        reason = USB_WAKE;
    else if (source & IOC_PEDR_BIT) {
        reason = IOC_WAKE;
    }
    set_remembered_wake_source(reason);
    printk("FPOW: wake source was %s\n",fpow_wake_src_to_str(g_fpow_wake_source));
}


void 
fiona_clear_wake_source(int gpio)
{
    g_fiona_hw.pwer &= ~PWER_GPIO(gpio);
}
void 
fiona_set_wake_source(int gpio)
{
    g_fiona_hw.pwer |= PWER_GPIO(gpio);
}

void 
fiona_set_rising_edge_wake_source(int gpio)
{
    g_fiona_hw.prer |= PRER_GPIO(gpio);
}
void 
fiona_set_falling_edge_wake_source(int gpio)
{
    g_fiona_hw.pfer |= PFER_GPIO(gpio);
}
void 
fiona_clear_rising_edge_wake_source(int gpio)
{
    g_fiona_hw.prer &= ~PRER_GPIO(gpio);
}
void 
fiona_clear_falling_edge_wake_source(int gpio)
{
    g_fiona_hw.pfer &= ~PFER_GPIO(gpio);
}


void fiona_set_power_switch_wake_source(int enable, int bothEdges)
{
    if (enable) {
        fiona_set_wake_source(g_fiona_hw.sleep_switch_gpio);

        if (g_fiona_hw.sleep_switch_polarity)  {
            // Off is the assert / rising edge, On is the falling edge
            fiona_set_falling_edge_wake_source(g_fiona_hw.sleep_switch_gpio);
            if (bothEdges == 0)
                fiona_clear_rising_edge_wake_source(g_fiona_hw.sleep_switch_gpio);
            else
                fiona_set_rising_edge_wake_source(g_fiona_hw.sleep_switch_gpio);
        }
        else {
            fiona_set_rising_edge_wake_source(g_fiona_hw.sleep_switch_gpio);
            if (bothEdges == 0)
                fiona_clear_falling_edge_wake_source(g_fiona_hw.sleep_switch_gpio);
            else
                fiona_set_falling_edge_wake_source(g_fiona_hw.sleep_switch_gpio);
        }
    }
    else  {
        fiona_clear_wake_source(g_fiona_hw.sleep_switch_gpio);
        fiona_clear_falling_edge_wake_source(g_fiona_hw.sleep_switch_gpio);
        fiona_clear_rising_edge_wake_source(g_fiona_hw.sleep_switch_gpio);
    }
}


void fiona_set_usb_wake_source(int enable)
{
    if (enable)  {
        fiona_set_wake_source(g_fiona_hw.usb_wake_gpio);
        if (g_fiona_hw.usb_wake_polarity) {
            fiona_clear_falling_edge_wake_source(g_fiona_hw.usb_wake_gpio);
            fiona_set_rising_edge_wake_source(g_fiona_hw.usb_wake_gpio);
        }
        else {
            fiona_clear_rising_edge_wake_source(g_fiona_hw.usb_wake_gpio);
            fiona_set_falling_edge_wake_source(g_fiona_hw.usb_wake_gpio);
        }
    } else {
        fiona_clear_wake_source(g_fiona_hw.usb_wake_gpio);
        fiona_clear_rising_edge_wake_source(g_fiona_hw.usb_wake_gpio);
        fiona_clear_falling_edge_wake_source(g_fiona_hw.usb_wake_gpio);
    }
}


void fiona_set_ioc_wake_source(int enable)
{
    if (enable)  {
        fiona_set_wake_source(g_fiona_hw.ioc_wake_gpio);
        if (g_fiona_hw.ioc_wake_polarity) {
            fiona_set_rising_edge_wake_source(g_fiona_hw.ioc_wake_gpio);
            fiona_clear_falling_edge_wake_source(g_fiona_hw.ioc_wake_gpio);
        }
        else {
            fiona_set_falling_edge_wake_source(g_fiona_hw.ioc_wake_gpio);
            fiona_clear_rising_edge_wake_source(g_fiona_hw.ioc_wake_gpio);
        }
    } else {
        fiona_clear_wake_source(g_fiona_hw.ioc_wake_gpio);
        fiona_clear_rising_edge_wake_source(g_fiona_hw.ioc_wake_gpio);
        fiona_clear_falling_edge_wake_source(g_fiona_hw.ioc_wake_gpio);
    }
}


FPOW_WAKE_SOURCE convert_passed_event_to_wake_src(int event)  {
    FPOW_WAKE_SOURCE wakeEvent = WAKE_INVALID;
    switch(event)  {
        case EVENT_WAN_RING:
            wakeEvent = WAN_RING_WAKE;
            break;
        case EVENT_POWER_SWITCH_ON:
            wakeEvent = POWER_ON_WAKE;
            break;
        case EVENT_POWER_SWITCH_OFF:
            wakeEvent = POWER_OFF_WAKE;
            break;
        case EVENT_WAN_SWITCH_ON:
            wakeEvent = WAN_ON_WAKE;
            break;
        case EVENT_WAN_SWITCH_OFF:
            wakeEvent = WAN_OFF_WAKE;
            break;
        case EVENT_KEY_WAKE:
            wakeEvent = KEY_PRESS_WAKE;
            break;
        case EVENT_AC_INSERTED:
            wakeEvent = AC_PLUG_INSERTED_WAKE;
            break;
        case EVENT_BATTERY_LOW:
            wakeEvent = LOW_BATTERY_WAKE;
            break;
        case EVENT_USB_WAKE:
            wakeEvent = USB_WAKE;
            break;
        case EVENT_WATERMARK_CHECK:
            wakeEvent = WATERMARK_WAKE;
            break;
        case ALT_FONT_KEY_COMBINATION:
            wakeEvent = KEY_PRESS_WAKE;
            break;
        case INVALID_EVENT:
            wakeEvent = WAKE_INVALID;
            break;
        default:
            break;
    }
    powdebug("convert_passed_event_to_wake_src(%s) returned %d (%s)\n",fpow_passed_event_to_str(event),wakeEvent,fpow_wake_src_to_str(wakeEvent));
    return(wakeEvent);
}



/***************************************************************************
 * Routine:     sniffer
 * Purpose:     Search the event queue to determine if the current sleep
 *              transition that's caused by the passed event has an event
 *              currently in the event queue that would have caused the device
 *              to wakeup from the target sleep mode should the event had
 *              come after we were sleeping (i.e. if processing our sleep event
 *              could have been done in real time).
 *
 *              If cancelling events are found and can be removed, they will be.
 *              
 * Paramters:   sniff_state - the target state type to sniff for 
 * Returns:     0 - No cancelling events found in queue
 *              1 - Cancelling event found in queue
 * Assumptions: none
 ***************************************************************************/
int sniffer(FPOW_SNIFF_STATES sniff_state) {
    int cancel = 0;
    int wakeEventFoundInQ = INVALID_EVENT;

    switch(sniff_state)  {
            // Look for events that would cancel a "Power Switch Off" operation
            // If a ON/OFF combination exists, remove it
        case OFF_STATE_SNIFF:
            cancel = sniff_remove_events_from_queue(OFF_STATE_SNIFF);     
            if (cancel)
                hlog("sniffer(OFF) found POWER_SWITCH_ON event\n");
            break;

            // Look for events that would cancel a "Power Switch On" operation
        case ON_STATE_SNIFF:
            cancel = sniff_remove_events_from_queue(ON_STATE_SNIFF);     
            if (cancel)
                hlog("sniffer(ON) found POWER_SWITCH_OFF event\n");
            break;

            // Look for events that would cancel a "Power Lockout" operation
        case POWER_LOCKOUT_STATE_SNIFF:
            cancel = sniff_remove_events_from_queue(POWER_LOCKOUT_STATE_SNIFF);     
            if (cancel)
                hlog("sniffer(POWER_LOCKOUT) found AC_IN event\n");
            break;

        case IDLE_SLEEP_STATE_SNIFF:
            // should check for off here, but it's checked by caller
            wakeEventFoundInQ = idle_wake_event_in_event_queue();
            if (wakeEventFoundInQ != INVALID_EVENT) {
                hlog("sniffer(IDLE_SLEEP) found wake event %s\n",fpow_wake_src_to_str(wakeEventFoundInQ));
                cancel = 1;
            }
            break;

        default:
            break;
    }
    return(cancel);
}


static int
check_calc_op_amp_offset(int verbose) {

#ifndef CONFIG_CALCULATE_OP_AMP_OFFSET

    return(get_op_amp_offset());

#else

    int             y = 0;
    int             sample_count = 0;
    int             op_amp_offset = get_op_amp_offset();    // Default to value in boot globals
    unsigned long   amp_sum = 0;
    unsigned long   amps[OP_AMP_OFFSET_CALCULATION_SAMPLE_COUNT];
    int             draw = 0;

    // Verify we have a new enough version of firmware
    // If AC plug is inserted, we can't calculate resistance, so bail...
    if (ac_adapter_installed()) {
        if (verbose)
            printk("check_calc_op_amp_offset(AC-in) returning default value %d\n",op_amp_offset);
        return(op_amp_offset);
    }

    printk("Check now...\n");
#ifdef RUN_WD_TEST
    for (y=0;y<80;y++)  
    {
        printk("%d Seconds... Here poochie poochie.... \n",y+1); 
        mdelay(1000);
    }
#endif

    // Grab some samples
    sample_count = 0;
    while (sample_count < OP_AMP_OFFSET_CALCULATION_SAMPLE_COUNT) {
        draw = fiona_get_current_draw();
        if (draw == 0) {
            if (verbose)
                printk("check_calc_op_amp_offset() ERRORED getting battery info from IOC..\n");
            return(op_amp_offset);
        }
        amps[sample_count++] = draw;
        mdelay(1000);
    }

    // Now average our samples
    amp_sum = 0;
    for (y=0;y<sample_count;y++) {
        if (verbose)
            printk("Op amp offset sample %d: %d mA\n",y+1,(int)amps[y]);
        amp_sum += amps[y];
    }
    amp_sum /= sample_count;

    // Determine our offset by subtracting read value from real value...
    op_amp_offset = (MAMP_DRAW_WHILE_IDLE - amp_sum);
    printk("Op amp offset calculated at %d mA\n",op_amp_offset);
    return(op_amp_offset);
#endif
}

#ifdef ENABLE_EVENT_TEST_PROC
static char *get_expected_results(int state)  {
    switch (state)  {
        case 1:
            return("Do: Turn power switch OFF\nExpected Result: Device comes back on in active state\n");
        case 2:
            return("Do: ALT+FONT to screen saver, T2 timeout to sleep\nExpected Results:  Device clears screen and goes to STATE_M\n");
        case 3:
            return("Do: WAN Switch ON, ALT+FONT to screen saver, T2 timeout to sleep\nExpected Results: Device shuts off wan and ends sleeping in STATE_I\n");
        case 4:
            return("Do: WAN Switch OFF, ALT+FONT to screen saver, T2 timeout to sleep\nExpected Results:  Device enables wan and ends sleeping in STATE_L\n");
        case 5:
            return("Do: ALT+FONT to screen saver, T2 timeout to sleep\nExpected Result: Device comes back on in active state\n");
        case 6:
            return("Do: Drain battery til CRITICAL_BATTERY_EVENT is generated\nExpected Result: Device comes back on briefly showing reverse transition in log which would show success\n");
        case 7:
            return("Do: Turn power switch OFF\nExpected Result: Device clears screen and goes to sleep after removing the four injected events from the event queue\n");
        case 8:
            return("Do: Turn power switch OFF\nExpected Result: Device should end awake and in an active state\n");
        case 9:
            return("Do: ALT+FONT to screen saver, T2 timeout to sleep\nExpected Results:  Device should end awake and in an active state\n");
        case 10:
             return("Do: ALT+FONT to screen saver, T2 timeout to sleep\nExpected Results:  Device clears screen and goes to STATE_M AFTER removing the first four injected events (PS_ON, WON, WOFF, PS_OFF) from queue\n");
        default:
            return("Unsupported test case\n");
    }
}
#endif

/*******************************************************************
 * Routine:     fiona_power_state
 * Purpose:     Send a power state change request to the FPOW components.
 *              (i.e. change the device state to that requested.
 * Paramters:   state - the FPOW state to transition to
 * Returns:     FPOW_ERR - One of:
 *  FPOW_NO_ERR                 :   Successfully changed state
 *  -EINVAL                     :   Bad parameter (out-of-range), state not changed
 *  -EAGAIN                     :   Could not freeze processes, state not changed
 *  -FPOW_CANCELLED_EVENT_ERR   :   User canceled event, FPOW will make it
 *                                  invisible and generate a wake event as
 *                                  if the device had gone to sleep and woken
 *                                  back up.
 *  -FPOW_NO_MATCHES_FOUND_ERR  :   There was not a particular class component
 *                                  registered in fpow, but state still changed
 *  -FPOW_COMPONENT_TRANSITION_ERR: A component couldn't transition 
 *  -FPOW_NOT_ENOUGH_POWER_ERR  :   Not enough power to complete operation
 *
 * Assumptions: none
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *******************************************************************/
static FPOW_ERR
fiona_power_state(FPOW_POWER_STATE state)
{
    int rc = FPOW_NO_ERR;
    int err = FPOW_NO_ERR;
    int sleep_err = FPOW_NO_ERR;
    int class;
    int target_state = state;
    int sleep_target_state = 0;
    int off_target_state = 0;
    int power_lockout_state = 0;
    int last_state = globs.system_state;
    int real_last_target_state = FPOW_STATE_INVALID;
    int reverse_order = 0;
    int we_slept = 0;
    int post_wake_source = 0;
    int wakeEvent = INVALID_EVENT;
    int abortReason = 0;
    int wakeEventFoundInQ = INVALID_EVENT;
    int unstalled_queue = 0;

#ifdef DEBUG_IOC_WAKE_TRANSACTIONS
    int orig = g_do_core_debug; 
#endif

    // Range-check parameter
    if (!((target_state >= FPOW_STATE_FIRST) && (target_state <= FPOW_STATE_LAST))) {
        powdebug("fiona_power_state(%d) - bad state %d !!\n",target_state,target_state);
        rc = -EINVAL;
        goto exit;
    }

    poweventdebug(" fiona_power_state(%s -> %s)\n",fpow_state_to_str(last_state),fpow_state_to_str(state));

    // If the CPU is going to sleep, but into a state where we're not supposed to wake and
    // check battery level,  make sure we disable the wakeup timer
    if (is_wakeup_timer_enabled_state(target_state)) {
        powdebug("fiona_power_state(%s -> %s) wakeup timer-enabled state, enabling.\n",fpow_state_to_str(globs.system_state),fpow_state_to_str(target_state));
        enable_wakeup_timer();
    } else {
        powdebug("fiona_power_state(%s -> %s) wakeup timer-disabled state, disabling.\n",fpow_state_to_str(globs.system_state),fpow_state_to_str(target_state));
        disable_wakeup_timer();
    }

    // If the CPU is going to sleep, make sure we put the processes to sleep,
    // and do the wakeup calling in reverse order...
    if (is_processor_sleep_state(target_state))  {
        sleep_target_state = 1;
        off_target_state = is_processor_off_state(target_state);
        power_lockout_state = is_power_lockout_state(target_state);

#ifdef FPOW_DO_PROCESS_SUSPEND_RESUME
        powdebug("fiona_power_state(): Freezing processes\n");
        pm_prepare_console();
        g_threads_are_stopped = 1;
        if (freeze_processes()) {
            rc = -EAGAIN;
            goto thaw;
        }
#endif
        
#ifdef ENABLE_EVENT_TEST_PROC
        // Processes are stopped, should be safe as no events will get sent
        // upstream yet...
        if (g_fiona_insert_event)  {
            int test_event = INVALID_EVENT;
            /*
             *      To execute any of these tests, echo the number into /proc/fpow/event_test
             *      prior to performing the "Do:" line and look for expected results.
             *
             *      Each of these will execute once and then test mechanism will reset.
             *
             *      1 - inject a POWER_ON event
             *      2 - inject a POWER_OFF event
             *      3 - inject a WAN_SWITCH_OFF event
             *      4 - inject a WAN_SWITCH_ON event 
             *      5 - inject an ALT+FONT key event sequence 
             *      6 - inject an AC_IN event 
             *      7 - inject a POWER_ON : WAN_ON : WAN_OFF : POWER_OFF events
             *      8 - inject a POWER_ON : WAN_ON : WAN_OFF : POWER_OFF : POWER_ON events
             *      9 - inject a POWER_OFF : WAN_ON : WAN_OFF : POWER_ON events
             *      10 - inject a POWER_OFF : WAN_ON : WAN_OFF : POWER_ON : POWER_OFF events
             *
             *      Here's the test senario for each mode listed above:
             *
             *      1   Do: Turn power switch OFF
             *          Expected Result: Device comes back on in active state
             *
             *          NOTE:  unless DO_PS_IGNORE_TEST is set in arch/arm/mach/fiona.c,
             *                  code will correct for discrepency with actual switch position !! 
             *                  (code's too smart for test :-)
             *
             *      2   Do: ALT+FONT to screen saver, T2 timeout to sleep
             *          Expected Results:  Device clears screen and goes to STATE_M
             *
             *      3   Do: WAN Switch ON, ALT+FONT to screen saver, T2 timeout to sleep
             *          Expected Results:  Device shuts off wan and ends sleeping in STATE_I
             *
             *      4   Do: WAN Switch OFF, ALT+FONT to screen saver, T2 timeout to sleep
             *          Expected Results:  Device enables wan and ends sleeping in STATE_L
             *
             *      5   Do: ALT+FONT to screen saver, T2 timeout to sleep
             *          Expected Result: Device comes back on in active state
             * 
             *      6   Do: Drain battery til CRITICAL_BATTERY_EVENT is generated
             *          Expected Result: Device comes back on briefly (should see 
             *          reverse transition in log which would show success, but it may
             *          shortly thereafter process another one since we never actually
             *          applied any power in this case (but pretended we did to test reverse
             *          engine)
             *
             *      7   Do: Turn power switch OFF
             *          Expected Result: Device clears screen and goes to sleep after removing
             *          the four injected events from the event queue
             *
             *      8   Do: Turn power switch OFF
             *          Expected Result: Device should end awake and in an active state
             *
             *      9   Do: ALT+FONT to screen saver, T2 timeout to sleep
             *          Expected Results:  Device should end awake and in an active state
             *
             *      10  Do: ALT+FONT to screen saver, T2 timeout to sleep
             *          Expected Results:  Device clears screen and goes to STATE_M AFTER removing
             *          the first four injected events (PS_ON, WON, WOFF, PS_OFF) from queue
             *
             */
            g_ps_ignore_next_wake = 0; // clear flag, only set for particular tests
            switch(g_fiona_insert_event)  {
                case 1:
                    printk("\n\nFPOW: test_event going to insert POWER_SW_ON.\n");
                    test_event = POWER_SW_ON_EVENTCODE;
                    g_ps_ignore_next_wake = 1;
                    break;

                case 2:
                    printk("\n\nFPOW: test_event going to insert POWER_SW_OFF\n");
                    test_event = POWER_SW_OFF_EVENTCODE;
                    g_ps_ignore_next_wake = 1; // Power Switch status will be off, don't correct for it
                    break;

                case 3:
                    printk("\n\nFPOW: test_event going to insert WAN_SW_OFF\n");
                    test_event = WAN_SW_OFF_EVENTCODE;
                    break;

                case 4:
                    printk("\n\nFPOW: test_event going to insert WAN_SW_ON\n");
                    test_event = WAN_SW_ON_EVENTCODE;
                    break;

                case 5:
                    // Insert ALT+FONT key events into queue
                    printk("Inserting ALT+FONT keys into event queue.\n");
                    handle_fiona_event(keycode_to_rawcode(ALT_KEY_PRESSED));
                    handle_fiona_event(keycode_to_rawcode(FONT_KEY_PRESSED));
                    handle_fiona_event(keycode_to_rawcode(FONT_KEY_PRESSED) | 0x80);
                    handle_fiona_event(keycode_to_rawcode(ALT_KEY_PRESSED)|0x80);
                    break;

                case 6:
                    printk("\n\nFPOW: test_event going to insert AC_IN\n");
                    test_event = AC_IN_EVENTCODE;
                    break;

                case 7:     // inject a POWER_ON : WAN_ON : WAN_OFF : POWER_OFF events
                    printk("Inserting POWER_ON : WAN_ON : WAN_OFF : POWER_OFF events into event queue.\n");
                    handle_fiona_event(POWER_SW_ON_EVENTCODE);
                    handle_fiona_event(WAN_SW_ON_EVENTCODE);
                    handle_fiona_event(WAN_SW_OFF_EVENTCODE);
                    handle_fiona_event(POWER_SW_OFF_EVENTCODE);
                    break;

                case 8:     // inject a POWER_ON : WAN_ON : WAN_OFF : POWER_OFF : POWER_ON events
                    printk("Inserting POWER_ON : WAN_ON : WAN_OFF : POWER_OFF : POWER_ON events into event queue.\n");
                    handle_fiona_event(POWER_SW_ON_EVENTCODE);
                    handle_fiona_event(WAN_SW_ON_EVENTCODE);
                    handle_fiona_event(WAN_SW_OFF_EVENTCODE);
                    handle_fiona_event(POWER_SW_OFF_EVENTCODE);
                    handle_fiona_event(POWER_SW_ON_EVENTCODE);
                    g_ps_ignore_next_wake = 1; // Power Switch status will be off, don't correct for it
                    break;

                case 9:     // inject a POWER_OFF : WAN_ON : WAN_OFF : POWER_ON events
                    printk("Inserting POWER_OFF : WAN_ON : WAN_OFF : POWER_ON events into event queue.\n");
                    handle_fiona_event(POWER_SW_OFF_EVENTCODE);
                    handle_fiona_event(WAN_SW_ON_EVENTCODE);
                    handle_fiona_event(WAN_SW_OFF_EVENTCODE);
                    handle_fiona_event(POWER_SW_ON_EVENTCODE);
                    break;

                case 10:    // inject a POWER_OFF : WAN_ON : WAN_OFF : POWER_ON : POWER_OFF events
                    printk("Inserting POWER_OFF : WAN_ON : WAN_OFF : POWER_ON : POWER_OFF events into event queue.\n");
                    handle_fiona_event(POWER_SW_OFF_EVENTCODE);
                    handle_fiona_event(WAN_SW_ON_EVENTCODE);
                    handle_fiona_event(WAN_SW_OFF_EVENTCODE);
                    handle_fiona_event(POWER_SW_ON_EVENTCODE);
                    handle_fiona_event(POWER_SW_OFF_EVENTCODE);
                    g_ps_ignore_next_wake = 1; // Power Switch status will be off, don't correct for it
                    break;
                default:
                    break;
            }
            printk("Running Test Case %d\n%s\n",g_fiona_insert_event,get_expected_results(g_fiona_insert_event));
            g_fiona_insert_event = 0;       // Execute once test...

            if (test_event != INVALID_EVENT) {
                printk("FPOW should identify situation and cancel sleep request\n");
                printk("Inserting test event %s into event queue.\n",fiona_event_keycode_to_str(test_event));
                handle_fiona_event(test_event);
            }
        }
#endif  // ENABLE_EVENT_TEST_PROC

        for (class = FPOW_CLASS_FIRST; class <= FPOW_CLASS_LAST; class++)  {

            // Make sure the user hasn't changed their mind.
            // If this is a "Power Off" request, check the queue to see if 
            // a "Power On" is pending.  IF there is a cancelling 
            // event pending, remove it and ignore the current event
            //  as they cancel out.
            if (power_lockout_state)  {
                // If there's a cancelling power lockout event pending, reverse engines...
                if (sniffer(POWER_LOCKOUT_STATE_SNIFF))  {
                    hlog("\n\nFPOW ABORT - # 1 \n\n");
                    hlog("Found AC insert in queue, reversing power_lockout...\n");
                    rc = -FPOW_CANCELLED_EVENT_ERR;     // we didn't change states
                    print_queue(NULL,0);
                    //strip_queue_for_power_lockout();
                    //sniffFixQueue(POWER_LOCKOUT_STATE_SNIFF);
                    abortReason = POWER_LOCKOUT_STATE_SNIFF;
                    print_queue(NULL,0);
                    unstallEventQueue();
                    unstalled_queue = 1;
                    break;
                }
            }
            else if (off_target_state)  {
                if (class == FPOW_CLASS_IOC) {
                    check_calc_op_amp_offset(0); // Calculate it if we haven't already
                }

                // If we don't have an ON left over pending in the queue, keep going
                if (sniffer(OFF_STATE_SNIFF)) {
                    hlog("\n\nFPOW ABORT - # 2 \n\n");
                    // We have an ON event left in the queue, so cancel our sleep request...
                    rc = -FPOW_CANCELLED_EVENT_ERR;     // we didn't change states
                    abortReason = OFF_STATE_SNIFF;
                    break;          // Abort our of outer for(class) loop
                }
            }
            else if (sleep_target_state)  {
                if (class == FPOW_CLASS_IOC) {
                    if (0 == is_WAN_state(target_state))
                        check_calc_op_amp_offset(0); // Calculate it if we haven't already
                }

                // If there's an OFF in the queue,
                //  cancel sleep request, 
                //  Simulate a "wake from power off switch" situation,
                //  return and let the power manager engine process the power off...
                if (find_event_in_queue(POWER_SWITCH_OFF_EVENT))  {
                    hlog("\n\nFPOW ABORT - # 4 \n\n");
                    // We have an OFF event left in the queue, so cancel our sleep request...
                    rc = -FPOW_CANCELLED_EVENT_ERR;     // we didn't change states
                    abortReason = IDLE_SLEEP_OFF_FOUND;
                    break;          // Abort our of outer for(class) loop
                }
                else if (sniffer(IDLE_SLEEP_STATE_SNIFF))  {
                    rc = -FPOW_CANCELLED_EVENT_ERR;     // we didn't change states
                    abortReason = IDLE_SLEEP_STATE_SNIFF;
                    break;
                }
            }
            if (class == FPOW_CLASS_CPU) {
                printk("FPOW (%s -> %s)\n",fpow_state_to_str(globs.system_state),fpow_state_to_str(target_state));

                // We're ready to go to sleep.  To keep the wakeup voltage check from being
                // skew'ed by averaging it in with samples that will be very old by the time
                // we do the next check, we clear the voltages list.  This gives a more accurate
                // idea of true voltage on wakeup.  Without this, it could take several wakeups
                // before the averages "catch up"...
                //
                ioc_reset_voltage_averages();

                print_queue("FPOW sleeping CPU, Queue Contents:\n",0);
                if (is_processor_off_state(target_state)) {
                    // Make the only wake source the Power switch
                    powdebug("OFF mode seen, disabling wake sources...\n");
                    fiona_set_usb_wake_source(0);           // Disable USB Wake source
                    fiona_set_ioc_wake_source(0);           // Disable IOC Wake
                    fiona_set_power_switch_wake_source(1,0);  // Enable Power Switch Wake
                    wake_on_wan_switch(0);                  // Disable WAN Switch Wake
                }
                else if (power_lockout_state) {
                    // Clear the queue
                    hlog("Clearing queue prior to power lockout...\n");
                    print_queue("Pre power lockout Queue:",0);
                    strip_queue_for_power_lockout();
                    print_queue("Post power lockout Queue:",0);

                    // Wake on AC Adapter event only (which comes from IOC)
                    fiona_set_usb_wake_source(0);           // Disable USB Wake source
                    fiona_set_ioc_wake_source(1);
                    fiona_set_power_switch_wake_source(0,0);
                    wake_on_wan_switch(0);
                }
                else {
                    fiona_set_usb_wake_source(1);           // Enable USB Wake source
                    fiona_set_ioc_wake_source(1);
                    fiona_set_power_switch_wake_source(1,1);
                    wake_on_wan_switch(1);
                }
            }
            rc = fpow_send_to_class(globs.power_state_xlate[target_state][class],class);

            // Turn on IOC debug if CPU is waking up
            if (class == FPOW_CLASS_CPU) {
#ifdef DEBUG_IOC_WAKE_TRANSACTIONS
                g_do_core_debug = 1;
#endif
                if (sleep_target_state && (rc == FPOW_NO_ERR))  {
                    // We just slept and woke up, make sure to clear the queue from stale events
                    hlog("PXA woke, clearing queue of stale events...\n");
                    post_sniff_remove_events_from_queue(WAKE_FROM_SLEEP_STATE_SNIFF);
                }
            }

            if ((rc != FPOW_NO_ERR) && (rc != -FPOW_NO_MATCHES_FOUND_ERR)) {
                powdebug("Power transition failed err=%d for class = %s\n",rc,fpow_class_to_str(class));
                break;
            }
        }

        // If something didn't failed going to sleep, remember the wake reason
        //  otherwise, we never really woke because we never really slept.
        if (rc == FPOW_NO_ERR)  {
            // Determine what the wake source was and send an event for it if needed...
            // unstallEventQueue();
            remember_wake_source();
            we_slept = 1;
            post_wake_source = 1;
        }
        else  {
            if (rc == -FPOW_CANCELLED_EVENT_ERR)  {
                switch (abortReason)  {
                    case OFF_STATE_SNIFF:
                        print_queue("FPOW found cancelling event for OFF operation, STOPPING SLEEP REQUEST\n",0);
                        powdebug("FPOW found cancelling event, fixing Q...\n");
                        unstalled_queue = postSniffFixUnstallQueue(OFF_STATE_SNIFF);
                        unstalled_queue = 1;
                        print_queue("FPOW found cancelling event, Done fixing Q.\n",0);
                        we_slept = 1;
                        rc = FPOW_NO_ERR;
                        break;

                    case IDLE_SLEEP_OFF_FOUND:
                        print_queue("FPOW found OFF while idle sleeping, STOPPING SLEEP REQUEST\n",0);
                        postSniffFixUnstallQueue(IDLE_SLEEP_OFF_FOUND); 
                        unstalled_queue = 1;
                        we_slept = 1;
                        rc = FPOW_NO_ERR;
                        break;

                    case POWER_LOCKOUT_STATE_SNIFF:
                        abortReason = POWER_LOCKOUT_STATE_SNIFF;
                        set_remembered_wake_source(AC_PLUG_INSERTED_WAKE);
                        we_slept = 1;
                        rc = FPOW_NO_ERR;
                        break;

                    case IDLE_SLEEP_STATE_SNIFF:
                        hlog("\n\nFPOW ABORT - # 8 fixing event queue\n\n");
                        wakeEventFoundInQ = find_fix_idle_wake_event_in_event_queue();
                        if (INVALID_EVENT != wakeEventFoundInQ)  {
                            hlog("FPOW ABORT - saw %s \n\n",fpow_wake_src_to_str(wakeEventFoundInQ));
                            print_queue("FPOW found wake event in queue, STOPPING SLEEP REQUEST\n",0);
                            hlog("FPOW found wake event %s (%d) in queue, STOPPING SLEEP REQUEST\n",fpow_event_to_str(wakeEventFoundInQ),wakeEventFoundInQ);
                            unstallEventQueue();
                            unstalled_queue = 1;
                            print_queue("FPOW found wake event in queue, Done fixing Q.\n",0);
                            switch(wakeEventFoundInQ)  {
                                case WAN_ON_WAKE:
                                case WAN_OFF_WAKE:
                                case ALT_FONT_KEY_COMBINATION:
                                    wakeEvent = wakeEventFoundInQ;
                                    break;
                                case EVENT_AC_INSERTED:
                                    wakeEvent = AC_PLUG_INSERTED_WAKE;
                                    break;
                                case EVENT_KEY_WAKE:
                                    wakeEvent = KEY_PRESS_WAKE;
                                    break;
                                case EVENT_USB_WAKE:
                                    wakeEvent = USB_WAKE;
                                    break;
                                case EVENT_WAN_RING:
                                    wakeEvent = WAN_RING_WAKE;
                                    break;
                                default:
                                    wakeEvent = INVALID_EVENT;
                                    powdebug("Unknown wakeEvent %s\n",fpow_wake_src_to_str(wakeEventFoundInQ));
                                    break;
                            }
                        }
                        if (INVALID_EVENT != wakeEvent) {
                            powdebug("wakeEvent triggered - remembering %s\n",fpow_wake_src_to_str(wakeEvent));
                            set_remembered_wake_source(wakeEvent);
                            we_slept = 1;
                            rc = FPOW_NO_ERR;
                        }
                        break;

                    default:
                        break;
                }
            }
            else {
                printk("Power transition failed err=%d for class = %s\n",rc,fpow_class_to_str(class));
                sleep_err = rc;
            }
        }

        // We just woke up, make sure to unstall the event queue if we had it stalled...
        // And to remove BAD events if any exist...
        // -njv 10/13/06  Don't do this here, but when the POWER_ON actually
        // gets inserted into the queue...
        if (isEventQueueStalled() && (unstalled_queue == 0)) {
            powdebug("fiona_power_state() - WOKE UP, queue stalled, calling unstallEventQueue\n");
            unstallEventQueue();
        }

        // NOTE: We transition here irregardless of error.  Can we actually
        // do anything on error?
        spin_lock(&globs.state_lock);
        globs.system_state = target_state;
        spin_unlock(&globs.state_lock);

        reverse_order = 1;
        // Remember what state we were in for our log at the end...
        real_last_target_state = target_state;
        if (off_target_state)
            target_state = FPOW_OFF_WAKE_STATE;
        else
            target_state = FPOW_IDLE_WAKE_STATE;
    }

    // For now, it request is for ON, and we were OFF (OFF or WAKE_ON_UPDATES),
    //  then send the request in the reverse order...
    else if ((FPOW_STATE_ON == target_state) &&  \
            ((globs.system_state == FPOW_STATE_OFF) || \
             (globs.system_state == FPOW_STATE_WAKE_ON_UPDATES)))  {

        powdebug("POWER:  We're in a funky state !!!\n");

        reverse_order = 1;
    }

    // If we need to call devices in reverse order (waking up), do it now
    if (reverse_order)  {
        powdebug("fiona_power_state(%s -> %s) in REVERSE order\n",fpow_state_to_str(globs.system_state),fpow_state_to_str(target_state));

        // Now call all interested components and tell them to wake up
        for (class = FPOW_CLASS_LAST; class >= FPOW_CLASS_FIRST; class--)  {
            rc = fpow_send_to_class(globs.power_state_xlate[target_state][class],class);
            if ((rc != FPOW_NO_ERR) && (rc != -FPOW_NO_MATCHES_FOUND_ERR)) {
                powdebug("Power transition reverse failed err=%d for class=%s\n",rc,fpow_class_to_str(class));
                powdebug("Trying one more time....\n");
                rc = fpow_send_to_class(globs.power_state_xlate[target_state][class],class);
                if ((rc != FPOW_NO_ERR) && (rc != -FPOW_NO_MATCHES_FOUND_ERR)) {
                    powdebug("Second attempt failed err=%d for class %s.\n",rc,fpow_class_to_str(class));
                    err |= rc;
                }
                else
                    powdebug("Second attempt succeeded for class %s.\n",fpow_class_to_str(class));
            }
        }
#ifdef FORCE_VIDEO_ON_UPON_WAKE
        fpow_send_to_class(FPOW_MODE_ON,FPOW_CLASS_VIDEO_DISPLAY);
#endif

        // If we had an error on waking up,  ignore it.
        // Since we can't put things back to sleep (won't fix the problem or make it better),
        // we need to live with a component that didn't wake up....
        if (err && we_slept)  {
            rc = FPOW_NO_ERR;
        }

        // Post the event that woke us up if needed
        if (we_slept && post_wake_source)
            post_wake_event();

#ifdef FPOW_DO_PROCESS_SUSPEND_RESUME
        powdebug("fiona_power_state(): Unfreezing processes...\n");
        thaw_processes();
        g_threads_are_stopped = 0;
        pm_restore_console();
#endif
    }

    // Send the power transition messages to all drivers
    else {
        powdebug("fiona_power_state(%s -> %s)\n",fpow_state_to_str(globs.system_state),fpow_state_to_str(target_state));
        for (class = FPOW_CLASS_FIRST; class <= FPOW_CLASS_LAST; class++)  {
            rc = fpow_send_to_class(globs.power_state_xlate[target_state][class],class);
            if ((rc != FPOW_NO_ERR) && (rc != -FPOW_NO_MATCHES_FOUND_ERR)) {
                powdebug("Power transition error=%d for class = %s\n",rc,fpow_class_to_str(class));
                break;
            }
        }

        // If we had an error, retransitions the components that we already changed
        // trying to get to this state.  This will leave us in the same state as when
        // we started.
        if (rc != FPOW_NO_ERR)  {
            // recall all interested components and tell them to go back to previous state
            for (class = FPOW_CLASS_LAST; class >= FPOW_CLASS_FIRST; class--)  {
                err = fpow_send_to_class(globs.power_state_xlate[last_state][class],class);
                if ((err != FPOW_NO_ERR) && (err != -FPOW_NO_MATCHES_FOUND_ERR)) {
                    powdebug("Power transition FIX failed (err=%d) for class = %s\n",err,fpow_class_to_str(class));
                    powdebug("Trying one more time....\n");
                    err = fpow_send_to_class(globs.power_state_xlate[target_state][class],class);
                    if ((err != FPOW_NO_ERR) && (err != -FPOW_NO_MATCHES_FOUND_ERR)) {
                        powdebug("Second attempt failed (err=%d) for class %s.\n",err,fpow_class_to_str(class));
                    }
                    else {
                        powdebug("Second attempt succeeded for class %s.\n",fpow_class_to_str(class));
                    }
                }
            }
        }
    }



    // NOTE: We transition here irregardless of error.  Can we actually
    // do anything on error?
    spin_lock(&globs.state_lock);
    globs.system_state = target_state;
    spin_unlock(&globs.state_lock);


    // FIXME - if we got here and it was an "OFF" request, then we're actually in
    // the process of waking up, so we need to send the wakeup to folks
    // (actually, we send the last state we were in before sleeping...) -njv

exit:
    // Make sure that we check to see if we had an error trying to sleep.
    // If we did, make sure to send back the error here.
    if (sleep_err != FPOW_NO_ERR)
        rc = sleep_err;

    // Tell the user
    if (target_state != globs.system_state) {
        powdebug("Error !! fiona_power_state(%s) EXIT with system_state = %s\n",fpow_state_to_str(target_state),fpow_state_to_str(globs.system_state));
    } else {
        // If it was sleep state and target changes since we started, use sleep state
        // as the "coming from" state
        if (real_last_target_state != FPOW_STATE_INVALID)
            last_state = real_last_target_state;
        poweventdebug("POWER_STATE: %s <--- %s\n",fpow_state_to_str(target_state),fpow_state_to_str(last_state));
        print_queue("Queue upon leaving fiona_power_state()\n",0);
    }

    if (rc != FPOW_NO_ERR)
        printk("ERROR: fiona_power_state() returning error %d\n",rc);

#ifdef DEBUG_IOC_WAKE_TRANSACTIONS
    g_do_core_debug = orig;
#endif

    return(rc);

#ifdef FPOW_DO_PROCESS_SUSPEND_RESUME
thaw:
    powdebug("fiona_power_state(): Unfreezing processes (thaw)...\n");
    thaw_processes();
    g_threads_are_stopped = 0;
    pm_restore_console();
    if (rc != FPOW_NO_ERR)
        printk("ERROR: fiona_power_state() returning error %d\n",rc);
    return(rc);
#endif
}

void fiona_pgsr_set(int gpio)
{
	if (gpio < 32) {
		g_fiona_hw.pgsr0_reg |= (1 << (gpio));
	} else if (gpio < 64) {
		g_fiona_hw.pgsr1_reg |= (1 << ((gpio) - 32));
	} else {
		g_fiona_hw.pgsr2_reg |= (1 << ((gpio) - 64));
	}
}

EXPORT_SYMBOL(fiona_pgsr_set);

void fiona_pgsr_clear(int gpio)
{
	if (gpio < 32) {
		g_fiona_hw.pgsr0_reg &= ~(1 << (gpio));
	} else if (gpio < 64) {
		g_fiona_hw.pgsr1_reg &= ~(1 << ((gpio) - 32));
	} else {
		g_fiona_hw.pgsr2_reg &= ~(1 << ((gpio) - 64));
	}
}

EXPORT_SYMBOL(fiona_pgsr_clear);

#ifdef CONFIG_FIONA_PM_CPU

static void inline
cpu_fpow_set_sleep_regs(void)
{
    PGSR0 = g_fiona_hw.pgsr0_reg;
    PGSR1 = g_fiona_hw.pgsr1_reg;
    PGSR2 = g_fiona_hw.pgsr2_reg;
    PWER = g_fiona_hw.pwer;
    PFER = g_fiona_hw.pfer;
    PRER = g_fiona_hw.prer;
}

#define CPUFREQ_USERSPACE_SETSPEED      "/sys/devices/system/cpu/cpu0/cpufreq/scaling_setspeed"

static int
cpu_fpow_set_cpu_speed(int speed)
{
    int ret = -1, fd;

    mm_segment_t old_fs = get_fs();
    set_fs(get_ds());

    fd = sys_open(CPUFREQ_USERSPACE_SETSPEED, O_WRONLY, 0);
    if (fd >= 0) {
        char buf[128];
        int len = sprintf(buf, "%d\n", speed);
        if (len > 0 && sys_write(fd, buf, len) == len) {
            ret = 0;
        }

        sys_close(fd);
    }

    set_fs(old_fs);

    return 0 /* keep fpow happy (ret) */;
}

static void print_wake_sources(void) {
    int displayed = 0;
    int found = 0;
    printk("FPOW WAKE SOURCES");
    if (PWER & (1<<g_fiona_hw.sleep_switch_gpio)) {
        found = 1;
        printk(" : Power Switch ");
        if (g_fiona_hw.sleep_switch_polarity == 1) {
            if (PRER && (1<<g_fiona_hw.sleep_switch_gpio)) {
                displayed = 1;
                printk("OFF");
            }
            if (PFER && (1<<g_fiona_hw.sleep_switch_gpio)) {
                if (displayed)
                    printk("/");
                printk("ON");
            }
        }
        else  {
            if (PFER && (1<<g_fiona_hw.sleep_switch_gpio)) {
                displayed = 1;
                printk("OFF");
            }
            if (PRER && (1<<g_fiona_hw.sleep_switch_gpio)) {
                if (displayed)
                    printk("/");
                printk("ON");
            }
        }
    }
    if (PWER & (1<<g_fiona_hw.ioc_wake_gpio)) {
        found = 1;
        printk(" : IOC");
    }
    if (PWER & (1<<g_fiona_hw.usb_wake_gpio)) {
        found = 1;
        printk(" : USB");
    }
    if (PWER & (1<<g_fiona_hw.wake_on_wan_gpio)) {
        found = 1;
        printk(" : WAN");
    }
    if (PWER & PWER_RTC) {
        found = 1;
        printk(" : Voltage Check (RTC)");
    }
    if (found == 0)
        printk(" : NONE ??\n");
    else
        printk("\n");
}
                
/*******************************************************************
 * Routine:     cpu_fpow_setmode
 *
 * Purpose:     Handles Fiona Power Management callbacks
 *              for doing CPU power management.
 *
 * Paramters:   private - value passed at registration - ignored
 *              state   - ignored
 *              mode    - Power mode to switch to
 *                        Currently supported Modes:
 *                          FPOW_MODE_400_MHZ
 *                          FPOW_MODE_200_MHZ
 *                          FPOW_MODE_SLEEP
 * Returns:     
 *  FPOW_NO_ERR                     - Transitioned to mode with no errors
 *  -FPOW_NOT_ENOUGH_POWER_ERR      - Not enough power to transition
 *  -FPOW_COMPONENT_TRANSITION_ERR  - Failed to transition to mode
 *  -EINVAL                         - Invalid mode
 *
 * Assumptions: None
 *******************************************************************/

static FPOW_ERR 
cpu_fpow_setmode(void *private, u32 state, u32 mode)
{
    int retVal = FPOW_NO_ERR;

    powdebug("cpu_fpow_setmode(%s)\n",fpow_mode_to_str(mode));

    // Switch to the appropriate power level
    switch (mode)  {

        case FPOW_MODE_SLEEP:
            // set the sleep states based on the current local settings
            cpu_fpow_set_sleep_regs();
            print_queue("FPOW Event Queue:",1);
            print_wake_sources();
            printk("FPOW PWER=%08X, PRER=%08X, PFER=%08X\n", PWER, PRER, PFER);
            printk("FPOW PGSR: 0=%08X 1=%08X 2=%08X\n", PGSR0, PGSR1, PGSR2);
            retVal = fiona_suspend_system();
            if (retVal != 0)  {
                powdebug("FPOW_CPU: MODE_SLEEP failed fiona_suspend_system with error %d\n",retVal);
            }
            break;


		case FPOW_MODE_400_MHZ:
            retVal = cpu_fpow_set_cpu_speed(PXA_400MHZ_SPEED);
            if (0 == retVal)
                g_cpu_power_mode = mode;
            else
                retVal = -FPOW_COMPONENT_TRANSITION_ERR;
            break;


		case FPOW_MODE_200_MHZ:
            retVal = cpu_fpow_set_cpu_speed(PXA_200MHZ_SPEED);
            if (0 == retVal)
                g_cpu_power_mode = mode;
            else
                retVal = -FPOW_COMPONENT_TRANSITION_ERR;
            break;


        default:
            retVal = -EINVAL;
            break;
    }

    powdebug("FPOW_CPU: CPU now in %s mode\n",fpow_mode_to_str(g_cpu_power_mode));
    return(retVal);
}

static int
cpu_fpow_getmode(void *private)
{
    powdebug("cpu_fpow_getmode(0x%x) returning %s\n",(int)private,fpow_mode_to_str(g_cpu_power_mode));
    return (g_cpu_power_mode);
}

static int
cpu_register_with_fpow(void *private, struct fpow_component **fpow_dev)
{
    struct fpow_registration_rec   reg;
    int rc = 0;

    powdebug("cpu_register_with_fpow(0x%x, 0x%x)\n",(int)private,(int)fpow_dev);
    memclr(&reg,sizeof(reg));

    strcpy(reg.name,"cpu");
    reg.device_class = FPOW_CLASS_CPU;
    reg.supported_modes = ( FPOW_MODE_SLEEP_SUPPORTED | \
                            FPOW_MODE_400_MHZ_SUPPORTED | \
                            FPOW_MODE_200_MHZ_SUPPORTED );
    reg.private = private;
    reg.getmode = cpu_fpow_getmode;
    reg.setmode = cpu_fpow_setmode;
    *fpow_dev = fpow_register(&reg);
    if (*fpow_dev == NULL) {
        printk("CPU_PM: Failed to register with fpow !!\n");
        rc = -1;
    }
    return(rc);
}

static inline void
cpu_unregister_with_fpow(struct fpow_component *fpow_dev)
{
    fpow_unregister(fpow_dev);
}


static int
cpu_pm_init(void)
{
    int retVal = 0;

    // Register CPU power support driver with Fiona Power Manager,
    // we're ignoring result here as we'll still allow driver to
    // work even if we can't do efficient power management...
    retVal = cpu_register_with_fpow(NULL, &g_cpu_fpow_component_ptr);
    if (retVal)
        printk("CPU_PM: Failed to register with fpow !!\n");
    return retVal;
}
#endif  // CONFIG_FIONA_PM_CPU

module_init(FPOWDriver_Init);
module_exit(FPOWDriver_Exit);

MODULE_AUTHOR("Nick Vaccaro");
MODULE_LICENSE("Proprietary");
